//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef CLIENTCREDITS_H
#define CLIENTCREDITS_H

#include "Types.h"		// Needed for uint16_t and uint32_t
#include "MD4Hash.h"	// Needed for CMD4Hash

#include "i2p/CI2PAddress.h"

#define	 MAXPUBKEYSIZE		80

#define CRYPT_CIP_REMOTECLIENT	10
#define CRYPT_CIP_LOCALCLIENT	20
#define CRYPT_CIP_NONECLIENT	30

struct CreditStruct
{
	CreditStruct();
	
	CMD4Hash	key;
	uint32_t		nLastSeen;
	wxULongLong		nDownloaded;	// downloaded from him
	wxULongLong		nUploaded;      // uploaded to him
	uint16_t		nReserved3;
	uint8_t		nKeySize;
	uint8_t		abySecureIdent[MAXPUBKEYSIZE];
};

enum EIdentState{
	IS_NOTAVAILABLE,
	IS_IDNEEDED,
	IS_IDENTIFIED,
	IS_IDFAILED,
	IS_IDBADGUY
};

class CClientCredits
{

public:
	CClientCredits(CreditStruct* in_credits);
	CClientCredits(const CMD4Hash& key);
	~CClientCredits();

	const CMD4Hash& GetKey() const 			{return m_pCredits->key;}
	const uint8_t*	GetSecureIdent() const	{return m_abyPublicKey;}
	uint8_t	GetSecIDKeyLen() const 			{return m_nPublicKeyLen;}
	const CreditStruct* GetDataStruct() const	{return m_pCredits;}
	void	ClearWaitStartTime();
	
	//void	AddDownloaded(uint32_t bytes, uint32_t dwForIP, bool cryptoavail);
	void	AddDownloaded(uint32_t bytes, const CI2PAddress & dwForDest, bool cryptoavail);
	//void	AddUploaded(uint32_t bytes, uint32_t dwForIP, bool cryptoavail);
	void	AddUploaded(uint32_t bytes, const CI2PAddress & dwForDest, bool cryptoavail);
	
	wxULongLong	GetUploadedTotal() const;
	wxULongLong	GetDownloadedTotal() const;
	//I2P:float	GetScoreRatio(uint32_t dwForIP, bool cryptoavail);
	float	GetScoreRatio(const CI2PAddress & dwForDest, bool cryptoavail);
	void	SetLastSeen();
	bool	SetSecureIdent(const uint8_t* pachIdent, uint8_t nIdentLen); // Public key cannot change, use only if there is not public key yet
	uint32_t	m_dwCryptRndChallengeFor;
	uint32_t	m_dwCryptRndChallengeFrom;
	//I2P:EIdentState	GetCurrentIdentState(uint32_t dwForIP) const; // can be != m_identState
	EIdentState	GetCurrentIdentState(const CI2PAddress & dwForDest) const; // can be != m_identState
	//I2P:uint32_t	GetSecureWaitStartTime(uint32_t dwForIP);
	uint32_t	GetSecureWaitStartTime(const CI2PAddress &  dwForDest);
	//I2P:void	SetSecWaitStartTime(uint32_t dwForIP);
	void	SetSecWaitStartTime(const CI2PAddress &  dwForDest);
	//I2P:void	Verified(uint32_t dwForIP);
	void	Verified(const CI2PAddress &  dwForDest);
	EIdentState GetIdentState() const { return m_identState; }
	void	SetIdentState(EIdentState state) { m_identState = state; }
	
private:
	EIdentState		m_identState;
	void			InitalizeIdent();
	CreditStruct*	m_pCredits;
	uint8_t			m_abyPublicKey[80];			// even keys which are not verified will be stored here, and - if verified - copied into the struct
	uint8_t			m_nPublicKeyLen;
	//uint32_t			m_dwIdentIP;
	CI2PAddress 		m_dwIdentDest;
	
	uint32_t			m_dwWaitTime;
	uint32_t			m_dwSecureWaitTime;
	uint32_t			m_dwUnSecureWaitTime;
	//uint32_t			m_dwWaitTimeIP;			   // client IP assigned to the waittime
	CI2PAddress			m_dwWaitTimeDest;			   // client IP assigned to the waittime
};

#endif // CLIENTCREDITS_H
