//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef CLIENTCREDITSLIST_H
#define CLIENTCREDITSLIST_H

#include "Types.h"		// Needed for uint16_t and uint32_t
#include "MD4Hash.h"	// Needed for CMD4Hash

#include <map>

#include "i2p/CI2PAddress.h"

class CClientCredits;

class CClientCreditsList
{
public:
	CClientCreditsList();
	~CClientCreditsList();
	
	// return signature size, 0 = Failed | use sigkey param for debug only
	uint8_t	CreateSignature(CClientCredits* pTarget, uint8_t* pachOutput, uint8_t nMaxSize, const CI2PAddress & ChallengeDest/*I2P:uint32_t ChallengeIP*/, uint8_t byChaIPKind, void* sigkey = NULL);
	bool	VerifyIdent(CClientCredits* pTarget, const uint8_t* pachSignature, uint8_t nInputSize, const CI2PAddress & dwForDest/*I2P:uint32_t dwForIP*/, uint8_t byChaIPKind);	

	CClientCredits* GetCredit(const CMD4Hash& key);
	void	Process();
	uint8_t	GetPubKeyLen() const 			{return m_nMyPublicKeyLen;}
	const uint8_t*	GetPublicKey() const		{return m_abyMyPublicKey;}
	bool	CryptoAvailable() const;
	void	SaveList();
protected:
	void	LoadList();
	void	InitalizeCrypting();
	bool	CreateKeyPair();
#ifdef _DEBUG
	bool	Debug_CheckCrypting();
#endif
private:
	typedef std::map<CMD4Hash, CClientCredits*> ClientMap;
	ClientMap m_mapClients;
	uint32_t			m_nLastSaved;
	// A void* to avoid having to include the large CryptoPP.h file
	void*		m_pSignkey;
	uint8_t			m_abyMyPublicKey[80];
	uint8_t			m_nMyPublicKeyLen;
};

#endif // CLIENTCREDITSLIST_H
