//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include "ClientTCPSocket.h"	// Interface declarations

#include <vector>

#include <wx/arrimpl.cpp>	// this is a magic incantation which must be done!
#include <wx/tokenzr.h> 	// Needed for wxStringTokenizer

#include "Proxy.h"		// Needed for CProxyData
#include "Preferences.h"	// Needed for thePrefs
#include "OPCodes.h"		// Needed for CONNECTION_TIMEOUT
#include "Packet.h"		// Needed for CPacket
#include "Statistics.h"		// Needed for theStats
#include "Logger.h"		// Neeed for logRemoteClient
#include "updownclient.h"	// Needed for CUpDownClient
#include <common/Format.h>		// Needed for CFormat
#include "imule.h"		// Needed for theApp
#include "SharedFileList.h"	// Needed for CSharedFileList
#include "ClientList.h"		// Needed for CClientList
#include "UploadQueue.h"	// Needed for CUploadQueue
#include "ClientUDPSocket.h" // Needed for CClientUDPSocket
#include "PartFile.h"		// Needed for CPartFile
#include "MemFile.h"		// Needed for CMemFile
#include "kademlia/kademlia/Kademlia.h" // Needed for CKademlia::Kademlia
#include "kademlia/kademlia/Prefs.h"	// Needed for CKademlia::CPrefs
#include "DownloadQueue.h"	// Needed for CDownloadQueue
#include "Server.h"		// Needed for CServer
#include "ServerList.h"		// Needed for CServerList
#include "IPFilter.h"		// Needed for CIPFilter
#include "ListenSocket.h"	// Needed for CListenSocket
#include "GuiEvents.h"          // Needed for Nofify_... macros
#include "SHA.h"                // Needed for CSHA
#include "i2p/wxI2PEvents.h"
#include "wx/socket.h"

// similar defines are in EMSocket.cpp for tracking errors in packet headers
// especially to dump sent packets : __PACKET_SEND_DUMP__
//#define __PACKET_RECV_DUMP__

//IMPLEMENT_DYNAMIC_CLASS ( CClientTCPSocket, CEMSocket )

//------------------------------------------------------------------------------
// CClientTCPSocketHandler
//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
// CClientTCPSocketHandler
//------------------------------------------------------------------------------

class CClientTCPSocketHandler: public wxEvtHandler
{
	public:
		CClientTCPSocketHandler() {};

	private:
        //void ClientTCPSocketHandler ( wxSocketEvent& event );
        void ClientTCPSocketHandler ( CI2PSocketEvent& event );
        DECLARE_EVENT_TABLE();
};


/**
 * Handles socket events
 * @param event the event
 */
//void CClientTCPSocketHandler::ClientTCPSocketHandler ( wxSocketEvent& event )
void CClientTCPSocketHandler::ClientTCPSocketHandler ( CI2PSocketEvent& event )
{
	CClientTCPSocket *socket = dynamic_cast<CClientTCPSocket *>( event.GetSocket() );
	wxASSERT ( socket );

	if ( !socket )
	{
		return;
	}

	if ( socket->OnDestroy() || socket->deletethis )
	{
		return;
	}

	switch ( event.GetSocketEvent() )
	{
		case wxSOCKET_LOST:
			socket->OnError ( 0xFEFF /* SOCKET_LOST is not an error */ );

			break;

		case wxSOCKET_INPUT:
			socket->OnReceive ( 0 );

			break;

		case wxSOCKET_OUTPUT:
			socket->OnSend ( 0 );

			break;

		case wxSOCKET_CONNECTION:
			// connection stablished, nothing to do about it?
			socket->OnConnect ( socket->Error() ? socket->LastError() : 0 );

			break;

		default:
			// Nothing should arrive here...
			wxASSERT ( 0 );

			break;
	}
}

//
// There can be only one. :)
//

static CClientTCPSocketHandler g_clientReqSocketHandler;


//------------------------------------------------------------------------------
// CClientTCPSocket
//------------------------------------------------------------------------------

WX_DEFINE_OBJARRAY ( ArrayOfwxStrings )

CClientTCPSocket::CClientTCPSocket ( CUpDownClient * in_client )
		: CEMSocket ()
{
	SetClient ( in_client );
	ResetTimeOutTimer();
	deletethis = false;

	SetEventHandler ( g_clientReqSocketHandler, CLIENTTCPSOCKET_HANDLER );
	SetNotify (
	    wxSOCKET_CONNECTION_FLAG |
	    wxSOCKET_INPUT_FLAG |
	    wxSOCKET_OUTPUT_FLAG |
	    wxSOCKET_LOST_FLAG );
	Notify ( true );

	theApp.listensocket->AddSocket ( this );
	theApp.listensocket->AddConnection();
}


CClientTCPSocket::~CClientTCPSocket()

{
    wxASSERT_MSG(OnDestroy(), wxT("CClientTCPSocket deleted without call to Destroy()"));
}

bool CClientTCPSocket::Destroy()
{
    wxCHECK_MSG(!OnDestroy(), true, wxT("CClientTCPSocket already OnDestroy()"));
    
    // remove event handler
    SetNotify ( 0 );
    Notify ( false );

    if ( m_client )
    {
        m_client->SetSocket ( NULL );
        m_client = NULL;
    }

    if ( theApp.listensocket/* && !theApp.listensocket->OnShutdown()*/ )
    {
        theApp.listensocket->RemoveSocket ( this );
    }
    return CEMSocket::Destroy() ;
}

void CClientTCPSocket::ResetTimeOutTimer()
{
	timeout_timer = ::GetTickCount();
}


bool CClientTCPSocket::CheckTimeOut()
{
	// 0.42x
	uint32_t uTimeout = GetTimeOut();

	if ( m_client )
	{

		/*NOTE: I2P = no buddy
		if ( m_client->GetKadState() == KS_CONNECTED_BUDDY )
		{
			//We originally ignored the timeout here for buddies.
			//This was a stupid idea on my part. There is now a ping/pong system
			//for buddies. This ping/pong system now prevents timeouts.
			//This release will allow lowID clients with KadVersion 0 to remain connected.
			//But a soon future version needs to allow these older clients to time out to prevent dead connections from continuing.
			//JOHNTODO: Don't forget to remove backward support in a future release.

			if ( m_client->GetKadVersion() == 0 )
			{
				return false;
			}

			uTimeout += MIN2MS ( 15 );
		}*/

		if ( m_client->GetChatState() != MS_NONE )
		{
			uTimeout = 2 * GetTimeOut();
		}
	}

	if ( ::GetTickCount() - timeout_timer > uTimeout )
	{
		timeout_timer = ::GetTickCount();
		Disconnect ( wxT ( "Timeout" ) );
		return true;
	}

	return false;
}


void CClientTCPSocket::SetClient ( CUpDownClient* pClient )
{
	m_client = pClient;

	if ( m_client )
	{
		m_client->SetSocket ( this );
	}
}


void CClientTCPSocket::OnClose ( int nErrorCode )
{
	// 0.42x
	wxASSERT ( theApp.listensocket->IsValidSocket ( this ) );
	CEMSocket::OnClose ( nErrorCode );

	if ( nErrorCode )
	{
		Disconnect ( wxString::Format ( wxT ( "Closed: %u" ), nErrorCode ) );
	}
	else
	{
		Disconnect ( wxT ( "Close" ) );
	}
}


void CClientTCPSocket::Disconnect ( const wxString& strReason )
{
	byConnected = ES_DISCONNECTED;

	if ( m_client )
	{
		m_client->Disconnected ( strReason, true );
	}
	else
		AddDebugLogLineM ( false, logRemoteClient, strReason );

	Safe_Delete();
}


void CClientTCPSocket::Safe_Delete()

{
	if ( !deletethis && !OnDestroy() )
	{
		// lfroen: first of all - stop handler
		deletethis = true;

		byConnected = ES_DISCONNECTED;

		Destroy();
	}
}


bool CClientTCPSocket::ProcessPacket ( CPacket * in_packet )
//const char* buffer, uint32_t size, uint8_t opcode)
{

	const char * buffer = in_packet->GetDataBuffer() ;

	uint32_t size = in_packet->GetPacketSize();

	uint8_t opcode = in_packet->GetOpCode() ;

	if ( !m_client && opcode != OP_HELLO )
	{
		throw wxString ( CFormat ( wxT ( "%x : Asks for something without saying hello" ) ) % ( long ) this );
	}
	else if ( m_client && opcode != OP_HELLO && opcode != OP_HELLOANSWER )
		m_client->CheckHandshakeFinished ( OP_EDONKEYPROT, opcode );

	switch ( opcode )
	{
		case OP_HELLOANSWER:
		{	// 0.43b
			theStats::AddDownOverheadOther ( size );
			m_client->ProcessHelloAnswer ( buffer, size );

			// start secure identification, if
			//  - we have received OP_EMULEINFO and OP_HELLOANSWER (old eMule)
			//	- we have received eMule-OP_HELLOANSWER (new eMule)

			if ( m_client->GetInfoPacketsReceived() == IP_BOTH )
			{
				m_client->InfoPacketsReceived();
			}

			// Socket might die because of sending in InfoPacketsReceived, so check
			if ( m_client )
			{
				m_client->ConnectionEstablished();
			}

			// Socket might die on ConnectionEstablished somehow. Check it.
			if ( m_client )
			{
				Notify_UploadCtrlRefreshClient ( m_client );
			}

			break;
		}

		case OP_HELLO:
		{	// 0.43b

			theStats::AddDownOverheadOther ( size );
			bool bNewClient = !m_client;

			if ( bNewClient )
			{
				// create new client to save standard informations
				m_client = new CUpDownClient ( this );
			}

			bool bIsMuleHello = false;

				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_HELLO" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			try
			{
				bIsMuleHello = m_client->ProcessHelloPacket ( buffer, size );
			}
			catch ( ... )
			{
				if ( bNewClient && m_client )
				{
					// Don't let CUpDownClient::Disconnected be processed for a client which is not in the list of clients.
					m_client->Safe_Delete();
					m_client = NULL;
				}

				throw;
			}

			// if IP is filtered, dont reply but disconnect...
			if ( theApp.ipfilter->IsFiltered ( m_client->GetTCPDest/*I2P:IP*/() ) )
			{
				if ( bNewClient )
				{
					m_client->Safe_Delete();
					m_client = NULL;
				}

				Disconnect ( wxT ( "IPFilter" ) );

				return false;
			}

			wxASSERT ( m_client );

			// now we check if we know this client already. if yes this socket will
			// be attached to the known client, the new client will be deleted
			// and the var. "client" will point to the known client.
			// if not we keep our new-constructed client ;)

			if ( theApp.clientlist->AttachToAlreadyKnown ( &m_client, this ) )
			{
				// update the old client informations
				bIsMuleHello = m_client->ProcessHelloPacket ( buffer, size );
			}
			else
			{
				theApp.clientlist->AddClient ( m_client );
				m_client->SetCommentDirty();
			}

			Notify_UploadCtrlRefreshClient ( m_client );

			// send a response packet with standart informations

			if ( ( m_client->GetHashType() == SO_EMULE ) && !bIsMuleHello )
			{
				m_client->SendMuleInfoPacket ( false );
			}

			// Client might die from Sending in SendMuleInfoPacket, so check
			if ( m_client )
			{
				m_client->SendHelloAnswer();
			}

			// Kry - If the other side supports it, send OS_INFO
			// Client might die from Sending in SendHelloAnswer, so check
			if ( m_client && m_client->GetOSInfoSupport() )
			{
				m_client->SendMuleInfoPacket ( false, true ); // Send the OS Info tag on the recycled Mule Info
			}

			// Client might die from Sending in SendMuleInfoPacket, so check
			if ( m_client )
			{
				m_client->ConnectionEstablished();
			}

			// start secure identification, if
			//	- we have received eMule-OP_HELLO (new eMule)
			if ( m_client && m_client->GetInfoPacketsReceived() == IP_BOTH )
			{
				m_client->InfoPacketsReceived();
			}

			break;
		}

		case OP_REQUESTFILENAME:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_REQUESTFILENAME" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );
			// IP banned, no answer for this request

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( size >= 16 )
			{
				if ( !m_client->GetWaitStartTime() )
				{
					m_client->SetWaitStartTime();
				}

				CMemFile data_in ( ( uint8_t* ) buffer, size );

				CMD4Hash reqfilehash = data_in.ReadHash();
				CKnownFile *reqfile = theApp.sharedfiles->GetFileByID ( reqfilehash );

				if ( reqfile == NULL )
				{
					reqfile = theApp.downloadqueue->GetFileByID ( reqfilehash );

					if ( ! ( reqfile != NULL && reqfile->GetFileSize() > PARTSIZE ) )
					{
						break;
					}
				}

				// if we are downloading this file, this could be a new source
				// no passive adding of files with only one part
				if ( reqfile->IsPartFile() && reqfile->GetFileSize() > PARTSIZE )
				{
					if ( thePrefs::GetMaxSourcePerFile() >
					        ( ( CPartFile* ) reqfile )->GetSourceCount() )
					{
						theApp.downloadqueue->CheckAndAddKnownSource ( ( CPartFile* ) reqfile, m_client );
					}
				}

				// check to see if this is a new file they are asking for
				if ( m_client->GetUploadFileID() != reqfilehash )
				{
					m_client->SetCommentDirty();
				}

				m_client->SetUploadFileID ( reqfile );

				m_client->ProcessExtendedInfo ( &data_in, reqfile );

				// send filename etc
				CMemFile data_out ( 128 );
				data_out.WriteHash ( reqfile->GetFileHash() );
				data_out.WriteString ( reqfile->GetFileName() );
				CPacket* packet = new CPacket ( &data_out );
				packet->SetOpCode ( OP_REQFILENAMEANSWER );

				theStats::AddUpOverheadFileRequest ( packet->GetPacketSize() );
				SendPacket ( packet, true );

				// SendPacket might kill the socket, so check

				if ( m_client )
					m_client->SendCommentInfo ( reqfile );

				break;
			}

			throw wxString ( wxT ( "Invalid OP_REQUESTFILENAME packet size" ) );

			break;
		}

		case OP_SETREQFILEID:
		{	// 0.43b EXCEPT track of bad clients
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_SETREQFILEID" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( m_client->IsBanned() )
			{
				break;
			}

			// DbT:FileRequest
			if ( size == 16 )
			{
				if ( !m_client->GetWaitStartTime() )
				{
					m_client->SetWaitStartTime();
				}

				const CMD4Hash fileID ( ( uint8_t* ) buffer );

				CKnownFile *reqfile = theApp.sharedfiles->GetFileByID ( fileID );

				if ( reqfile == NULL )
				{
					reqfile = theApp.downloadqueue->GetFileByID ( fileID );

					if ( ! ( reqfile  != NULL && reqfile->GetFileSize() > PARTSIZE ) )
					{
						CPacket* replypacket = new CPacket ( OP_FILEREQANSNOFIL, 16 );
						replypacket->Copy16ToDataBuffer ( fileID.GetHash() );
						theStats::AddUpOverheadFileRequest ( replypacket->GetPacketSize() );
						SendPacket ( replypacket, true );
						break;
					}
				}

				// check to see if this is a new file they are asking for
				if ( m_client->GetUploadFileID() != fileID )
				{
					m_client->SetCommentDirty();
				}

				m_client->SetUploadFileID ( reqfile );

				// send filestatus
				CMemFile data ( 16 + 16 );
				data.WriteHash ( reqfile->GetFileHash() );

				if ( reqfile->IsPartFile() )
				{
					( ( CPartFile* ) reqfile )->WritePartStatus ( &data );
				}
				else
				{	
					//! write part count
					data.WriteUInt64 ( 0 );
				}

				CPacket* packet = new CPacket ( &data );

				packet->SetOpCode ( OP_FILESTATUS );

				theStats::AddUpOverheadFileRequest ( packet->GetPacketSize() );
				SendPacket ( packet, true );
				break;
			}

			throw wxString ( wxT ( "Invalid OP_FILEREQUEST packet size" ) );

			break;
			// DbT:End
		}

		case OP_FILEREQANSNOFIL:
		{	// 0.43b protocol, lacks ZZ's download manager on swap
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_FILEREQANSNOFIL" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( size == 16 )
			{
				// if that client does not have my file maybe has another different
				CPartFile* reqfile = theApp.downloadqueue->GetFileByID ( CMD4Hash ( ( uint8_t* ) buffer ) );

				if ( reqfile )
				{
					reqfile->AddDeadSource ( m_client );
				}
				else
				{
					break;
				}

				// we try to swap to another file ignoring no needed parts files
				switch ( m_client->GetDownloadState() )
				{
					case DS_CONNECTED:
					case DS_ONQUEUE:
					case DS_NONEEDEDPARTS:
						if ( !m_client->SwapToAnotherFile ( true, true, true, NULL ) )
						{
							theApp.downloadqueue->RemoveSource ( m_client );
						}

						break;
				}

				break;
			}

			throw wxString ( wxT ( "Invalid OP_FILEREQUEST packet size" ) );

			break;
		}

		case OP_REQFILENAMEANSWER:
		{	// 0.43b except check for bad clients
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_REQFILENAMEANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			CMD4Hash hash = data.ReadHash();

			const CPartFile* file = theApp.downloadqueue->GetFileByID ( hash );

			m_client->ProcessFileInfo ( &data, file );

			break;
		}

		case OP_FILESTATUS:
		{		// 0.43b except check for bad clients
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_FILESTATUS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			CMD4Hash hash = data.ReadHash();

			const CPartFile* file = theApp.downloadqueue->GetFileByID ( hash );

			m_client->ProcessFileStatus ( false, &data, file );

			break;
		}

		case OP_STARTUPLOADREQ:
		{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_STARTUPLOADREQ" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( !m_client->CheckHandshakeFinished ( OP_EDONKEYPROT, opcode ) )
			{
				break;
			}

			m_client->CheckForAggressive();

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( size == 16 )
			{
				const CMD4Hash fileID ( ( uint8_t* ) buffer );

				CKnownFile* reqfile = theApp.sharedfiles->GetFileByID ( fileID );

				if ( reqfile )
				{
					if ( m_client->GetUploadFileID() != fileID )
					{
						m_client->SetCommentDirty();
					}

					m_client->SetUploadFileID ( reqfile );

					m_client->SendCommentInfo ( reqfile );

					// Socket might die because of SendCommentInfo, so check

					if ( m_client )
						theApp.uploadqueue->AddClientToQueue ( m_client );
				}
			}

			break;
		}

		case OP_QUEUERANK:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_QUEUERANK" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			uint32_t rank = data.ReadUInt32();

			wxASSERT( rank <= (uint16_t) -1 );
			m_client->SetRemoteQueueRank ( (uint16_t) rank );
			break;
		}

		case OP_ACCEPTUPLOADREQ:
		{	// 0.42e (xcept khaos stats)
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ACCEPTUPLOADREQ" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( m_client->GetRequestFile() && !m_client->GetRequestFile()->IsStopped() && ( m_client->GetRequestFile()->GetStatus() == PS_READY || m_client->GetRequestFile()->GetStatus() == PS_EMPTY ) )
			{
				if ( m_client->GetDownloadState() == DS_ONQUEUE )
				{
					m_client->SetDownloadState ( DS_DOWNLOADING );
					m_client->SetLastPartAsked ( CPartFile::NO_PART ); // Reset current downloaded Chunk // Maella -Enhanced Chunk Selection- (based on jicxicmic)
					m_client->SendBlockRequests();
				}
			}
			else
			{
				m_client->SendCancelTransfer() ;
				
				if ( m_client )
					m_client->SetDownloadState ( ( m_client->GetRequestFile() == NULL || m_client->GetRequestFile()->IsStopped() ) ? DS_NONE : DS_ONQUEUE );
			}

			break;
		}

		case OP_REQUESTPARTS:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_REQUESTPARTS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			CMemFile data ( ( uint8_t* ) buffer, size );

			CMD4Hash reqfilehash = data.ReadHash();

			wxULongLong auStartOffsets[3];
			auStartOffsets[0] = data.ReadUInt64();
			auStartOffsets[1] = data.ReadUInt64();
			auStartOffsets[2] = data.ReadUInt64();

			wxULongLong auEndOffsets[3];
			auEndOffsets[0] = data.ReadUInt64();
			auEndOffsets[1] = data.ReadUInt64();
			auEndOffsets[2] = data.ReadUInt64();


			for ( int i = 0; i < ARRSIZE ( auStartOffsets ); i++ )
			{
				if ( auEndOffsets[i] > auStartOffsets[i] )
				{
					Requested_Block_Struct* reqblock = new Requested_Block_Struct;
					reqblock->StartOffset = auStartOffsets[i];
					reqblock->EndOffset = auEndOffsets[i];
					md4cpy ( reqblock->FileID, reqfilehash.GetHash() );
					reqblock->transferred = 0;
					m_client->AddReqBlock ( reqblock );
				}
				else
				{
					if ( CLogger::IsEnabled ( logRemoteClient ) )
					{
						if ( auEndOffsets[i] != 0 || auStartOffsets[i] != 0 )
						{
							AddDebugLogLineM( false, logRemoteClient, CFormat( _("Client request %u is invalid. File block %llu-%llu (%lld bytes) from client %s")) 
								% i 
								% (auStartOffsets[i].GetValue())
								% (auEndOffsets[i].GetValue())
								% (auEndOffsets[i].GetValue() - auStartOffsets[i].GetValue())
								% m_client->GetTCPDest().humanReadable() );
						}
					}
				}
			}

			break;
		}

		case OP_CANCELTRANSFER:
		{
			theStats::AddDownOverheadFileRequest ( size );
			theApp.uploadqueue->RemoveFromUploadQueue ( m_client );

			AddDebugLogLineM ( false, logClient, CFormat ( wxT ( "%x : Upload session to client %s has ended on his request." ) ) 
								% ( long ) m_client % m_client->GetUserName() );

			break;
		}

		case OP_END_OF_DOWNLOAD:
		{ // 0.43b except check for bad clients
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_END_OF_DOWNLOAD" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( size >= 16 && m_client->GetUploadFileID() == CMD4Hash ( ( uint8_t* ) buffer ) )
			{
				theApp.uploadqueue->RemoveFromUploadQueue ( m_client );

				if ( CLogger::IsEnabled ( logClient ) )
				{
					AddDebugLogLineM ( false, logClient, CFormat ( wxT ( "%x :" ) ) % ( long ) m_client + m_client->GetUserName() + CFormat ( wxT ( ": Upload session ended due ended transfer." ) ) );
				}
			}

			break;
		}

		case OP_HASHSETREQUEST:
		{		// 0.43b except check for bad clients
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_HASHSETREQUEST" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );


			theStats::AddDownOverheadFileRequest ( size );

			if ( size != 16 )
			{
				throw wxString ( wxT ( "Invalid OP_HASHSETREQUEST packet size" ) );
			}

			m_client->SendHashsetPacket ( CMD4Hash ( ( uint8_t* ) buffer ) );

			break;
		}

		case OP_HASHSETANSWER:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_HASHSETANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );
			m_client->ProcessHashSet ( buffer, size );
			break;
		}

		case OP_SENDINGPART:
		{			// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_SENDINGPART" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			if ( m_client->GetRequestFile() &&
			        !m_client->GetRequestFile()->IsStopped() &&
			        ( m_client->GetRequestFile()->GetStatus() == PS_READY || m_client->GetRequestFile()->GetStatus() == PS_EMPTY ) )
			{
				m_client->ProcessBlockPacket ( buffer, size );

				if (	m_client &&
				        ( m_client->GetRequestFile()->IsStopped() ||
				          m_client->GetRequestFile()->GetStatus() == PS_PAUSED ||
				          m_client->GetRequestFile()->GetStatus() == PS_ERROR ) )
				{
					m_client->SendCancelTransfer() ;
					
					if ( m_client )
						m_client->SetDownloadState ( m_client->GetRequestFile()->IsStopped() ? DS_NONE : DS_ONQUEUE );
				}
			}
			else
			{
				m_client->SendCancelTransfer() ;
				if (m_client)
					m_client->SetDownloadState ( ( m_client->GetRequestFile() == NULL || m_client->GetRequestFile()->IsStopped() ) ? DS_NONE : DS_ONQUEUE );
			}

			break;
		}

		case OP_OUTOFPARTREQS:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_OUTOFPARTREQS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( m_client->GetDownloadState() == DS_DOWNLOADING )
			{
				m_client->SetDownloadState ( DS_ONQUEUE );
			}

			break;
		}

		case OP_CHANGE_CLIENT_ID:
		{ 	// Kad reviewed
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_CHANGE_CLIENT_ID" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			CI2PAddress nNewUserDest = data.ReadAddress();
			CI2PAddress nNewServerDest = data.ReadAddress();

			if ( nNewUserDest == m_client->GetTCPDest() )
			{
				CServer* pNewServer = theApp.serverlist->GetServerByDest ( nNewServerDest );

				if ( pNewServer != NULL )
				{
					m_client->SetServerDest ( nNewServerDest );
				}
			}

			break;
		}

		case OP_CHANGE_SLOT:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_CHANGE_SLOT" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			// sometimes sent by Hybrid
			theStats::AddDownOverheadOther ( size );
			break;
		}

		case OP_MESSAGE:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_MESSAGE" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );

			CMemFile message_file ( ( uint8_t* ) buffer, size );

			wxString message = message_file.ReadString ( m_client->GetUnicodeSupport() );

			if ( IsMessageFiltered ( message, m_client ) )
			{
				AddLogLineM ( true, CFormat ( _ ( "Message filtered from '%s' (dest:%x)" ) ) % m_client->GetUserName() % m_client->GetTCPDest/*IP*/().hashCode() );
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "New message from '%s' (dest:%x)" ) ) % m_client->GetUserName() % m_client->GetTCPDest/*IP*/().hashCode() );

				Notify_ChatProcessMsg ( /*GUI_ID(*/m_client->GetTCPDest() /*I2P:IP(),m_client->GetUserPort())*/, m_client->GetUserName() + wxT ( "|" ) + message );
			}

			break;
		}

		case OP_ASKSHAREDFILES:
		{	// 0.43b (well, er, it does the same, but in our own way)
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDFILES" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			// client wants to know what we have in share, let's see if we allow him to know that
			theStats::AddDownOverheadOther ( size );
			// IP banned, no answer for this request

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( thePrefs::CanSeeShares() == vsfaEverybody || ( thePrefs::CanSeeShares() == vsfaFriends && m_client->IsFriend() ) )
			{
				AddDebugLogLineM ( false, logRemoteClient, CFormat ( _ ( "User %s (%x) requested your sharedfiles-list -> Accepted" ) )
				              % m_client->GetUserName()
				              %  m_client->GetTCPDest().hashCode() );

				std::vector<CKnownFile*> list;
				theApp.sharedfiles->CopyFileList ( list );

				CMemFile tempfile ( 80 );
				tempfile.WriteUInt32 ( list.size() );

				for ( unsigned i = 0; i < list.size(); ++i )
				{
					theApp.sharedfiles->CreateOfferedFilePacket ( list[i], &tempfile, NULL, m_client );
				}

				// create a packet and send it
				CPacket* replypacket = new CPacket ( &tempfile );

				replypacket->SetOpCode ( OP_ASKSHAREDFILESANSWER );

				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );

				SendPacket ( replypacket, true, true );
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) requested your sharedfiles-list -> Denied" ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode() );

				CPacket* replypacket = new CPacket ( OP_ASKSHAREDDENIEDANS, 0 );
				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
				SendPacket ( replypacket, true, true );
			}

			break;
		}

		case OP_ASKSHAREDFILESANSWER:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDFILESANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );
			wxString EmptyStr;
			m_client->ProcessSharedFileList ( buffer, size, EmptyStr );
			break;
		}

		case OP_ASKSHAREDDIRS:
		{ 		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDDIRS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );
			wxASSERT ( size == 0 );
			// IP banned, no answer for this request

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( ( thePrefs::CanSeeShares() == vsfaEverybody ) || ( ( thePrefs::CanSeeShares() == vsfaFriends ) && m_client->IsFriend() ) )
			{
				AddDebugLogLineM ( false, logRemoteClient, CFormat ( _ ( "User %s (%x) requested your shareddirectories-list -> Accepted" ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode() );

				// Kry - This new code from eMule will avoid duplicated folders
				ArrayOfwxStrings folders_to_send = getListofSharedFolders();

				// Send packet.
				CMemFile tempfile ( 80 );

				unsigned int uDirs = folders_to_send.GetCount();

				tempfile.WriteUInt32 ( uDirs );

				for ( uint32_t iDir = 0; iDir < uDirs; iDir++ )
				{
					AddDebugLogLineM ( true, logRemoteClient, CFormat ( _ ( "User %s (%x) receives directory code %s, corresponding to %s" ) )
									% m_client->GetUserName()
									% m_client->GetTCPDest().hashCode()
									% folders_to_send[iDir]
									% encodeDirName(folders_to_send[iDir])
							 );
					tempfile.WriteString ( encodeDirName(folders_to_send[iDir]) );
				}

				CPacket* replypacket = new CPacket ( &tempfile );

				replypacket->SetOpCode ( OP_ASKSHAREDDIRSANS );
				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
				SendPacket ( replypacket, true, true );
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) requested your shareddirectories-list -> Denied" ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode() );

				CPacket* replypacket = new CPacket ( OP_ASKSHAREDDENIEDANS, 0 );
				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
				SendPacket ( replypacket, true, true );
			}

			break;
		}

		case OP_ASKSHAREDFILESDIR:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDFILESDIR" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );
			// IP banned, no answer for this request

			if ( m_client->IsBanned() )
			{
				break;
			}

			CMemFile data ( ( uint8_t* ) buffer, size );

			wxString strReqDir = data.ReadString ( m_client->GetUnicodeSupport() );

			if ( thePrefs::CanSeeShares() == vsfaEverybody || ( thePrefs::CanSeeShares() == vsfaFriends && m_client->IsFriend() ) )
			{
				AddLogLineM ( false, CFormat ( _ ( "User %s (%x) requested your sharedfiles-list for directory %s -> accepted" ) ) % m_client->GetUserName()
				              %  m_client->GetTCPDest().hashCode() % strReqDir );
				wxASSERT ( data.GetPosition() == data.GetLength() );
				
				
				// look for the directory name (the request is encoded)
				
				ArrayOfwxStrings folders_available = getListofSharedFolders();
				uint32_t uDirs = folders_available.GetCount() ;
				wxString realReqDir ;
				for (unsigned int i = 0; i<uDirs; i++)
					if (encodeDirName(folders_available[i]) == strReqDir)
				{
					realReqDir = folders_available[i];
					break;
				}
				
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) requested your sharedfiles-list for directory %s -> accepted" ) ) % m_client->GetUserName()
						%  m_client->GetTCPDest().hashCode() % realReqDir );
				
				CTypedPtrList<CPtrList, CKnownFile*> list;

				if ( realReqDir == OP_INCOMPLETE_SHARED_FILES )
				{
					// get all shared files from download queue
					int iQueuedFiles = theApp.downloadqueue->GetFileCount();

					for ( int i = 0; i < iQueuedFiles; i++ )
					{
						CPartFile* pFile = theApp.downloadqueue->GetFileByIndex ( i );

						if ( pFile == NULL || pFile->GetStatus ( true ) != PS_READY )
						{
							break;
						}

						list.AddTail ( pFile );
					}
				}
				else
				{
					theApp.sharedfiles->GetSharedFilesByDirectory ( realReqDir, list );
				}

				CMemFile tempfile ( 80 );

				tempfile.WriteString ( strReqDir );
				tempfile.WriteUInt32 ( list.GetCount() );

				while ( list.GetCount() )
				{
					theApp.sharedfiles->CreateOfferedFilePacket ( list.GetHead(), &tempfile, NULL, m_client );
					list.RemoveHead();
				}

				CPacket* replypacket = new CPacket ( &tempfile );

				replypacket->SetOpCode ( OP_ASKSHAREDFILESDIRANS );
				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
				SendPacket ( replypacket, true, true );
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) requested your sharedfiles-list for directory %s -> denied" ) ) % m_client->GetUserName() % m_client->GetTCPDest().hashCode() % strReqDir );

				CPacket* replypacket = new CPacket ( OP_ASKSHAREDDENIEDANS, 0 );
				theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
				SendPacket ( replypacket, true, true );
			}

			break;
		}

		case OP_ASKSHAREDDIRSANS:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDDIRSANS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );

			if ( m_client->GetFileListRequested() == 1 )
			{
				CMemFile data ( ( uint8_t* ) buffer, size );
				uint32_t uDirs = data.ReadUInt32();

				for ( uint32_t i = 0; i < uDirs; i++ )
				{
					wxString strDir = data.ReadString ( m_client->GetUnicodeSupport() );
					AddLogLineM ( true, CFormat ( _ ( "User %s (%x) shares directory %s" ) )
					              % m_client->GetUserName()
					              % m_client->GetTCPDest().hashCode()
					              % strDir );

					CMemFile tempfile ( 80 );
					tempfile.WriteString ( strDir );
					CPacket* replypacket = new CPacket ( &tempfile );
					replypacket->SetOpCode ( OP_ASKSHAREDFILESDIR );
					theStats::AddUpOverheadOther ( replypacket->GetPacketSize() );
					SendPacket ( replypacket, true, true );
				}

				wxASSERT ( data.GetPosition() == data.GetLength() );

				m_client->SetFileListRequested ( uDirs );
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) sent unrequested shared dirs." ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode() );
			}

			break;
		}

		case OP_ASKSHAREDFILESDIRANS:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDFILESDIRANS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			wxString strDir = data.ReadString ( m_client->GetUnicodeSupport() );

			if ( m_client->GetFileListRequested() > 0 )
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) sent sharedfiles-list for directory %s" ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode()
				              % strDir );

				m_client->ProcessSharedFileList ( buffer + (size_t) data.GetPosition().GetValue(), (uint32_t) (size - data.GetPosition()).GetValue(), strDir );

				if ( m_client->GetFileListRequested() == 0 )
				{
					AddLogLineM ( true, CFormat ( _ ( "User %s (%x) finished sending sharedfiles-list" ) )
					              % m_client->GetUserName()
					              % m_client->GetTCPDest().hashCode() );
				}
			}
			else
			{
				AddLogLineM ( true, CFormat ( _ ( "User %s (%x) sent unwanted sharedfiles-list" ) )
				              % m_client->GetUserName()
				              % m_client->GetTCPDest().hashCode() );
			}

			break;
		}

		case OP_ASKSHAREDDENIEDANS:
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ASKSHAREDDENIEDANS" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );

			wxASSERT ( size == 0 );

			AddLogLineM ( true, CFormat ( _ ( "User %s (%x) denied access to shared directories/files list" ) )
			              % m_client->GetUserName()
			              % m_client->GetTCPDest().hashCode() );

			m_client->SetFileListRequested ( 0 );

			break;

		default:
			theStats::AddDownOverheadOther ( size );

				//AddDebugLogLineM ( false, logRemoteClient, wxString::Format ( wxT ( "Edonkey packet: unknown opcode: %i %x" ), opcode, opcode ) );

			return false;
	}

	return true;
}







ArrayOfwxStrings CClientTCPSocket::getListofSharedFolders()
{
	ArrayOfwxStrings folders_to_send ;
	
	uint32_t uDirs = theApp.glob_prefs->shareddir_list.GetCount();

				// the shared folders

	for ( uint32_t iDir = 0; iDir < uDirs; iDir++ )
	{
		folders_to_send.Add ( wxString ( theApp.glob_prefs->shareddir_list[iDir] ) );
	}

	bool bFoundFolder = false;

	wxString char_ptrDir;
				// ... the categories folders ... (category 0 -> incoming)

	for ( uint32_t ix = 0;ix < theApp.glob_prefs->GetCatCount();ix++ )
	{
		char_ptrDir = theApp.glob_prefs->GetCategory ( ix )->incomingpath;
		bFoundFolder = false;

		for ( uint32_t iDir = 0; iDir < ( uint32_t ) folders_to_send.GetCount(); iDir++ )
		{
			if ( folders_to_send[iDir].CmpNoCase ( char_ptrDir ) == 0 )
			{
				bFoundFolder = true;
				break;
			}
		}

		if ( !bFoundFolder )
		{
			folders_to_send.Add ( wxString ( char_ptrDir ) );
		}
	}

				// ... and the Magic thing from the eDonkey Hybrids...
	bFoundFolder = false;

	for ( uint32_t iDir = 0; iDir < ( uint32_t ) folders_to_send.GetCount(); iDir++ )
	{
		if ( folders_to_send[iDir].CmpNoCase ( OP_INCOMPLETE_SHARED_FILES ) == 0 )
		{
			bFoundFolder = true;
			break;
		}
	}

	if ( !bFoundFolder )
	{
		folders_to_send.Add ( wxString ( OP_INCOMPLETE_SHARED_FILES ) );
	}
	
	return folders_to_send;
}







wxString CClientTCPSocket::encodeDirName(wxString dirName)
{
	static unsigned char * key = NULL ; // key for encoding
	
	if (key == NULL)
	{
		key = new uint8_t[512] ;
		srand(time(NULL));
		for (int i=0; i<512; i++)
			key[i] = (uint8_t) (rand() % 512);
	}
	unsigned int len = 512 + strlen( (const char *) dirName.mb_str());
	char buffer[len];
	memcpy(buffer, key, 512);
	memcpy(buffer + 512, dirName.mb_str(), strlen( (const char *) dirName.mb_str()) );
	
	CSHA hasher;
	hasher.Add(buffer, len) ;
	CAICHHash hash;
	hasher.Finish(hash);
	wxString ret = hash.GetString();
	
	return ret ;
}


bool CClientTCPSocket::ProcessExtPacket ( CPacket * in_packet )
{
	const char* buffer = in_packet->GetDataBuffer();

	uint32_t size = in_packet->GetPacketSize();

	uint8_t opcode = in_packet->GetOpCode();


	// 0.42e - except the catchs on mem exception and file exception
	if ( !m_client )
	{
		throw wxString ( wxT ( "Unknown clients sends extended protocol packet" ) );
	}

	/*
	if (!client->CheckHandshakeFinished(OP_EMULEPROT, opcode)) {
		// Here comes a extended packet without finishing the hanshake.
		// IMHO, we should disconnect the client.
		throw wxString(wxT("Client send extended packet before finishing handshake"));
	}
	*/
	switch ( opcode )
	{
		case OP_MULTIPACKET:
		{	// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_MULTIPACKET" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_MULTIPACKET before finishing handshake" ) );
			}

			CMemFile data_in ( ( uint8_t* ) buffer, size );

			CMD4Hash reqfilehash = data_in.ReadHash();
			CKnownFile* reqfile = theApp.sharedfiles->GetFileByID ( reqfilehash );

			if ( reqfile == NULL )
			{
				reqfile = theApp.downloadqueue->GetFileByID ( reqfilehash );

				if ( ! ( reqfile != NULL && reqfile->GetFileSize() > PARTSIZE ) )
				{
					CPacket* replypacket = new CPacket ( OP_FILEREQANSNOFIL, 16 );
					replypacket->Copy16ToDataBuffer ( reqfilehash.GetHash() );
					theStats::AddUpOverheadFileRequest ( replypacket->GetPacketSize() );
					SendPacket ( replypacket, true );
					break;
				}
			}

			if ( !m_client->GetWaitStartTime() )
			{
				m_client->SetWaitStartTime();
			}

			// if we are downloading this file, this could be a new source
			// no passive adding of files with only one part
			if ( reqfile->IsPartFile() && reqfile->GetFileSize() > PARTSIZE )
			{
				if ( thePrefs::GetMaxSourcePerFile() > ( ( CPartFile* ) reqfile )->GetSourceCount() )
				{
					theApp.downloadqueue->CheckAndAddKnownSource ( ( CPartFile* ) reqfile, m_client );
				}
			}

			// check to see if this is a new file they are asking for
			if ( m_client->GetUploadFileID() != reqfilehash )
			{
				m_client->SetCommentDirty();
			}

			m_client->SetUploadFileID ( reqfile );

			CMemFile data_out ( 128 );
			data_out.WriteHash ( reqfile->GetFileHash() );

			while ( data_in.GetLength() - data_in.GetPosition() > 0 )
			{
				if ( !m_client )
				{
					throw wxString ( wxT ( "Client suddenly disconnected" ) );
				}

				uint8_t opcode_in = data_in.ReadUInt8();

				AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): OP_MULTIPACKET - %s" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % opstr(opcode_in) );
				
				switch ( opcode_in )
				{
					case OP_REQUESTFILENAME:
					{
						m_client->ProcessExtendedInfo ( &data_in, reqfile );
						data_out.WriteUInt8 ( OP_REQFILENAMEANSWER );
						data_out.WriteString ( reqfile->GetFileName() );
						break;
					}

					case OP_AICHFILEHASHREQ:
					{
						if ( m_client->IsSupportingAICH() && reqfile->GetAICHHashset()->GetStatus() == AICH_HASHSETCOMPLETE
						        && reqfile->GetAICHHashset()->HasValidMasterHash() )
						{
							data_out.WriteUInt8 ( OP_AICHFILEHASHANS );
							reqfile->GetAICHHashset()->GetMasterHash().Write ( &data_out );
						}

						break;
					}

					case OP_SETREQFILEID:
					{
						data_out.WriteUInt8 ( OP_FILESTATUS );

						if ( reqfile->IsPartFile() )
						{
							( ( CPartFile* ) reqfile )->WritePartStatus ( &data_out );
						}
						else
						{
							//! write parts count
							data_out.WriteUInt64 ( 0 );
						}

						break;
					}

					//We still send the source packet seperately..
					//We could send it within this packet.. If agreeded, I will fix it..
					case OP_REQUESTSOURCES:
					{
						//Although this shouldn't happen, it's a just in case to any Mods that mess with version numbers.

						if ( m_client->GetSourceExchangeVersion() > 1 )
						{
							//data_out.WriteUInt8(OP_ANSWERSOURCES);
							uint32_t dwTimePassed = ::GetTickCount() - m_client->GetLastSrcReqTime() + CONNECTION_LATENCY;
							bool bNeverAskedBefore = m_client->GetLastSrcReqTime() == 0;

							if (
							    //if not complete and file is rare
							    ( reqfile->IsPartFile()
							      && ( bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASKS )
							      && ( ( CPartFile* ) reqfile )->GetSourceCount() <= RARE_FILE
							    ) ||
							    //OR if file is not rare or if file is complete
							    ( ( bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASKS * MINCOMMONPENALTY ) )
							)
							{
								m_client->SetLastSrcReqTime();
								CPacket* tosend = reqfile->CreateSrcInfoPacket ( m_client );

								if ( tosend )
								{
									theStats::AddUpOverheadSourceExchange ( tosend->GetPacketSize() );
									SendPacket ( tosend, true );
								}
							}
						}

						break;
					}
				}

			}

			if ( data_out.GetLength() > 16 )
			{
				CPacket* reply = new CPacket ( &data_out, OP_EMULEPROT );
				reply->SetOpCode ( OP_MULTIPACKETANSWER );
				theStats::AddUpOverheadFileRequest ( reply->GetPacketSize() );
				SendPacket ( reply, true );
			}

			break;
		}

		case OP_MULTIPACKETANSWER:
		{	// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_MULTIPACKETANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			

			theStats::AddDownOverheadFileRequest ( size );

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_MULTIPACKETANSWER before finishing handshake" ) );
			}

			CMemFile data_in ( ( uint8_t* ) buffer, size );

			CMD4Hash reqfilehash = data_in.ReadHash();

			const CPartFile *reqfile = theApp.downloadqueue->GetFileByID ( reqfilehash );

			//Make sure we are downloading this file.
			if ( !reqfile )
			{
				throw wxString ( wxT ( " Wrong File ID: (OP_MULTIPACKETANSWER; reqfile==NULL)" ) );
			}

			if ( !m_client->GetRequestFile() )
			{
				throw wxString ( wxT ( " Wrong File ID: OP_MULTIPACKETANSWER; client->reqfile==NULL)" ) );
			}

			if ( reqfile != m_client->GetRequestFile() )
			{
				throw wxString ( wxT ( " Wrong File ID: OP_MULTIPACKETANSWER; reqfile!=client->reqfile)" ) );
			}

			while ( data_in.GetLength() != data_in.GetPosition() )
			{
				// Some of the cases down there can actually send a packet and lose the client

				if ( !m_client )
				{
					throw wxString ( wxT ( "Client suddenly disconnected" ) );
				}

				uint8_t opcode_in = data_in.ReadUInt8();

				AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): OP_MULTIPACKETANSWER - %s" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % opstr(opcode_in) );

				switch ( opcode_in )
				{
					case OP_REQFILENAMEANSWER:
					{
						m_client->ProcessFileInfo ( &data_in, reqfile );
						break;
					}

					case OP_FILESTATUS:
					{
						m_client->ProcessFileStatus ( false, &data_in, reqfile );
						break;
					}

					case OP_AICHFILEHASHANS:
					{
						m_client->ProcessAICHFileHash ( &data_in, reqfile );
						break;
					}
				}
			}

			break;
		}

		case OP_EMULEINFO:
		{	// 0.43b
			theStats::AddDownOverheadOther ( size );

			if ( !m_client->ProcessMuleInfoPacket ( buffer, size ) )
			{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_EMULEINFO" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

				// If it's not a OS Info packet, is an old client
				// start secure identification, if
				//  - we have received eD2K and eMule info (old eMule)

				if ( m_client->GetInfoPacketsReceived() == IP_BOTH )
				{
					m_client->InfoPacketsReceived();
				}

				m_client->SendMuleInfoPacket ( true );
			}
			else
			{
				AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_EMULEINFO is an OS_INFO" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			}

			break;
		}

		case OP_EMULEINFOANSWER:
		{	// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_EMULEINFOANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			theStats::AddDownOverheadOther ( size );

			m_client->ProcessMuleInfoPacket ( buffer, size );
			// start secure identification, if
			//  - we have received eD2K and eMule info (old eMule)

			if ( m_client->GetInfoPacketsReceived() == IP_BOTH )
			{
				m_client->InfoPacketsReceived();
			}

			break;
		}

		case OP_SECIDENTSTATE:
		{		// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_SECIDENTSTATE" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_SECIDENTSTATE before finishing handshake" ) );
			}

			m_client->ProcessSecIdentStatePacket ( ( uint8_t* ) buffer, size );

			// ProcessSecIdentStatePacket() might cause the socket to die, so check

			if ( m_client )
			{
				int SecureIdentState = m_client->GetSecureIdentState();

				if ( SecureIdentState == IS_SIGNATURENEEDED )
				{
					m_client->SendSignaturePacket();
				}
				else if ( SecureIdentState == IS_KEYANDSIGNEEDED )
				{
					m_client->SendPublicKeyPacket();
					// SendPublicKeyPacket() might cause the socket to die, so check

					if ( m_client )
					{
						m_client->SendSignaturePacket();
					}
				}
			}

			break;
		}

		case OP_PUBLICKEY:
		{		// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_PUBLICKEY" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			if ( m_client->IsBanned() )
			{
				break;
			}

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_PUBLICKEY before finishing handshake" ) );
			}

			m_client->ProcessPublicKeyPacket ( ( uint8_t* ) buffer, size );

			break;
		}

		case OP_SIGNATURE:
		{			// 0.43b
			//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_SIGNATURE" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_COMPRESSEDPART before finishing handshake" ) );
			}

			m_client->ProcessSignaturePacket ( ( uint8_t* ) buffer, size );

			break;
		}

		case OP_COMPRESSEDPART:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_COMPRESSEDPART" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_COMPRESSEDPART before finishing handshake" ) );
			}

			if ( m_client->GetRequestFile() && !m_client->GetRequestFile()->IsStopped() && ( m_client->GetRequestFile()->GetStatus() == PS_READY || m_client->GetRequestFile()->GetStatus() == PS_EMPTY ) )
			{
				m_client->ProcessBlockPacket ( buffer, size, true );

				if ( m_client && (
				            m_client->GetRequestFile()->IsStopped() ||
				            m_client->GetRequestFile()->GetStatus() == PS_PAUSED ||
				            m_client->GetRequestFile()->GetStatus() == PS_ERROR ) )
				{
					m_client->SendCancelTransfer() ;

					if ( m_client )
					{
						m_client->SetDownloadState ( m_client->GetRequestFile()->IsStopped() ? DS_NONE : DS_ONQUEUE );
					}
				}
			}
			else
			{
				m_client->SendCancelTransfer() ;

				if ( m_client )
				{
					m_client->SetDownloadState ( ( m_client->GetRequestFile() == NULL || m_client->GetRequestFile()->IsStopped() ) ? DS_NONE : DS_ONQUEUE );
				}
			}

			break;
		}

		case OP_QUEUERANKING:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_QUEUERANKING" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadOther ( size );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_QUEUERANKING before finishing handshake" ) );
			}

			if ( size != 12 )
			{
				throw wxString ( wxT ( "Invalid size (OP_QUEUERANKING)" ) );
			}

			uint16_t newrank = PeekUInt16 ( buffer );

			m_client->SetRemoteQueueFull ( false );
			m_client->SetRemoteQueueRank ( newrank );
			break;
		}

		case OP_REQUESTSOURCES:
		{	// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_REQUESTSOURCES" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadSourceExchange ( size );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_REQUESTSOURCES before finishing handshake" ) );
			}

			if ( m_client->GetSourceExchangeVersion() >= 1 )
			{
				if ( size != 16 )
				{
					throw wxString ( wxT ( "Invalid size (OP_QUEUERANKING)" ) );
				}

				//first check shared file list, then download list
				const CMD4Hash fileID ( ( uint8_t* ) buffer );

				CKnownFile* file = theApp.sharedfiles->GetFileByID ( fileID );

				if ( !file )
				{
					file = theApp.downloadqueue->GetFileByID ( fileID );
				}

				if ( file )
				{
					uint32_t dwTimePassed = ::GetTickCount() - m_client->GetLastSrcReqTime() + CONNECTION_LATENCY;
					bool bNeverAskedBefore = m_client->GetLastSrcReqTime() == 0;

					if (
					    //if not complete and file is rare, allow once every 40 minutes
					    ( file->IsPartFile() &&
					      ( ( CPartFile* ) file )->GetSourceCount() <= RARE_FILE &&
					      ( bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASKS )
					    ) ||
					    //OR if file is not rare or if file is complete, allow every 90 minutes
					    ( ( bNeverAskedBefore || dwTimePassed > SOURCECLIENTREASKS * MINCOMMONPENALTY ) )
					)
					{
						m_client->SetLastSrcReqTime();
						CPacket* tosend = file->CreateSrcInfoPacket ( m_client );

						if ( tosend )
						{
							theStats::AddUpOverheadSourceExchange ( tosend->GetPacketSize() );
							SendPacket ( tosend, true, true );
						}
					}
				}
			}

			break;
		}

		case OP_ANSWERSOURCES:
		{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_ANSWERSOURCES" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadSourceExchange ( size );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_ANSWERSOURCES before finishing handshake" ) );
			}

			CMemFile data ( ( uint8_t* ) buffer, size );

			CMD4Hash hash = data.ReadHash();

			const CKnownFile* file = theApp.downloadqueue->GetFileByID ( hash );

			if ( file )
			{
				if ( file->IsPartFile() )
				{
					//set the client's answer time
					m_client->SetLastSrcAnswerTime();
					//and set the file's last answer time
					( ( CPartFile* ) file )->SetLastAnsweredTime();

					( ( CPartFile* ) file )->AddClientSources ( &data, m_client->GetSourceExchangeVersion(), SF_SOURCE_EXCHANGE );
				}
			}

			break;
		}

		case OP_FILEDESC:
		{		// 0.43b
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_FILEDESC" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );

			theStats::AddDownOverheadFileRequest ( size );

			if ( !m_client->CheckHandshakeFinished ( OP_EMULEPROT, opcode ) )
			{
				// Here comes a extended packet without finishing the hanshake.
				// IMHO, we should disconnect the client.
				throw wxString ( wxT ( "Client send OP_FILEDESC before finishing handshake" ) );
			}

			m_client->ProcessMuleCommentPacket ( buffer, size );

			break;
		}

		// Unsupported
		case OP_REQUESTPREVIEW:
		{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_REQUESTPREVIEW" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			break;
		}

		// Unsupported
		case OP_PREVIEWANSWER:
		{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_PREVIEWANSWER" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			break;
		}

		case OP_PUBLICIP_ANSWER:
		{
			theStats::AddDownOverheadOther ( size );
			m_client->ProcessPublicIPAnswer ( ( uint8_t* ) buffer, size );
			break;
		}

		case OP_PUBLICIP_REQ:
		{
			theStats::AddDownOverheadOther ( size );
			//CPacket* pPacket = new CPacket(OP_PUBLICIP_ANSWER, 4, OP_EMULEPROT);
			//pPacket->CopyUInt32ToDataBuffer(m_client->GetTCPDest/*I2P:IP*/());
			char payload[m_client->GetTCPDest().size() ];
			m_client->GetTCPDest().writeBytes ( payload );
			CPacket* pPacket = new CPacket ( OP_PUBLICIP_ANSWER,
			                                 sizeof ( payload ), OP_EMULEPROT );
			pPacket->CopyToDataBuffer ( 0, payload, sizeof ( payload ) );
			theStats::AddUpOverheadOther ( pPacket->GetPacketSize() );
			SendPacket ( pPacket );
			break;
		}

		case OP_AICHANSWER:
		{
			theStats::AddDownOverheadOther ( size );
			m_client->ProcessAICHAnswer ( buffer, size );
			break;
		}

		case OP_AICHREQUEST:
		{
			theStats::AddDownOverheadOther ( size );
			m_client->ProcessAICHRequest ( buffer, size );
			break;
		}

		case OP_AICHFILEHASHANS:
		{
			// those should not be received normally, since we should only get those in MULTIPACKET
			theStats::AddDownOverheadOther ( size );
			CMemFile data ( ( uint8_t* ) buffer, size );
			m_client->ProcessAICHFileHash ( &data, NULL );
			break;
		}

		case OP_AICHFILEHASHREQ:
		{
			// those should not be received normally, since we should only get those in MULTIPACKET
			CMemFile data ( ( uint8_t* ) buffer, size );
			CMD4Hash hash = data.ReadHash();
			CKnownFile* pPartFile = theApp.sharedfiles->GetFileByID ( hash );

			if ( pPartFile == NULL )
			{
				break;
			}

			if ( m_client->IsSupportingAICH() && pPartFile->GetAICHHashset()->GetStatus() == AICH_HASHSETCOMPLETE
			        && pPartFile->GetAICHHashset()->HasValidMasterHash() )
			{
				CMemFile data_out;
				data_out.WriteHash ( hash );
				pPartFile->GetAICHHashset()->GetMasterHash().Write ( &data_out );
				CPacket* packet = new CPacket ( &data_out, OP_EMULEPROT, OP_AICHFILEHASHANS );
				theStats::AddUpOverheadOther ( packet->GetPacketSize() );
				SendPacket ( packet );
			}

			break;
		}

		case OP_CALLBACK:
		{
				//AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x OP_CALLBACK" ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client );
			theStats::AddDownOverheadFileRequest ( size );

			if ( !Kademlia::CKademlia::isRunning() )
			{
				break;
			}

			CMemFile data ( ( uint8_t* ) buffer, size );

			CUInt128 check = data.ReadUInt128();
			check.XOR ( Kademlia::CUInt128 ( true ) );

			if ( check.compareTo ( Kademlia::CKademlia::getPrefs()->getKadID() ) )
			{
				break;
			}

			CUInt128 fileid = data.ReadUInt128();

			uint8_t fileid2[16];
			fileid.toByteArray ( fileid2 );

			const CMD4Hash fileHash ( fileid2 );

			if ( theApp.sharedfiles->GetFileByID ( fileHash ) == NULL )
			{
				if ( theApp.downloadqueue->GetFileByID ( fileHash ) == NULL )
				{
					break;
				}
			}

			CI2PAddress dest = data.ReadAddress();

			CUpDownClient* callback;

			callback = theApp.clientlist->FindClientByDest ( dest );

			if ( callback == NULL )
			{
				//TODO #warning Do we actually have to check friend status here?
				callback = new CUpDownClient ( dest, CI2PAddress::null, NULL, false, false );
				theApp.clientlist->AddClient ( callback );
			}

			callback->TryToConnect ( true );

			break;
		}

		default:
			theStats::AddDownOverheadOther ( size );

				//AddDebugLogLineM ( false, logRemoteClient, wxString::Format ( wxT ( "eMule packet : unknown opcode: %i %x" ), opcode, opcode ) );

			break;
	}

	return true;
}


void CClientTCPSocket::OnConnect ( int nErrorCode )
{
	AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s --> %s): Client: %x : Processing connection event" ) )
	                   % ( long ) this
	                   % GetLocal().humanReadable()
	                   % GetPeer().humanReadable()
	                   % ( long ) m_client );

	if ( nErrorCode )
	{
		AddDebugLogLineM ( false, logRemoteClient, CFormat ( wxT ( "%x (%s --> %s): Client: %x : Processing connection event : error" ) )
		                   % ( long ) this
		                   % GetLocal().humanReadable()
		                   % GetPeer().humanReadable()
		                   % ( long ) m_client );
		OnError ( nErrorCode );
	}
	else if ( !m_client )
	{
		// and now? Disconnect? not?
		AddDebugLogLineM ( false, logLocalClient, CFormat ( wxT ( "%x (%s --> %s): Client: %x : Couldn't send hello packet (Client deleted!)" ) )
		                   % ( long ) this
		                   % GetLocal().humanReadable()
		                   % GetPeer().humanReadable()
		                   % ( long ) m_client );
	}
	else if ( !m_client->SendHelloPacket() )
	{
		// and now? Disconnect? not?
		AddDebugLogLineM ( false, logLocalClient, CFormat ( wxT ( "%x (%s --> %s): Client: %x : Couldn't send hello packet (Client deleted by SendHelloPacket!)" ) )
		                   % ( long ) this
		                   % GetLocal().humanReadable()
		                   % GetPeer().humanReadable()
		                   % ( long ) m_client );
	}
	else
	{
		AddDebugLogLineM ( false, logLocalClient, CFormat ( wxT ( "%x (%s --> %s): Client: %x : Hello packet sent" ) )
		                   % ( long ) this
		                   % GetLocal().humanReadable()
		                   % GetPeer().humanReadable()
		                   % ( long ) m_client );
		ResetTimeOutTimer();
	}
}


void CClientTCPSocket::OnSend ( int nErrorCode )
{
	ResetTimeOutTimer();
	CEMSocket::OnSend ( nErrorCode );
}


void CClientTCPSocket::OnReceive ( int nErrorCode )
{
	ResetTimeOutTimer();
	CEMSocket::OnReceive ( nErrorCode );
}


void CClientTCPSocket::OnError ( int nErrorCode )
{
	// 0.42e + Kry changes for handling of socket lost events
	wxString strError;

	if ( ( nErrorCode == 0 ) || ( nErrorCode == 7 ) || ( nErrorCode == 0xFEFF ) )
	{
		if ( m_client )
		{
			if ( !m_client->GetUserName().IsEmpty() )
			{
				strError = CFormat ( wxT ( "Client '%s'" ) ) % m_client->GetUserName() ;
			}
			else
			{
				strError = CFormat ( wxT ( "An unnamed client" ) );
			}

			strError += CFormat ( wxT ( " (hashDest: %x)" ) ) % m_client->GetTCPDest().hashCode();
		}
		else
		{
			strError = CFormat ( wxT ( "A client " ) );
		}

		if ( nErrorCode == 0 )
		{
			strError += CFormat ( wxT ( "closed connection." ) );
		}
		else if ( nErrorCode == 0xFEFF )
		{
			strError += CFormat ( wxT ( " caused a wxSOCKET_LOST event." ) );
		}
		else
		{
			strError += CFormat ( wxT ( "caused a socket blocking error." ) );
		}
	}
	else
	{
		if ( CLogger::IsEnabled ( logClient ) && ( nErrorCode != 107 ) )
		{
			// 0    -> No Error / Disconect
			// 107  -> Transport endpoint is not connected

			if ( m_client )
			{
				if ( m_client->GetUserName() )
				{
					strError = CFormat ( wxT ( "%x : OnError: Client '%s' (IP:%x) caused an error: %u. Disconnecting client!\n" ) )
						% ( long ) this 
						% m_client->GetUserName()
						%  m_client->GetTCPDest().hashCode()
						% nErrorCode;
				}
				else
				{
					strError = CFormat ( wxT ( "%x : OnError: Unknown client (IP:%x) caused an error: %u. Disconnecting client!" ) ) % ( long ) this % m_client->GetTCPDest().hashCode() % nErrorCode;
				}
			}
			else
			{
				strError = CFormat ( wxT ( "%x : OnError: A client caused an error or did something bad (error %u). Disconnecting client !" ) ) % ( long ) this % nErrorCode;
			}

		}
		else
		{
			strError = CFormat ( wxT ( "%x : Error 107 (Transport endpoint is not connected)" ) ) % ( long ) this;
		}
	}

	Disconnect ( strError );
}


bool CClientTCPSocket::PacketReceived ( CPacket* packet )
{
#ifdef __PACKET_RECV_DUMP__

	if ( CLogger::IsEnabled ( logRemoteClient ) )
	{

		printf ( "%lx (%s <-- %s): Recv: OPCODE %x SIZE %d\n", ( unsigned long ) this, ( const char* ) unicode2char ( GetLocal().humanReadable().c_str() ), ( const char* ) unicode2char ( GetPeer().humanReadable().c_str() ), ( unsigned ) packet->GetOpCode(), packet->GetPacketSize() );

		DumpMem ( packet->GetPacket(), packet->GetPacketSize() + sizeof ( Header_Struct ) );
	}

#endif
	AddDebugLogLineM ( false, logRemoteClient,
	                   CFormat ( wxT ( "%x : Packet (%s, %s), size %u, received from %s" ) )
	                   % ( long ) this
	                   % opstr(packet->GetProtocol())
	                   % opstr(packet->GetOpCode())
	                   % packet->GetPacketSize()
	                   % ( m_client ? m_client->GetTCPDest().humanReadable() : wxString ( wxT ( "Unknown Client" ) ) )
	                 );

	// 0.42e
	bool bResult = false;

	uint32_t uRawSize = packet->GetPacketSize();


	wxString exceptionString;

	try
	{
		switch ( packet->GetProtocol() )
		{
			case OP_EDONKEYPROT:
				bResult = ProcessPacket ( packet );

				break;

			case OP_PACKEDPROT:
				if ( !packet->UnPackPacket() )
				{
					AddDebugLogLineM ( false, logZLib, CFormat ( wxT ( "Failed to decompress client TCP packet." ) ) );
					bResult = false;
					break;
				}
				else
				{
				    AddDebugLogLineM ( false, logRemoteClient,
					    CFormat ( wxT ( "%x : Packet unpacked (%s, %s), size %u" ) )
						    % ( long ) this
						    % opstr(packet->GetProtocol())
						    % opstr(packet->GetOpCode())
						    % packet->GetPacketSize()
						     );
				}

			case OP_EMULEPROT:
				bResult = ProcessExtPacket ( packet );//->GetDataBuffer(), packet->GetPacketSize(), packet->GetOpCode());

				break;

			default:
			{
				theStats::AddDownOverheadOther ( uRawSize );

				if ( m_client )
				{
					m_client->SetDownloadState ( DS_ERROR );
				}

				Disconnect ( wxT ( "Unknown protocol" ) );

				bResult = false;
			}
		}
	}
	catch ( const CEOFException& err )
	{
		exceptionString = CFormat ( wxT ( "EOF exception: " ) ) + err.what();
	}
	catch ( const CInvalidPacket& err )
	{
		exceptionString = CFormat ( wxT ( "InvalidPacket exception: " ) ) + err.what();
	}
	catch ( const wxString& error )
	{
		exceptionString = CFormat ( wxT ( "error: " ) ) + ( error.IsEmpty() ? CFormat ( wxT ( "Unknown error" ) ) : error );
	}

	if ( !exceptionString.IsEmpty() )
	{
		AddDebugLogLineM ( false, logPacketErrors,
		                   CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x : " ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client +
		                   CFormat ( wxT ( "Caught %s\n"
		                                   "On packet (%s, %s), size %u"
		                                   "\tClientData: %s\n" ) )
		                   % exceptionString
		                   % opstr(packet->GetProtocol())
		                   % opstr(packet->GetOpCode())
		                   % packet->GetPacketSize()
		                   % ( m_client ? m_client->GetClientFullInfo() : CFormat ( wxT ( "Unknown" ) ) )
		                 );

		if ( m_client )
		{
			m_client->SetDownloadState ( DS_ERROR );
		}

		AddDebugLogLineM ( false, logRemoteClient,

		                   CFormat ( wxT ( "%x (%s <-- %s): Remote client: %x : " ) ) % ( long ) this % GetLocal().humanReadable() % GetPeer().humanReadable() % ( long ) m_client + CFormat ( wxT ( "Client '%s' (Dest: %s) caused an error (%s). Disconnecting client!" ) )
		                   % ( m_client ? m_client->GetUserName() : wxString ( wxT ( "Unknown" ) ) )
		                   % ( m_client ? m_client->GetTCPDest().humanReadable() : wxString ( wxT ( "Unknown" ) ) )
		                   % exceptionString
		                 );

		Disconnect ( wxT ( "Caught exception on CClientTCPSocket::ProcessPacket\n" ) );
	}

	return bResult;
}


bool CClientTCPSocket::IsMessageFiltered ( const wxString& Message, CUpDownClient* client )
{

	bool filtered = false;
	// If we're chatting to the guy, we don't want to filter!

	if ( client->GetChatState() != MS_CHATTING )
	{
		if ( thePrefs::MsgOnlyFriends() && !client->IsFriend() )
		{
			filtered = true;
		}
		else if ( thePrefs::MsgOnlySecure() && client->GetUserName().IsEmpty() )
		{
			filtered = true;
		}
		else if ( thePrefs::MustFilterMessages() )
		{
			filtered = thePrefs::IsMessageFiltered ( Message );
		}
	}

	return filtered;
}

SocketSentBytes CClientTCPSocket::SendControlData ( uint32_t maxNumberOfBytesToSend, uint32_t overchargeMaxBytesToSend )

{
	SocketSentBytes returnStatus = CEMSocket::SendControlData ( maxNumberOfBytesToSend, overchargeMaxBytesToSend );

	if ( returnStatus.success && ( returnStatus.sentBytesControlPackets > 0 || returnStatus.sentBytesStandardPackets > 0 ) )
	{
		ResetTimeOutTimer();
	}

	return returnStatus;
}


SocketSentBytes CClientTCPSocket::SendFileAndControlData ( uint32_t maxNumberOfBytesToSend, uint32_t overchargeMaxBytesToSend )
{
	SocketSentBytes returnStatus = CEMSocket::SendFileAndControlData ( maxNumberOfBytesToSend, overchargeMaxBytesToSend );

	if ( returnStatus.success && ( returnStatus.sentBytesControlPackets > 0 || returnStatus.sentBytesStandardPackets > 0 ) )
	{
		ResetTimeOutTimer();
	}

	return returnStatus;
}


void CClientTCPSocket::SendPacket ( CPacket* packet, bool delpacket, bool controlpacket, uint32_t actualPayloadSize )
{
	AddDebugLogLineM ( false, logLocalClient,
	                   CFormat ( wxT ( "%x : Packet (%s, %s), size %u queud to %s" ) )
	                   % ( long ) this
	                   % opstr(packet->GetProtocol())
	                   % opstr(packet->GetOpCode())
	                   % packet->GetPacketSize()
	                   % ( m_client ? m_client->GetTCPDest().humanReadable() : wxT ( "Unknown Client" ) )
	                 );

#ifdef __TCPPACKET_SEND_DUMP__

	if ( CLogger::isEnabled ( logLocalClient ) )
	{
		DumpMem ( packet->GetPacket(), packet->GetPacketSize() + sizeof ( Header_Struct ) );
	}

#endif
	ResetTimeOutTimer();

	CEMSocket::SendPacket ( packet, delpacket, controlpacket, actualPayloadSize );
}

BEGIN_EVENT_TABLE ( CClientTCPSocketHandler, wxEvtHandler )
        //EVT_SOCKET ( CLIENTTCPSOCKET_HANDLER, CClientTCPSocketHandler::ClientTCPSocketHandler )
        EVT_I2PSOCKET ( CLIENTTCPSOCKET_HANDLER, CClientTCPSocketHandler::ClientTCPSocketHandler )
END_EVENT_TABLE()
