//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef EMSOCKET_H
#define EMSOCKET_H

//#include "Proxy.h"				// Needed for CSocketClientProxy
#include "i2p/wxI2PSocketClient.h"
#include "Types.h"				// Needed for uint8_t and uint32_t
#include "ThrottledSocket.h"	// Needed for ThrottledFileSocket

#include <list>


class CPacket;


#define ERR_WRONGHEADER		0x01
#define ERR_TOOBIG			0x02

#define	ES_DISCONNECTED		0xFF
#define	ES_NOTCONNECTED		0x00
#define	ES_CONNECTED		0x01


const int32_t PACKET_HEADER_SIZE	= 6;


class CEMSocket : public wxI2PSocketClient, public ThrottledFileSocket
{
//	DECLARE_DYNAMIC_CLASS(CEMSocket)
	
public:
	CEMSocket();
	~CEMSocket();
	
	virtual void 	SendPacket(CPacket* packet, bool delpacket = true, bool controlpacket = true, uint32_t actualPayloadSize = 0);
    bool    HasQueues();
	bool	IsConnected() { return byConnected==ES_CONNECTED;};
	uint8_t	GetConState()	{return byConnected;}
	void	SetDownloadLimit(uint32_t limit);
	void	DisableDownloadLimit();

	virtual uint32_t	GetTimeOut() const;
	virtual void	SetTimeOut(uint32_t uTimeOut);
	
    uint32_t	GetLastCalledSend() { return lastCalledSend; }

    wxULongLong	GetSentBytesCompleteFileSinceLastCallAndReset();
    wxULongLong	GetSentBytesPartFileSinceLastCallAndReset();
    wxULongLong	GetSentBytesControlPacketSinceLastCallAndReset();
    wxULongLong	GetSentPayloadSinceLastCallAndReset();
    void	TruncateQueues();

    virtual SocketSentBytes SendControlData(uint32_t maxNumberOfBytesToSend, uint32_t minFragSize) { return Send(maxNumberOfBytesToSend, minFragSize, true); };
    virtual SocketSentBytes SendFileAndControlData(uint32_t maxNumberOfBytesToSend, uint32_t minFragSize) { return Send(maxNumberOfBytesToSend, minFragSize, false); };

    uint32_t	GetNeededBytes();
	
	bool Destroy();
	bool OnDestroy() { return DoingDestroy; };
	
	//protected:
	// these functions are public on our code because of the imuleDlg::socketHandler
	virtual void	OnError(int WXUNUSED(nErrorCode)) { };
	virtual void	OnSend(int nErrorCode);	
	virtual void	OnReceive(int nErrorCode);
	
protected:
	bool RecievePending() { return pendingOnReceive; }
	virtual bool	PacketReceived(CPacket* WXUNUSED(packet)) { return false; };
	virtual void	OnClose(int nErrorCode);
	
	uint8_t	byConnected;
	uint32_t	m_uTimeOut;

private:
    virtual SocketSentBytes Send(uint32_t maxNumberOfBytesToSend, uint32_t minFragSize, bool onlyAllowedToSendControlPacket);
	void	ClearQueues();	

    uint32_t	GetNextFragSize(uint32_t current, uint32_t minFragSize);
    bool    HasSent() { return m_hasSent; }
	
	// Download (pseudo) rate control	
	uint32_t	downloadLimit;
	bool	downloadLimitEnable;
	bool	pendingOnReceive;

	// Download partial header
	// actually, this holds only 'PACKET_HEADER_SIZE-1' bytes.
	char	pendingHeader[PACKET_HEADER_SIZE];
	uint32_t	pendingHeaderSize;

	// Download partial packet
	CPacket* pendingPacket;
	uint32_t  pendingPacketSize;

	// Upload control
	char*	sendbuffer;
	uint32_t	sendblen;
	uint32_t	sent;

	typedef std::list<CPacket*> CPacketQueue;
	CPacketQueue m_control_queue;

	struct StandardPacketQueueEntry
	{
		uint32_t actualPayloadSize;
		CPacket* packet;
	};

	typedef	std::list<StandardPacketQueueEntry> CStdPacketQueue;
	CStdPacketQueue m_standard_queue;
	
    bool m_currentPacket_is_controlpacket;

	wxMutex	m_sendLocker;

	wxULongLong m_numberOfSentBytesCompleteFile;
    wxULongLong m_numberOfSentBytesPartFile;
    wxULongLong m_numberOfSentBytesControlPacket;
    bool m_currentPackageIsFromPartFile;

	bool	m_bAccelerateUpload;
	uint32_t	lastCalledSend;
    uint32_t	lastSent;
	uint32_t	lastFinishedStandard;

    uint32_t m_actualPayloadSize;
    uint32_t m_actualPayloadSizeSent;

    bool m_bBusy;
    bool m_hasSent;

	bool DoingDestroy;
};


#endif // EMSOCKET_H
