//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Tiku
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include <wx/defs.h>

#ifdef __WXMSW__
	#include <wx/msw/winundef.h>
#endif

#include <wx/intl.h>
#include <wx/wfstream.h>
#include <wx/filesys.h>
#include <wx/url.h>
#include <wx/app.h>

#include <cmath>

#include "HTTPDownload.h"				// Interface declarations
#include <common/StringFunctions.h>		// Needed for unicode2char
#include "OtherFunctions.h"				// Needed for CastChild
#include "Logger.h"						// Needed for AddLogLine*
#include <common/Format.h>				// Needed for CFormat
#include "InternalEvents.h"				// Needed for CMuleInternalEvent
#include "Proxy.h"						// Needed for CProxyData
#include "Preferences.h"


#ifndef IMULE_DAEMON
#include "inetdownload.h"	// Needed for inetDownload
#include "muuli_wdr.h"		// Needed for ID_CANCEL: Let it here or will fail on win32
#include "MuleGifCtrl.h"
#include <wx/gauge.h>

#ifdef __WXMSW__
typedef wxGauge95 wxGaugeControl;
#else
typedef wxGauge wxGaugeControl;
#endif


DECLARE_LOCAL_EVENT_TYPE(wxEVT_HTTP_PROGRESS, wxANY_ID)
DECLARE_LOCAL_EVENT_TYPE(wxEVT_HTTP_SHUTDOWN, wxANY_ID)


class CHTTPDownloadDialog : public wxDialog
{
public:
	wxMutex mutex;
	wxCondition condition;
	
	bool m_threadRunning;
	bool m_stopThread;
	
	CHTTPDownloadDialog(CHTTPDownloadThread* thread)
	  : wxDialog(wxTheApp->GetTopWindow(), -1, _("Downloading..."),
			wxDefaultPosition, wxDefaultSize, wxDEFAULT_DIALOG_STYLE | wxSYSTEM_MENU),
			    condition(mutex)
   	{
		downloadDlg(this, true)->Show(this, true);
	
		m_progressbar = CastChild(ID_HTTPDOWNLOADPROGRESS, wxGaugeControl);
		m_progressbar->SetRange(100);

		m_ani = CastChild(ID_ANIMATE, MuleGifCtrl);
		m_ani->LoadData((const char*)inetDownload, sizeof(inetDownload));
		m_ani->Start();
	
		m_thread = thread;
		m_stopThread = false;
	}

	~CHTTPDownloadDialog() {
	}

	void UpdateGauge(int total, int current) {
		CFormat label(_("( %s / %s )"));
		
		label % CastItoXBytes(current);
		if (total > 0) {
			label % CastItoXBytes(total);
		} else {
			label % _("Unknown");
		}
	
		CastChild(IDC_DOWNLOADSIZE, wxStaticText)->SetLabel(label.GetString());
	
		if (total and (total != m_progressbar->GetRange())) {
			m_progressbar->SetRange(total);
		}
	
		if (current && (current <= total)) {
			m_progressbar->SetValue(current);
		}
	
		Layout();
	}

private:
	DECLARE_EVENT_TABLE();

	void OnBtnCancel(wxCommandEvent& WXUNUSED(evt)) {
		m_stopThread = true ;
		printf("HTTP download cancelled\n");
		Show(false);
		mutex.Lock();
		while (m_threadRunning) condition.Wait();
		mutex.Unlock();
	}
	
	void OnProgress(CMuleInternalEvent& evt) {
		UpdateGauge(evt.GetExtraLong(), evt.GetInt());
	}

	void OnShutdown(CMuleInternalEvent& WXUNUSED(evt)) {
		Show(false);
		Destroy();
	}
  
	wxThread*	m_thread;
	MuleGifCtrl* 	m_ani;
	wxGaugeControl* m_progressbar;
};


BEGIN_EVENT_TABLE(CHTTPDownloadDialog, wxDialog)
	EVT_BUTTON(ID_HTTPCANCEL, CHTTPDownloadDialog::OnBtnCancel)
	EVT_MULE_INTERNAL(wxEVT_HTTP_PROGRESS, -1, CHTTPDownloadDialog::OnProgress)
	EVT_MULE_INTERNAL(wxEVT_HTTP_SHUTDOWN, -1, CHTTPDownloadDialog::OnShutdown)
END_EVENT_TABLE()

DEFINE_LOCAL_EVENT_TYPE(wxEVT_HTTP_PROGRESS)
DEFINE_LOCAL_EVENT_TYPE(wxEVT_HTTP_SHUTDOWN)
	
#endif


CHTTPDownloadThread::CHTTPDownloadThread(const wxString& url, const wxString& filename, HTTP_Download_File file_id, bool showDialog, uint32_t delay)
//#ifdef IMULE_DAEMON
	: wxThread(wxTHREAD_DETACHED),
//#else
//	: wxThread(showDialog ? wxTHREAD_JOINABLE : wxTHREAD_DETACHED),
//#endif
	  m_url(url),
	  m_tempfile(filename),
	  m_result(1),
	  m_file_id(file_id),
	  m_companion(NULL),
	m_delay(delay)
{
	if (showDialog) {
#ifndef IMULE_DAEMON
		CHTTPDownloadDialog* dialog = new CHTTPDownloadDialog(this);
		dialog->Show(true);
		m_companion = dialog;
#endif
	}
}


wxThread::ExitCode CHTTPDownloadThread::Entry()
{
	if (m_companion) {
		wxMutexLocker lock( ((CHTTPDownloadDialog*)m_companion)->mutex);
		((CHTTPDownloadDialog*)m_companion)->m_threadRunning = true;
		if (((CHTTPDownloadDialog*)m_companion)->m_stopThread) {
		    ((CHTTPDownloadDialog*)m_companion)->m_threadRunning = false;
			// Thread going down...
			m_result = -1;
			return NULL;
		}
	}

	if (TestDestroy())
	{
	    m_result = -1;
	    return NULL;
	}
	
	wxInputStream* url_read_stream = NULL;
	
	printf("HTTP download thread started\n");
	while ( m_delay-- ) wxThread::Sleep( 1000 ) ;
	//const CProxyData* proxy_data = thePrefs::GetProxyData();
	//bool use_proxy = proxy_data != NULL && proxy_data->m_proxyEnable;
	
	try {	
		wxFFileOutputStream outfile(m_tempfile);
			
		/*if (!use_proxy)
			throw wxString(wxT("This is ANONYMOUS imule :"))
				+ wxString(wxT("you have to set a http I2P proxy to access the Web.\n"))
				+ wxString(wxT("Go to preferences/Proxy and set the http proxy !\n"))
				+ wxString(wxT("the I2P Proxy usually is : IP=127.0.0.1, port=4444\n"));
		*/
		if (!outfile.Ok()) {
			throw wxString(CFormat(wxT("Unable to create destination file %s for download!\n")) % m_tempfile);
		}
			
		if ( m_url.IsEmpty() ) {
			// Nowhere to download from!
			throw wxString(wxT("The URL to download can't be empty\n"));
		}

		//wxString proxyAddr = proxy_data->m_proxyHostName;
		//proxyAddr << wxT(":") << proxy_data->m_proxyPort;
		
		wxString proxyAddr;
		wxURL urlhandler(m_url);
		
		proxyAddr << thePrefs::GetI2PServerIP() << wxT(":") << thePrefs::GetI2PProxyPort() ;
		
		if (urlhandler.IsOk())
		{
			urlhandler.SetProxy(proxyAddr);
			url_read_stream = urlhandler.GetInputStream();
		}
		if (!url_read_stream) {
			throw wxString(CFormat(wxT("The URL %s returned: Error (%i)!")) % m_url % urlhandler.GetError());
		}
		
		int download_size = url_read_stream->GetSize();
		printf("Download size: %i\n",download_size);
		
		const unsigned MAX_HTTP_READ = 4092;
		
		char buffer[MAX_HTTP_READ];
		int current_read = 0;
		int total_read = 0;
		do {
			url_read_stream->Read(buffer, MAX_HTTP_READ);
			current_read = url_read_stream->LastRead();
			if (current_read) {
				total_read += current_read;
				outfile.Write(buffer,current_read);
				int current_write = outfile.LastWrite();
				if (current_read != current_write) {
					throw wxString(wxT("Critical error while writing downloaded file"));
				} else if (m_companion) {
#ifndef IMULE_DAEMON
					CMuleInternalEvent evt(wxEVT_HTTP_PROGRESS);
					evt.SetInt(total_read);
					evt.SetExtraLong(download_size);
					wxPostEvent(m_companion, evt);
#endif
				}
			}
			if (m_companion) 
			    if (((CHTTPDownloadDialog*)m_companion)->m_stopThread) break ;
				
		} while (current_read && !TestDestroy());
		
	} catch (const wxString& error) {
		if (wxFileExists(m_tempfile)) {
			wxRemoveFile(m_tempfile);
		}

		m_result = -1;		
		AddLogLineM(false, error);
	}

	printf("HTTP download thread ended\n");
	
	if (m_companion) {
	    wxMutexLocker lock(((CHTTPDownloadDialog*)m_companion)->mutex);
	    ((CHTTPDownloadDialog*)m_companion)->m_threadRunning = false;
	    ((CHTTPDownloadDialog*)m_companion)->condition.Broadcast();
	}
//	CI2PClient::DetachThread();
	return 0;
}


void CHTTPDownloadThread::OnExit() 
{
#ifndef IMULE_DAEMON
	if (m_companion) {
		CMuleInternalEvent termEvent(wxEVT_HTTP_SHUTDOWN);
		wxPostEvent(m_companion, termEvent);	
	}
#endif
	
	// Notice the app that the file finished download
	CMuleInternalEvent evt(wxEVT_CORE_FINISHED_HTTP_DOWNLOAD);
	evt.SetInt((int)m_file_id);
	evt.SetExtraLong((long)m_result);
	wxPostEvent(wxTheApp, evt);
}
