//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org / http://www.imule.i2p )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#if defined(__GNUG__) && !defined(NO_GCC_PRAGMA)
    #pragma implementation "KadDlg.cpp"
#endif

// For compilers that support precompilation, includes "wx/wx.h".
#include "wx/wxprec.h"

#ifdef __BORLANDC__
    #pragma hdrstop
#endif

#include "KadDlg.h"

#include <wx/msgdlg.h>
#include <wx/filedlg.h>
//#include <wx/listctrl.h>
#include "kadNodesListView.h"


#include "common/Format.h"
#include "MD4Hash.h"
#include "OScopeCtrl.h"
#include "OtherFunctions.h"
#include "HTTPDownload.h"
#include "Logger.h"
#include "imule.h"
#include "Preferences.h"
#include "Statistics.h"
#include "StatisticsDlg.h"
#include "ColorFrameCtrl.h"
#include "MuleTextCtrl.h" // Needed for CMuleTextCtrl

#include "kademlia/kademlia/Kademlia.h"
#include "kademlia/routing/RoutingZone.h"

#ifndef CLIENT_GUI
#include "NetworkFunctions.h"
#endif

#include <utility>

BEGIN_EVENT_TABLE(CKadDlg, wxPanel)
    /*EVT_TEXT(ID_NODE_IP1, CKadDlg::OnFieldsChange)
    EVT_TEXT(ID_NODE_IP2, CKadDlg::OnFieldsChange)
    EVT_TEXT(ID_NODE_IP3, CKadDlg::OnFieldsChange)  
    EVT_TEXT(ID_NODE_IP4, CKadDlg::OnFieldsChange)
    EVT_TEXT(ID_NODE_PORT, CKadDlg::OnFieldsChange)
    */
    EVT_TEXT(ID_NODE_DEST, CKadDlg::OnFieldsChange)
        
    EVT_TEXT_ENTER(IDC_NODESLISTURL ,CKadDlg::OnBnClickedUpdateNodeList)
    
    EVT_BUTTON(ID_MYDESTTOCLIPBOARD, CKadDlg::OnBnClickedCopyMyDestToClipboard)
    EVT_BUTTON(ID_NODECONNECT, CKadDlg::OnBnClickedBootstrapClient)
    EVT_BUTTON(ID_UPDATEKADLIST, CKadDlg::OnBnClickedUpdateNodeList)
    EVT_BUTTON(ID_NODEEXPORT, CKadDlg::OnBnClickedNodeExport)
END_EVENT_TABLE();



CKadDlg::CKadDlg ( wxWindow* pParent, const int & id, const wxPoint, wxSize size, int opt )
        : wxPanel ( pParent, id, wxDefaultPosition, wxDefaultSize, opt, wxT ( "kadwnd" ) )
{
    m_kad_scope = NULL;
    m_nodes_list = NULL;
    if ( KadDlg ( this ) )
    {
        m_nodes_box  = ( wxStaticBoxSizer * ) kadDlgNodesGraphFrame ;
        m_nodes_list = GetListctrl();
        m_kad_scope  = GetKadscope();
    }
}







void CKadDlg::Init()
{
    m_kad_scope->SetRanges(0.0, thePrefs::GetStatsMax());
    m_kad_scope->SetYUnits(wxT("Nodes"));

    SetUpdatePeriod(thePrefs::GetTrafficOMeterInterval());
    SetGraphColors();
}


void CKadDlg::SetUpdatePeriod(int step)
{
    // this gets called after the value in Preferences/Statistics/Update delay has been changed
    if (step == 0) {
        m_kad_scope->Stop();
    } else {
        m_kad_scope->Reset(step);
    }
}


void CKadDlg::SetGraphColors()
{
    static const char aTrend[] = { 2,      1,        0        };
    static const int aRes[]    = { IDC_C0, IDC_C0_3, IDC_C0_2 };
    
    m_kad_scope->SetBackgroundColor(CStatisticsDlg::getColors(0));
    m_kad_scope->SetGridColor(CStatisticsDlg::getColors(1));
    
    for (size_t i = 0; i < 3; ++i) {    
        m_kad_scope->SetPlotColor(CStatisticsDlg::getColors(12 + i), aTrend[i]);
        
        CColorFrameCtrl* ctrl = CastChild(aRes[i], CColorFrameCtrl);
        ctrl->SetBackgroundColor(CStatisticsDlg::getColors(12 + i));
        ctrl->SetFrameColor(*wxBLACK);
    }
}


void CKadDlg::UpdateGraph(bool bStatsVisible, const GraphUpdateInfo& update)
{   
    const float* apfKad[] = { &update.kadnodes[0], &update.kadnodes[1], &update.kadnodes[2] };
    unsigned nodeCount = static_cast<unsigned>(update.kadnodes[2]);
    
    if (!bStatsVisible) {
        m_kad_scope->DelayPoints();
    } else {
        // Check the current node-count to see if we should increase the graph height
        if (m_kad_scope->GetUpperLimit() < update.kadnodes[2]) {
            // Grow the limit by 10 sized increments.
            m_kad_scope->SetRanges(0.0, (float)(((nodeCount + 9) / 10) * 10));
        }

        m_kad_scope->AppendPoints(update.timestamp, apfKad);
    }

    wxStaticText* label = CastChild( wxT("nodesListLabel"), wxStaticText );
    wxCHECK_RET(label, wxT("Failed to find kad-nodes label"));

    label->SetLabel(wxString::Format(_("Nodes (%u)"), nodeCount));
    label->GetParent()->Layout();
}


// Enables or disables the node connect button depending on the conents of the text fields
void CKadDlg::OnFieldsChange(wxCommandEvent& WXUNUSED(evt))
{
    // Enable the node connect button if all fields contain text
    wxString txtaddr = ((wxTextCtrl*)FindWindowById( ID_NODE_DEST ))->GetValue() ;
    CI2PAddress dest = CI2PAddress::fromString(txtaddr);
    if (dest.isValid())
        FindWindowById(ID_NODECONNECT)->Enable( true );
    else
        FindWindowById(ID_NODECONNECT)->Enable( false );
}


void CKadDlg::OnBnClickedCopyMyDestToClipboard(wxCommandEvent& WXUNUSED(evt))
{
#ifndef CLIENT_GUI
    if (FindWindowById(ID_MYDESTTOCLIPBOARD)->IsEnabled())
#endif
    {
    theApp.CopyTextToClipboard(theApp.GetUdpDest().toString());
    }
}

    
void CKadDlg::OnBnClickedBootstrapClient(wxCommandEvent& WXUNUSED(evt))
{
    if (FindWindowById(ID_NODECONNECT)->IsEnabled()) {
        //TODO #warning TODO EC
        #ifndef CLIENT_GUI
        // Ip is reversed since StringIPtoUint32 returns anti-host and kad expects host order
        /*uint32_t ip = StringIPtoUint32(
                    ((wxTextCtrl*)FindWindowById( ID_NODE_IP4 ))->GetValue() +
                    wxT(".") + 
                    ((wxTextCtrl*)FindWindowById( ID_NODE_IP3 ))->GetValue() +
                    wxT(".") + 
                    ((wxTextCtrl*)FindWindowById( ID_NODE_IP2 ))->GetValue() +
                    wxT(".") + 
                    ((wxTextCtrl*)FindWindowById( ID_NODE_IP1 ))->GetValue() );*/
        
        wxString txtaddr = ((wxTextCtrl*)FindWindowById( ID_NODE_DEST ))->GetValue() ;
        CI2PAddress dest = CI2PAddress::fromString(txtaddr);
        
        AddDebugLogLineM(false, logKadPrefs, CFormat(wxT("bootstrap on %s ")) % dest.humanReadable());
        
        //if (ip == 0) {
        if (!dest.isValid()) {
            wxMessageBox(_("Invalid ip to bootstrap"), _("Warning"), wxOK | wxICON_EXCLAMATION, this);
        } else {
            /*unsigned long port;
            if (((wxTextCtrl*)FindWindowById( ID_NODE_PORT ))->GetValue().ToULong(&port)) {
                if ( !Kademlia::CKademlia::isRunning() ) {
                    Kademlia::CKademlia::start();
                    theApp.ShowConnectionState();
                }
                Kademlia::CKademlia::bootstrap(ip, port);               
            } else {
                wxMessageBox(_("Invalid port to bootstrap"), _("Warning"), wxOK | wxICON_EXCLAMATION, this);
            }*/
            Kademlia::CKademlia::bootstrap(dest);
        }
        #else
            wxMessageBox(_("You can't bootstrap an specific ip from remote GUI yet."), _("Message"), wxOK | wxICON_INFORMATION, this);
        #endif      
    } else {
        wxMessageBox(_("Please fill all fields required"), _("Message"), wxOK | wxICON_INFORMATION, this);
    }
}

void CKadDlg::OnBnClickedNodeExport(wxCommandEvent& WXUNUSED(evt))
{
    if ( !Kademlia::CKademlia::isRunning() )
    {
        wxMessageDialog md(NULL,
            _("Kademlia is not running.\n No known node !"),
            _("Export nodes destinations"));
        md.ShowModal();
        return ;
    }
    wxString name = wxFileSelector(_("Select file to export to"));
    if (!name.empty())
    {
        Kademlia::CKademlia::getRoutingZone()->exportFile(name) ;
    }
}

void CKadDlg::OnBnClickedBootstrapKnown(wxCommandEvent& WXUNUSED(evt))
{
    theApp.StartKad();
}


void CKadDlg::OnBnClickedDisconnectKad(wxCommandEvent& WXUNUSED(evt))
{
    theApp.StopKad();
}


void CKadDlg::OnBnClickedUpdateNodeList(wxCommandEvent& WXUNUSED(evt))
{
    //TODO #warning TODO EC
    #ifndef CLIENT_GUI
    if ( wxMessageBox( wxString(_("Are you sure you want to download a new nodes.dat file?\n")) +
                        _("Doing so will remove your current nodes and restart Kademlia connection.")
                    , _("Continue?"), wxICON_EXCLAMATION | wxYES_NO, this) == wxYES ) {
        wxString strURL = ((wxTextCtrl*)FindWindowById( IDC_NODESLISTURL ))->GetValue();
        if (strURL.Find(wxT("://")) == -1) {
            AddLogLineM(true, _("Invalid URL"));
            return;
        }
        wxString strTempFilename(theApp.ConfigDir + wxT("nodes.dat.download"));
        CHTTPDownloadThread *downloader = new CHTTPDownloadThread(strURL,strTempFilename, HTTP_NodesDat);
        downloader->Create();
        downloader->Run();
    }
    #else
    wxMessageBox(_("You can't update server.met from remote GUI yet."), _("Message"), wxOK | wxICON_INFORMATION, this);
    #endif      
}


void CKadDlg::OnNodeEvent(const GUIEvent & evt)
{
    m_nodes_list->OnNodeEvent(evt);
    m_nodes_box->GetStaticBox()->SetLabel(wxString::Format(_("Nodes (%u)"), theStats::GetKadNodesCur()));
}

