//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef KNOWNFILE_H
#define KNOWNFILE_H

#include <set>

#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/dynarray.h>

#include "MD4Hash.h"
#include "SHAHashSet.h"
#include "CTypedPtrList.h"
#include "Types.h"		// Needed for uint8_t, uint8_t, uint16_t, uint32_t and wxULongLong
#include "OPCodes.h"		// Needed for PARTSIZE

#include "kademlia/kademlia/SearchManager.h"
#include "kademlia/kademlia/Indexed.h"

#ifdef CLIENT_GUI
#include "ec/ECSpecialTags.h"
#endif

#define	PS_READY			0
#define	PS_EMPTY			1
#define PS_WAITINGFORHASH		2
#define PS_HASHING			3
#define PS_ERROR			4
#define	PS_INSUFFICIENT			5
#define	PS_UNKNOWN			6
#define PS_PAUSED			7
#define PS_COMPLETING			8
#define PS_COMPLETE			9

#define PR_VERYLOW			4 // I Had to change this because it didn't save negative number correctly.. Had to modify the sort function for this change..
#define PR_LOW				0 //*
#define PR_NORMAL			1 // Don't change this - needed for edonkey clients and server!
#define	PR_HIGH				2 //*
#define PR_VERYHIGH			3
#define PR_AUTO				5
#define PR_POWERSHARE                   6 //added for powershare (deltaHF)

class CUpDownClient;
class CFileDataIO;
class CPacket;
class CTag;

namespace Kademlia{
	class CEntry;
};

WX_DECLARE_OBJARRAY(CMD4Hash, ArrayOfCMD4Hash);

//WX_DECLARE_OBJARRAY(CTag*, ArrayOfCTag);

class CFileStatistic {
	friend class CKnownFile;
	friend class CSharedFilesRem;
public:
	CFileStatistic();
	void	AddRequest();
	void	AddAccepted();
	void    AddTransferred(wxULongLong bytes);
	uint16_t	GetRequests() const			{return requested;}
	uint16_t	GetAccepts() const			{return accepted;}
	wxULongLong  GetTransfered() const			{return transfered;}
	uint32_t	GetAllTimeRequests() const	{return alltimerequested;}
	void	SetAllTimeRequests(uint32_t new_value) { alltimerequested = new_value; };
	uint32_t	GetAllTimeAccepts() const		{return alltimeaccepted;}
	void	SetAllTimeAccepts(uint32_t new_value) { alltimeaccepted = new_value; };	
	wxULongLong	GetAllTimeTransfered() const	{return alltimetransferred;}
	void	SetAllTimeTransfered(wxULongLong new_value) { alltimetransferred = new_value; };
	CKnownFile* fileParent;
	
private:
	uint16_t requested;
	wxULongLong transfered;
	uint16_t accepted;
	uint32_t alltimerequested;
	wxULongLong alltimetransferred;
	uint32_t alltimeaccepted;
};

/*
					   CPartFile
					 /
		  CKnownFile
		/
CAbstractFile 
		\ 
		  CSearchFile
*/

typedef std::list<Kademlia::CEntry> CKadEntriesList;

class CAbstractFile
{
public:
	CAbstractFile();
	virtual ~CAbstractFile() {};

	virtual const wxString&	GetFileName() const		{return m_strFileName;}
	const CMD4Hash&	GetFileHash() const	{return m_abyFileHash;}
	wxULongLong	GetFileSize() const			{return m_nFileSize;}
	virtual void 	SetFileSize(wxULongLong nFileSize) { m_nFileSize = nFileSize; }
	
	virtual void	SetFileName(const wxString& strmakeFilename);

	/* Tags and Notes handling */
	wxULongLong GetIntTagValue(uint8_t tagname) const;
	//uint32_t GetIntTagValue(const char* tagname) const;
	bool GetIntTagValue(uint8_t tagname, wxULongLong& ruValue) const;
	const wxString& GetStrTagValue(uint8_t tagname) const;
	//const wxString& GetStrTagValue(const char* tagname) const;
	//CTag* GetTag(const char* tagname) const;	
	//CTag* GetTag(const char* tagname, uint8_t tagtype) const;
	const CTag& GetTag(uint8_t tagname) const;
	const CTag& GetTag(uint8_t tagname, uint8_t tagtype) const;
	void AddTagUnique(const CTag& pTag);
	//const ArrayOfCTag& GetTags() const { return taglist; }
	const TagList& GetTags() const { return m_taglist; }
	bool AddNote(Kademlia::CEntry& pEntry);
	const CKadEntriesList& getNotes() const { return m_kadNotes; }

	/* Comment and rating */	
	virtual const wxString&	GetFileComment() const { return m_strComment; }
	virtual uint8_t	GetFileRating() 		const { return m_iRating; }	
	
	bool	HasComment() const		{ return m_hasComment; }
	bool	HasRating() const		{ return m_iUserRating; }
	bool	HasBadRating() const		{ return (m_iUserRating == 1); }
	uint8_t	UserRating() const 		{ return m_iUserRating; }
	void	UpdateFileRatingCommentAvail();

	uint32_t GetSourceCount() const;
	uint32_t GetCompleteSourceCount() const;

protected:
	wxString	m_strFileName;
	CMD4Hash	m_abyFileHash;
	wxULongLong	m_nFileSize;
	wxString	m_strComment;
	uint8_t		m_iRating;
	bool		m_hasComment;
	uint8_t		m_iUserRating;
	//ArrayOfCTag taglist;
	TagList         m_taglist;
	CKadEntriesList m_kadNotes;
};


class CKnownFile : public CAbstractFile
{
friend class CAddFileThread;
public:
	CKnownFile();
#ifdef CLIENT_GUI
	CKnownFile(CEC_SharedFile_Tag *);
	friend class CSharedFilesRem;
#endif

	virtual ~CKnownFile();

	#ifndef CLIENT_GUI	
	virtual void SetFileName(const wxString& strmakeFilename);
	#endif
		
	void SetFilePath(const wxString& strFilePath);
	const wxString& GetFilePath() const { return m_strFilePath; }
	
	virtual	bool	IsPartFile() const	{return false;}
	virtual bool	LoadFromFile(CFileDataIO* file);	//load date, hashset and tags from a .met file
	virtual uint8_t	GetStatus(bool WXUNUSED(ignorepause) = false) const { return PS_COMPLETE; }
	bool	WriteToFile(CFileDataIO* file);	
	uint32_t	GetFileDate() const	{return date;}

		
	virtual void SetFileSize(wxULongLong nFileSize);

	// local available part hashs
	size_t	GetHashCount() const	{return hashlist.GetCount();}
	const CMD4Hash&	GetPartHash(wxULongLong part) const;

	// nr. of part hashs according the file size wrt ED2K protocol
	inline size_t GetED2KPartHashCount() const { return m_iED2KPartHashCount; }

	// nr. of 9MB parts (file data)
	inline size_t GetPartCount() const { return m_iPartCount; }

	// nr. of 9MB parts according the file size wrt ED2K protocol (OP_FILESTATUS)
	inline size_t GetED2KPartCount() const { return m_iED2KPartCount; }
	
	// file upload priority
	uint8_t	GetUpPriority()	 const		{return m_iUpPriority;}
	void	SetUpPriority(uint8_t newUpPriority, bool bSave=true);
	bool	IsAutoUpPriority() const		{return m_bAutoUpPriority;}
	void	SetAutoUpPriority(bool flag)	{m_bAutoUpPriority = flag;}
	void	UpdateAutoUpPriority();
	size_t	GetQueuedCount() const {return m_ClientUploadList.size();}

	bool	LoadHashsetFromFile(CFileDataIO* file, bool checkhash);
	void	AddUploadingClient(CUpDownClient* client);
	void	RemoveUploadingClient(CUpDownClient* client);
	
	// comment 
	const wxString&	GetFileComment() { if (!m_bCommentLoaded) LoadComment(); return m_strComment; } 
	uint8_t	GetFileRating() 		{ if (!m_bCommentLoaded) LoadComment(); return m_iRating; }

	void	SetFileComment(const wxString& strNewComment);
	void	SetFileRating(uint8_t iNewRating); 
	void	SetPublishedED2K( bool val );
	bool	GetPublishedED2K() const	{return m_PublishedED2K;}

	/* Kad stuff */ 
	uint32_t	GetKadFileSearchID() const { return kadFileSearchID; }
	// KAD TODO: This must be used on KadSearchListCtrl too once imported
	void	SetKadFileSearchID(uint32_t id) { kadFileSearchID = id; } // John - Don't use this unless you know what your are DOING!! (Hopefully I do.. :)
	const Kademlia::WordList& GetKadKeywords() const { return wordlist; }
	// KAD TODO: If we add the proper column to SharedFilesCtrl, this is the funtion.
	uint32_t	GetLastPublishTimeKadSrc() const { return m_lastPublishTimeKadSrc; }	
	void	SetLastPublishTimeKadSrc(uint32_t time) { m_lastPublishTimeKadSrc = time; }
	uint32_t GetNextPublishTimeKadSrc() {return GetLastPublishTimeKadSrc() + KADEMLIAREPUBLISHTIMES;}
		
	
	void	SetLastPublishTimeKadNotes(uint32_t time) {m_lastPublishTimeKadNotes = time;}
	uint32_t	GetLastPublishTimeKadNotes() const { return m_lastPublishTimeKadNotes; }	
	
	bool	PublishSrc();
	bool	PublishNotes();	
	
	// TODO: This must be implemented if we ever want to have metadata.
	uint32_t	GetMetaDataVer() const { return /*m_uMetaDataVer*/ 0; }
	
	// file sharing
	virtual	CPacket*	CreateSrcInfoPacket(const CUpDownClient* forClient);
	
	virtual void	UpdatePartsInfo();	

	uint32_t	date;
	
	CFileStatistic statistic;
	
	time_t m_nCompleteSourcesTime;
	uint16_t m_nCompleteSourcesCount;
	uint16_t m_nCompleteSourcesCountLo;
	uint16_t m_nCompleteSourcesCountHi;
	
	// Maybe find a common place for this typedef?
	typedef std::set<CUpDownClient*> SourceSet;
	SourceSet m_ClientUploadList;
	ArrayOfUInts16 m_AvailPartFrequency;
	
	bool	CreateAICHHashSetOnly();
	// aich
	CAICHHashSet*	GetAICHHashset() const							{return m_pAICHHashSet;}
	void			SetAICHHashset(CAICHHashSet* val)				{m_pAICHHashSet = val;}		

	/**
	 * Updates the requency of uploading parts from with the data the client provides.
	 *
	 * @param client The clients whoose uploading parts should be considered.
	 * @param increment If true, the counts are incremented, otherwise they are decremented.
	 *
	 * This functions updates the frequency list of file-upparts, using the clients 
	 * upparts-status. This function should be called by clients every time they update their
	 * upparts-status, or when they are added or removed from the file.
	 */
	void UpdateUpPartsFrequency( CUpDownClient* client, bool increment );

	void	CreateHashFromString(uint8_t* in_string, uint32_t Length, uint8_t* Output, CAICHHashTree* pShaHashOut = NULL)	{CreateHashFromInput(NULL, Length,Output,in_string,pShaHashOut);}
	
	void	ClearPriority();
	
protected:
	bool	LoadTagsFromFile(CFileDataIO* file);
	bool	LoadDateFromFile(CFileDataIO* file);
	void	CreateHashFromFile(CFileDataIO* file, uint32_t Length, uint8_t* Output, CAICHHashTree* pShaHashOut = NULL) const 
		{ 
			CreateHashFromInput(file, Length, Output, NULL, pShaHashOut); 
		}	
	void	LoadComment();//comment
	ArrayOfCMD4Hash hashlist;
	wxString m_strFilePath;	
	CAICHHashSet*			m_pAICHHashSet;

	void	CreateHashFromInput(CFileDataIO* file, uint32_t Length, uint8_t* Output, uint8_t* in_string, CAICHHashTree* pShaHashOut) const;
	bool	m_bCommentLoaded;
	size_t	m_iPartCount;
	size_t	m_iED2KPartCount;
	size_t	m_iED2KPartHashCount;
	uint8_t	m_iUpPriority;
	bool	m_bAutoUpPriority;
	bool	m_PublishedED2K;
	
	/* Kad stuff */
	Kademlia::WordList wordlist;
	uint32_t	kadFileSearchID;
	uint32_t	m_lastPublishTimeKadSrc;
	uint32_t	m_lastPublishTimeKadNotes;
	
private:
	/** Common initializations for constructors. */
	void Init();
};

#endif // KNOWNFILE_H
