//
// This file is part of the imule Project.
//
// Copyright (C) 2005-2006imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//


#include "MuleUDPSocket.h"	// Interface declarations
#include "i2p/wxI2PDatagramSocket.h"
#include "i2p/CSamDefines.h"
#include "Logger.h"			// Needed for AddDebugLogLineM
#include "imule.h"			// Needed for theApp
#include "GetTickCount.h"	// Needed for GetTickCount()
#include "OPCodes.h"		// Needed for SEC2MS
#include "Packet.h"			// Needed for CPacket
#include <common/StringFunctions.h>// Needed for unicode2char
#include "Logger.h"			// Needed for AddDebugLogLineM
#include "common/Format.h"
#include "UploadBandwidthThrottler.h"
#include <wx/apptrait.h>


const unsigned UDP_BUFFER_SIZE = SAM_DGRAM_PAYLOAD_MAX;


static uint16_t packetHash ( const char * buf, int size )
{
	uint16_t h = 0;

	while ( size-- > 0 )
		h = (uint16_t) (h + * ( buf++ ));

	return h ;
}


//CMuleUDPSocket::CMuleUDPSocket(const wxString& name, int id, const & address, const CProxyData* ProxyData)
/*CMuleUDPSocket::CMuleUDPSocket ( const wxString name, int id, const wxIPV4address ip_port,
                                 wxString key,        const CProxyData* ProxyData )*/
CMuleUDPSocket::CMuleUDPSocket ( const wxString name, int id,
                                 wxString key,        const CProxyData* ProxyData )
		: m_busy ( false ),
		m_name ( name ),
		m_id ( id ),
		m_privKey ( key ),
		//m_addr( ip_port ),
		m_proxy ( ProxyData ),
		m_socket ( NULL )
{}



CMuleUDPSocket::~CMuleUDPSocket()

{
    if ( m_socket )
    {
	    if (theApp.uploadBandwidthThrottler)
	        theApp.uploadBandwidthThrottler->RemoveFromAllQueues ( this );

	    DestroySocket();
    }
}


void CMuleUDPSocket::CreateSocket()
{
	if (! m_socket)
	{
        m_socket = ::new wxI2PDatagramSocket( m_privKey, wxSOCKET_NOWAIT | wxSOCKET_REUSEADDR );
        //test blocking : m_socket = ::new wxI2PDatagramSocket( m_addr, m_privKey, wxSOCKET_WAITALL | wxSOCKET_BLOCK );
	}
	SetClientData ( m_socket_data );
	SetEventHandler( *m_socket_handler, m_socket_id );
	SetNotify( m_socket_flags );
	Notify( m_socket_notify );
}

wxString CMuleUDPSocket::GetPrivKey()
{
    if (! m_socket)
        return wxT("");
    else
        return m_socket->GetPrivKey() ;
}

void CMuleUDPSocket::SetClientData( void *data )
{
	m_socket_data = data ;
	if ( m_socket )
		m_socket->SetClientData ( data );
}

void CMuleUDPSocket::SetEventHandler( wxEvtHandler& handler, int id )
{
	m_socket_handler = &handler ;
	m_socket_id      = id ;
	if ( m_socket )
		m_socket->SetEventHandler ( handler, id );
}

void CMuleUDPSocket::SetNotify( wxSocketEventFlags flags )
{
	m_socket_flags = flags ;
	if ( m_socket )
		m_socket->SetNotify ( flags );
}

void CMuleUDPSocket::Notify( bool notify )
{
	m_socket_notify = notify ;
	if ( m_socket )
		m_socket->Notify ( notify );
}

bool CMuleUDPSocket::Destroy()
{
    DestroySocket();

    // schedule this object for deletion
    wxAppTraits *traits = wxTheApp ? wxTheApp->GetTraits() : NULL;
    if ( traits )
    {
      // let the traits object decide what to do with us
        traits->ScheduleForDestroy(this);
    }
    else // no app or no traits
    {
      // in wxBase we might have no app object at all, don't leak memory
        delete this;
    }

    return true;
}

void CMuleUDPSocket::DestroySocket()
{
	wxCHECK_RET ( m_socket, wxT ( "Cannot destroy non-existing socket." ) );

	AddDebugLogLineM ( false, logMuleUDP, wxT ( "Shutting down " ) + m_name );
	m_socket->SetNotify ( 0 );
	m_socket->Notify ( false );
	m_socket->Close();
	m_socket->Destroy();
	m_socket = NULL;
}


void CMuleUDPSocket::Open()
{

	wxMutexLocker lock ( m_mutex );

	if ( ! m_socket )
	{
		CreateSocket();
	}
	//m_socket->Start();
}


void CMuleUDPSocket::Close()
{

	wxMutexLocker lock ( m_mutex );

	DestroySocket();
}


void CMuleUDPSocket::OnSend ( int errorCode )
{
	if ( errorCode )
	{
		return;
	}

	{

		wxMutexLocker lock ( m_mutex );

		m_busy = false;

		if ( m_queue.empty() )
		{
			return;
		}
	}

	theApp.uploadBandwidthThrottler->QueueForSendingControlPacket ( this );
}





void CMuleUDPSocket::OnReceive ( int errorCode )
{
	if ( errorCode or !Ok() )
	{
		return;
	}

	char buffer[UDP_BUFFER_SIZE];

	//imuleIPV4Address addr;
	CI2PAddress addr;
	unsigned length = 0;
	bool error = false;
	int lastError = 0;

	{

		wxMutexLocker lock ( m_mutex );

		length = m_socket->RecvFrom ( addr, buffer, UDP_BUFFER_SIZE ).LastCount();

		error = m_socket->Error();

		lastError = m_socket->LastError();

	}

	if ( error )
	{
		AddDebugLogLineM ( false, logMuleUDP, ( m_name + wxT ( ": Error while reading: " ) ) << lastError );
		return;
	}
	else if ( length < 2 )
	{
		// 2 bytes (protocol and opcode) is the smallets possible packet.
		AddDebugLogLineM ( false, logMuleUDP, m_name + wxT ( ": Invalid Packet received by " ) );
		return;
	}
	else if ( !addr/*StringIPtoUint32(addr.IPAddress())*/ )
	{
		// wxASSERT(0);

		printf ( "Unknown dest receiving on UDP packet! Ignoring: '%s'\n", ( const char* ) unicode2char ( addr.toString() /*.IPAddress()*/ ) );

		return;
	}/* else if (!addr.Service()) {

			// wxASSERT(0);
			printf("Unknown port receiving an UDP packet! Ignoring\n");
			return;
		}*/

	AddDebugLogLineM ( false, logMuleUDP, CFormat ( wxT ( "%s : Packet received from %x (size=%db, hash=%x)" ) )
	                   % m_name % addr.hashCode() % length % packetHash ( buffer + 2, length - 2 )
	                 );

	//uint16_t h = packetHash ( buffer + 2, length - 2 );

	if ( length > UDP_BUFFER_SIZE ) 
		AddLogLineM( true, CFormat(_("Please report this : length(%d) > UDP_BUFFER_SIZE(%d) in"
				" CMuleUDPSocket::OnReceive. protocol was %s, opcode was %s."))
				% length % UDP_BUFFER_SIZE %opstr(buffer[0]) % opstr(buffer[0], buffer[1]) );

	OnPacketReceived ( addr, ( uint8_t* ) buffer, length );
}


void CMuleUDPSocket::SendPacket ( CPacket* packet, const CI2PAddress & dest )

//I2P:uint32_t IP, uint16_t port)
{
	wxCHECK_RET ( packet, wxT ( "Invalid packet." ) );
	//wxCHECK_RET(port, wxT("Invalid port."));
	//wxCHECK_RET(IP, wxT("Invalid IP."));
	wxCHECK_RET ( dest.isValid(), wxT ( "Invalid Destination." ) );

    if ( packet->GetPacketSize() > UDP_BUFFER_SIZE )
    {
        wxASSERT(false);
    }
	wxCHECK_RET ( packet->GetPacketSize() <= UDP_BUFFER_SIZE , (wxString) (CFormat(wxT( "Packet too big : %ub (max: %ub"))
		% packet->GetPacketSize() % UDP_BUFFER_SIZE) ) ;

	if ( !Ok() )
	{
		AddDebugLogLineM ( false, logMuleUDP, CFormat ( wxT ( "%s: Packet discarded (socket not Ok) from %x (size=%db)" ) )
		                   % m_name % dest.hashCode() % packet->GetPacketSize() );

		delete packet;
		return;
	}

	AddDebugLogLineM ( false, logMuleUDP, CFormat ( wxT ( "%s: Packet queued to %x (size=%db, hash=%x)" ) )

	                   % m_name % dest.hashCode() % packet->GetPacketSize() % packetHash ( packet->GetDataBuffer(), packet->GetPacketSize() ) );


	UDPPack newpending;
	//newpending.IP = IP;
	//newpending.port = port;
	newpending.dest = dest;
	newpending.packet = packet;
	newpending.time = GetTickCount();

	{

		wxMutexLocker lock ( m_mutex );

		m_queue.push_back ( newpending );
	}

	theApp.uploadBandwidthThrottler->QueueForSendingControlPacket ( this );
}


bool CMuleUDPSocket::Ok()

{

	wxMutexLocker lock ( m_mutex );

	return m_socket and m_socket->IsOk();
}

const uint32_t & CMuleUDPSocket::maxPacketDataSize() {
	static const uint32_t l = CPacket::UDPPacketMaxDataSize();
	return l ;
}

SocketSentBytes CMuleUDPSocket::SendControlData ( uint32_t maxNumberOfBytesToSend, uint32_t WXUNUSED ( minFragSize ) )
{

	wxMutexLocker lock ( m_mutex );

	uint32_t sentBytes = 0;

	while ( !m_queue.empty() && !m_busy && ( sentBytes < maxNumberOfBytesToSend ) )
	{
		UDPPack item = m_queue.front();
		CPacket* packet = item.packet;

		if ( GetTickCount() - item.time < UDPMAXQUEUETIME )
		{
			char sendbuffer[packet->GetPacketSize() + 2];
			memcpy ( sendbuffer, packet->GetUDPHeader(), 2 );
			memcpy ( sendbuffer + 2, packet->GetDataBuffer(), packet->GetPacketSize() );

			if ( SendTo ( sendbuffer, packet->GetPacketSize() + 2, item.dest/*IP, item.port*/ ) )
			{
				sentBytes += packet->GetPacketSize() + 2;

				m_queue.pop_front();
				delete packet;
			}
			else
			{
				// TODO: Needs better error handling, see SentTo
				break;
			}
		}
		else
		{
			m_queue.pop_front();
			delete packet;
		}
	}

	if ( !m_busy && !m_queue.empty() )
	{
		theApp.uploadBandwidthThrottler->QueueForSendingControlPacket ( this );
	}

	SocketSentBytes returnVal = { true, 0, sentBytes };
	return returnVal;
}


bool CMuleUDPSocket::SendTo ( char* buffer, uint32_t length, const CI2PAddress & addr )
//I2P:uint32_t ip, uint16_t port)
{
	// Just pretend that we sent the packet in order to avoid infinete loops.

	if ( ! ( m_socket and m_socket->IsOk() ) )
	{
		return true;
	}

	//I2P:imuleIPV4Address addr;
	//addr.Hostname(ip);
	//addr.Service(port);

	// We better clear this flag here, status might have been changed
	// between the U.B.T. adition and the real sending happening later
	m_busy = false;

	bool sent = false;

	m_socket->SendTo ( addr, buffer, length );

	if ( m_socket->Error() )
	{
		wxSocketError error = m_socket->LastError();

		switch ( error )
		{
			case wxSOCKET_WOULDBLOCK:
				// Socket is busy and can't send this data right now,
				// so we just return not sent and set the wouldblock
				// flag so it gets resent when socket is ready.
				m_busy = true;

				break;

			case wxSOCKET_INVSOCK:
				// Once in a while, wxDatagram sockets have a tendency
				// to suddenly become invalid. In order to work around
				// this problem, we simply create a replacement socket.
				AddDebugLogLineM ( true, logMuleUDP, wxT ( "Socket died. Recreating socket." ) );

				DestroySocket();

				CreateSocket();

				break;

			default:
				// An error which we can't handle happended, so we drop
				// the packet rather than risk entering an infinite loop.
				printf ( "WARNING! %s discarded packet due to errors (%i) while sending.\n", ( const char* ) unicode2char ( m_name ), error );

				sent = true;

				break;

		}
	}
	else
	{
		AddDebugLogLineM ( false, logMuleUDP, CFormat ( wxT ( "%s: Packet sent to %x (size=%db)" ) )
		                   % m_name % addr.hashCode() % length );

		sent = true;
	}

	return sent;
}

bool CMuleUDPSocket::GetLocal(CI2PAddress & ret)
{
	if (m_socket) {
		m_socket->GetLocal(ret);
		return true;
	}
	return false;
}
