//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2004-2006 Marcelo Jimenez ( phoenix@amule.org )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef __PROXY_H__
#define __PROXY_H__

#include <wx/wx.h>


//------------------------------------------------------------------------------
// CProxyType
//------------------------------------------------------------------------------

/*
 * These constants must match the integer values saved in the configuration file,
 * DO NOT CHANGE THIS ORDER!!!
 */
enum CProxyType {
	PROXY_NONE = -1,
	PROXY_SOCKS5,
	PROXY_SOCKS4,
	PROXY_HTTP,
	PROXY_SOCKS4a,
};


//------------------------------------------------------------------------------
// CProxyData
//------------------------------------------------------------------------------
/**
 * The ProxyData class will hold information about the proxy server to be used.
 */
class CProxyData
{
public:
	/**
	 * Default constructor.
	 */
	CProxyData();
	/**
	 * Constructor.
	 * 
	 * @param proxyEnable	Whether proxy is enabled or not.
	 * @param proxyType	The type of the proxy server.
	 * @param proxyHostName	The proxy host name or IP address.
	 * @param proxyPort	The proxy port number.
	 * @param enablePassword Whether authentication should be performed.
	 * @param userName	The user name to authenticate to the server.
	 * @param password	The password to authenticate to the server.
	 */
	CProxyData(
		bool		proxyEnable,
		CProxyType	proxyType,
		const wxString	& proxyHostName,
		unsigned short	proxyPort,
		bool		enablePassword,
		const wxString	&userName,
		const wxString	&password
	);
	/**
	 * Clears the object contents.
	 */
	void Clear();

public:
	//! Whether proxy is enabled or not.
	bool		m_proxyEnable;
	//! The type of the proxy server.
	CProxyType	m_proxyType;
	//! The proxy host name or IP address.
	wxString	m_proxyHostName;
	//! The proxy port number.
	unsigned short	m_proxyPort;
	//! Whether authentication should be performed.
	bool		m_enablePassword;
	//! The user name to authenticate to the server.
	wxString	m_userName;
	//! The password to authenticate to the server.
	wxString	m_password;
};


#endif /* __PROXY_H__ */

