//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef SERVER_H
#define SERVER_H

#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/list.h>		// Needed for WX_DECLARE_LIST

#include <list>

#include "Types.h"		// Needed for uint8_t, uint16_t and uint32_t

#include "i2p/CI2PAddress.h"
#include "Tag.h"

#ifdef CLIENT_GUI
#include <ec/ECSpecialTags.h>
#endif

class ServerMet_Struct;
class CFileDataIO;

// Server TCP flags
#define SRV_TCPFLG_COMPRESSION          0x00000001
#define SRV_TCPFLG_NEWTAGS                      0x00000008
#define SRV_TCPFLG_UNICODE                      0x00000010
#define SRV_TCPFLG_RELATEDSEARCH        0x00000040
#define SRV_TCPFLG_TYPETAGINTEGER       0x00000080

// Server UDP flags
#define SRV_UDPFLG_EXT_GETSOURCES       0x00000001
#define SRV_UDPFLG_EXT_GETFILES         0x00000002
#define SRV_UDPFLG_NEWTAGS                      0x00000008
#define SRV_UDPFLG_UNICODE                      0x00000010
#define SRV_UDPFLG_EXT_GETSOURCES2      0x00000020

// Server priority
#define SRV_PR_LOW                      2
#define SRV_PR_NORMAL                   0
#define SRV_PR_HIGH                     1
// Server priority max and min values
#define SRV_PR_MAX			2
#define SRV_PR_MID			1
#define SRV_PR_MIN			0

//typedef std::list<CTag*> TagList;

class CServer
{
public:
	CServer(ServerMet_Struct* in_data);
	//CServer(uint16_t in_port, const wxString i_addr);//I2P
	CServer(const CI2PAddress & in_dest);
	CServer(CServer* pOld);

#ifdef CLIENT_GUI
	CServer(CEC_Server_Tag *);
#endif

	~CServer();
	void  AddTag(CTag & in_tag)	{m_taglist.push_back(in_tag);}
	const wxString &GetListName() const	{return listname;}
	
	/* I2P
	const wxString &GetFullIP() const	{return ipfull;}
	
	const wxString &GetAddress() const {
	if (!dynip.IsEmpty()) {
	return dynip;
} else {
	return ipfull;
}
}	
	
	// the official port
	uint16_t  GetPort() const			{return realport ? realport : port;}
	// the connection port
	uint16_t  GetConnPort() const		{return port;}
	void    SetPort(uint32_t val)		{realport = val;}
	*/
	const wxString GetAddress() const 	{
		if (dyndest) 	return dest.getAlias(); 
		else 		return dest.toString();}
	// the official port
	const CI2PAddress & GetDest() const 	{return realdest.isValid() ? realdest : dest;}
	// the connection port
	const CI2PAddress & GetConnDest() const	{return dest;}
	void    SetDest( const CI2PAddress & val)	{realdest = val;}

	bool    AddTags ( const TagList & taglist );
	void	SetListName(const wxString& newname);
	void	SetDescription(const wxString& newdescription);
	/* I2P
	uint32_t	GetIP() const			{return ip;}
	*/
	
	uint32_t	GetFiles() const		{return files;} 
	uint32_t	GetUsers() const		{return users;} 
	const wxString	&GetDescription() const	{return description;} 
	uint32_t	GetPing() const			{return ping;} 
	uint32_t	GetPreferences() const		{return preferences;} 
	uint32_t	GetMaxUsers() const		{return maxusers;}
	void	SetMaxUsers(uint32_t in_maxusers) {maxusers = in_maxusers;}
	void	SetUserCount(uint32_t in_users)	{users = in_users;}
	void	SetFileCount(uint32_t in_files)	{files = in_files;}
	void	ResetFailedCount()		{failedcount = 0;} 
	void	AddFailedCount()		{failedcount++;} 
	uint32_t	GetFailedCount() const		{return failedcount;} 
	
	/* I2P
	void	SetID(uint32_t newip);
	const wxString &GetDynIP() const	{return dynip;}
	bool	HasDynIP() const		{return dynip.Length();}
	void	SetDynIP(const wxString& newdynip);
	*/
	void	SetDestID(const CI2PAddress & newdest);
	bool	HasDynDest() const		{return dyndest;}
	void	SetDynDest(const wxString& newdyndest);
	const wxString & GetDynDest() const	{return dest.getAlias();}
	
	uint32_t	GetLastPinged() const		{return lastpinged;}
	void	SetLastPinged(uint32_t in_lastpinged) {lastpinged = in_lastpinged;}
	void	SetPing(uint32_t in_ping)		{ping = in_ping;}
	void	SetPreference(uint32_t in_preferences) {preferences = in_preferences;}
	void	SetIsStaticMember(bool in)	{staticservermember=in;}
	bool	IsStaticMember() const		{return staticservermember;}
	uint32_t	GetChallenge() const		{return challenge;}
	void	SetChallenge(uint32_t in_challenge) {challenge = in_challenge;}
	uint32_t	GetSoftFiles() const		{return softfiles;}
	void	SetSoftFiles(uint32_t in_softfiles) {softfiles = in_softfiles;}
	uint32_t	GetHardFiles() const		{return hardfiles;}
	void	SetHardFiles(uint32_t in_hardfiles) {hardfiles = in_hardfiles;}
	const	wxString &GetVersion() const	{return m_strVersion;}
	void	SetVersion(const wxString &pszVersion)	{m_strVersion = pszVersion;}
	void	SetTCPFlags(uint32_t uFlags)	{m_uTCPFlags = uFlags;}
	uint32_t	GetTCPFlags() const		{return m_uTCPFlags;}
	void	SetUDPFlags(uint32_t uFlags)	{m_uUDPFlags = uFlags;}
	uint32_t	GetUDPFlags() const		{return m_uUDPFlags;}
	//uint32_t	GetLowIDUsers() const		{return m_uLowIDUsers;}
	//void	SetLowIDUsers(uint32_t uLowIDUsers) {m_uLowIDUsers = uLowIDUsers;}
	uint32_t	GetDescReqChallenge() const	{return m_uDescReqChallenge;}
	void	SetDescReqChallenge(uint32_t uDescReqChallenge) {m_uDescReqChallenge = uDescReqChallenge;}
	uint8_t	GetLastDescPingedCount() const	{return lastdescpingedcout;}
	void	SetLastDescPingedCount(bool reset);
	bool	GetUnicodeSupport() const				{return GetTCPFlags() & SRV_TCPFLG_UNICODE;}
	const wxString& GetAuxPortsList() const	{return m_auxPorts;}
	void	SetAuxPortsList(const wxString& val)	{m_auxPorts = val;}
	
	wxULongLong GetLastDNSSolve() const { return m_lastdnssolve; }
	void SetLastDNSSolve(wxULongLong value) { m_lastdnssolve = value; }
	
	bool GetDNSError() const { return m_dnsfailure; }
	void SetDNSError(bool value) { m_dnsfailure = value; }
	
private:
	uint32_t		challenge;
	uint32_t		lastpinged;
	uint32_t		files;
	uint32_t		users;
	uint32_t		maxusers;
	uint32_t		softfiles;
	uint32_t		hardfiles;
	uint32_t		preferences;
	uint32_t		ping;
	wxString	description;
	wxString	listname;
	//wxString	dynip;
	bool		dyndest;
	uint32_t		tagcount;
	//CI2PAddress	destfull; // I2P : ipfull;
	/* I2P
	uint32_t		ip;
	uint16_t		port;
	uint16_t		realport;
	*/
	CI2PAddress	dest;
	CI2PAddress	realdest;
	
	uint32_t		failedcount; 
	uint32_t		m_uDescReqChallenge;
	uint8_t		lastdescpingedcout;
	TagList		m_taglist;
	//CTypedPtrList<CPtrList, CTag*>*	taglist;
	uint8_t		staticservermember;
	wxString	m_strVersion;
	uint32_t		m_uTCPFlags;
	uint32_t		m_uUDPFlags;
	//uint32_t		m_uLowIDUsers;
	wxString	m_auxPorts;
	
	wxULongLong		m_lastdnssolve;
	bool		m_dnsfailure;
	
	void Init();

};

#endif // SERVER_H
