//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

//
// Client to Server communication
// 

#ifndef SERVERCONNECT_H
#define SERVERCONNECT_H

#include <list>			// Needed for std::list
#include <map>			// Needed for std::map

#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/filename.h>

#include "Types.h"		// Needed for uint8_t, uint8_t, uint16_t and uint32_t
#include "imuleIPV4Address.h"	// Needed for imuleIPV4Address
#include "Timer.h"		// Needed for CTimer


class CServerList;
class CServerSocket;
class CServer;
class CPacket;
class CServerUDPSocket;

#define CS_FATALERROR	-5
#define CS_DISCONNECTED	-4
#define CS_SERVERDEAD	-3
#define	CS_ERROR	-2
#define CS_SERVERFULL	-1
#define	CS_NOTCONNECTED	0
#define	CS_CONNECTING	1
#define	CS_CONNECTED	2
#define	CS_WAITFORLOGIN	3
#define CS_RETRYCONNECTTIME  30 // seconds

class CServerConnect {
public:
//I2P:	CServerConnect(CServerList* in_serverlist, imuleIPV4Address &address);
	CServerConnect(CServerList* in_serverlist, const wxFileName & destFile);
	~CServerConnect();
	
	void	ConnectionFailed(CServerSocket* sender);
	void	ConnectionEstablished(CServerSocket* sender);
	
	void	ConnectToAnyServer(bool prioSort = true);
	void	ConnectToServer(CServer* toconnect, bool multiconnect = false);
	void	StopConnectionTry();
	void	CheckForTimeout();
	
	// safe socket closure and destruction
	void	DestroySocket(CServerSocket* pSck);
	bool	SendPacket(CPacket* packet,bool delpacket = true, CServerSocket* to = 0);

	// Creteil Begin
	bool	IsUDPSocketAvailable() const { return serverudpsocket != NULL; }
	// Creteil End

	bool	SendUDPPacket(CPacket* packet,CServer* host, bool delpacket = false );
	bool	Disconnect();
	bool	IsConnecting()	{ return connecting; }
	bool	IsConnected()	{ return connected; }
	/*I2P:uint32_t	GetClientID()	{ return clientid; }*/
	CI2PAddress	GetClientDest()	{ return clientdest; }
	CServer*GetCurrentServer();
	/*I2P:uint32_t	clientid;*/
	CI2PAddress	clientdest;
	uint8_t	pendingConnects;
	//I2P:bool	IsLowID()	{ return ::IsLowID(clientid); }
	//bool	IsLowID()	{ return ::IsLowID(clientdest); }
	//I2P:void	SetClientID(uint32_t newid);
	void	SetClientDest(const CI2PAddress & newdest);
	/*I2P:bool	IsLocalServer(uint32_t dwIP, uint16_t nPort);*/
	bool	IsLocalServer(const CI2PAddress & nDest);
	void	TryAnotherConnectionrequest();
	bool	IsSingleConnect()	{ return singleconnecting; }
	void	KeepConnectionAlive();

private:
	bool	connecting;
	bool	singleconnecting;
	bool	connected;
	uint8_t	max_simcons;
	CServerSocket*	connectedsocket;
	CServerList*	used_list;
	CServerUDPSocket*	serverudpsocket;
	
	// list of currently opened sockets
	typedef	std::list<CServerSocket*>	SocketsList;
	SocketsList	m_lstOpenSockets;
	CTimer	m_idRetryTimer;

	std::map<uint32_t, CServerSocket*> connectionattemps;
};

#endif // SERVERCONNECT_H
