//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef SERVERLIST_H
#define SERVERLIST_H

#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/timer.h>		// Needed for wxTimer

#include <list>
#include <vector>

#include "Types.h"		// Needed for uint8_t, uint16_t and uint32_t

#include "ObservableQueue.h"

#include "i2p/CI2PAddress.h"
class CServer;
class CPacket;

class CServerList : public CObservableQueue<CServer*>
{
	friend class CServerListCtrl;

public:
	CServerList();
	~CServerList();
	bool		Init();
	bool		AddServer(CServer* in_server, bool fromUser = false);
	void		RemoveServer(CServer* in_server);
	void		RemoveAllServers();
	void		RemoveDeadServers();	
	bool		LoadServerMet(const wxString& strFile);
	bool		SaveServerMet();
	void		ServerStats();
	void		ResetServerPos()	{m_serverpos = m_servers.begin();}
	CServer*	GetNextServer();
	size_t		GetServerCount()	{return m_servers.size();}
	/*I2P:
	CServer*	GetServerByAddress(const wxString& address, uint16_t port);
	CServer*	GetServerByIP(uint32_t nIP);
	CServer*	GetServerByIP(uint32_t nIP, uint16_t nPort);	
	*/
	CServer*	GetServerByAddress(const wxString& address);
	CServer*	GetServerByDest(const CI2PAddress & nDest);
	
	void		GetStatus(uint32_t &failed, uint32_t &user, uint32_t &file, uint32_t &tuser, uint32_t &tfile, float &occ);
	void		GetUserFileStatus( uint32_t &user, uint32_t &file);
	void		Sort();
	void 		UpdateServerMetFromURL(const wxString& strURL);	
	void		DownloadFinished(uint32_t result);	
	void		AutoDownloadFinished(uint32_t result);	
	uint32_t		GetAvgFile() const;

	std::vector<const CServer*> CopySnapshot() const;
	
private:
	virtual void 	ObserverAdded( ObserverType* );
	void		AutoUpdate();
	CServer*	GetNextStatServer();

	void		LoadStaticServers( const wxString& filename );
	uint8_t		current_url_index;

	typedef std::list<CServer*>	CInternalList;
	CInternalList			m_servers;
	CInternalList::const_iterator	m_serverpos;
	CInternalList::const_iterator	m_statserverpos;

	uint32_t		m_nLastED2KServerLinkCheck;// emanuelw(20030924) added
	wxString	URLUpdate;
	wxString	URLAutoUpdate;
};

#endif // SERVERLIST_H
