//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#include <zlib.h>		// Needed for uLongf

#include "Tag.h"				// Interface declarations

#include <wx/string.h>

#include <common/MuleDebug.h>		// Needed for CInvalidPacket
#include <common/Format.h>
#include <Logger.h>
#include "SafeFile.h"		// Needed for CFileDataIO
#include "MemFile.h"
#include "MD4Hash.h"			// Needed for CMD4Hash

#define _DEBUG_TAGS
///////////////////////////////////////////////////////////////////////////////
// CTag


const CTag CTag::null;

CTag::CTag()
{
	m_uType = TAGTYPE_INVALID;
}

bool CTag::isValid() const
{
	return m_uType!=TAGTYPE_INVALID;
}

CTag::CTag(uint8_t uName, wxULongLong uVal)
{
	m_uType = TAGTYPE_UINT64;
	m_uName = uName;
//	m_pszName = NULL;
	m_uVal = uVal;
	m_nBlobSize = 0;
}

/*CTag::CTag(char* pszName, const wxString& rstrVal)
{
	m_uType = TAGTYPE_STRING;
	m_uName = 0;
	m_pszName = nstrdup(pszName);
	m_pstrVal = new wxString(rstrVal);
	m_nBlobSize = 0;
}*/

CTag::CTag(uint8_t uName, const wxString& rstrVal)
{
	m_uType = TAGTYPE_STRING;
	m_uName = uName;
//	m_pszName = NULL;
	m_pstrVal = new wxString(rstrVal);
	m_nBlobSize = 0;
}

/*CTag::CTag(char* pszName, const CI2PAddress& raddrVal)
{
	m_uType = TAGTYPE_ADDRESS;
	m_uName = 0;
	m_pszName = nstrdup(pszName);
	m_paddrVal = new CI2PAddress(raddrVal);
	m_nBlobSize = 0;
}*/

CTag::CTag(uint8_t uName, const CI2PAddress& raddrVal)
{
	m_uType = TAGTYPE_ADDRESS;
	m_uName = uName;
//	m_pszName = NULL;
	m_paddrVal = new CI2PAddress(raddrVal);
	m_nBlobSize = 0;
}

CTag::CTag(uint8_t uName, const CMD4Hash& hash)
{
	m_uType = TAGTYPE_HASH;
	m_uName = uName;
//	m_pszName = NULL;
	m_hashVal = new CMD4Hash(hash);
	m_nBlobSize = 0;
}

CTag::CTag(uint8_t uName, uint32_t nSize, const unsigned char* pucData){
	m_uType = TAGTYPE_BLOB;
	m_uName = uName;
//	m_pszName = NULL;
	m_pData = new unsigned char[nSize];
	memcpy(m_pData, pucData, nSize);
	m_nBlobSize = nSize;
}

CTag::CTag(uint8_t uName, wxULongLong first, wxULongLong last){
	m_uType = TAGTYPE_UINT64UINT64;
	m_uName = uName;
//	m_pszName = NULL;
	m_u64u64.first = first;
	m_u64u64.last  = last;
}

CTag::CTag(const CTag& rTag)
{
	CTag();
	(*this) = rTag;
}

CTag & CTag::operator=(const CTag& rTag)
{
	m_uType = rTag.m_uType;
	m_uName = rTag.m_uName;
	m_nBlobSize = 0;
	
	if (rTag.IsAddr()) {
		m_paddrVal = new CI2PAddress(rTag.GetAddr());
	} else if (rTag.IsStr()) {
		m_pstrVal = new wxString(rTag.GetStr());
	} else if (rTag.IsInt()) {
		m_uVal = rTag.GetInt();
	} else if (rTag.IsFloat()) {
		m_fVal = rTag.GetFloat();
	} else if (rTag.IsHash()) {
		m_hashVal = new CMD4Hash(rTag.GetHash());
	} else if (rTag.IsBlob()) {
		m_nBlobSize = rTag.GetBlobSize();
		m_pData = new unsigned char[rTag.GetBlobSize()];
		memcpy(m_pData, rTag.GetBlob(), rTag.GetBlobSize());
	} else if (rTag.IsUint64Uint64()) {
		m_u64u64.first = rTag.GetUint64Uint64().first;
		m_u64u64.last = rTag.GetUint64Uint64().last;
	} else {
		wxASSERT(0);
		m_uVal = 0;
	}
	return (*this);
}


CTag::CTag(CFileDataIO& data/*, bool bOptUTF8*/)
{
#ifdef _DEBUG_TAGS
	uint32_t ntag ;
#endif
	// Zero variables to allow for safe deletion
	m_nBlobSize = m_uName = m_uType = 0;
//	m_pszName = NULL;
	m_pData = NULL;	
	
	try {
		m_uType = data.ReadUInt8();
		if (m_uType & 0x80) {
			m_uType =  (uint8_t) (m_uType & 0x7F);
			m_uName = data.ReadUInt8();
//			m_pszName = NULL;
		} else {
			uint16_t length = data.ReadUInt16();
			if (length == 1) {
				m_uName = data.ReadUInt8();
//				m_pszName = NULL;
			} else {
				m_uName = 0;
//				m_pszName = new char[length + 1];
//				data.Read(m_pszName, length);
//				m_pszName[length] = '\0';
			}
		}
	
		// NOTE: It's very important that we read the *entire* packet data,
		// even if we do not use each tag. Otherwise we will get in trouble
		// when the packets are returned in a list - like the search results
		// from a server. If we cannot do this, then we throw an exception.
#ifdef _DEBUG_TAGS
		ntag = data.ReadUInt32() ;
#endif
		switch (m_uType) {
			case TAGTYPE_ADDRESS:
				m_paddrVal = new CI2PAddress(data.ReadAddress());
				break;
			
			case TAGTYPE_STRING:
				m_pstrVal = new wxString(data.ReadString(true/*bOptUTF8*/));
				break;
			
			case TAGTYPE_UINT64:
				m_uVal = data.ReadUInt64();
				break;
			
			case TAGTYPE_UINT32:
				m_uVal = data.ReadUInt32();
				m_uType = TAGTYPE_UINT64;
				break;
			
			case TAGTYPE_UINT16:
				m_uVal = data.ReadUInt16();
				m_uType = TAGTYPE_UINT64;
				break;
			
			case TAGTYPE_UINT8:
				m_uVal = data.ReadUInt8();
				m_uType = TAGTYPE_UINT64;
				break;
			
			case TAGTYPE_UINT64UINT64:
				m_u64u64.first = CTag(data).GetInt();
				m_u64u64.last  = CTag(data).GetInt();
				break;
			
			case TAGTYPE_FLOAT32:
				//TODO #warning Endianess problem?
				data.Read(&m_fVal, 4);
				break;
			
			case TAGTYPE_HASH:
				m_hashVal = new CMD4Hash(data.ReadHash());
				break;
			
			case TAGTYPE_BOOL:
//				printf("***NOTE: %s; Reading BOOL tag\n", __FUNCTION__);
				data.ReadUInt8();
				break;
			
			case TAGTYPE_BOOLARRAY: {
//				printf("***NOTE: %s; Reading BOOL Array tag\n", __FUNCTION__);
				uint16_t len = data.ReadUInt16();
			
				// 07-Apr-2004: eMule versions prior to 0.42e.29 used the formula "(len+7)/8"!
				//TODO #warning This seems to be off by one! 8 / 8 + 1 == 2, etc.
				//data.Seek((len / 8) + 1, wxFromCurrent);
				data.Seek((len+7) / 8, wxFromCurrent);
				break;
			}
	
			case TAGTYPE_BLOB:
				// 07-Apr-2004: eMule versions prior to 0.42e.29 handled the "len" as uint16_t!
				m_nBlobSize = data.ReadUInt32();
				
				// Since the length is 32b, this check is needed to avoid
				// huge allocations in case of bad tags.
				if (m_nBlobSize > data.GetLength() - data.GetPosition()) {
					throw CInvalidPacket(wxT("Malformed tag"));
				}
					
				m_pData = new unsigned char[m_nBlobSize];
				data.Read(m_pData, m_nBlobSize);
				break;
		
			default:
				if (m_uType >= TAGTYPE_STR1 && m_uType <= TAGTYPE_STR16) {
					uint8_t length = (uint8_t) (m_uType - TAGTYPE_STR1 + 1);
					m_pstrVal = new wxString(data.ReadOnlyString(true /*bOptUTF8*/, length));
					m_uType = TAGTYPE_STRING;
				} else {
					// Since we cannot determine the length of this tag, we
					// simply have to abort reading the file.
					throw CInvalidPacket(CFormat(wxT("Unknown tag type (0x%x) encounted, cannot proceed")) % m_uType);
				}
		}
	} catch (...) {
		reinit() ;
		//wxASSERT(false);

		throw wxString(wxT("Unable to read a Tag"));
	}
	AddDebugLogLineM( false, logCTag, CFormat(wxT("Received #%d, type %s , name %s : %s "))
			% ntag
					% tagtypeStr(m_uType)
					% tagnameStr(m_uName)
					% GetFullInfo() );
}

void CTag::reinit()
{
	if (IsAddr()) {
		delete m_paddrVal;
	} else if (IsStr()) {
		delete m_pstrVal;
	} else if (IsHash()) {
		delete m_hashVal;
	} else if (IsBlob()) {
		delete[] m_pData;
	}
}

CTag::~CTag()
{
	reinit();
}


#define CHECK_TAG_TYPE(check, expected) \
	if (!(check)) { \
		throw CInvalidPacket(wxT(#expected) wxT(" tag expected, but found ") + GetFullInfo()); \
	}

wxULongLong CTag::GetInt() const
{
	CHECK_TAG_TYPE(IsInt(), Integer);
	
	return m_uVal; 
}

CTag::U64U64 CTag::GetUint64Uint64() const
{
	CHECK_TAG_TYPE(IsUint64Uint64(), Uint64Uint64);
	
	return m_u64u64; 
}


const CI2PAddress& CTag::GetAddr() const
{
	CHECK_TAG_TYPE(IsAddr(), Addr);
	
	return *m_paddrVal;
}


const wxString& CTag::GetStr() const
{
	CHECK_TAG_TYPE(IsStr(), String);
	
	return *m_pstrVal; 	
}


float CTag::GetFloat() const
{
	CHECK_TAG_TYPE(IsFloat(), Float);

	return m_fVal;
}


const CMD4Hash& CTag::GetHash() const
{
	CHECK_TAG_TYPE(IsHash(), Hash);
	
	return *m_hashVal;
}
	

uint32_t CTag::GetBlobSize() const
{
	CHECK_TAG_TYPE(IsBlob(), Blob);
	
	return m_nBlobSize;
}
	

const uint8_t* CTag::GetBlob() const
{
	CHECK_TAG_TYPE(IsBlob(), Blob);
	
	return m_pData;
}

void CTag::SetInt(wxULongLong uVal)
{
	CHECK_TAG_TYPE(IsInt(), Integer);
		
	m_uVal = uVal;
}


bool CTag::WriteTo(CFileDataIO* data) const
{

	// Write tag type
	uint8_t uType;
	
	if (IsInt()) {
		if (m_uVal <= 0xFF) {
			uType = TAGTYPE_UINT8;
		} else if (m_uVal <= 0xFFFF) {
			uType = TAGTYPE_UINT16;
		} else if (m_uVal <= 0xFFFFFFFF) {
			uType = TAGTYPE_UINT32;
		} else
			uType = TAGTYPE_UINT64;
	} else if (IsStr()) {
		size_t uStrValLen = strlen( m_pstrVal->mb_str(wxConvUTF8));
		if (uStrValLen >= 1 && uStrValLen <= 16) {
			uType = (uint8_t) (TAGTYPE_STR1 + uStrValLen - 1);
		} else {
			uType = TAGTYPE_STRING;
		}
	} else {
		uType = m_uType;
	}

	// Write tag name
	/*if (m_pszName) {
		data->WriteUInt8(uType);
		uint16_t uTagNameLen = strlen(m_pszName);
		data->WriteUInt16(uTagNameLen);
		data->Write(m_pszName, uTagNameLen);
	} else */{
		wxASSERT( m_uName != 0 );
		data->WriteUInt8(uType | 0x80);
		data->WriteUInt8(m_uName);
#ifdef _DEBUG_TAGS
		static uint32_t ntag = 0;
		ntag++;
		data->WriteUInt32(ntag);
		AddDebugLogLineM( false, logCTag, CFormat(wxT("Sent     #%d, type %s , name %s : %s"))
						% ntag
						% tagtypeStr(m_uType)
						% tagnameStr(m_uName)
						% GetFullInfo() ) ;
#endif
	}

	// Write tag data
	switch (uType) {
		case TAGTYPE_ADDRESS:
			data->WriteAddress(*m_paddrVal);
			break;
		case TAGTYPE_STRING:
			data->WriteString(*m_pstrVal);
			break;
		case TAGTYPE_UINT64:
			data->WriteUInt64(m_uVal);
			break;
		case TAGTYPE_UINT32:
			data->WriteUInt32((uint32_t)m_uVal.GetValue());
			break;
		case TAGTYPE_UINT16:
			data->WriteUInt16((uint16_t)m_uVal.GetValue());
			break;
		case TAGTYPE_UINT8:
			data->WriteUInt8((uint8_t)m_uVal.GetValue());
			break;
		case TAGTYPE_FLOAT32:
			//TODO #warning Endianess problem?
			data->Write(&m_fVal, 4);
			break;
		case TAGTYPE_HASH:
			data->WriteHash(*m_hashVal);
			break;
		case TAGTYPE_BLOB:
			data->WriteUInt32(m_nBlobSize);
			data->Write(m_pData, m_nBlobSize);
			break;
		case TAGTYPE_UINT64UINT64:
			CTag( m_uName, m_u64u64.first ).WriteTo(data);
			CTag( m_uName, m_u64u64.last  ).WriteTo(data);
			break;
		default:
			// See comment on the default: of CTag::CTag(const CFileDataIO& data, bool bOptUTF8)
			if (uType >= TAGTYPE_STR1 && uType <= TAGTYPE_STR16) {
				// Sending '0' as len size makes it not write the len on the IO file.
				// This is because this tag types send the len as their type.
				data->WriteString(*m_pstrVal,utf8strRaw/*eStrEncode*/,0); 
			} else {
				printf("%s; Unknown tag: type=0x%02X\n", __FUNCTION__, uType);
				wxASSERT(0);
				return false;
			}
			break;
	}

	return true;
}



TagList::TagList(CFileDataIO& data)
	: std::list<CTag>()
{
	wxULongLong pos = data.GetPosition() ;
	uint8_t n = data.ReadByte() ;
	AddDebugLogLineM( false, logCTag, CFormat(wxT("TagList : pos = %s, %d tags to read")) % pos.ToString() % n ) ;
	while (n--) push_back( CTag(data) );
}

bool TagList::WriteTo(CFileDataIO* file) const
{
	CMemFile temp ;
	uint8_t n = 0;
	wxASSERT( size() < 0xFF );
	for (const_iterator it=begin(); it!=end(); it++)
		if (it->WriteTo(&temp)) n++ ;
	
	temp.Seek(0, wxFromStart);
	char buffer[ temp.GetLength().GetValue() ];
	temp.Read( buffer, (size_t) temp.GetLength().GetValue() );
	
	//uint32_t pos = file->GetPosition() ;
	file->WriteByte( n );
	file->Write( buffer, (size_t) temp.GetLength().GetValue() );
	AddDebugLogLineM( false, logCTag, CFormat(wxT("TagList : %d tags sent")) % n ) ;
	//uint32_t end = file->GetPosition() ;
	//file->Seek( pos );
	//n = file->ReadByte() ;
	//AddDebugLogLineM( false, logCTag, CFormat(wxT("TagList : pos = %d, n = %d")) %pos % n ) ;
	//file->Seek( end );
	
	return true ;
}

void TagList::deleteTags()
{
	clear() ;
}


wxString CTag::GetFullInfo() const
{
	wxString strTag;

	strTag = tagnameStr(m_uName);
	
	strTag += wxT(" = ");
	if (m_uType == TAGTYPE_ADDRESS) {
		strTag += wxT("\"");
		strTag += m_paddrVal->humanReadable();
		strTag += wxT("\"");
	} else if (m_uType == TAGTYPE_STRING) {
		strTag += wxT("\"");
		strTag += *m_pstrVal;
		strTag += wxT("\"");
	} else if (m_uType >= TAGTYPE_STR1 && m_uType <= TAGTYPE_STR16) {
		strTag += wxString::Format(wxT("(Str%u)\""), m_uType - TAGTYPE_STR1 + 1)
					+  *m_pstrVal + wxT("\"");
	} else if (m_uType == TAGTYPE_UINT64) {
		strTag += wxString::Format(wxT("(UInt64)%s"), m_uVal.ToString().c_str());
	} else if (m_uType == TAGTYPE_UINT32) {
		strTag += wxString::Format(wxT("(UInt32)%s"), m_uVal.ToString().c_str());
	} else if (m_uType == TAGTYPE_UINT16) {
		strTag += wxString::Format(wxT("(UInt16)%s"), m_uVal.ToString().c_str());
	} else if (m_uType == TAGTYPE_UINT8) {
		strTag += wxString::Format(wxT("(UInt8)%s"), m_uVal.ToString().c_str());
	} else if (m_uType == TAGTYPE_FLOAT32) {
		strTag += wxString::Format(wxT("(Float32)%f"), m_fVal);
	} else if (m_uType == TAGTYPE_BLOB) {
		strTag += wxString::Format(wxT("(Blob)%u"), m_nBlobSize);
	} else {
		strTag += wxString::Format(wxT("Type=%u"), m_uType);
	}
	return strTag;
}

