//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef TAG_H
#define TAG_H

#include <list>

#include "Types.h"		// Needed for uint8_t, uint32_t, uint8_t and uint32_t
#include "OPCodes.h"		// Needed for TAGTYPE_*
#include <common/StringFunctions.h>	// Needed for EUtf8Str

class CMD4Hash;
class CFileDataIO;
class wxString;
class CI2PAddress;

///////////////////////////////////////////////////////////////////////////////
// CTag

class CTag
{
public:
	typedef struct {wxULongLong first; wxULongLong last;} U64U64;
	CTag();
	bool isValid() const;
	//CTag(char* pszName, uint32_t uVal);
	CTag(uint8_t uName, wxULongLong uVal);
	//CTag(char* pszName, const wxString& rstrVal);
	CTag(uint8_t uName, const wxString& rstrVal);
	//CTag(char* pszName, const CI2PAddress& raddrVal); //I2P
	CTag(uint8_t uName, const CI2PAddress& raddrVal); //I2P
	CTag(uint8_t uName, const CMD4Hash& hash);
	CTag(uint8_t uName, uint32_t nSize, const unsigned char* pucData);
	//CTag(uint8_t uName, uint16_t nI16, uint32_t nI32);
	CTag(uint8_t uName, wxULongLong first, wxULongLong last);
	CTag(const CTag& rTag);
	CTag& operator=(const CTag&);
	CTag(CFileDataIO& data/*, bool bOptUTF8*/);
	void reinit();
	~CTag();

	uint8_t GetType() const			{ return m_uType; }
	uint8_t GetID() const			{ return m_uName; }
	//char* GetName() const			{ return m_pszName; }
	
	bool IsAddr() const				{ return m_uType == TAGTYPE_ADDRESS; }
	bool IsStr() const				{ return m_uType == TAGTYPE_STRING; }
	bool IsInt() const				{ return m_uType == TAGTYPE_UINT64; }
	bool IsFloat() const			        { return m_uType == TAGTYPE_FLOAT32; }
	bool IsHash() const				{ return m_uType == TAGTYPE_HASH; }
	bool IsBlob() const				{ return m_uType == TAGTYPE_BLOB; }
	bool IsUint64Uint64() const                     { return m_uType == TAGTYPE_UINT64UINT64; }
	
	wxULongLong GetInt() const;
	const wxString& GetStr() const;
	const CI2PAddress& GetAddr() const;
	float GetFloat() const;
	const CMD4Hash& GetHash() const;
	uint32_t GetBlobSize() const;
	const uint8_t* GetBlob() const;
	U64U64 GetUint64Uint64() const;

	void SetInt(wxULongLong uVal);
	
	CTag* CloneTag()				{ return new CTag(*this); }
	
	//bool WriteTagToFile(CFileDataIO* file, EUtf8Str eStrEncode = utf8strNone) const;	// old eD2K tags
	//bool WriteNewEd2kTag(CFileDataIO* file, EUtf8Str eStrEncode = utf8strNone) const;	// new eD2K tags
	bool WriteTo(CFileDataIO* file) const;
	
	wxString GetFullInfo() const;

	static const CTag null;
private:
	//! CTag is not assignable.
	
	uint8_t	m_uType;
	uint8_t	m_uName;
	//char*	m_pszName;
	uint32_t	m_nBlobSize;

	
	  CMD4Hash*	 m_hashVal;
	  wxString*	 m_pstrVal;
	  CI2PAddress*	 m_paddrVal;
	  wxULongLong	 m_uVal;
	  float		 m_fVal;
	  unsigned char* m_pData;
	  U64U64         m_u64u64;
	
};

struct TagList : public std::list<CTag>
{
	TagList() : std::list<CTag>() {};
	TagList(CFileDataIO& data);
	bool WriteTo(CFileDataIO* file) const;
	void deleteTags() ;
};

///////////////////////////////////////////////////////////////////////////////
// CTag and tag string helpers

inline int CmpED2KTagName(const char* pszTagName1, const char* pszTagName2){
	// string compare is independant from any codepage and/or LC_CTYPE setting.
	return strcasecmp(pszTagName1, pszTagName2);
}
void ConvertED2KTag(CTag*& pTag);

bool WriteOptED2KUTF8Tag(CFileDataIO* data, const wchar_t* pwsz, uint8_t uTagName);

#endif // TAG_H
