//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifdef TRUE
#undef TRUE
#define TRUE 1
#endif

#ifdef FALSE
#undef FALSE
#define FALSE 1
#endif

#include <iostream>
#include <cerrno>
#include <cmath>
#include <csignal>
#include <unistd.h>			// Needed for close(2) and sleep(3), getuid(2)

#include <sys/types.h>

#include <wx/defs.h>
#include <wx/process.h>
#include <wx/sstream.h>
#include <wx/filefn.h>
#include <wx/ffile.h>
#include <wx/file.h>
#include <wx/log.h>
#include <wx/timer.h>
#include <wx/config.h>
#include <wx/fileconf.h>
#include <wx/socket.h>			// Needed for wxSocket
#include <wx/intl.h>			// Needed for i18n
#include <wx/mimetype.h>		// For launching default browser
#include <wx/cmdline.h>			// Needed for wxCmdLineParser
#include <wx/wfstream.h>
#include <wx/tokenzr.h>
#include <wx/filename.h>
#include <wx/snglinst.h>
#include <wx/thread.h>

#include "i2p/CI2PAddress.h"

#ifdef HAVE_CONFIG_H
#include "config.h"		        // Needed for HAVE_GETRLIMIT, HAVE_SETRLIMIT,
// HAVE_SYS_RESOURCE_H, HAVE_SYS_STATVFS_H and VERSION
#endif
#include "imule.h"			// Interface declarations.
#include "GetTickCount.h"		// Needed for GetTickCount
#include "HTTPDownload.h"		// Needed for CHTTPDownloadThread
#include "Server.h"			// Needed for GetListName
#include "OtherFunctions.h"		// Needed for GetTickCount
#include "IPFilter.h"			// Needed for CIPFilter
#include "UploadQueue.h"		// Needed for CUploadQueue
#include "DownloadQueue.h"		// Needed for CDownloadQueue
#include "ClientCreditsList.h"		// Needed for CClientCreditsList
#include "ServerSocket.h"		// Needed for CServerSocket
#include "SharedFileList.h"		// Needed for CSharedFileList
#include "ServerConnect.h"		// Needed for CServerConnect
#include "ServerList.h"			// Needed for CServerList
#include "KnownFileList.h"		// Needed for CKnownFileList
#include "SearchList.h"			// Needed for CSearchList
#include "ClientList.h"			// Needed for CClientList
#include "Preferences.h"		// Needed for CPreferences
#include "i2p/CI2PRouter.h"             // Needed for CI2PRouter
#include "i2p/CSamLauncher.h"		// Needed for CSamLauncher
#include "ListenSocket.h"		// Needed for CListenSocket
#include "ExternalConn.h"		// Needed for ExternalConn & MuleConnection
#include "ServerUDPSocket.h"		// Needed for CServerUDPSocket
#include "ClientUDPSocket.h"		// Needed for CClientUDPSocket & CMuleUDPSocket
#include "PartFile.h"			// Needed for CPartFile
#include "AddFileThread.h"		// Needed for CAddFileThread
#include "FriendList.h"			// Needed for CFriendList
#include "updownclient.h"		// Needed for CUpDownClient
#include <common/StringFunctions.h>	// Needed for validateURI
#include "Packet.h"
#include "Statistics.h"			// Needed for CStatistics
#include "AICHSyncThread.h"
#include "Logger.h"
#include <common/Format.h>		// Needed for CFormat
#include <common/MD5Sum.h>		// Needed for MD5Sum (no kidding ?)
#include "UploadBandwidthThrottler.h"
#include "InternalEvents.h"		// Needed for CMuleInternalEvent
#include "FileFunctions.h"		// Needed for CDirIterator
#include "kademlia/kademlia/Kademlia.h"
#include "kademlia/kademlia/Prefs.h"
#include "Timer.h"
#include "I2PConnectionManager.h"

#ifndef IMULE_DAEMON
#ifdef __WXMAC__
#include <CoreFoundation/CFBundle.h>
#include <wx/mac/corefoundation/cfstring.h>
#endif
#endif


#ifdef HAVE_SYS_RESOURCE_H
#include <sys/resource.h>
#endif

#ifdef HAVE_SYS_STATVFS_H
#include <sys/statvfs.h>
#endif


#ifdef __GLIBC__
# define RLIMIT_RESOURCE __rlimit_resource
#else
# define RLIMIT_RESOURCE int
#endif


static void UnlimitResource ( RLIMIT_RESOURCE resType )
{
#if defined(HAVE_GETRLIMIT) && defined(HAVE_SETRLIMIT)

	struct rlimit rl;
	getrlimit ( resType, &rl );
	rl.rlim_cur = rl.rlim_max;
	setrlimit ( resType, &rl );
#endif
}


static void SetResourceLimits()
{
#ifdef HAVE_SYS_RESOURCE_H
	UnlimitResource ( RLIMIT_DATA );
	UnlimitResource ( RLIMIT_FSIZE );
	UnlimitResource ( RLIMIT_NOFILE );
#ifdef RLIMIT_RSS
	UnlimitResource ( RLIMIT_RSS );
#endif
#endif
}

// We store the received signal in order to avoid race-conditions
// in the signal handler.
bool g_shutdownSignal = false;

void OnShutdownSignal ( int /* sig */ )
{
	signal ( SIGINT, SIG_DFL );
	signal ( SIGTERM, SIG_DFL );

	g_shutdownSignal = true;

#ifdef IMULE_DAEMON
	theApp.ExitMainLoop();
#endif
}




wxString CimuleApp::GetMuleVersion()

{
	wxString ver ( wxT ( VERSION ) );

	ver += wxT ( " using " );


	// Figure out the wx build-type
#ifdef __WXGTK__
#ifdef __WXGTK20__
	ver += wxT ( "wxGTK2" );
#else
	ver += wxT ( "wxGTK1" );
#endif
#elif defined(__WXMAC__)
	ver += wxT ( "wxMac" );
#elif defined(__WXMSW__)
	ver += wxT ( "wxMSW" );
#elif defined(__WXCOCOA__)
	ver += wxT ( "wxCocoa" );
#endif

	ver += wxString::Format ( wxT ( " v%d.%d.%d" ), wxMAJOR_VERSION, wxMINOR_VERSION, wxRELEASE_NUMBER );

#if wxUSE_UNICODE && defined(__WXDEBUG__)
	ver += wxT ( " (Unicoded, Debugging)" );
#elif wxUSE_UNICODE
	ver += wxT ( " (Unicoded)" );
#elif defined(__WXDEBUG__)
	ver += wxT ( " (Debugging)" );
#endif

#ifdef CVSDATE
	ver += wxString::Format ( wxT ( " (Snapshot: %s)" ), wxT ( CVSDATE ) );
#endif

	return ver;
}


wxString CimuleApp::GetFullMuleVersion()
{
#ifdef IMULE_DAEMON
	wxString app = wxT ( "imuled" );
#elif defined(CLIENT_GUI)
	wxString app = wxT ( "Remote imule-GUI" );
#else
	wxString app = wxT ( "imule" );
#endif

	return app + wxT ( " " ) + GetMuleVersion();
}





CimuleApp::CimuleApp()

{
	// Madcat - Initialize timer as the VERY FIRST thing to avoid any issues later.
	// Kry - I love to init the vars on init, even before timer.
	StartTickTimer();

	// Initialization
	m_app_state       = APP_STATE_STARTING;
	
	m_connection_manager = new I2PConnectionManager();

	clientlist	= NULL;
	searchlist	= NULL;
	knownfiles	= NULL;
	serverlist	= NULL;
	serverconnect	= NULL;
	sharedfiles	= NULL;
#if INTERNAL_ROUTER
	i2prouter       = NULL;
#endif
	listensocket	= NULL;
	clientudp	= NULL;
	clientcredits	= NULL;
	friendlist	= NULL;
	downloadqueue	= NULL;
	uploadqueue	= NULL;
	ipfilter	= NULL;
	ECServerHandler = NULL;
	m_singleInstance = NULL;
	glob_prefs	= NULL;
	statistics = NULL;
	uploadBandwidthThrottler = NULL;
	core_timer = NULL;
	applog = NULL;

	//m_localip		= 0;
	m_localdest		= CI2PAddress::null;
	//I2P:m_dwPublicIP	= 0;
	m_dwPublicDest	= CI2PAddress::null;
	webserver_pid	= 0;

	enable_stdout_log = false;
	enable_daemon_fork = false;


	// Apprently needed for *BSD
	SetResourceLimits();
	networkStarted = false ;
	theApp.networkStarting = false ;
}

CimuleApp::~CimuleApp()
{
	// Closing the log-file as the very last thing, since
	// wxWidgets log-events are saved in it as well.
	delete applog;
	applog = NULL;
}

int CimuleApp::OnExit()
{
	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "Now, exiting main app...\n" );
	}

	// From wxWidgets docs, wxConfigBase:
	// ...
	// Note that you must delete this object (usually in wxApp::OnExit)
	// in order to avoid memory leaks, wxWidgets won't do it automatically.
	//
	// As it happens, you may even further simplify the procedure described
	// above: you may forget about calling Set(). When Get() is called and
	// there is no current object, it will create one using Create() function.
	// To disable this behaviour DontCreateOnDemand() is provided.
	delete wxConfigBase::Set ( ( wxConfigBase * ) NULL );

	// Save credits
	clientcredits->SaveList();

	// Kill imuleweb if running
	if ( webserver_pid )
	{
		printf ( "Killing imuleweb instance...\n" );
		wxKillError rc;
		wxKill ( webserver_pid, wxSIGKILL, &rc );
		printf ( "Killed!\n" );
	}


	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: Terminating core.\n" );
	}

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: serverlist.\n" );
	}

	delete serverlist;

	serverlist = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: searchlist.\n" );
	}

	delete searchlist;

	searchlist = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: clientcredits.\n" );
	}

	delete clientcredits;

	clientcredits = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: friendlist.\n" );
	}

	delete friendlist;

	friendlist = NULL;

	// Destroying CDownloadQueue calls destructor for CPartFile
	// calling CSharedFileList::SafeAddKFile occasionally.

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: sharedfiles.\n" );
	}

	delete sharedfiles;

	sharedfiles = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: serverconnect.\n" );
	}

	delete serverconnect;

	serverconnect = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: listensocket.\n" );
	}

	delete listensocket;

	listensocket = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: clientudp.\n" );
	}

	delete clientudp;

	clientudp = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: knownfiles.\n" );
	}

	delete knownfiles;

	knownfiles = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: clientlist.\n" );
	}

	delete clientlist;

	clientlist = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: uploadqueue.\n" );
	}

	delete uploadqueue;

	uploadqueue = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: downloadqueue.\n" );
	}

	delete downloadqueue;

	downloadqueue = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: ipfilter.\n" );
	}

	delete ipfilter;

	ipfilter = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: ECServerHandler.\n" );
	}

	delete ECServerHandler;

	ECServerHandler = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: statistics.\n" );
	}

	delete statistics;

	statistics = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: glob_prefs.\n" );
	}

	delete glob_prefs;

	glob_prefs = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: CPreferences::EraseItemList.\n" );
	}

	CPreferences::EraseItemList();

#if defined(__WXMAC__) && defined(IMULE_DAEMON)
#warning TODO: fix wxSingleInstanceChecker for imuled on Mac (wx link problems)
#else

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: m_singleInstance.\n" );
	}

	delete m_singleInstance;

#endif
	m_singleInstance = NULL;

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule OnExit: uploadBandwidthThrottler.\n" );
	}

	delete uploadBandwidthThrottler;

	uploadBandwidthThrottler = NULL;

#if INTERNAL_ROUTER

	if ( m_app_state != APP_STATE_STARTING )
	{
		if ( i2prouter != NULL )
			printf ( "imule OnExit: i2prouter.\n" );
	}

	if ( i2prouter != NULL ) delete i2prouter;

	i2prouter = NULL;

#endif

	if ( m_app_state != APP_STATE_STARTING )
	{
		printf ( "imule shutdown completed.\n" );
	}

#if wxUSE_MEMORY_TRACING
	printf ( "Memory debug results for imule exit:\n" );

	// Log mem debug mesages to wxLogStderr
	wxLog* oldLog = wxLog::SetActiveTarget ( new wxLogStderr );


	/*
	printf("**************Classes**************\n");
	wxDebugContext::PrintClasses();
	*/
	//printf("***************Dump***************\n");
	//wxDebugContext::Dump();
	printf ( "***************Stats**************\n" );

	wxDebugContext::PrintStatistics ( true );

	// Set back to wxLogGui
	delete wxLog::SetActiveTarget ( oldLog );

#endif

	StopTickTimer();

	// Return 0 for succesful program termination
//	CI2PClient::DetachThread();
	return IMULE_APP_BASE::OnExit();
}


int CimuleApp::InitGui ( bool, wxString & )
{
	return 0;
}


/**
 * Checks permissions on a imule directory, creating if needed.
 *
 * @param desc A description of the directory in question, used for error messages.
 * @param directory The directory in question.
 * @param fallback If the dir specified with 'directory' could not be created, try this instead.
 * @return The bool is false on error. The wxString contains the used path.
 */

std::pair<bool, wxString> CheckMuleDirectory ( const wxString& desc, const wxString& dir, const wxString& fallback )
{
	wxFileName directory;
	directory.AssignDir(dir) ;
	
	if ( not ( directory.DirExists() ) )
	{
		if ( not directory.Mkdir(0777, wxPATH_MKDIR_FULL) )
		{
			wxString msg = wxString()
			               << wxT ( "Could not create the imule " ) << desc << wxT ( " directory\n" )
			               << wxT ( "at location '" ) << dir << wxT ( "'.\n" );

			// For the temp and incoming folder, we may be able to fall back
			// to the default values, in case of bad overrides.

			if ( fallback.Length() and ( dir != fallback ) )
			{
				wxFileName fb(fallback);
				if ( fb.DirExists() or fb.Mkdir(0777,wxPATH_MKDIR_FULL) )
				{
					msg << wxT ( "Using default directory at location \n'" )
					<< fallback << wxT ( "'." );
					theApp.ShowAlert ( msg, wxT ( "FATAL ERROR" ), wxICON_ERROR | wxOK );

					return std::pair<bool, wxString> ( true, fallback );
				}
			}

			msg << wxT ( "Please check permissions and restart imule." );

			theApp.ShowAlert ( msg, wxT ( "FATAL ERROR" ), wxICON_ERROR | wxOK );

			return std::pair<bool, wxString> ( false, wxEmptyString );
		}
	}
	else if ( not ( directory.IsDirReadable() && directory.IsDirWritable() ) )
	{
		wxString msg;
		msg << wxT ( "Permissions on the imule " ) << desc << wxT ( " directory too strict!\n" )
		<< wxT ( "imule cannot proceed. To fix this, you must set read/write/exec\n" )
		<< wxT ( "permissions for the folder '" ) << dir << wxT ( "'" );

		theApp.ShowAlert ( msg, wxT ( "FATAL ERROR" ), wxICON_ERROR | wxOK );

		return std::pair<bool, wxString> ( false, wxEmptyString );
	}

	return std::pair<bool, wxString> ( true, dir );
}


//
// Static variables initialization and application initialization
//
const wxString CimuleApp::FullMuleVersion = GetFullMuleVersion();

const wxString CimuleApp::OSDescription = wxGetOsDescription();

char *CimuleApp::strFullMuleVersion = new char[GetFullMuleVersion().Length() +1];

char *CimuleApp::strOSDescription = new char[wxGetOsDescription().Length() +1];


bool CimuleApp::OnInit()
{
	wxSocketBase::Initialize();

	IMULE_APP_BASE::OnInit();

#if wxUSE_MEMORY_TRACING
	printf ( "Checkpoint set on app init for memory debug\n" );
	wxDebugContext::SetCheckpoint();
#endif

	// Forward wxLog events to CLogger
	wxLog::SetActiveTarget ( new CLoggerTarget );

	// no more for I2P
	//m_localip = StringHosttoUint32_t(::wxGetFullHostName());

#ifndef __WXMSW__
	// get rid of sigpipe
	signal ( SIGPIPE, SIG_IGN );
	// Handle sigint and sigterm
	signal ( SIGINT, OnShutdownSignal );
	signal ( SIGTERM, OnShutdownSignal );
#endif

	// Handle uncaught exceptions
	InstallMuleExceptionHandler();

	// This can't be on constructor or wx2.4.2 doesn't set it.
	SetVendorName ( wxT ( "MKVore" ) );
	SetAppName ( wxT ( "iMule" ) );

	ConfigDir = CConfigDir::GetConfigDir();

	strcpy ( strFullMuleVersion, ( const char * ) unicode2char ( FullMuleVersion ) );

	strcpy ( strOSDescription, ( const char * ) unicode2char ( OSDescription ) );

	OSType = OSDescription.BeforeFirst ( wxT ( ' ' ) );

	if ( OSType.IsEmpty() )
	{
		OSType = wxT ( "Unknown" );
	}

	// Parse cmdline arguments.
	wxCmdLineParser cmdline ( IMULE_APP_BASE::argc, IMULE_APP_BASE::argv );

	// Handle these arguments.
	cmdline.AddSwitch ( wxT ( "v" ), wxT ( "version" ), wxT ( "Displays the current version number." ) );

	cmdline.AddSwitch ( wxT ( "h" ), wxT ( "help" ), wxT ( "Displays this information." ) );

	cmdline.AddSwitch ( wxT ( "i" ), wxT ( "enable-stdin" ), wxT ( "Does not disable stdin." ) );

#ifdef IMULE_DAEMON
	cmdline.AddSwitch ( wxT ( "f" ), wxT ( "full-daemon" ), wxT ( "Fork to background." ) );

#else
	cmdline.AddOption ( wxT ( "geometry" ), wxEmptyString,
	                    wxT (	"Sets the geometry of the app.\n"
	                          "\t\t\t<str> uses the same format as standard X11 apps:\n"
	                          "\t\t\t[=][<width>{xX}<height>][{+-}<xoffset>{+-}<yoffset>]" ) );

#endif

	cmdline.AddSwitch ( wxT ( "d" ), wxT ( "disable-fatal" ), wxT ( "Does not handle fatal exception." ) );

	cmdline.AddSwitch ( wxT ( "o" ), wxT ( "log-stdout" ), wxT ( "Print log messages to stdout." ) );

	cmdline.AddSwitch ( wxT ( "r" ), wxT ( "reset-config" ), wxT ( "Resets config to default values." ) );

	// Show help on --help or invalid commands
	if ( cmdline.Parse() )
	{
		return false;
	}
	else if ( cmdline.Found ( wxT ( "help" ) ) )
	{
		cmdline.Usage();
		return false;
	}

	if ( !cmdline.Found ( wxT ( "disable-fatal" ) ) )
	{
#ifndef __WXMSW__
		// catch fatal exceptions
		wxHandleFatalExceptions ( true );
#endif
	}

	bool reset_config = cmdline.Found ( wxT ( "reset-config" ) );

	enable_stdout_log = cmdline.Found ( wxT ( "log-stdout" ) );
#ifdef IMULE_DAEMON
	enable_daemon_fork = cmdline.Found ( wxT ( "full-daemon" ) );
#else
	enable_daemon_fork = false;
#endif

	if ( enable_stdout_log )
	{
		if ( enable_daemon_fork )
		{
			printf ( "Daemon will fork to background - log to stdout disabled\n" );
			enable_stdout_log = false;
		}
		else
		{
			printf ( "Logging to stdout enabled\n" );
		}
	}

	if ( cmdline.Found ( wxT ( "version" ) ) )
	{
		printf ( "%s (OS: %s)\n",

		         ( const char* ) unicode2char ( GetFullMuleVersion() ),
		         ( const char* ) unicode2char ( OSType ) );

		return false;
	}

	// Default geometry of the GUI. Can be changed with a cmdline argument...
	bool geometry_enabled = false;

	wxString geom_string;

#ifndef IMULE_DAEMON
	if ( cmdline.Found ( wxT ( "geometry" ), &geom_string ) )
	{
		geometry_enabled = true;
	}

#endif


	printf ( "Initialising imule\n" );

	// Ensure that "~/.imule/" is accessible.
	if ( not CheckMuleDirectory ( wxT ( "configuration" ), ConfigDir, wxEmptyString ).first )
	{
		return false;
	}

	// Check for old imule configs, or for old lmule/xmule config if no imule configs found.
	if ( !reset_config )
	{
		// Check for old imule configs, or for old lmule/xmule config if no imule configs found.
		CheckConfig();
	}
	else
	{
		// Make a backup first.
		wxRemoveFile ( ConfigDir + wxT ( "imule.conf.backup" ) );
		wxRenameFile ( ConfigDir + wxT ( "imule.conf" ), ConfigDir + wxT ( "imule.conf.backup" ) );
		printf ( "Your settings have ben resetted to default values.\nOld config file has been saved as imule.conf.backup\n" );
	}


#if defined(__WXMAC__) && defined(IMULE_DAEMON)
#warning TODO: fix wxSingleInstanceChecker for imuled on Mac (wx link problems)
	printf ( "WARNING: The check for other instances is currently disabled in imuled.\n"
	         "Please make sure that no other instance of imule is running or your files might be corrupted.\n" );

#else
	printf ( "Checking if there is an instance already running...\n" );

	m_singleInstance = new wxSingleInstanceChecker ( wxT ( "muleLock" ), ConfigDir );

	if ( m_singleInstance->IsAnotherRunning() )
	{
		printf ( "There seems to be an instance of imule already running\n" );
		printf ( "If this is not the case, you may have to remove\n" );
		printf ( "the file ~/.iMule/muleLock, to allow imule to run.\n" );

		// This is very tricky. The most secure way to communicate is via ED2K links file
		wxTextFile ed2kFile ( ConfigDir + wxT ( "ED2KLinks" ) );

		if ( !ed2kFile.Exists() )
		{
			ed2kFile.Create();
		}

		if ( ed2kFile.Open() )
		{
			ed2kFile.AddLine ( wxT ( "RAISE_DIALOG" ) );
			ed2kFile.Write();

			printf ( "Attempting to raise current running instance.\n" );
		}
		else
		{
			printf ( "Failed to open 'ED2KFile', cannot signal running instance.\n" );
		}

		return false;
	}
	else
	{
		printf ( "No other instances are running.\n" );
	}

#endif

	// Close standard-input
	if ( !cmdline.Found ( wxT ( "enable-stdin" ) ) )
	{
		// The full daemon will close all std file-descriptors by itself,
		// so closing it here would lead to the closing on the first open
		// file, which is the logfile opened below

		if ( !enable_daemon_fork )
		{
			close ( 0 );
		}
	}

	// This creates the CFG file we shall use
	wxConfigBase* cfg = new wxFileConfig ( wxEmptyString, wxEmptyString, ConfigDir + wxT ( "imule.conf" ) );

	// Set the config object as the global cfg file
	wxConfig::Set ( cfg );

	// Make a backup of the log file
	wxString logfileName ( ConfigDir + wxT ( "logfile" ) );

	if ( wxFileName::FileExists ( logfileName ) )
	{
		wxString logBackupName ( ConfigDir + wxT ( "logfile.bak" ) );
		UTF8_MoveFile ( logfileName, logBackupName );
	}

	// Open the log file
	applog = new wxFFileOutputStream ( logfileName );

	if ( !applog->Ok() )
	{
		// use std err as last resolt to indicate problem
		fputs ( "ERROR: unable to open log file\n", stderr );
		delete applog;
		applog = NULL;
		// failure to open log is serious problem
		return false;
	}

	// Load Preferences
	CPreferences::BuildItemList ( ConfigDir );

	CPreferences::LoadAllItems ( wxConfigBase::Get() );

	glob_prefs = new CPreferences();

	std::pair<bool, wxString> checkResult;

	checkResult = CheckMuleDirectory ( wxT ( "temp" ), thePrefs::GetTempDir(), ConfigDir + wxT ( "Temp" ) );

	if ( checkResult.first )
	{
		thePrefs::SetTempDir ( checkResult.second );
	}
	else
	{
		return false;
	}

	checkResult = CheckMuleDirectory ( wxT ( "incoming" ), thePrefs::GetIncomingDir(), ConfigDir + wxT ( "Incoming" ) );

	if ( checkResult.first )
	{
		thePrefs::SetIncomingDir ( checkResult.second );
	}
	else
	{
		return false;
	}

	// Some sanity check
	if ( !thePrefs::UseTrayIcon() )
	{
		thePrefs::SetMinToTray ( false );
	}

	// Build the filenames for the two OS files
	SetOSFiles ( thePrefs::GetOSDir() );

	// Load localization settings
	Localize_mule();

	// Display notification on new version or first run
	wxTextFile vfile ( ConfigDir + wxT ( "lastversion" ) );

	wxString newMule ( wxT ( VERSION ) );

	if ( !wxFileExists ( vfile.GetName() ) )
	{
		vfile.Create();
	}

#ifndef __WXMSW__
	if ( getuid() == 0 )
	{
		wxString msg =
		    wxT ( "Warning! You are running imule as root.\n" )
		    wxT ( "Doing so is not recommended for security reasons,\n" )
		    wxT ( "and you are advised to run imule as an normal\n" )
		    wxT ( "user instead." );

		ShowAlert ( msg, _ ( "Warning" ), wxCENTRE | wxOK | wxICON_ERROR );

		fprintf ( stderr, "\n--------------------------------------------------\n" );

		fprintf ( stderr, "%s", ( const char* ) unicode2UTF8 ( msg ) );

		fprintf ( stderr, "\n--------------------------------------------------\n\n" );
	}

#endif


	if ( vfile.Open() )
	{
		// Check if this version has been run before
		bool found = false;

		for ( size_t i = 0; i < vfile.GetLineCount(); i++ )
		{
			// Check if this version has been run before

			if ( vfile.GetLine ( i ) == newMule )
			{
				found = true;
				break;
			}
		}

		// We havent run this version before?
		if ( !found )
		{
			// Insert new at top to provide faster searches
			vfile.InsertLine ( newMule, 0 );

			Trigger_New_version ( newMule );
		}

		// Keep at most 10 entires
		while ( vfile.GetLineCount() > 10 )
			vfile.RemoveLine ( vfile.GetLineCount() - 1 );

		vfile.Write();

		vfile.Close();
	}

	// Check if we have the old style locale config
	wxString langId = thePrefs::GetLanguageID();

	if ( !langId.IsEmpty() && ( langId.GetChar ( 0 ) >= '0' && langId.GetChar ( 0 ) <= '9' ) )
	{
		wxString info ( _ ( "Your locale has been changed to System Default due to a configuration change. Sorry." ) );
		thePrefs::SetLanguageID ( wxLang2Str ( wxLANGUAGE_DEFAULT ) );
		ShowAlert ( info, _ ( "Info" ), wxCENTRE | wxOK | wxICON_ERROR );
	}


	// start java VM for I2P objects
	wxSetWorkingDirectory ( CConfigDir::GetConfigDir() );

	statistics = new CStatistics();

	// Ready file-hasher
	CAddFileThread::Start();

	clientlist	= new CClientList();

	friendlist	= new CFriendList();

	searchlist	= new CSearchList();

	knownfiles	= new CKnownFileList();

	serverlist	= new CServerList();

	sharedfiles	= new CSharedFileList ( knownfiles );

	clientcredits	= new CClientCreditsList();

	downloadqueue	= new CDownloadQueue();

	uploadqueue	= new CUploadQueue();

	ipfilter	= new CIPFilter();

	// Create main dialog
	InitGui ( geometry_enabled, geom_string );

	uploadBandwidthThrottler = new UploadBandwidthThrottler();

	// Load and update IPFilter.dat files. Loading the
	// existing ipfilter.dat file should be done before
	// loading the server.met, to ensure that servers are
	// filtered.
	AddLogLineM ( false, _ ( "Loading ipfilter.dat files." ) );

	ipfilter->Reload();

	serverlist->Init();

	// init downloadqueue
	downloadqueue->LoadMetFiles ( thePrefs::GetTempDir() );

	// reload shared files
	sharedfiles->Reload ( true );


	if ( thePrefs::IPFilterAutoLoad() )
	{
		ipfilter->Update ( thePrefs::IPFilterURL() );
	}


	// Ensure that the up/down ratio is used
	CPreferences::CheckUlDlRatio();

	// The user can start pressing buttons like mad if he feels like it.
	m_app_state = APP_STATE_RUNNING;

	// Kry - Load the sources seeds on app init
	if ( thePrefs::GetSrcSeedsOn() )
	{
		downloadqueue->LoadSourceSeeds();
	}

	// open connections to I2P
	wxString msg ;
	ReinitializeNetwork( &msg ) ;
	if ( ! msg.IsEmpty() )
		AddLogLine( msg );

	return true;
}


/**
 * Finds the path from which the application was loaded
 */

wxString wxFindAppPath ( const wxString& argv0, const wxString& cwd, const wxString& appVariableName )
{
	wxString str;

	// Try appVariableName

	if ( !appVariableName.IsEmpty() )
	{
		str = wxGetenv ( appVariableName );

		if ( !str.IsEmpty() )
			return str;
	}

#if defined(__WXMAC__) && !defined(__DARWIN__)
	// On Mac, the current directory is the relevant one when
	// the application starts.
	return cwd;

#endif

	if ( wxIsAbsolutePath ( argv0 ) )
		return wxPathOnly ( argv0 );
	else
	{
		// Is it a relative path?
		wxString currentDir ( cwd );

		if ( currentDir.Last() != wxFILE_SEP_PATH )
			currentDir += wxFILE_SEP_PATH;

		str = currentDir + argv0;

		if ( wxFileExists ( str ) )
			return wxPathOnly ( str );
	}

	// OK, it's neither an absolute path nor a relative path.
	// Search PATH.

	wxPathList pathList;

	pathList.AddEnvList ( wxT ( "PATH" ) );

	str = pathList.FindAbsoluteValidPath ( argv0 );

	if ( !str.IsEmpty() )
		return wxPathOnly ( str );

	// Failed
	return wxEmptyString;
}



// Returns a ed2k file URL

wxString CimuleApp::CreateED2kLink ( const CAbstractFile *f )
{
	// Construct URL like this: ed2k://|file|<filename>|<size>|<hash>|/
	wxString strURL	= wxString ( wxT ( "ed2k://|file|" ) ) <<
	                  CleanupFilename ( f->GetFileName(), true ) << wxT ( "|" ) <<
	                  f->GetFileSize() << wxT ( "|" ) <<
	                  f->GetFileHash().Encode() << wxT ( "|/" );
	return strURL;
}

// Returns a ed2k source URL
wxString CimuleApp::CreateED2kSourceLink ( const CAbstractFile *f )
{
	if ( !IsConnected() /*|| IsFirewalled()*/ )
	{
		return wxEmptyString;
	}

	CI2PAddress clientDest = GetTcpDest();

	// Create the first part of the URL
	// And append the source information: "|sources,<ip>:<port>|/"
	wxString strURL = CreateED2kLink ( f ) <<
	                  wxT ( "|sources," ) << clientDest.toString() << wxT ( "|/" );
	// Result is "ed2k://|file|<filename>|<size>|<hash>|/|sources,<ip>:<port>|/"
	return strURL;
}

// Returns a ed2k link with AICH info if available
wxString CimuleApp::CreateED2kAICHLink ( const CKnownFile* f )
{
	// Create the first part of the URL
	wxString strURL = CreateED2kLink ( f );
	// Append the AICH info

	if ( f->GetAICHHashset()->HasValidMasterHash() &&
	        (
	            f->GetAICHHashset()->GetStatus() == AICH_VERIFIED ||
	            f->GetAICHHashset()->GetStatus() == AICH_HASHSETCOMPLETE
	        ) )
	{
		strURL << wxT ( "|h=" ) << f->GetAICHHashset()->GetMasterHash().GetString();
	}

	strURL << wxT ( "|/" );

	// Result is "ed2k://|file|<filename>|<size>|<hash>|/|h=<AICH master hash>|/"
	return strURL;
}

// Returns a ed2k source URL using a hostname rather than IP. Currently, the
// hostname doesn't appear to be set, thus this function wont work as intended.
wxString CimuleApp::CreateED2kHostnameSourceLink ( const CAbstractFile* f )
{
	// Create the first part of the URL
	// Append the source information: "|sources,<host>:port|/"
	return CreateED2kSourceLink ( f );
	/*
	wxString strURL = CreateED2kLink(f) <<
		wxT("|sources,") <<
		thePrefs::GetYourHostname() << wxT(":") <<
		thePrefs::GetPort() << wxT("|/");
	// Result is "ed2k://|file|<filename>|<size>|<hash>|/|sources,<host>:<port>|/"
	return strURL;*/
}


/* Original implementation by Bouc7 of the eMule Project.
   imule Signature idea was designed by BigBob and implemented
   by Un-Thesis, with design inputs and suggestions from bothie.
*/
void CimuleApp::OnlineSig ( bool zero /* reset stats (used on shutdown) */ )
{
	// Do not do anything if online signature is disabled in Preferences

	if ( !thePrefs::IsOnlineSignatureEnabled() || m_emulesig_path.IsEmpty() )
	{
		// We do not need to check m_imulesig_path because if m_emulesig_path is empty,
		// that means m_imulesig_path is empty too.
		return;
	}

	// Remove old signature files
	if ( wxFileExists ( m_emulesig_path ) ) { wxRemoveFile ( m_emulesig_path ); }

	if ( wxFileExists ( m_imulesig_path ) ) { wxRemoveFile ( m_imulesig_path ); }


	wxTextFile imulesig_out;

	wxTextFile emulesig_out;

	// Open both files if needed

	if ( !emulesig_out.Create ( m_emulesig_path ) )
	{
		AddLogLineM ( true, _ ( "Failed to create OnlineSig File" ) );
		// Will never try again.
		m_imulesig_path.Clear();
		m_emulesig_path.Clear();
		return;
	}

	if ( !imulesig_out.Create ( m_imulesig_path ) )
	{
		AddLogLineM ( true, _ ( "Failed to create imule OnlineSig File" ) );
		// Will never try again.
		m_imulesig_path.Clear();
		m_emulesig_path.Clear();
		return;
	}

	wxString emulesig_string;

	wxString temp;

	if ( zero )
	{
		emulesig_string = wxT ( "0\xA0.0|0.0|0" );
		imulesig_out.AddLine ( wxT ( "0\n0\n0\n0\n0\n0\n0.0\n0.0\n0\n0" ) );
	}
	else
	{
		if ( IsConnectedED2K() )
		{

			// temp = wxString::Format(wxT("%d"),serverconnect->GetCurrentServer()->GetPort()); //I2P
			temp = serverconnect->GetCurrentServer()->GetDest().toString();

			// We are online
			emulesig_string =
			    // Connected
			    wxT ( "1|" )
			    //Server name
			    + serverconnect->GetCurrentServer()->GetListName()
			    + wxT ( "|" )
			    // IP and port of the server
			    + serverconnect->GetCurrentServer()->GetDest().toString() //I2P : GetFullIP()
			    + wxT ( "|" )
			    + temp;


			// Now for imule sig

			// Connected. State 1, full info
			imulesig_out.AddLine ( wxT ( "1" ) );
			// Server Name
			imulesig_out.AddLine ( serverconnect->GetCurrentServer()->GetListName() );
			// Server IP
			imulesig_out.AddLine ( serverconnect->GetCurrentServer()->GetDest().toString() );//I2P:IP());
			// Server Port
			imulesig_out.AddLine ( temp );

			/*if ( serverconnect->IsLowID() )
			{
				imulesig_out.AddLine ( wxT ( "L" ) );
			}
			else
			{
				imulesig_out.AddLine ( wxT ( "H" ) );
			}*/

		}
		else if ( serverconnect->IsConnecting() )
		{
			emulesig_string = wxT ( "0" );

			// Connecting. State 2, No info.
			imulesig_out.AddLine ( wxT ( "2\n0\n0\n0\n0" ) );
		}
		else
		{
			// Not connected to a server
			emulesig_string = wxT ( "0" );

			// Not connected, state 0, no info
			imulesig_out.AddLine ( wxT ( "0\n0\n0\n0\n0" ) );
		}

		if ( IsConnectedKad() )
		{
			//if ( Kademlia::CKademlia::isFirewalled() )
			//{
			//	// Connected. Firewalled. State 1.
			//	imulesig_out.AddLine ( wxT ( "1" ) );
			//}
			//else
			//{
			// Connected. State 2.
			imulesig_out.AddLine ( wxT ( "2" ) );
			//}
		}
		else
		{
			// Not connected.State 0.
			imulesig_out.AddLine ( wxT ( "0" ) );
		}

		emulesig_string += wxT ( "\xA" );

		// Datarate for downloads
		temp = wxString::Format ( wxT ( "%.1f" ), theStats::GetDownloadRate() / 1024.0 );

		emulesig_string += temp + wxT ( "|" );
		imulesig_out.AddLine ( temp );

		// Datarate for uploads
		temp = wxString::Format ( wxT ( "%.1f" ), theStats::GetUploadRate() / 1024.0 );

		emulesig_string += temp + wxT ( "|" );
		imulesig_out.AddLine ( temp );

		// Number of users waiting for upload
		temp = wxString::Format ( wxT ( "%d" ), theStats::GetWaitingUserCount() );

		emulesig_string += temp;
		imulesig_out.AddLine ( temp );

		// Number of shared files (not on eMule)
		imulesig_out.AddLine ( wxString::Format ( wxT ( "%s" ), theStats::GetSharedFileCount().ToString().c_str() ) );
	}

	// eMule signature finished here. Write the line to the wxTextFile.
	emulesig_out.AddLine ( emulesig_string );

	// Now for imule signature extras

	// Nick on the network
	imulesig_out.AddLine ( thePrefs::GetUserNick() );

	// Total received in bytes
	imulesig_out.AddLine ( CFormat ( wxT ( "%s" ) ) % ( theStats::GetSessionReceivedBytes() + thePrefs::GetTotalDownloaded() ).ToString() );

	// Total sent in bytes
	imulesig_out.AddLine ( CFormat ( wxT ( "%s" ) ) % ( theStats::GetSessionSentBytes() + thePrefs::GetTotalUploaded() ).ToString() );

	// imule version
#ifdef CVSDATE
	imulesig_out.AddLine ( wxT ( VERSION " " CVSDATE ) );

#else
	imulesig_out.AddLine ( wxT ( VERSION ) );

#endif

	if ( zero )
	{
		imulesig_out.AddLine ( wxT ( "0" ) );
		imulesig_out.AddLine ( wxT ( "0" ) );
		imulesig_out.AddLine ( wxT ( "0" ) );
	}
	else
	{
		// Total received bytes in session
		imulesig_out.AddLine ( CFormat ( wxT ( "%s" ) ) % theStats::GetSessionReceivedBytes().ToString() );

		// Total sent bytes in session
		imulesig_out.AddLine ( CFormat ( wxT ( "%s" ) ) % theStats::GetSessionSentBytes().ToString() );

		// Uptime
		imulesig_out.AddLine ( CFormat ( wxT ( "%s" ) ) % theStats::GetUptimeSeconds().ToString() );
	}

	// Flush the files
	emulesig_out.Write();

	imulesig_out.Write();
} //End Added By Bouc7


// Gracefully handle fatal exceptions and print backtrace if possible
void CimuleApp::OnFatalException()
{
	/* Print the backtrace */
	fprintf ( stderr, "\n--------------------------------------------------------------------------------\n" );
	fprintf ( stderr, "A fatal error has occurred and imule has crashed.\n" );
	fprintf ( stderr, "Please assist us in fixing this problem by posting the backtrace below in our\n" );
	fprintf ( stderr, "'imule Crashes' forum and include as much information as possible regarding the\n" );
	fprintf ( stderr, "circumstances of this crash. The forum is located here:\n" );
	fprintf ( stderr, "    http://forum.amule.org/board.php?boardid=67\n" );
	fprintf ( stderr, "If possible, please try to generate a real backtrace of this crash:\n" );
	fprintf ( stderr, "    http://www.imule.i2p/wiki/index.php/Backtraces\n\n" );
	fprintf ( stderr, "----------------------------=| BACKTRACE FOLLOWS: |=----------------------------\n" );
	fprintf ( stderr, "Current version is: %s\n", strFullMuleVersion );
	fprintf ( stderr, "Running on: %s\n\n", strOSDescription );

	print_backtrace ( 1 ); // 1 == skip this function.

	fprintf ( stderr, "\n--------------------------------------------------------------------------------\n" );
}


// Sets the localization of imule
void CimuleApp::Localize_mule()
{
	InitCustomLanguages();
	InitLocale ( m_locale, StrLang2wx ( thePrefs::GetLanguageID() ) );

	if ( !m_locale.IsOk() )
	{
		AddLogLineM ( false, _ ( "The selected locale seems not to be installed on your box."
		                         " (Note: I'll try to set it anyway)" ) );
	}
}


// Displays information related to important changes in imule.
// Is called when the user runs a new version of imule
void CimuleApp::Trigger_New_version ( wxString new_version )
{
	wxString info = wxT ( " --- " ) + CFormat ( _ ( "This is the first time you run imule %s" ) ) % new_version + wxT ( " ---\n\n" );

	info += wxT ( "\n" );

#if INTERNAL_ROUTER
	info += _ ( 
	"If you are running the java I2P router,\n"
	"\n"
	"(1) PLEASE DISABLE the \"internal I2P router\"\n"
	"     in iMule Preferences tab !\n"
	"\n"
	"(2) ENABLE the SAMBridge client in your router configuration tab\n"
	"     (probably by browsing http://localhost:7657/configclients.jsp)\n"
	"     (you may have to restart your I2P router after this change)\n" );
#else
	info += _ ( 
	"This version does NOT include an internal I2P router.\n"
	"In order to run iMule, you have to install and run the java I2P router.\n"
	"You can fetch it here :\n"
	"\n"
	" http://www.i2p2.de/download\n"
	"\n"
	"PLEASE ENABLE the SAMBridge client in your router configuration tab\n"
	"     (probably by browsing http://localhost:7657/configclients.jsp)\n"
	"     (you may have to restart your I2P router after this change)\n" );
#endif
	
	if ( new_version == wxT ( "CVS" ) )
	{
		info += wxT ( "\n" );
		info += _ ( "This version is a testing version, updated daily, and\n" );
		info += _ ( "we give no warranty it won't break anything, burn your house,\n" );
		info += _ ( "or kill your dog. But it *should* be safe to use anyway.\n" );
	}

	// General info
	info += wxT ( "\n" );

	info += _ ( "More information, support and new releases can found at our homepage,\n" );

	info += _ ( "at www.imule.i2p.\n" );

	info += wxT ( "\n" );

	info += _ ( "Feel free to report any bugs to http://forum.i2p/viewforum.php?f=30" );

	ShowAlert ( info, _ ( "Info" ), wxCENTRE | wxOK | wxICON_ERROR );
}


void CimuleApp::SetOSFiles ( const wxString new_path )

{
	if ( thePrefs::IsOnlineSignatureEnabled() )
	{
		if ( ::wxDirExists ( new_path ) )
		{
			m_emulesig_path = JoinPaths ( new_path, wxT ( "onlinesig.dat" ) );
			m_imulesig_path = JoinPaths ( new_path, wxT ( "imulesig.dat" ) );
		}
		else
		{
			ShowAlert ( _ ( "The folder for Online Signature files you specified is INVALID!\n OnlineSignature will be DISABLED until you fix it on preferences." ), _ ( "Error" ), wxOK | wxICON_ERROR );
			m_emulesig_path.Clear();
			m_imulesig_path.Clear();
		}
	}
	else
	{
		m_emulesig_path.Clear();
		m_imulesig_path.Clear();
	}
}


#ifdef __WXDEBUG__
#ifndef wxUSE_STACKWALKER
#define wxUSE_STACKWALKER 0
#endif

void CimuleApp::OnAssert ( const wxChar *file, int line,
                           const wxChar *cond, const wxChar *msg )
{
	if ( !wxUSE_STACKWALKER || !wxThread::IsMain() || !IsRunning() )
	{
		wxString errmsg = CFormat ( wxT ( "%s:%d: Assertion '%s' failed. %s" ) )
		                  % file % line % cond % ( msg ? msg : wxT ( "" ) );

		fprintf ( stderr, "Assertion failed: %s\n", ( const char* ) unicode2char ( errmsg ) );

		// Skip the function-calls directly related to the assert call.
		fprintf ( stderr, "\nBacktrace follows:\n" );

		print_backtrace ( 3 );

		fprintf ( stderr, "\n" );
	}

	if ( wxThread::IsMain() && IsRunning() )
	{
		IMULE_APP_BASE::OnAssert ( file, line, cond, msg );
	}
	else
	{
		// Abort, allows gdb to catch the assertion
		raise ( SIGABRT );
	}
}

#endif


void CimuleApp::OnUDPDnsDone ( CMuleInternalEvent& evt )
{
	CServerUDPSocket* socket = ( CServerUDPSocket* ) evt.GetClientData();
	socket->OnHostnameResolved ( CI2PAddress::fromString ( evt.GetString() ) /*ExtraLong()*/ );
}


void CimuleApp::OnSourceDnsDone ( CMuleInternalEvent& evt )

{
	//downloadqueue->OnHostnameResolved(evt.GetExtraLong());
	downloadqueue->OnHostnameResolved ( CI2PAddress::fromString ( evt.GetString() ) );
}


void CimuleApp::OnServerDnsDone ( CMuleInternalEvent& evt )
{
	printf ( "Server hostname notified\n" );
	CServerSocket* socket = ( CServerSocket* ) evt.GetClientData();
	socket->OnHostnameResolved ( CI2PAddress::fromString ( evt.GetString() /*GetExtraLong()*/ ) );
}


void CimuleApp::OnNotifyEvent ( wxEvent& e )
{
	NotifyEvent ( dynamic_cast<GUIEvent&> ( e ) );
}


void CimuleApp::OnTCPTimer ( CMuleInternalEvent& WXUNUSED ( evt ) )
{
	if ( !IsRunning() )
	{
		return;
	}

	serverconnect->StopConnectionTry();

	if ( IsConnectedED2K() )
	{
		return;
	}

	serverconnect->ConnectToAnyServer();
}

#if INTERNAL_ROUTER
void CimuleApp::UpdateRouterStatus()
{
	wxString info ;
	if ( i2prouter != NULL && networkStarted )
	{
		info =  i2prouter->getInfo() ;
	}
	else if (!thePrefs::GetI2PServerInternal())
	{
		
		info = _("<center><b>You are not running the internal I2P router</b></center>") ;
	}
	else
	{
		info = _("<center><b>The network is not ready.</b></center>");
	}
	UpdateRouterStatus ( info );
}

#ifndef CLIENT_GUI
void CimuleGuiApp::UpdateRouterStatus()
{
	CimuleApp::UpdateRouterStatus();
}
#endif // CLIENT_GUI

#endif // INTERNAL_ROUTER



void CimuleApp::OnCoreTimer ( CMuleInternalEvent& WXUNUSED ( evt ) )
{
	// Former TimerProc section

	static wxULongLong	msPrev1, msPrev5, msPrevSave, msPrevHist, msPrevOS, msPrevKnownMet;

	wxULongLong msCur = theStats::GetUptimeMillis();

	if ( !IsRunning() )
	{
		return;
	}

#ifndef IMULE_DAEMON
	// Check if we should terminate the app
	if ( g_shutdownSignal )
	{
		wxWindow* top = GetTopWindow();

		if ( top )
		{
			top->Close ( true );
		}
		else
		{
			// No top-window, have to force termination.
			wxExit();
		}
	}

#endif

	CLogger::FlushPendingEntries();

//	if ( networkStarted )
	{
		uploadqueue->Process();

		downloadqueue->Process();
	}

	//theApp.clientcredits->Process();
	theStats::CalculateRates();

	if ( msCur - msPrevHist > 1000 )
	{
		// unlike the other loop counters in this function this one will sometimes
		// produce two calls in quick succession (if there was a gap of more than one
		// second between calls to TimerProc) - this is intentional!  This way the
		// history list keeps an average of one node per second and gets thinned out
		// correctly as time progresses.
		msPrevHist += 1000;

		statistics->RecordHistory();

	}


	if ( networkStarted && ( msCur - msPrev1 > 1000 ) )
	{  // approximately every second
		msPrev1 = msCur;
		clientcredits->Process();
		clientlist->Process();

		// Publish files if needed.
		sharedfiles->Process();

		if ( Kademlia::CKademlia::isRunning() )
		{
			Kademlia::CKademlia::process();

			if ( Kademlia::CKademlia::getPrefs()->hasLostConnection() )
			{ // restart Kademlia
				StopKad();
				StartKad();
			}
		}

// 		if ( serverconnect->IsConnecting() && !serverconnect->IsSingleConnect() )
// 		{
// 			serverconnect->TryAnotherConnectionrequest();
// 		}
// 
// 		if ( serverconnect->IsConnecting() )
// 		{
// 			serverconnect->CheckForTimeout();
// 		}

		listensocket->UpdateConnectionsStatus();
	}


	if ( msCur - msPrev5 > 5000 )
	{  // every 5 seconds
		msPrev5 = msCur;

		if ( networkStarted )
		{
			listensocket->Process();
		}
	}

	if ( msCur - msPrevSave >= 60000 )
	{
		msPrevSave = msCur;
		wxString buffer;

		// Save total upload/download to preferences
		wxConfigBase* cfg = wxConfigBase::Get();
		buffer = CFormat ( wxT ( "%s" ) ) % ( theStats::GetSessionReceivedBytes() + thePrefs::GetTotalDownloaded() ).ToString();
		cfg->Write ( wxT ( "/Statistics/TotalDownloadedBytes" ), buffer );

		buffer = CFormat ( wxT ( "%s" ) ) % ( theStats::GetSessionSentBytes() + thePrefs::GetTotalUploaded() ).ToString();
		cfg->Write ( wxT ( "/Statistics/TotalUploadedBytes" ), buffer );

		// Write changes to file
		cfg->Flush();

	}

	// Special
	if ( msCur - msPrevOS >= thePrefs::GetOSUpdate() * 1000 )
	{
		OnlineSig(); // Added By Bouc7
		msPrevOS = msCur;
	}

	if ( msCur - msPrevKnownMet >= 30*60*1000/*There must be a prefs option for this*/ )
	{
		// Save Shared Files data
		knownfiles->Save();
		msPrevKnownMet = msCur;
	}


	// Recomended by lugdunummaster himself - from emule 0.30c
	if ( networkStarted ) serverconnect->KeepConnectionAlive();

}


void CimuleApp::OnHashingShutdown ( CMuleInternalEvent& WXUNUSED ( evt ) )

{
	if ( m_app_state != APP_STATE_SHUTINGDOWN )
	{
		// Save the known.met file
		knownfiles->Save();

		// Known.met changed, AICH sync thread start
		RunAICHThread();
	}
}


void CimuleApp::OnFinishedHashing ( CMuleInternalEvent& evt )
{
	CKnownFile* result = ( CKnownFile* ) evt.GetClientData();

	if ( evt.GetExtraLong() )
	{
		CPartFile* requester = ( CPartFile* ) evt.GetExtraLong();

		if ( downloadqueue->IsPartFile ( requester ) )
		{
			requester->PartFileHashFinished ( result );
		}
	}
	else
	{
		static int filecount;

		static wxULongLong bytecount;

		if ( knownfiles->SafeAddKFile ( result ) )
		{
			AddDebugLogLineM ( false, logKnownFiles, wxT ( "Safe adding file to sharedlist: " ) + result->GetFileName() );
			sharedfiles->SafeAddKFile ( result );

			filecount++;
			bytecount += result->GetFileSize();
			// If we have added 30 files or files with a total size of ~300mb

			if ( ( filecount == 30 ) || ( bytecount >= 314572800 ) )
			{
				AddDebugLogLineM ( false, logKnownFiles, wxT ( "Failsafe for crash on file hashing creation" ) );

				if ( m_app_state != APP_STATE_SHUTINGDOWN )
				{
					knownfiles->Save();
					filecount = 0;
					bytecount = 0;
				}
			}
		}
		else
		{
			AddDebugLogLineM ( false, logKnownFiles, wxT ( "File not added to sharedlist: " ) + result->GetFileName() );
			delete result;
		}
	}
}


void CimuleApp::OnFinishedCompletion ( CMuleInternalEvent& evt )
{
	CPartFile* completed = ( CPartFile* ) evt.GetClientData();
	wxCHECK_RET ( completed, wxT ( "Completion event sent for unspecified file" ) );
	completed->CompleteFileEnded ( evt.GetInt(), ( wxString* ) evt.GetExtraLong() );

	// Check if we should execute an script/app/whatever.

	if ( thePrefs::CommandOnCompletion() and not evt.GetInt() )
	{
		wxString command = thePrefs::GetCommandOnCompletion();

		command.Replace ( wxT ( "%FILE" ), completed->GetFullName() );
		command.Replace ( wxT ( "%HASH" ), completed->GetFileHash().Encode() );

		if ( ::wxExecute ( command, wxEXEC_ASYNC ) == 0 )
		{
			AddLogLineM ( true, CFormat ( _ ( "Failed to execute on-completion command. Template is: %s" ) ) % command );
		}
	}
}


void CimuleApp::ShutDown()
{
	// Signal the hashing thread to terminate
	m_app_state = APP_STATE_SHUTINGDOWN;

	StopKad();

	// Kry - Save the sources seeds on app exit

	if ( thePrefs::GetSrcSeedsOn() )
	{
		downloadqueue->SaveSourceSeeds();
	}

	OnlineSig ( true ); // Added By Bouc7

	// Close sockets to avoid new clients coming in

	if ( listensocket )
	{
		listensocket->Close();
		listensocket->KillAllSockets();
	}

	if ( serverconnect )
	{
		serverconnect->Disconnect();
	}

	// saving data & stuff
	if ( knownfiles )
	{
		knownfiles->Save();
	}

	thePrefs::Add2TotalDownloaded ( theStats::GetSessionReceivedBytes() );

	thePrefs::Add2TotalUploaded ( theStats::GetSessionSentBytes() );

	if ( glob_prefs )
	{
		glob_prefs->Save();
	}

	if ( clientlist )
	{
		clientlist->DeleteAll();
	}

	if ( CAddFileThread::IsRunning() )
	{
		CAddFileThread::Stop();
	}

	if ( CAICHSyncThread::IsRunning() )
	{
		CAICHSyncThread::Stop();
	}

	theApp.uploadBandwidthThrottler->EndThread();
}

bool CimuleApp::AddServer ( CServer *srv, bool fromUser )
{
	if ( serverlist->AddServer ( srv, fromUser ) )
	{
		Notify_ServerAdd ( srv );
		return true;
	}

	return false;
}


void CimuleApp::AddLogLine ( const wxString &msg )
{
	// At most one trailing new-line, which we add
	wxString message = msg;

	while ( !message.IsEmpty() && message.Last() == wxT ( '\n' ) )
	{
		message.RemoveLast();
	}

	wxString full_line = wxDateTime::Now().FormatISODate() + wxT ( " " ) +

	                     wxDateTime::Now().FormatISOTime() + wxT ( ": " ) + message + wxT ( "\n" );

	wxStringInputStream stream ( full_line );

	( *applog ) << stream;
	applog->Sync();

	if ( enable_stdout_log )
	{

		printf ( "%s", ( const char* ) unicode2char ( full_line ) );
	}
}


CI2PAddress CimuleApp::GetPublicDest ( bool ignorelocal ) const
{
	if ( !m_dwPublicDest )
	{
		if ( Kademlia::CKademlia::isConnected() && Kademlia::CKademlia::getDestAddress().isValid() )
		{
			return Kademlia::CKademlia::getDestAddress();
		}
		else
		{
			return ignorelocal ? CI2PAddress::null : m_localdest;
		}
	}

	return m_dwPublicDest;
}


void CimuleApp::SetPublicDest ( const CI2PAddress & dwDest )

{
	//wxASSERT ( ( !dwDest ) or !IsLowID ( dwDest ) );
	wxASSERT ( dwDest.isValid() );
	m_dwPublicDest = dwDest;
}

wxString CimuleApp::GetLog ( bool reset )

{
	ConfigDir = CConfigDir::GetConfigDir();
	wxFile logfile;
	logfile.Open ( ConfigDir + wxT ( "logfile" ) );

	if ( !logfile.IsOpened() )
	{
		return wxTRANSLATE ( "ERROR: can't open logfile" );
	}

	wxFileOffset len_check = logfile.Length();

	if ( len_check == 0 )
	{
		return wxTRANSLATE ( "WARNING: logfile is empty. Something is wrong." );
	}

	wxASSERT( len_check <= std::numeric_limits<size_t>::max() - sizeof(wxChar) );

	size_t len = (size_t) len_check ;
	char *tmp_buffer = new char[(size_t) len + sizeof ( wxChar ) ];

	logfile.Read ( tmp_buffer, len );
	memset ( tmp_buffer + len, 0, sizeof ( wxChar ) );
#if wxUSE_UNICODE
	// try to guess file format
	wxString str;

	if ( tmp_buffer[0] && tmp_buffer[1] )
	{
		str = wxString ( UTF82unicode ( tmp_buffer ) );
	}
	else
	{
		str = wxString ( ( wxWCharBuffer& ) tmp_buffer );
	}

#else
	wxString str ( tmp_buffer );

#endif
	delete [] tmp_buffer;

	if ( reset )
	{
		delete applog;
		applog = new wxFFileOutputStream ( ConfigDir + wxT ( "logfile" ) );

		if ( applog->Ok() )
		{
			AddLogLine ( _ ( "Log has been reset" ) );
		}
		else
		{
			delete applog;
			applog = 0;
		}
	}

	return str;
}


wxString CimuleApp::GetServerLog ( bool reset )
{
	wxString ret = server_msg;

	if ( reset )
	{
		server_msg.Clear();
	}

	return ret;
}

wxString CimuleApp::GetDebugLog ( bool reset )
{
	return GetLog ( reset );
}


void CimuleApp::AddServerMessageLine ( wxString &msg )
{
	server_msg += msg + wxT ( "\n" );
	AddLogLine ( CFormat ( _ ( "ServerMessage: %s" ) ) % msg );
}


void CimuleApp::RunAICHThread()
{
	if ( !CAICHSyncThread::IsRunning() )
		CAICHSyncThread::Start();
}

void CimuleApp::OnFinishedHTTPDownload ( CMuleInternalEvent& event )
{
	switch ( event.GetInt() )
	{
		case HTTP_IPFilter:
			ipfilter->DownloadFinished ( event.GetExtraLong() );

			break;

		case HTTP_ServerMet:
			serverlist->DownloadFinished ( event.GetExtraLong() );

			break;

		case HTTP_ServerMetAuto:
			serverlist->AutoDownloadFinished ( event.GetExtraLong() );

			break;

		case HTTP_VersionCheck:
			CheckNewVersion ( event.GetExtraLong() );

			break;

		case HTTP_NodesDat:
			if ( event.GetExtraLong() != -1 )
			{

				wxString file = ConfigDir + wxT ( "nodes.dat" );

				if ( wxFileExists ( file ) )
				{
					wxRemoveFile ( file );
				}

				if ( Kademlia::CKademlia::isRunning() )
				{
					Kademlia::CKademlia::stop();
				}

				wxRenameFile ( file + wxT ( ".download" ), file );

				Kademlia::CKademlia::start();
				theApp.ShowConnectionState();

			}
			else
			{
				AddLogLineM ( true, _ ( "Failed to download the nodes list." ) );
			}

			break;
	}
}

void CimuleApp::CheckNewVersion ( uint32_t result )
{
	if ( result == 1 )
	{
		wxString filename = ConfigDir + wxT ( "last_version_check" );
		wxTextFile file;

		if ( !file.Open ( filename ) )
		{
			AddLogLineM ( true, _ ( "Failed to open the downloaded version check file" ) );
			return;
		}
		else if ( !file.GetLineCount() )
		{
			AddLogLineM ( true, _ ( "Corrupted version check file" ) );
		}
		else
		{
			wxString versionLine = file.GetFirstLine();
			wxStringTokenizer tkz ( versionLine, wxT ( "." ) );

			AddDebugLogLineM ( false, logGeneral, wxString ( wxT ( "Running: " ) ) + wxT ( VERSION ) + wxT ( ", Version check: " ) + versionLine );

			long fields[] = {0, 0, 0};

			for ( int i = 0; i < 3; ++i )
			{
				if ( !tkz.HasMoreTokens() )
				{
					AddLogLineM ( true, _ ( "Corrupted version check file" ) );
					return;
				}
				else
				{
					wxString token = tkz.GetNextToken();

					if ( !token.ToLong ( &fields[i] ) )
					{
						AddLogLineM ( true, _ ( "Corrupted version check file" ) );
						return;
					}
				}
			}

			long curVer = get_full_ed2k_version ( );

			long newVer = make_full_ed2k_version ( fields[0], fields[1], fields[2] );

			if ( curVer < newVer )
			{
				AddLogLineM ( true, _ ( "You are using an outdated version of imule!" ) );
#ifdef PACKAGE_VERSION
				AddLogLineM ( false, wxString::Format ( _ ( "Your imule version is %s and the latest version is %li.%li.%li" ), wxT ( PACKAGE_VERSION ), fields[0], fields[1], fields[2] ) );
				AddLogLineM ( false, _ ( "The latest version can always be found at http://www.imule.i2p" ) );
#else
				AddLogLineM ( false, wxString::Format ( _ ( "Your imule version is %li.%li.%li and the latest version is %li.%li.%li" ), VERSION_MJR, VERSION_MIN, VERSION_UPDATE, fields[0], fields[1], fields[2] ) );
				AddLogLineM ( false, _ ( "The latest version can always be found at http://www.imule.i2p" ) );
#endif

#ifdef IMULE_DAEMON

#ifdef PACKAGE_VERSION

				printf ( "%s\n", ( const char* ) unicode2UTF8 ( wxString::Format (
				             _ ( "WARNING: Your imuled version is outdated: %s < %li.%li.%li" ),
				             wxT ( PACKAGE_VERSION ), fields[0], fields[1], fields[2] ) ) );

#else
				printf ( "%s\n", ( const char* ) unicode2UTF8 ( wxString::Format (
				             _ ( "WARNING: Your imuled version is outdated: %li.%li.%li < %li.%li.%li" ),
				             VERSION_MJR, VERSION_MIN, VERSION_UPDATE, fields[0], fields[1], fields[2] ) ) );

#endif

#endif
			}
			else
			{
				AddLogLineM ( false, _ ( "Your copy of imule is up to date." ) );
			}
		}

		file.Close();

		wxRemoveFile ( filename );
	}
	else
	{
		AddLogLineM ( true, _ ( "Failed to download the version check file" ) );
	}

}


bool CimuleApp::IsConnected()
{
	return ( IsConnectedED2K() || IsConnectedKad() );
}


bool CimuleApp::IsConnectedED2K()
{
	return serverconnect && serverconnect->IsConnected();
}


bool CimuleApp::IsConnectedKad()
{
	return Kademlia::CKademlia::isConnected();
}


bool CimuleApp::IsKadRunning()
{
	return Kademlia::CKademlia::isRunning();
}


void CimuleApp::ShowUserCount()
{
	uint32_t totaluser = 0, totalfile = 0;

	theApp.serverlist->GetUserFileStatus ( totaluser, totalfile );

	wxString buffer =
	    CFormat ( _ ( "Users: E: %s K: %s | Files E: %s K: %s" ) ) % CastItoIShort ( totaluser ) %
	    CastItoIShort ( Kademlia::CKademlia::getKademliaUsers() ) % CastItoIShort ( totalfile ) % CastItoIShort ( Kademlia::CKademlia::getKademliaFiles() );

	Notify_ShowUserCount ( buffer );
}


void CimuleApp::ListenSocketHandler ( CI2PSocketEvent& event )
        //void CimuleApp::ListenSocketHandler ( wxSocketEvent& event )
{
    if ( listensocket==NULL and event.GetSocketEvent()==wxSOCKET_LOST )
        // this is the case when connection to the I2P router has been lost
        // and listensocket has already been nulled
        return ;

	wxCHECK_RET ( listensocket, wxT ( "Connection-event for NULL'd listen-socket" ) );
	
	wxCHECK_RET ( dynamic_cast<CListenSocket *>(event.GetEventObject()) == listensocket, wxT( "CimuleApp::ListenSocketHandler received an unattended event" ) );
	
	
	switch ( event.GetSocketEvent() )
	{
		case wxSOCKET_CONNECTION :
			if ( m_app_state == APP_STATE_RUNNING )
			{
				listensocket->OnAccept ( 0 );
			}
			else if ( m_app_state == APP_STATE_STARTING )
			{
				// When starting up, connection may be made before we are able
				// to handle them. However, if these are ignored, no futher
				// connection-events will be triggered, so we have to accept it.
				wxI2PSocketClient * socket = listensocket->Accept ( false );
	
				wxCHECK_RET ( socket, wxT ( "NULL returned by Accept() during startup" ) );
	
				socket->Destroy();
			}
			break ;

		case wxSOCKET_LOST :
			if ( m_app_state == APP_STATE_RUNNING )
			{
				AddLogLineM( true, _("iMule has just been disconnected from the I2P router. Restarting network...") );
				StopNetwork() ;
				StartNetwork() ;
			}

		default:
			break;
	}
}


void CimuleApp::ShowConnectionState()
{
	static uint8_t old_state = ( 1 << 7 ); // This flag doesn't exist

	uint8_t state = 0;

// 	if ( theApp.->IsConnected() )
// 	{
// 		state |= CONNECTED_ED2K;
// 	}

	if ( Kademlia::CKademlia::isRunning() )
	{
		if ( Kademlia::CKademlia::isConnected() )
		{
			state |= CONNECTED_KAD_OK;
		}
		else
		{
			state |= CONNECTED_KAD_NOT;
		}
	}

	Notify_ShowConnState ( state );

	if ( old_state != state )
	{

		// Get the changed value

		int changed_flags = old_state ^ state;

		if ( changed_flags & CONNECTED_ED2K )
		{
			// ED2K status changed
			wxString connected_server;
			CServer* ed2k_server = theApp.serverconnect->GetCurrentServer();

			if ( ed2k_server )
			{
				connected_server = ed2k_server->GetListName();
			}

			if ( state & CONNECTED_ED2K )
			{
				// We connected to some server

				//const wxString id = theApp.serverconnect->IsLowID() ? _ ( "with LowID" ) : _ ( "with HighID" );

				AddLogLine ( CFormat ( _ ( "Connected to %s" ) ) % connected_server /*% id*/ );
			}
			else
			{
				if ( theApp.serverconnect->IsConnecting() )
				{
					AddLogLine ( CFormat ( _ ( "Connecting to %s" ) ) % connected_server );
				}
				else
				{
					AddLogLine ( _ ( "Disconnected from ED2K" ) );
				}
			}
		}

		if ( changed_flags & CONNECTED_KAD_NOT )
		{
			if ( state & CONNECTED_KAD_NOT )
			{
				AddLogLine ( _ ( "Kad started." ) );
			}
			else
			{
				AddLogLine ( _ ( "Kad stopped." ) );
			}
		}

		if ( changed_flags & ( CONNECTED_KAD_OK /*| CONNECTED_KAD_FIREWALLED*/ ) )
		{
			if ( state & ( CONNECTED_KAD_OK /*| CONNECTED_KAD_FIREWALLED*/ ) )
			{
				//if ( state & CONNECTED_KAD_OK )
				//{
				AddLogLine ( _ ( "Connected to Kad (ok)" ) );
				//}
				//else
				//{
				//	AddLogLine ( _ ( "Connected to Kad (firewalled)" ) );
				//}
			}
			else
			{
				AddLogLine ( _ ( "Disconnected from Kad" ) );
			}
		}

		old_state = state;
	}

	ShowUserCount();
}


//void CimuleApp::UDPSocketHandler ( wxSocketEvent& event )
void CimuleApp::UDPSocketHandler ( CI2PSocketEvent& event )
{
	CMuleUDPSocket* socket = ( CMuleUDPSocket* ) ( event.GetClientData() );
	wxCHECK_RET ( socket, wxT ( "No socket owner specified." ) );

	if ( !IsRunning() && !IsOnShutDown() )
	{
		AddDebugLogLineM( false, logGeneral, wxT("CimuleApp::UDPSocketHandler received an event, but App is not running.") ) ;
		return;
	}

	switch ( event.GetSocketEvent() )
	{
		case wxSOCKET_INPUT:
			socket->OnReceive ( 0 );

			break;

		case wxSOCKET_OUTPUT:
			socket->OnSend ( 0 );

			break;

		case wxSOCKET_LOST:
			AddDebugLogLineM( true, logGeneral, wxT("UDP socket lost. Restarting network.") ) ;
			theApp.StopNetwork() ;
			theApp.StartNetwork() ;
			break;
		default:
			wxASSERT ( 0 );

			break;
	}
}


void CimuleApp::OnUnhandledException()
{
	// Call the generic exception-handler.

	fprintf ( stderr, "\tiMule Version: %s\n", ( const char* ) unicode2char ( GetFullMuleVersion() ) );

	::OnUnhandledException();
}

int CimuleApp::OnRun()
{
	try
	{
		// call to the base class version
		return IMULE_APP_BASE::OnRun() ;
	}
	catch ( ... )
	{
		OnUnhandledException() ;
		return -1 ;
	}
}


void CimuleApp::StartKad()
{
	if ( !Kademlia::CKademlia::isRunning() && thePrefs::GetNetworkKademlia() )
	{
		// Kad makes no sense without the Client-UDP socket.

		if ( !thePrefs::IsUDPDisabled() )
		{
			Kademlia::CKademlia::start();
		}
		else
		{
			AddLogLineM ( true, _ ( "Kad network cannot be used if UDP port is disabled on preferences, not starting." ) );
		}
	}
	else if ( !thePrefs::GetNetworkKademlia() )
	{
		AddLogLineM ( true, _ ( "Kad network disabled on preferences, not connecting." ) );
	}
}

void CimuleApp::StopKad()
{
	// Stop Kad if it's running

	if ( Kademlia::CKademlia::isRunning() )
	{
		Kademlia::CKademlia::stop();
	}
}

void CimuleApp::DisconnectED2K()
{
	// Stop ED2K if it's running

	if ( IsConnectedED2K() )
	{
		serverconnect->Disconnect();
	}
}

bool CimuleApp::CryptoAvailable() const
{
	return clientcredits && clientcredits->CryptoAvailable();
}


/*I2P:
uint32_t CimuleApp::GetED2KID() const {
	return serverconnect ? serverconnect->GetClientID() : 0;
}*/
CI2PAddress CimuleApp::GetED2KDest() const
{
	if ( serverconnect )
		return serverconnect->GetClientDest();
	else
		return CI2PAddress::null;
}

CI2PAddress CimuleApp::GetUdpDest() const

{
	CI2PAddress dest;
	clientudp->GetLocal ( dest );
	return dest;
}

CI2PAddress CimuleApp::GetTcpDest() const

{
	CI2PAddress dest;
	listensocket->GetLocal ( dest );
	return dest;
}

bool CimuleApp::ReinitializeNetwork ( wxString* msg )
{
	bool ok = true;

	static bool firstTime = true;

	if ( !firstTime )
	{
		// TODO: Destroy previously created sockets
	}

	firstTime = false;

	// Some sanity checks first

	if ( thePrefs::ECPort() == thePrefs::GetI2PServerPort()
		    || thePrefs::ECPort() == thePrefs::GetI2PServerI2PTcpPort()
		    || thePrefs::ECPort() == thePrefs::GetI2PServerI2PUdpPort()
	   )
	{
		// Select a random usable port in the range 1025 ... 2^16 - 1
		uint16_t port = thePrefs::ECPort();

		while ( port < 1024
				      || port == thePrefs::GetI2PServerPort()
				      || port == thePrefs::GetI2PServerI2PUdpPort()
				      || port == thePrefs::GetI2PServerI2PTcpPort()
		      )
		{
			port = ( uint16_t ) rand();
		}

		thePrefs::SetECPort ( port );

		wxString err = wxT ( "Network configuration failed! You cannot use the same port\n"
				"for the I2P server ports and the External Connections port.\n"
				"The EC port has been changed to avoid conflict, see the\n"
				"preferences for the new value.\n" );

		*msg << err;

		AddLogLineM ( false, wxEmptyString );
		AddLogLineM ( true, err );
		AddLogLineM ( false, wxEmptyString );

		ok = false;
	}


	// Create the address where we are going to listen
	// TODO: read this from configuration file
	imuleIPV4Address myaddr;

	//myaddr.AnyAddress();
	if ( thePrefs::GetECAddress().IsEmpty() || !myaddr.Hostname ( thePrefs::GetECAddress() ) )
	{
		myaddr.AnyAddress();
	}

	myaddr.Service ( thePrefs::ECPort() );

	// Get ready to handle connections from apps like imulecmd
	ECServerHandler = new ExternalConn ( myaddr, msg );

						// Start the web server
					// ( No webserver on Windows yet )
#ifndef __WXMSW__
					// Run webserver?
	if ( thePrefs::GetWSIsEnabled() )
	{
		wxString imuleConfigFile = theApp.ConfigDir + wxT ( "imule.conf" );
		wxString imulewebPath = wxT ( "imuleweb" );

#if defined(__WXMAC__) && !defined(IMULE_DAEMON)
					// For the Mac GUI application, look for imuleweb in the bundle
		CFURLRef imulewebUrl = CFBundleCopyAuxiliaryExecutableURL (
				CFBundleGetMainBundle(), CFSTR ( "imuleweb" ) );

		if ( imulewebUrl )
		{
			CFURLRef absoluteUrl = CFURLCopyAbsoluteURL ( imulewebUrl );
			CFRelease ( imulewebUrl );

			if ( absoluteUrl )
			{
				CFStringRef imulewebCfstr = CFURLCopyFileSystemPath ( absoluteUrl, kCFURLPOSIXPathStyle );
				CFRelease ( absoluteUrl );
				imulewebPath = wxMacCFStringHolder ( imulewebCfstr ).AsString ( wxLocale::GetSystemEncoding() );
			}
		}

#endif

		theApp.webserver_pid = wxExecute ( wxT ( "'" ) + imulewebPath + wxT ( "' '--imule-config-file=" ) + imuleConfigFile + wxT ( "'" ) );

		if ( theApp.webserver_pid )
		{
			AddLogLineM ( true, CFormat ( _ ( "webserver running on pid %d" ) ) % theApp.webserver_pid );
		}
		else
		{
			theApp.ShowAlert ( _ (
					   "You requested to run webserver from startup, "
					"but the imuleweb binary cannot be run. "
					"Please install the package containing imule webserver, "
					"or compile imule using --enable-webserver and run make install" ),
     _ ( "Error" ), wxOK | wxICON_ERROR );
		}
	}

#endif /* ! __WXMSW__ */

	// connects to I2P network
	if ( thePrefs::DoAutoConnect() )
	{
		StartNetwork();
	}
	return ok;
}

void CimuleApp::StartNetwork()
{
	m_connection_manager->start() ;
	theApp.ShowConnectionState();
}

void CimuleApp::StopNetwork()
{
	//disconnect the ED2K servers
	if ( thePrefs::GetNetworkED2K() )
	{
		if ( theApp.serverconnect->IsConnecting() )
		{
			theApp.serverconnect->StopConnectionTry();
		}
		else
		{
			theApp.serverconnect->Disconnect();
		}
	}
	
	// Disconnect Kad
	if ( thePrefs::GetNetworkKademlia() )
	{
		theApp.StopKad();
	}
	
	// stop and kill connections
	m_connection_manager->stop() ;
	
//	clientlist->Process();
	
	// update display
	theApp.ShowConnectionState();
}

void CimuleApp::StartSearchNetwork()
{
	if ( thePrefs::GetNetworkKademlia() ) 
	{
		AddLogLineM ( false, _ ( "Autoconnect to the Kad indexing network (see Preferences if you want to disable it)." ) );
		
		theApp.StartKad();
	}
	if ( thePrefs::GetNetworkED2K() )
	{
		AddLogLineM ( false, _ ( "Autoconnect to the ED2K servers (see Preferences if you want to disable it)." ) );
		
		theApp.serverconnect->ConnectToAnyServer();
	}
}

void CimuleApp::FetchIfNewVersionIsAvailable()
{
    // Test if there's any new version
    if ( thePrefs::CheckNewVersion() )
    {
        // We use the thread base because I don't want a dialog to pop up.
        CHTTPDownloadThread* version_check =
                new CHTTPDownloadThread ( wxT ( "http://www.imule.i2p/lastversion" ),
                                          theApp.ConfigDir + wxT ( "last_version_check" ), HTTP_VersionCheck, false, MIN2S ( 5 ) );
        version_check->Create();
        version_check->Run();
    }
}

DEFINE_LOCAL_EVENT_TYPE ( wxEVT_MULE_NOTIFY_EVENT )

DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_FILE_HASHING_FINISHED )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_FILE_HASHING_SHUTDOWN )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_FINISHED_FILE_COMPLETION )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_FINISHED_HTTP_DOWNLOAD )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_SOURCE_DNS_DONE )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_UDP_DNS_DONE )
DEFINE_LOCAL_EVENT_TYPE ( wxEVT_CORE_SERVER_DNS_DONE )
