//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef IMULE_H
#define IMULE_H

#include "config.h"
#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/app.h>		// Needed for wxApp
#include <wx/intl.h>		// Needed for wxLocale
#include <wx/string.h>		// Needed for wxString

#include "Types.h"		// Needed for uint32_t, uint16_t and wxULongLong
#include "GuiEvents.h"		// Needed for GUIEvent

#include <list>			// Needed for std::list

//#include <wx/event.h>
#include "i2p/wxI2PEvents.h"
//#include "i2p/wxI2PClient.h"
#include "i2p/CI2PAddress.h"

class wxCloseEvent;

class CAbstractFile;

class CKnownFile;

class ExternalConn;

class CimuleDlg;

class CPreferences;

class CDownloadQueue;

class CUploadQueue;

class CServerConnect;

class CSharedFileList;

class CServer;

class CFriend;

class CMD4Hash;

class CServerList;

class CI2PRouter;

class CListenSocket;

class CClientList;

class CKnownFileList;

class CSearchList;

class CClientCreditsList;

class CFriendList;

class CClientUDPSocket;

class CIPFilter;

class UploadBandwidthThrottler;

class CStatistics;

class wxServer;

class wxSocketEvent;

class wxCommandEvent;

class wxFFileOutputStream;

class wxFile;

class CUpDownClient;

class CTimer;

class wxTimerEvent;

class wxSingleInstanceChecker;

class wxExecuteData;

class CMuleInternalEvent;

class I2PConnectionManager ;

#define theApp wxGetApp()

enum APPState {
    APP_STATE_RUNNING = 0,
    APP_STATE_SHUTINGDOWN,
    APP_STATE_STARTING
};

#ifdef IMULE_DAEMON
#define IMULE_APP_BASE wxAppConsole
#else
#define IMULE_APP_BASE wxApp
#endif

#define CONNECTED_ED2K (1<<0)
#define CONNECTED_KAD_NOT (1<<1)
#define CONNECTED_KAD_OK (1<<2)
//#define CONNECTED_KAD_FIREWALLED (1<<3)

class CimuleApp : public IMULE_APP_BASE
{
	friend class I2PConnectionManager ;
	
	public:
		/**
		 * Returns a description of the version of imule being used.
		 *
		 * @return A detailed description of the imule version, including wx information.
		 *
		 * Use this rather than just using the VERSION or CURRENT_VERSION_LONG
		 * constants, when displaying information to the user. The purpose is to
		 * help with debugging.
		 */

		static wxString GetMuleVersion();

		/**
		 * This functions is like the GetMuleVersion function above, with the exception
		 * that it also includes the name of the application. This can be one of the
		 * following:
		 *
		 *  - imule
		 *  - imuled
		 *  - Remote imule-GUI
		 */
		static wxString GetFullMuleVersion();


		CimuleApp();

		virtual	 ~CimuleApp();

		virtual int	OnRun() ;

		virtual bool	OnInit();

		int		OnExit();

		void		OnFatalException();

		bool		ReinitializeNetwork ( wxString *msg );
		
		void        StopNetwork() ;
		
		void        StartNetwork();
		
		void        StartSearchNetwork();

        void        FetchIfNewVersionIsAvailable();

		// derived classes may override those
		virtual int InitGui ( bool geometry_enable, wxString &geometry_string );

		// Socket handlers
        void ListenSocketHandler ( CI2PSocketEvent& event );
        //void ListenSocketHandler ( wxSocketEvent& event );

		// FIXME: seems not implemented ?
		//void ServerSocketHandler(CI2PSocketEvent& event);
        //void UDPSocketHandler ( wxSocketEvent& event );
        void UDPSocketHandler ( CI2PSocketEvent& event );

		virtual void NotifyEvent ( const GUIEvent& event ) = 0;

		virtual void ShowAlert ( wxString msg, wxString title, int flags ) = 0;

		// Barry - To find out if app is running or shutting/shut down
		bool IsRunning() const { return ( m_app_state == APP_STATE_RUNNING ); }

		bool IsOnShutDown() const { return ( m_app_state == APP_STATE_SHUTINGDOWN ); }

		bool NetworkStarting() {return networkStarting;}

		bool NetworkStarted() {return networkStarted;}

		// NOTE : i2P : no firewall, no LowID
		// Check ED2K and Kademlia state
		//bool IsFirewalled();
		// Check Kad state
		//bool IsFirewalledKad();
		// Check if we should callback this client
		//bool DoCallback( CUpDownClient *client );

		// Connection to ED2K
		bool IsConnectedED2K();
		// Connection to Kad
		bool IsConnectedKad();
		// Are we connected to at least one network?
		bool IsConnected();

		// What about Kad? Is it running?
		bool IsKadRunning();

		// ed2k URL functions

		wxString	CreateED2kLink ( const CAbstractFile* f );

		wxString	CreateED2kSourceLink ( const CAbstractFile* f );

		wxString	CreateED2kAICHLink ( const CKnownFile* f );

		wxString	CreateED2kHostnameSourceLink ( const CAbstractFile* f );

		void RunAICHThread();

		// Misc functions
		void		OnlineSig ( bool zero = false );

		void		Localize_mule();

		void		Trigger_New_version ( wxString newMule );

		// shakraw - new EC code using wxSocketBase
		ExternalConn*	ECServerHandler;

		// return current (valid) public IP or 0 if unknown
		// If ignorelocal is true, don't use m_localip
		//uint32_t	GetPublicIP(bool ignorelocal = false) const;
		//void		SetPublicIP(const uint32_t dwIP);
		CI2PAddress		GetPublicDest ( bool ignorelocal = false ) const;

		void		SetPublicDest ( const CI2PAddress & dwDest );

		//I2P:uint32_t	GetED2KID() const;
		CI2PAddress	GetED2KDest() const;

		CI2PAddress	GetUdpDest() const;

		CI2PAddress	GetTcpDest() const;

		// Other parts of the interface and such
		CPreferences*		glob_prefs;

		CDownloadQueue*		downloadqueue;

		CUploadQueue*		uploadqueue;

		// No server yet in imule
		CServerConnect*		serverconnect;

		CSharedFileList*	sharedfiles;

		CServerList*		serverlist;

#if INTERNAL_ROUTER
		CI2PRouter*             i2prouter;

#endif
		CListenSocket*		listensocket;

		CClientList*		clientlist;

		CKnownFileList*		knownfiles;

		CSearchList*		searchlist;

		CClientCreditsList*	clientcredits;

		CFriendList*		friendlist;

		CClientUDPSocket*	clientudp;

		CStatistics*		statistics;

		CIPFilter*		ipfilter;

		UploadBandwidthThrottler* uploadBandwidthThrottler;

		wxLocale m_locale;

		void ShutDown();

		wxString GetLog ( bool reset = false );

		wxString GetServerLog ( bool reset = false );

		wxString GetDebugLog ( bool reset = false );

		bool AddServer ( CServer *srv, bool fromUser = false );

		void  AddServerMessageLine( wxString &msg );

#if INTERNAL_ROUTER
		virtual void UpdateRouterStatus( wxString & msg ) {}
		
		virtual void UpdateRouterStatus();
#endif

#ifdef __DEBUG__
		void AddSocketDeleteDebug ( uint32_t socket_pointer, uint32_t creation_time );

#endif
		void SetOSFiles ( const wxString new_path );

		wxString ConfigDir;

		void AddLogLine ( const wxString &msg );

		const wxString& GetOSType() const { return OSType; }

		void ShowUserCount();

		void ShowConnectionState();

		void StartKad();
		void StopKad();

		void DisconnectED2K();

		bool CryptoAvailable() const;

		//! TODO: Move to CLogger
		wxFFileOutputStream* applog;

	protected:
		// Used to detect a previous running instance of imule
		wxSingleInstanceChecker*	m_singleInstance;

#ifdef __WXDEBUG__
		/**
		 * Handles asserts in a thread-safe manner.
		 */
		virtual void OnAssert ( const wxChar *file, int line, const wxChar *cond, const wxChar *msg );

#endif

		/**
		 * This class is used to contain log messages that are to be displayed
		 * on the GUI, when it is currently impossible to do so. This is in order
		 * to allows us to queue messages till after the dialog has been created.
		 */

		struct QueuedLogLine
		{
			//! The text line to be displayed
			wxString 	line;
			//! True if the line should be shown on the status bar, false otherwise.
			bool		show;
		};

		void OnUDPDnsDone ( CMuleInternalEvent& evt );

		void OnSourceDnsDone ( CMuleInternalEvent& evt );

		void OnServerDnsDone ( CMuleInternalEvent& evt );

		void OnTCPTimer ( CMuleInternalEvent& evt );

		void OnCoreTimer ( CMuleInternalEvent& evt );

		void OnFinishedHashing ( CMuleInternalEvent& evt );

		void OnFinishedCompletion ( CMuleInternalEvent& evt );

		void OnFinishedHTTPDownload ( CMuleInternalEvent& evt );

		void OnHashingShutdown ( CMuleInternalEvent& );

		void OnNotifyEvent ( wxEvent& evt );

		void SetTimeOnTransfer();

		std::list<QueuedLogLine> m_logLines;

		APPState m_app_state;
		
		I2PConnectionManager * m_connection_manager ;
		
		wxString m_emulesig_path;

		wxString m_imulesig_path;

		static const wxString FullMuleVersion;

		static const wxString OSDescription;

		static char *strFullMuleVersion;

		static char *strOSDescription;

		wxString OSType;

		//I2P:uint32_t m_dwPublicIP;
		CI2PAddress m_dwPublicDest;

		long webserver_pid;

		bool enable_stdout_log;

		bool enable_daemon_fork;

		wxString server_msg;

		CTimer* core_timer;


	private:
		virtual void OnUnhandledException();

		void CheckNewVersion ( uint32_t result );

		// I2P
		CI2PAddress m_localdest;

		bool networkStarted ;
		
		bool networkStarting ;
};

#ifndef IMULE_DAEMON

class CimuleGuiBase
{
	public:
		CimuleGuiBase();
		virtual	 ~CimuleGuiBase();

		wxString	m_FrameTitle;
		CimuleDlg*	imuledlg;

		bool CopyTextToClipboard ( wxString strText );

		virtual void NotifyEvent ( const GUIEvent& event );

		virtual int InitGui ( bool geometry_enable, wxString &geometry_string );

		virtual void ShowAlert ( wxString msg, wxString title, int flags );
};

#ifndef CLIENT_GUI

class CimuleGuiApp : public CimuleApp, public CimuleGuiBase
{

		virtual int InitGui ( bool geometry_enable, wxString &geometry_string );

		int OnExit();
		bool OnInit();

	public:

		virtual void ShowAlert ( wxString msg, wxString title, int flags );

		void ShutDown ( wxCloseEvent &evt );

		virtual void NotifyEvent ( const GUIEvent& event );

		wxString GetLog ( bool reset = false );

		wxString GetServerLog ( bool reset = false );

		void AddServerMessageLine ( wxString &msg );
		
#if INTERNAL_ROUTER
		void UpdateRouterStatus( wxString & msg );
		
		void UpdateRouterStatus();
#endif

		DECLARE_EVENT_TABLE()
};

DECLARE_APP ( CimuleGuiApp )

#else /* !CLIENT_GUI */

#include "imule-remote-gui.h"

#endif // CLIENT_GUI

#define CALL_APP_DATA_LOCK

#else /* ! IMULE_DAEMON */

#include <wx/apptrait.h>
#include <wx/socket.h>

class CSocketSet;

class CimuledGSocketFuncTable : public GSocketGUIFunctionsTable
{
		CSocketSet *m_in_set, *m_out_set;

		wxMutex m_lock;
	public:
		CimuledGSocketFuncTable();

		void AddSocket ( GSocket *socket, GSocketEvent event );
		void RemoveSocket ( GSocket *socket, GSocketEvent event );
		void RunSelect();

		virtual bool OnInit();
		virtual void OnExit();
		virtual bool CanUseEventLoop();
		virtual bool Init_Socket ( GSocket *socket );
		virtual void Destroy_Socket ( GSocket *socket );
		virtual void Install_Callback ( GSocket *socket, GSocketEvent event );
		virtual void Uninstall_Callback ( GSocket *socket, GSocketEvent event );
		virtual void Enable_Events ( GSocket *socket );
		virtual void Disable_Events ( GSocket *socket );
};

class CDaemonAppTraits : public wxConsoleAppTraits
{
		CimuledGSocketFuncTable *m_table;

		wxMutex m_lock;
		std::list<wxObject *> m_sched_delete;
	public:
		CDaemonAppTraits ( CimuledGSocketFuncTable *table );
		virtual GSocketGUIFunctionsTable* GetSocketGUIFunctionsTable();
		virtual void ScheduleForDestroy ( wxObject *object );
		virtual void RemoveFromPendingDelete ( wxObject *object );

		void DeletePending();

#ifndef __WXMSW__
		virtual int WaitForChild ( wxExecuteData& execData );
#endif

#ifdef __WXMAC__
		virtual wxStandardPathsBase& GetStandardPaths();
#endif
};

class CimuleDaemonApp : public CimuleApp
{
		bool m_Exit;

		bool OnInit();
		int OnRun();
		int OnExit();

		virtual int InitGui ( bool geometry_enable, wxString &geometry_string );

		CimuledGSocketFuncTable *m_table;
	public:
		CimuleDaemonApp();

		void ExitMainLoop() { m_Exit = true; }

		virtual void NotifyEvent ( const GUIEvent& event );

		bool CopyTextToClipboard ( wxString strText );

		virtual void ShowAlert ( wxString msg, wxString title, int flags );

		DECLARE_EVENT_TABLE()

		wxAppTraits *CreateTraits();

};

DECLARE_APP ( CimuleDaemonApp )

#endif /* ! IMULE_DAEMON */

#endif // IMULE_H
