//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

//#include "stdafx.h"
//#include "../utils/MiscUtils.h"
#include "Indexed.h"
#include "Kademlia.h"
#include "../../OPCodes.h"
#include "Defines.h"
#include "Prefs.h"
#include "../routing/RoutingZone.h"
#include "../routing/Contact.h"
#include "../net/KademliaUDPListener.h"
#include "../utils/UInt128.h"
#include "../../OtherFunctions.h"
#include "../../MemFile.h"
#include "../../CFile.h"
#include "../../Packet.h"
#include "imule.h"
//#include "Preferences.h"
#include "Logger.h"
#include "common/Format.h"

#include <wx/arrstr.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif


////////////////////////////////////////

using namespace Kademlia;

////////////////////////////////////////

wxString CIndexed::m_kfilename;

wxString CIndexed::m_sfilename;

wxString CIndexed::m_loadfilename;

CIndexed::CIndexed()
{
	m_sfilename = theApp.ConfigDir + wxT ( "src_index.dat" );
	m_kfilename = theApp.ConfigDir + wxT ( "key_index.dat" );
	m_loadfilename = theApp.ConfigDir + wxT ( "load_index.dat" );
	m_lastClean = time ( NULL ) + ( 60 * 30 );
	m_totalIndexSource = 0;
	m_totalIndexKeyword = 0;
	readFile();
}


void CIndexed::readFile ( void )

{
	try
	{
		uint32_t totalLoad = 0;
		uint32_t totalSource = 0;
		uint32_t totalKeyword = 0;
		uint32_t numKeys = 0;
		uint32_t numSource = 0;
		uint32_t numLoad = 0;

		//CBufferedFileIO load_file;
		CFile load_file ;

		if ( load_file.Open ( m_loadfilename, CFile::read ) )
		{
			uint32_t version = load_file.ReadUInt32();
			AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: version = %d" ) ) % version );

			if ( version < 2 )
			{
				time_t savetime = load_file.ReadUInt32(); //  Savetime is unused now
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: savetime = %d" ) ) % savetime );

				numLoad = load_file.ReadUInt32();
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numLoad = %d" ) ) % numLoad );

				while ( numLoad )
				{
					CUInt128 keyID = load_file.ReadUInt128();
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: keyID = %s" ) ) % keyID.toHexString() );
					uint32_t thisload = load_file.ReadUInt32();
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: thisload = %d" ) ) % thisload );

					if ( AddLoad ( keyID, thisload ) )
					{
						totalLoad++;
					}

					numLoad--;
				}
			}

			load_file.Close();
		}

		// keywords
		//CBufferedFileIO k_file;
		CFile k_file ;

		if ( k_file.Open ( m_kfilename, CFile::read ) )
		{
			try
			{

				uint32_t version = k_file.ReadUInt32();
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: version = %d" ) ) % version );

				if ( version < 2 )
				{
					time_t savetime = k_file.ReadUInt32();
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: savetime = %d" ) ) % savetime );

					if ( savetime > time ( NULL ) )
					{
						CUInt128 id = k_file.ReadUInt128();
						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: id = %s" ) ) % id.toHexString() );

						if ( !Kademlia::CKademlia::getPrefs()->getKadID().compareTo ( id ) )
						{
							numKeys = k_file.ReadUInt32();
							AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numKeys = %d" ) ) % numKeys );

							while ( numKeys )
							{
								CUInt128 keyID = k_file.ReadUInt128();
								AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: keyID = %s" ) ) % keyID.toHexString() );
								numSource = k_file.ReadUInt32();
								AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numSources = %d" ) ) % numSource );

								while ( numSource )
								{
									CUInt128 sourceID = k_file.ReadUInt128();
									AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: sourceID = %s" ) ) % sourceID.toHexString() );

									//numName = k_file.ReadUInt32();
									//AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numName = %d" ) ) % numName );


									//while ( numName )
									{
										Kademlia::CEntry toaddN;
										toaddN.source = false;
										uint32_t expire = k_file.ReadUInt32();
										AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: expire = %d" ) ) % expire );

										toaddN.lifetime = expire;
										toaddN.taglist = TagList ( k_file );

										for ( TagList::const_iterator it = toaddN.taglist.begin();
										        it != toaddN.taglist.end();
										        it++ )
										{
											const CTag & tag = *it;

											if ( tag.isValid() )
											{
												if ( tag.GetID() == TAG_FILENAME )
												{
													toaddN.fileName = tag.GetStr();
													toaddN.fileName.MakeLower(); // Make lowercase, the search code expects lower case strings!
													// NOTE: always add the 'name' tag, even if it's stored separately in 'fileName'. the tag is still needed for answering search request
												}
												else if ( tag.GetID() == TAG_FILESIZE )
												{
													toaddN.size = tag.GetInt();
													// NOTE: always add the 'size' tag, even if it's stored separately in 'size'. the tag is still needed for answering search request
												}
												else if ( tag.GetID() == TAG_SOURCEDEST )
												{
													toaddN.tcpdest = tag.GetAddr();
												}
												else if ( tag.GetID() == TAG_SOURCEUDEST )
												{
													toaddN.udpdest = tag.GetAddr();
												}
											}
										}

										toaddN.keyID.setValue ( keyID );

										toaddN.sourceID.setValue ( sourceID );
										uint8_t load = 0;

										if ( AddKeyword ( keyID, sourceID, toaddN, load ) )
										{
											totalKeyword++;
										}

										//numName--;
									}

									numSource--;
								}

								numKeys--;
							}
						}
					}
				}
			}
			catch ( wxString s )
			{
				AddLogLineM ( true, CFormat ( _ ( "Kad index file %s is buggy and has not been loaded completely.\n Message : " ) )
				              % m_kfilename % s );
			}

			k_file.Close();
		}

		// sources
		//CBufferedFileIO s_file;
		CFile s_file;

		if ( s_file.Open ( m_sfilename, CFile::read ) )
		{
			try
			{
				uint32_t version = s_file.ReadUInt32();
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: version = %d" ) ) % version );

				if ( version < 2 )
				{
					time_t savetime = s_file.ReadUInt32();
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: savetime = %d" ) ) % savetime );

					if ( savetime > time ( NULL ) )
					{
						numKeys = s_file.ReadUInt32();
						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numKeys = %d" ) ) % numKeys );

						CUInt128 id;

						while ( numKeys )
						{
							CUInt128 keyID = s_file.ReadUInt128();
							AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: keyID = %s" ) ) % keyID.toHexString() );

							numSource = s_file.ReadUInt32();
							AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numSources = %d" ) ) % numSource );

							while ( numSource )
							{
								CUInt128 sourceID = s_file.ReadUInt128();
								AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: sourceID = %s" ) ) % sourceID.toHexString() );

								//numName = s_file.ReadUInt32();
								//AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: numName = %d" ) ) % numName );

								//while ( numName )
								{
									Kademlia::CEntry toaddN;
									toaddN.source = true;
									uint32_t test = s_file.ReadUInt32();
									AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "readFile: test = %d" ) ) % test );

									toaddN.lifetime = test;

									toaddN.taglist = TagList ( s_file );

									for ( TagList::const_iterator it = toaddN.taglist.begin();
									        it != toaddN.taglist.end();
									        it++ )
									{
										const CTag & tag = *it ;

										if ( tag.isValid() )
										{
											//if (tag.GetID()==TAG_SOURCEIP)
											//{
											//	toaddN.ip = tag.GetInt();
											//	toaddN.taglist.push_back(tag);
											//} else

											if ( tag.GetID() == TAG_SOURCEDEST )
											{
												toaddN.tcpdest = tag.GetAddr();
											}
											else if ( tag.GetID() == TAG_SOURCEUDEST )
											{
												toaddN.udpdest = tag.GetAddr();
											}
										}
									}

									toaddN.keyID.setValue ( keyID );

									toaddN.sourceID.setValue ( sourceID );
									uint8_t load = 0;

									if ( AddSources ( keyID, sourceID, toaddN, load ) )
									{
										totalSource++;
									}

									//numName--;
								}

								numSource--;
							}

							numKeys--;
						}
					}
				}
			}
			catch ( wxString s )
			{
				AddLogLineM ( true, CFormat ( _ ( "Kad index file %s is buggy and has not been loaded completely.\n Message : " ) )
				              % m_sfilename % s );
			}

			s_file.Close();

			m_totalIndexSource = totalSource;
			m_totalIndexKeyword = totalKeyword;
			AddDebugLogLineM ( false, logKadIndex, wxString::Format ( wxT ( "Read %u source, %u keyword, and %u load entries" ), totalSource, totalKeyword, totalLoad ) );
		}
	}
	catch ( const CSafeIOException& ioe )
	{
		AddDebugLogLineM ( true, logKadIndex, CFormat ( wxT ( "Exception in CIndexed::readFile (%s)" ) ) % ioe.what() );
	}
	catch ( const wxString& e )
	{
		AddDebugLogLineM ( true, logKadIndex, wxT ( "Exception in CIndexed::readFile: " ) + e );
	}
}

CIndexed::~CIndexed()
{
	try
	{
		uint32_t s_total = 0;
		uint32_t k_total = 0;
		uint32_t l_total = 0;

		//CBufferedFileIO load_file;
		CFile load_file;

		if ( load_file.Open ( m_loadfilename, CFile::write ) )
		{
			uint32_t version = 1;
			load_file.WriteUInt32 ( version );
			load_file.WriteUInt32 ( time ( NULL ) );
			load_file.WriteUInt32 ( m_Load_map.size() );
			LoadMap::iterator it = m_Load_map.begin();

			for ( ; it != m_Load_map.end(); ++it )
			{
				load_file.WriteUInt128 ( it->first );
				load_file.WriteUInt32 ( it->second.time );
				l_total++;
			}

			load_file.Close();
		}

		//CBufferedFileIO s_file;
		CFile s_file;

		if ( s_file.Open ( m_sfilename, CFile::write ) )
		{

			uint32_t version = 1;
			s_file.WriteUInt32 ( version );

			s_file.WriteUInt32 ( time ( NULL ) + KADEMLIAREPUBLISHTIMES );

			s_file.WriteUInt32 ( m_Sources_map.size() );
			SrcHashMap::iterator itSrcHash = m_Sources_map.begin();

			for ( ; itSrcHash != m_Sources_map.end(); ++itSrcHash )
			{
				SrcHash & currSrcHash = itSrcHash->second;
				s_file.WriteUInt128 ( itSrcHash->first );

				CSourceKeyMap & KeyHashSrcMap = currSrcHash.m_Source_map;
				s_file.WriteUInt32 ( KeyHashSrcMap.size() );

				CSourceKeyMap::iterator itSource = KeyHashSrcMap.begin();

				for ( ; itSource != KeyHashSrcMap.end(); ++itSource )
				{
					//Source* currSource = *itSource;
					s_file.WriteUInt128 ( itSource->first );
					Kademlia::CEntry & currName = itSource->second;
					s_file.WriteUInt32 ( currName.lifetime );
					currName.taglist.WriteTo ( &s_file );
					s_total++;
				}
			}

			s_file.Close();
		}

		//CBufferedFileIO k_file;
		CFile k_file;

		if ( k_file.Open ( m_kfilename, CFile::write ) )
		{

			uint32_t version = 1;
			k_file.WriteUInt32 ( version );

			k_file.WriteUInt32 ( time ( NULL ) + KADEMLIAREPUBLISHTIMEK );

			k_file.WriteUInt128 ( Kademlia::CKademlia::getPrefs()->getKadID() );

			k_file.WriteUInt32 ( m_Keyword_map.size() );
			KeyHashMap::iterator itKeyHash = m_Keyword_map.begin();

			for ( ; itKeyHash != m_Keyword_map.end(); ++itKeyHash )
			{
				//loop on keywords map
				k_file.WriteUInt128 ( itKeyHash->first );             // key ID

				KeyHash & currKeyHash = itKeyHash->second;            // keyHash class
				CSourceKeyMap& KeyHashSrcMap = currKeyHash.m_Source_map;
				k_file.WriteUInt32 ( KeyHashSrcMap.size() );

				CSourceKeyMap::iterator itSource = KeyHashSrcMap.begin();

				for ( ; itSource != KeyHashSrcMap.end(); ++itSource )
				{
					k_file.WriteUInt128 ( itSource->first );          // source ID

					Kademlia::CEntry & currName = itSource->second;
					k_file.WriteUInt32 ( currName.lifetime );
					currName.taglist.WriteTo ( &k_file );
					k_total++;
				}
			}

			k_file.Close();
		}

		AddDebugLogLineM ( false, logKadIndex, wxString::Format ( wxT ( "Wrote %u source, %u keyword, and %u load entries" ), s_total, k_total, l_total ) );

	}
	catch ( const CSafeIOException & ioe )
	{
		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "Exception in CIndexed::~CIndexed (IO error(%s))" ) ) % ioe.what() ) ;
	}
}


void CIndexed::clean ( void )
{
	if ( m_lastClean > time ( NULL ) )
	{
		return;
	}

	uint32_t k_Removed = 0;

	uint32_t s_Removed = 0;
	uint32_t s_Total = 0;
	uint32_t k_Total = 0;
	time_t tNow = time ( NULL );

	KeyHashMap::iterator itKeyHash = m_Keyword_map.begin();

	while ( itKeyHash != m_Keyword_map.end() )
	{
		KeyHashMap::iterator curr_itKeyHash = itKeyHash++; // Don't change this to a ++it!

		KeyHash& currKeyHash = curr_itKeyHash->second;
		CSourceKeyMap::iterator itSource = currKeyHash.m_Source_map.begin();

		while ( itSource != currKeyHash.m_Source_map.end() )
		{
			CSourceKeyMap::iterator curr_itSource = itSource++; // Don't change this to a ++it!
			k_Total++;

			Kademlia::CEntry& currName = curr_itSource->second;

			if ( !currName.source && currName.lifetime < tNow )
			{
				k_Removed++;
				currKeyHash.m_Source_list.remove ( curr_itSource );
				currKeyHash.m_Source_map.erase ( curr_itSource );
			}
		}

		if ( currKeyHash.m_Source_map.empty() )
		{
			m_Keyword_map.erase ( curr_itKeyHash );
		}
	}

	SrcHashMap::iterator itSrcHash = m_Sources_map.begin();

	while ( itSrcHash != m_Sources_map.end() )
	{
		SrcHashMap::iterator curr_itSrcHash = itSrcHash++; // Don't change this to a ++it!
		SrcHash& currSrcHash = curr_itSrcHash->second;

		CSourceKeyMap::iterator itSource = currSrcHash.m_Source_map.begin();

		while ( itSource != currSrcHash.m_Source_map.end() )
		{
			CSourceKeyMap::iterator curr_itSource = itSource++; // Don't change this to a ++it!
			s_Total++;

			Kademlia::CEntry& currName = curr_itSource->second;

			if ( currName.lifetime < tNow )
			{
				s_Removed++;
				currSrcHash.m_Source_list.remove ( curr_itSource );
				currSrcHash.m_Source_map.erase ( curr_itSource );
			}
		}

		if ( currSrcHash.m_Source_map.empty() )
		{
			m_Sources_map.erase ( curr_itSrcHash );
		}
	}

	m_totalIndexSource = s_Total;

	m_totalIndexKeyword = k_Total;
	AddDebugLogLineM ( false, logKadIndex, wxString::Format ( wxT ( "Removed %u keyword out of %u and %u source out of %u" ),  k_Removed, k_Total, s_Removed, s_Total ) );
	m_lastClean = time ( NULL ) + MIN2S ( 30 );
}

bool CIndexed::AddKeyword ( const CUInt128& keyID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load )
{
	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "CIndexed::AddKeyword %s for file %s : filename %s" ) )
	                   % keyID.toHexString()
	                   % sourceID.toHexString()
	                   % entry.fileName );

	if ( m_totalIndexKeyword >= KADEMLIAMAXENTRIES )
	{
		load = 100;
		return false;
	}

	if ( entry.size == 0 || entry.fileName.IsEmpty() || entry.taglist.size() == 0 || entry.lifetime < time ( NULL ) )
	{
		return false;
	}


	if ( m_Keyword_map.find ( keyID ) == m_Keyword_map.end() )
	{
		// the key is not indexed yet
		m_Keyword_map[keyID]               = KeyHash();
	}

	KeyHashMap::iterator itKeyHash = m_Keyword_map.find ( keyID );

	// the key is in the index
	KeyHash & currKeyHash = itKeyHash->second;
	uint32_t indexTotal = currKeyHash.m_Source_map.size();

	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "CIndexed::AddKeyword %s for file %s (existing key)" ) )
	                   % keyID.toHexString()
	                   % sourceID.toHexString() );

	CSourceKeyMap::iterator itSource = currKeyHash.m_Source_map.find ( sourceID );

	if ( itSource != currKeyHash.m_Source_map.end() )
	{
		// the source is in the index for this key
		// update the current entry
		currKeyHash.m_Source_map[sourceID] = entry ;
		currKeyHash.m_Source_list.push_front ( currKeyHash.m_Source_map.find ( sourceID ) );

		if ( indexTotal > KADEMLIAMAXINDEX - 5000 )
		{
			load = 100;
			//We are in a hot node.. If we continued to update all the publishes
			//while this index is full, popular files will be the only thing you index.
			//So let's pretend we have not updated anything !
			return false;
		}

		return true;
	}
	else
	{
		// the source is not in the index for this key
		currKeyHash.m_Source_map[sourceID] = entry ;
		currKeyHash.m_Source_list.push_front ( currKeyHash.m_Source_map.find ( sourceID ) );
		m_totalIndexKeyword++;
		load = (uint8_t) (( indexTotal * 100 ) / KADEMLIAMAXINDEX);
		return true;
	}

}

uint32_t CIndexed::GetSourceCount ( const CUInt128& keyID ) const

{
	SrcHashMap::const_iterator itSrcHash = m_Sources_map.find ( keyID );

	if ( itSrcHash == m_Sources_map.end() )
	{
		return 0;
	}
	else
		return itSrcHash->second.m_Source_map.size();
}

uint32_t CIndexed::GetCompleteSourceCount ( const CUInt128& keyID ) const
{
	SrcHashMap::const_iterator itSrcHash = m_Sources_map.find ( keyID );

	if ( itSrcHash == m_Sources_map.end() )
	{
		return 0;
	}

	const SrcHash& currSrcHash = itSrcHash->second;

	// counting complete sources
	uint32_t complete = 0;

	for ( CSourceKeyMap::const_iterator itSource = currSrcHash.m_Source_map.begin();
	        itSource != currSrcHash.m_Source_map.end();
	        ++itSource )
	{
		uint32_t flags = (uint32_t) itSource->second.GetIntTagValue ( TAG_FLAGS ).GetValue() ;
		complete += ( ( flags & TAG_FLAGS_COMPLETEFILE ) ? 1 : 0 );
	}

	return complete ;
}



bool CIndexed::AddSources ( const CUInt128& keyID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load )

{
	AddDebugLogLineM ( false, logKadIndex,
	                   CFormat ( wxT ( "CIndexed::AddSources for key %s : tcp=%x, udp=%x, tagcount=%d, lifetime=%ld (now=%ld)" ) )
	                   % keyID.toHexString()
	                   % entry.tcpdest.hashCode()
	                   % entry.udpdest.hashCode()
	                   % entry.taglist.size()
	                   % entry.lifetime
	                   % time ( NULL ) ) ;

	if ( !entry.tcpdest || !entry.udpdest || entry.taglist.size() == 0 || entry.lifetime < time ( NULL ) )
	{
		return false;
	}

	if ( m_Sources_map.find ( keyID ) == m_Sources_map.end() )
	{
		// the file is not indexed for the moment
		m_Sources_map[keyID] =  SrcHash();
	}

	SrcHash & currSrcHash = m_Sources_map.find ( keyID )->second;

	uint32_t size = currSrcHash.m_Source_map.size();

	CSourceKeyMap::iterator itSource = currSrcHash.m_Source_map.find ( sourceID );

	if ( itSource == currSrcHash.m_Source_map.end() )
	{
		// the source is not in the index for this key

		if ( size >= KADEMLIAMAXSOUCEPERFILE )
		{
			// first remove the oldest entry if the index is full
			itSource = currSrcHash . m_Source_list . back();
			currSrcHash . m_Source_list . pop_back();
			currSrcHash . m_Source_map  . erase ( itSource );

			m_totalIndexSource--;
		}
	}

	// add the new entry (or replace an existing one)
	currSrcHash.m_Source_map[sourceID] = entry ;

	currSrcHash.m_Source_list.push_front ( currSrcHash.m_Source_map.find ( sourceID ) );

	m_totalIndexSource++;

	size++;

	load = (uint8_t) (( size * 100 ) / KADEMLIAMAXSOUCEPERFILE);

	return true;
}


bool CIndexed::AddNotes ( const CUInt128& keyID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load )

{
	if ( ( !entry.tcpdest && !entry.udpdest ) || entry.taglist.size() == 0 )
	{
		return false;
	}

	if ( m_Notes_map.find ( keyID ) == m_Notes_map.end() )
	{
		// the file is not indexed for the moment
		m_Notes_map[keyID] =  SrcHash();
	}

	SrcHash & currNoteHash = m_Sources_map.find ( keyID )->second;

	uint32_t size = currNoteHash.m_Source_map.size();

	CSourceKeyMap::iterator itSource = currNoteHash.m_Source_map.find ( sourceID );

	if ( itSource == currNoteHash.m_Source_map.end() )
	{
		// the source is not in the index for this key

		if ( size >= KADEMLIAMAXNOTESPERFILE )
		{
			// first remove the oldest entry if the index is full
			itSource = currNoteHash . m_Source_list . back();
			currNoteHash . m_Source_list . pop_back();
			currNoteHash . m_Source_map  . erase ( itSource );

			m_totalIndexSource--;
		}
	}

	// add the new entry (or replace an existing one)
	currNoteHash.m_Source_map[sourceID] = entry ;

	currNoteHash.m_Source_list.push_front ( currNoteHash.m_Source_map.find ( sourceID ) );

	size++;

	load = (uint8_t)(( size * 100 ) / KADEMLIAMAXNOTESPERFILE);

	return true;

}

bool CIndexed::AddLoad ( const CUInt128& keyID, uint32_t timet )

{
	if ( ( uint32_t ) time ( NULL ) > timet )
	{
		return false;
	}

	LoadMap::iterator it = m_Load_map.find ( keyID );

	if ( it != m_Load_map.end() )
	{
		wxASSERT ( 0 );
		return false;
	}

	Load load;

	load.time = timet;
	m_Load_map[keyID] = load;
	return true;
}

bool SearchTermsMatch ( const SSearchTerm* pSearchTerm, const Kademlia::CEntry* item/*, CStringArray& astrFileNameTokens*/ )

{
	// boolean operators

	if ( pSearchTerm->type == SSearchTerm::AND )
	{
		return SearchTermsMatch ( pSearchTerm->left, item/*, astrFileNameTokens*/ ) && SearchTermsMatch ( pSearchTerm->right, item/*, astrFileNameTokens*/ );
	}

	if ( pSearchTerm->type == SSearchTerm::OR )
	{
		return SearchTermsMatch ( pSearchTerm->left, item/*, astrFileNameTokens*/ ) || SearchTermsMatch ( pSearchTerm->right, item/*, astrFileNameTokens*/ );
	}

	if ( pSearchTerm->type == SSearchTerm::NAND )
	{
		return SearchTermsMatch ( pSearchTerm->left, item/*, astrFileNameTokens*/ ) && !SearchTermsMatch ( pSearchTerm->right, item/*, astrFileNameTokens*/ );
	}

	// word which is to be searched in the file name (and in additional meta data as done by some ed2k servers???)
	if ( pSearchTerm->type == SSearchTerm::String )
	{
		int iStrSearchTerms = pSearchTerm->astr->GetCount();

		if ( iStrSearchTerms == 0 )
		{
			return false;
		}

#if 0
		//TODO: Use a pre-tokenized list for better performance.
		// tokenize the filename (very expensive) only once per search expression and only if really needed
		if ( astrFileNameTokens.GetCount() == 0 )
		{
			int iPosTok = 0;
			CString strTok ( item->fileName.Tokenize ( _aszInvKadKeywordChars, iPosTok ) );

			while ( !strTok.IsEmpty() )
			{
				astrFileNameTokens.Add ( strTok );
				strTok = item->fileName.Tokenize ( _aszInvKadKeywordChars, iPosTok );
			}
		}

		if ( astrFileNameTokens.GetCount() == 0 )
			return false;

		// if there are more than one search strings specified (e.g. "aaa bbb ccc") the entire string is handled
		// like "aaa AND bbb AND ccc". search all strings from the string search term in the tokenized list of
		// the file name. all strings of string search term have to be found (AND)
		for ( int iSearchTerm = 0; iSearchTerm < iStrSearchTerms; iSearchTerm++ )
		{
			bool bFoundSearchTerm = false;

			for ( int i = 0; i < astrFileNameTokens.GetCount(); i++ )
			{
				// the file name string was already stored in lowercase
				// the string search term was already stored in lowercase

				if ( strcmp ( astrFileNameTokens.GetAt ( i ), pSearchTerm->astr->GetAt ( iSearchTerm ) ) == 0 )
				{
					bFoundSearchTerm = true;
					break;
				}
			}

			if ( !bFoundSearchTerm )
				return false;
		}

#else
		// if there are more than one search strings specified (e.g. "aaa bbb ccc") the entire string is handled
		// like "aaa AND bbb AND ccc". search all strings from the string search term in the tokenized list of
		// the file name. all strings of string search term have to be found (AND)
		for ( int iSearchTerm = 0; iSearchTerm < iStrSearchTerms; iSearchTerm++ )
		{
			// this will not give the same results as when tokenizing the filename string, but it is 20 times faster.

			if ( item->fileName.Find ( ( * ( pSearchTerm->astr ) ) [iSearchTerm] ) == -1 )
			{
				return false;
			}
		}

#endif
		return true;

		// search string value in all string meta tags (this includes also the filename)
		// although this would work, I am no longer sure if it's the correct way to process the search requests..
		/*const Kademlia::CTag *tag;
		TagList::const_iterator it;
		for (it = item->taglist.begin(); it != item->taglist.end(); it++)
		{
			tag = *it;
			if (tag.m_type == 2)
			{
				//TODO: Use a pre-tokenized list for better performance.
				int iPos = 0;
				CString strTok(static_cast<const CTagStr *>(tag)->m_value.Tokenize(_aszInvKadKeywordChars, iPos));
				while (!strTok.IsEmpty()){
					if (stricmp(strTok, *(pSearchTerm->str)) == 0)
						return true;
					strTok = static_cast<const CTagStr *>(tag)->m_value.Tokenize(_aszInvKadKeywordChars, iPos);
				}
			}
		}
		return false;*/
	}

	if ( pSearchTerm->type == SSearchTerm::MetaTag )
	{
		if ( pSearchTerm->tag.IsStr() )
		{
			// meta tags with string values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsStr() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetStr().CmpNoCase ( pSearchTerm->tag.GetStr() ) == 0;
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpGreaterEqual )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() >= pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() >= pSearchTerm->tag.GetFloat();
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpLessEqual )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() <= pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() <= pSearchTerm->tag.GetFloat();
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpGreater )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() > pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); it++ )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() > pSearchTerm->tag.GetFloat();
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpLess )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() < pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); ++it )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() < pSearchTerm->tag.GetFloat();
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpEqual )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); it++ )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() == pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); it++ )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() == pSearchTerm->tag.GetFloat();
				}
			}
		}
	}
	else if ( pSearchTerm->type == SSearchTerm::OpNotEqual )
	{
		if ( pSearchTerm->tag.IsInt() )
		{
			// meta tags with integer values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); it++ )
			{
				const CTag& tag = *it;

				if ( tag.IsInt() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetInt() != pSearchTerm->tag.GetInt();
				}
			}
		}
		else if ( pSearchTerm->tag.IsFloat() )
		{
			// meta tags with float values
			TagList::const_iterator it;

			for ( it = item->taglist.begin(); it != item->taglist.end(); it++ )
			{
				const CTag& tag = *it;

				if ( tag.IsFloat() && pSearchTerm->tag.GetID() == tag.GetID() )
				{
					return tag.GetFloat() != pSearchTerm->tag.GetFloat();
				}
			}
		}
	}

	return false;
}

//bool SearchTermsMatch(const SSearchTerm* pSearchTerm, const Kademlia::CEntry* item)
//{
//	// tokenize the filename (very expensive) only once per search expression and only if really needed
//	CStringArray astrFileNameTokens;
//	return SearchTermsMatch(pSearchTerm, item, astrFileNameTokens);
//}

void CIndexed::SendResults ( const CUInt128& keyID, std::list<CSourceKeyList::iterator> results, const CI2PAddress & dest )
{
    bool initPacket = true ;
    std::list<CSourceKeyList::iterator>::const_iterator ititSource = results.begin();
    std::list<CSourceKeyList::iterator>::const_iterator ititLastWritten ;
    CMemFile bio( CPacket::UDPPacketMaxDataSize() ) ;
    uint16_t count ;
    wxULongLong posNresults ;
    
    while ( ititSource != results.end() ) 
    {
        // packet initialization
        if (initPacket) {
            bio.SetLength( 0 );
            bio.WriteUInt128 ( keyID );
            posNresults = bio.GetPosition() ;
            bio.WriteUInt16 ( 0 ); // number of results. Will be overriden by "count"
            count = 0;
            initPacket = false ;   // unset initPacket flag
        }
        size_t posLastResult = bio.GetPosition().GetLo() ; // packet length before writing the result
        Kademlia::CEntry & currName = ( **ititSource )->second;
        bio.WriteUInt128 ( currName.sourceID );
        currName.taglist.WriteTo ( &bio );

        // test : if packet size is too big, revert last write
        //        and reset initPacket flag
        if (bio.GetPosition() > CPacket::UDPPacketMaxDataSize()) {
            initPacket = true ;
            bio.SetLength( posLastResult ) ;
        } else {
            count++;
        }

        // test : send packet if it is full or if we reached its max size
        ititLastWritten = ititSource ;
        ++ititSource ;
        if( initPacket or (ititSource==results.end()) ) {
            bio.Seek ( posNresults.GetValue() );
            bio.WriteUInt16 ( count );
            AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "KadSearchRes to %x" ) ) % dest.hashCode() );
            CKademlia::getUDPListener()->sendPacket ( &bio, KADEMLIA_SEARCH_RES, dest );
        }
        // loop to previous result if it has been discarded
        if ( initPacket ) {
            ititSource = ititLastWritten ;
            AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "KadSearchRes to %x : send more results" ) ) % dest.hashCode() );
        }
    }
}

void CIndexed::SendValidKeywordResult ( const CUInt128& keyID, const SSearchTerm* pSearchTerms, const CI2PAddress & dest )
{
    // TODO limit packet size here
	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "CIndexed::SendValidKeywordResult for key %s to dest %s" ) ) % keyID.toHexString() % dest.humanReadable() ) ;

	std::list<CSourceKeyList::iterator> results ;

	KeyHashMap::iterator itKeyHash = m_Keyword_map.find ( keyID );

	if ( itKeyHash != m_Keyword_map.end() )
	{
		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "CIndexed::SendValidKeywordResult :   key %s FOUND" ) )
		                   % keyID.toHexString() ) ;

		KeyHash & currKeyHash =  itKeyHash->second ;
		CSourceKeyList::iterator itSource = currKeyHash.m_Source_list.begin();

		for ( ; itSource != currKeyHash.m_Source_list.end(); ++itSource )
		{
			Kademlia::CEntry & currName = ( *itSource )->second;

			if ( !pSearchTerms || SearchTermsMatch ( pSearchTerms, &currName ) )
			{
				results.push_back ( itSource ) ;
			}
		}

		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "CIndexed::SendValidKeywordResult :   key %s -> %u files found" ) )

		                   % keyID.toHexString() % results.size() ) ;
	}
	else
	{
		AddDebugLogLineM ( false, logKadIndex,
		                   CFormat ( wxT ( "CIndexed::SendValidKeywordResult :   key %s NOT FOUND" ) )
		                   % keyID.toHexString() ) ;
	}

	SendResults ( keyID, results, dest ) ;

	clean();
}

/**
 * Sends :
 * - keyID (128)
 * - number n of sent entries (16)
 * - number of complete sources (32)
 * - total number of sources (32)
 * - n * (entryID(128) , taglist)
 */

void CIndexed::SendValidSourceResult ( const CUInt128& keyID, const CI2PAddress & dest )

{
	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "SendValidSourceResult key=%s to %x" ) ) % keyID.toHexString() % dest.hashCode() );

	SrcHashMap::iterator itSrcHash = m_Sources_map.find ( keyID );
	std::list<CSourceKeyList::iterator> results ;

	if ( itSrcHash != m_Sources_map.end() )
	{
		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "SendValidSourceResult key=%s found" ) ) % keyID.toHexString() );
		SrcHash& currSrcHash = itSrcHash->second;
		CSourceKeyList::iterator itSource = currSrcHash.m_Source_list.begin();

		for ( ; itSource != currSrcHash.m_Source_list.end(); ++itSource )
		{
			results.push_back ( itSource ) ;
		}
	}
	else
	{
		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "SendValidSourceResult key=%s NOT FOUND" ) ) % keyID.toHexString() );
	}

	SendResults ( keyID, results, dest ) ;

	clean();
}

void CIndexed::SendValidNoteResult ( const CUInt128& keyID, const CUInt128& WXUNUSED ( sourceID ), const CI2PAddress & dest )

{
	std::list<CSourceKeyList::iterator> results ;
	SrcHashMap::iterator itNote = m_Notes_map.find ( keyID );

	if ( itNote != m_Notes_map.end() )
	{
		SrcHash & currNoteHash = itNote->second;

		CSourceKeyList::iterator itSource = currNoteHash.m_Source_list.begin();

		for ( ; itSource != currNoteHash.m_Source_list.end(); ++itSource )
		{
			results.push_back ( itSource ) ;
		}
	}

	SendResults ( keyID, results, dest ) ;

	//clean(); //Not needed at the moment.
}

bool CIndexed::checkNotInLoadStatus ( const CUInt128& keyID )

{
	LoadMap::iterator it = m_Load_map.find ( keyID );

	if ( it != m_Load_map.end() )
	{
		Load& load = it->second;

		if ( load.time < ( uint32_t ) time ( NULL ) )
		{
			m_Load_map.erase ( it );
			return true;
		}

		return false;
	}

	return true;
}

SSearchTerm::SSearchTerm()

{
	type = AND;
	left = NULL;
	right = NULL;
}

SSearchTerm::~SSearchTerm()
{
	if ( type == String )
	{
		delete astr;
	}
}
