//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#ifndef __INDEXED_H__
#define __INDEXED_H__

#include <list>
#include <map>

#include "../../Types.h"
#include "SearchManager.h"
#include "../routing/Maps.h"
#include "../utils/UInt128.h"
#include "Entry.h"

class wxArrayString;

/*
// entry list (each one associates a key with a source (either a file or a client) )
// note : if it is a file, it can be duplicated with as many clients that share it

typedef std::list<Kademlia::CEntry*> CKadEntryPtrList;

// Source = (sourceID, entry list)
struct Source
{
	Kademlia::CUInt128 sourceID;
	CKadEntryPtrList entryList;
};

typedef std::list<Source*> CKadSourcePtrList;
// list of sources = list of (source key, one Source with this key as sourceID)
typedef std::map<Kademlia::CUInt128,Source*> CSourceKeyMap;
*/

//file for a keyword = (file key, file list)

typedef std::map<Kademlia::CUInt128, Kademlia::CEntry> CSourceKeyMap ;
typedef std::list<CSourceKeyMap::iterator>             CSourceKeyList;

struct KeyHash
{
	//Kademlia::CUInt128 keyID;
	//CSourceKeyMap m_Source_map;
	CSourceKeyMap         m_Source_map;    // maps source IDs to entries
	CSourceKeyList        m_Source_list;   // list of the map references ordered by inverse time of indexing
};


struct SrcHash
{
	//Kademlia::CUInt128 keyID;
	//CKadSourcePtrList m_Source_map;
	CSourceKeyMap         m_Source_map;    // maps source IDs to entries
	CSourceKeyList        m_Source_list;   // list of the map references ordered by inverse time of indexing
};

struct Load
{
//	Kademlia::CUInt128 keyID;
	uint32_t time;
};

// keywords list = list of (key, file)
typedef std::map<Kademlia::CUInt128,KeyHash> KeyHashMap;
typedef std::map<Kademlia::CUInt128,SrcHash> SrcHashMap;
typedef std::map<Kademlia::CUInt128,Load>    LoadMap;


struct SSearchTerm
{
	SSearchTerm();
	~SSearchTerm();
	
	enum ESearchTermType {
		AND,
		OR,
		NAND,
		String,
		MetaTag,
		OpGreaterEqual,
		OpLessEqual,
		OpGreater,
		OpLess,
		OpEqual,
		OpNotEqual
	} type;
	
	CTag           tag;
	wxArrayString* astr;

	SSearchTerm* left;
	SSearchTerm* right;
};

////////////////////////////////////////
namespace Kademlia {
////////////////////////////////////////

class CIndexed
{

public:
	CIndexed();
	~CIndexed();

	bool AddKeyword(const CUInt128& keyWordID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load);
	bool AddSources(const CUInt128& keyWordID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load);
	bool AddNotes(const CUInt128& keyID, const CUInt128& sourceID, Kademlia::CEntry& entry, uint8_t& load);
	bool AddLoad(const CUInt128& keyID, uint32_t time);
	size_t GetIndexedCount() {return m_Keyword_map.size() + m_Sources_map.size() + m_Notes_map.size();}
	size_t GetFileKeyCount() {return m_Sources_map.size();}

	void SendResults ( const CUInt128& keyID, std::list<CSourceKeyList::iterator> results, const CI2PAddress & dest ) ;
	void SendValidKeywordResult(const CUInt128& keyID, const SSearchTerm* pSearchTerms, const CI2PAddress & dest);
	void SendValidSourceResult(const CUInt128& keyID, const CI2PAddress & dest);
	void SendValidNoteResult(const CUInt128& keyID, const CUInt128& CheckID, const CI2PAddress & dest);

	bool checkNotInLoadStatus(const CUInt128& keyID);
	uint32_t GetSourceCount(const CUInt128& keyID) const;
	uint32_t GetCompleteSourceCount(const CUInt128& keyID) const;
	uint32_t m_totalIndexSource;
	uint32_t m_totalIndexKeyword;

private:
	time_t m_lastClean;
	KeyHashMap m_Keyword_map;
	SrcHashMap m_Sources_map;
	SrcHashMap m_Notes_map;
	LoadMap m_Load_map;
	static wxString m_sfilename;
	static wxString m_kfilename;
	static wxString m_loadfilename;
	void readFile(void);
	void clean(void);
};

} // End namespace

#endif //__INDEXED_H__
