//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/


#include "../../Tag.h"

#include "Search.h"
#include "Kademlia.h"
#include "../../OPCodes.h"
#include "Defines.h"
#include "Prefs.h"
#include "Indexed.h"
#include "../routing/RoutingZone.h"
#include "../routing/Contact.h"
#include "../net/KademliaUDPListener.h"
#include "../../imule.h"
#include "../../SharedFileList.h"
#include "../../OtherFunctions.h"
#include "../../KnownFile.h"
#include "DownloadQueue.h"
#include "PartFile.h"
#include "SearchList.h"
#include "MemFile.h"
#include "ServerConnect.h"
#include "Server.h"
#include "ClientList.h"
#include "updownclient.h"
#include "GuiEvents.h"          // Needed for Nofify_... macros
#include "Logger.h"
#include "../../Preferences.h"
#include "common/Format.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif


////////////////////////////////////////

using namespace Kademlia;

////////////////////////////////////////

CSearch::CSearch()
{
	m_created = time ( NULL );
	m_searchTerms = NULL;
	m_type = ( uint32_t ) - 1;
	m_completedPeers = 0;
	m_searchResults = 0;
	m_totalRequestAnswers = 0;
	m_kadPacketSent = 0;
	m_searchID = ( uint32_t ) - 1;
	m_keywordPublish = 0;
	( void ) m_fileName;
	m_stoping = false;
	m_totalLoad = 0;
	m_totalLoadResponses = 0;
	m_bio = NULL;
	m_lastResponse = time ( NULL );
}

CSearch::~CSearch()

{

	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i:  deleted search" ) ) % m_searchID ) ;

	CPartFile* temp = theApp.downloadqueue->GetFileByKadFileSearchID ( getSearchID() );

	if ( temp )
	{
		temp->SetKadFileSearchID ( 0 );
	}

	delete m_searchTerms;

// 	ContactList::const_iterator it2;
// 
// 	for ( it2 = m_delete.begin(); it2 != m_delete.end(); it2++ )
// 	{
// 		delete *it2;
// 	}

	if ( m_bio ) delete m_bio;

	if ( CKademlia::isRunning() && getNodeLoad() > 20 )
	{
		switch ( getSearchTypes() )
		{
			case CSearch::STOREKEYWORD:
				Kademlia::CKademlia::getIndexed()->AddLoad ( getTarget(), ( ( uint32_t ) ( DAY2S ( 7 ) * ( ( double ) getNodeLoad() / 100.0 ) ) + ( uint32_t ) time ( NULL ) ) );

				break;

			default:
				;

				// WTF?
		}
	}
	
	// add keywords if a file has been published
	// and updates the item in the sharedfileslist : new source counts can have been received
	if ( getSearchTypes() == CSearch::STOREFILE 
		    && getCompletedPeers() > 0 )
	{
		CKnownFile* file = NULL ;
		uint8_t fileid[16];
		m_target.toByteArray ( fileid );
		file = theApp.sharedfiles->GetFileByID ( CMD4Hash ( fileid ) );
		if (file ) {
			theApp.sharedfiles->AddKeywords( file ) ;
			Notify_SharedFilesUpdateItem(file);
		}
	}
}

void CSearch::go ( void )
{
	wxString debugStr;
#ifdef __DEBUG__
#define DEBSTR(X) debugStr << (X) << wxT(" ")
#else
#define DEBSTR(X)
#endif
	DEBSTR ( 0 );
	// Start with a lot of possible contacts, this is a fallback in case search stalls due to dead contacts

	wxASSERT( m_possible.empty() && m_responded.empty() );
	
	//if ( m_possible.empty() && m_responded.empty() )
	{
		DEBSTR ( 1 );
		CUInt128 distance ( CKademlia::getPrefs()->getKadID() );
		distance.XOR ( m_target );
		CKademlia::getRoutingZone()->getClosestTo ( 3, m_target, distance, 50, m_possible, true );
		// fall back to type 4 contacts if still empty

		if ( m_possible.empty() )
		{
			DEBSTR ( 1.1 );
			CKademlia::getRoutingZone()->getClosestTo ( 4, m_target, distance, 50, m_possible, true );
		}
	}

	DEBSTR ( 2 );


	//if ( m_possible.empty() && m_responded.empty() )
	if ( m_possible.empty() )
	{
		DEBSTR ( 3 );
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: nobody to contact while searching %s (debug steps=%s)" ) )
		                   % m_searchID  % m_target.toHexString() % debugStr ) ;
		return;
	}

	ContactMap::iterator it;

	// Let's keep our contact list entries in mind so that they don't get deleted during the search

	for ( it = m_possible.begin(); it != m_possible.end(); ++it )
	{
		m_delete.push_back( it->second );
		DEBSTR ( 4 );
	}

	wxASSERT ( m_possible.size() == m_delete.size() );

	// Take top ALPHA_QUERY(==3) possible
	CContact c;

	while ( m_possible.size() && m_best.size() < ALPHA_QUERY )
	{
		DEBSTR ( 5 );
		it = m_possible.begin();
		c = it->second;
		// Add to ALPHA_QUERY best list
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: GO ! %s added to m_best list searching %s (debug steps=%s)" ) )
		                   % m_searchID  % c.getUDPDest().humanReadable() % m_target.toHexString() % debugStr ) ;
		m_best [it->first] = c;
		// Add to tried list
		m_tried[it->first] = c;
		m_possible.erase ( it );
		m_lastResponse = time ( NULL );

		// Send request
		sendFindPeersForTarget ( c );

		if ( m_type == NODE )
		{
			break;
		}
	}
}



void CSearch::sendFindPeersForTarget ( CContact contact )
{
	try
	{
		if ( m_stoping )
		{
			return;
		}

		CMemFile bio ( 1/*m_type*/ + 16 /*m_target*/ + 16 /*dest ID (check)*/ + 16 /*my ID*/ );

		switch ( m_type )
		{
			case NODE:
			case NODECOMPLETE:
				bio.WriteUInt8 ( KADEMLIA_FIND_NODE );

				break;

			case FILE:
			case KEYWORD:
				//case FINDSOURCE:
			case NOTES:
				bio.WriteUInt8 ( KADEMLIA_FIND_VALUE );

				break;

				//case FINDBUDDY:
			case STOREFILE:
			case STOREKEYWORD:
			case STORENOTES:
				bio.WriteUInt8 ( KADEMLIA_STORE );

				break;

			default:
				AddDebugLogLineM ( false, logKadSearch, wxT ( "Invalid search type. (CSearch::sendFindPeersForTarget)" ) );

				return;
		}

		bio.WriteUInt128 ( m_target );

		bio.WriteUInt128 ( contact.getClientID() );
		bio.WriteUInt128 ( CKademlia::getPrefs()->getKadID() );

		m_kadPacketSent++;
#ifdef __DEBUG__
		wxString Type;

		switch ( m_type )
		{
			case NODE:
				Type = wxT ( "KadReqFindNode" );

				break;

			case NODECOMPLETE:
				Type = wxT ( "KadReqFindNodeCompl" );

				//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("%s")) % get_backtrace(1));
				break;

			case FILE:
				Type = wxT ( "KadReqFindFile" );

				break;

			case KEYWORD:
				Type = wxT ( "KadReqFindKeyw" );

				break;

			case STOREFILE:
				Type = wxT ( "KadReqStoreFile" );

				break;

			case STOREKEYWORD:
				Type = wxT ( "KadReqStoreKeyw" );

				break;

			case STORENOTES:
				Type = wxT ( "KadReqStoreNotes" );

				break;

			case NOTES:
				Type = wxT ( "KadReqNotes" );

				break;

			default:
				Type = wxT ( "KadReqUnknown" );
		}

		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: sendFindPeersForTarget to %s (type %s)" ) )

		                   % m_searchID % contact.getUDPDest().humanReadable() % Type );
#endif

		// send the request, but first signal the routing zone that the contact will be sent something
		// so that the type of contact will increase if it does not respond
		contact.checkingType() ;
		CKademlia::getUDPListener()->sendPacket ( &bio, KADEMLIA_REQ, contact.getUDPDest() );
	}
	catch ( const CSafeIOException & ioe )
	{
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "Exception in CSearch::sendFindPeersForTarget (IO error(%s))" ) ) % ioe.what() );
	}
}




//If we allow about a 15 sec delay before deleting, we won't miss a lot of delayed returning packets.

// stop looking for peers, but gives time for waiting for responses

void CSearch::stopLookingForPeers()
{
	if ( m_stoping )
	{
		return;
	}

	m_created = time ( NULL ) ;

	m_stoping = true;
}

bool CSearch::jumpStart ( void )

{
	// m_lastResponse is first set in CSearch::go
	// then at every reeceived response in CSearch::processResponse

	time_t now = time ( NULL );

	if ( ( time_t ) ( m_lastResponse + SEC ( 3 ) ) > now )
	{
		return true;
	}

	AddDebugLogLineM ( false, logKadSearch, 
			   CFormat ( wxT ( "#%i: jumpStart. searchAnswers: %d , maxAnswers: %d , tried: %d , answers: %d , possible: %d , responded: %d" ) ) 
					   % m_searchID 
					   % getSearchAnswers() % getSearchMaxAnswers() 
					   % m_tried.size() % getCompletedPeers() % m_possible.size() % m_responded.size() );
	// kill this search if max answers have been received
	// or if every possible peers have already answered with results
	if ( getSearchAnswers() >= getSearchMaxAnswers() ||
		    ( m_tried.size() == getCompletedPeers() && m_possible.empty() && m_responded.empty() ) )
	{
		if ( m_type == NODECOMPLETE && getSearchAnswers() != 0 ) CKademlia::getPrefs()->setPublish ( true );

		return false ;
	}

	// if timeout is reached,
	// first time : stop looking for peers (side effect : m_created <- now)
	//    but still waits for answers if needed
	// second time : kill this search
	if ( m_created + getSearchLifetime() < now )
	{
		if ( m_type == NODECOMPLETE && getSearchAnswers() != 0 ) CKademlia::getPrefs()->setPublish ( true );

		if ( ! m_stoping && m_tried.size() > getCompletedPeers() )
			stopLookingForPeers() ;
		else
			return false ;
	}


	//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("#%i:  CSearch::jumpStart. %d possible, %d responded")) % m_searchID  % m_possible.size() % m_responded.size()) ;

	while ( !m_responded.empty() )
	{
		// send request to contact that have responded
		sendRequestToSelectedPeers();
		m_responded.erase ( m_responded.begin() );

	}

	if ( !m_possible.empty() )
	{
		CContact c = m_possible.begin()->second;
		// Move to tried
		m_tried[m_possible.begin()->first] = c;
		m_possible.erase ( m_possible.begin() );
		// Send request
		sendFindPeersForTarget ( c );
		// and quit, waiting for the new requesting process to time out
	}

	return true ;
}

void CSearch::processGotPeerListForTarget ( const CI2PAddress & fromDest, ContactList *results )
{
	//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Process search response from %x, count=%d")) % fromDest.hashCode() % results->size());

	if ( results )
	{
		m_lastResponse = time ( NULL );
	}

	// Remember the contacts to be deleted if not used when the search will be deleted finished
	m_delete.insert(m_delete.end(), results->begin(), results->end());
	
	
	ContactMap::const_iterator tried;

	CContact c;
	CContact from;
	CUInt128 distance;
	CUInt128 fromDistance;

	//Find contact that is responding amongst tried contacts

	//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Looking for %x in the list of tried clients")) % fromDest.hashCode());

	for ( tried = m_tried.begin(); tried != m_tried.end(); ++tried )
	{
		fromDistance = tried->first;
		from = tried->second;

		//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Looking ... %x = %x ?")) % fromDest.hashCode() % from->getUDPDest().hashCode());

		if ( ( from.getUDPDest() == fromDest ) )
		{
			//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("Looking ... %x = %x yes")) % fromDest.hashCode() % from->getUDPDest().hashCode());

			// Add to list of people who responded

			// Not interested in responses for FIND_NODE, has already been added to contacts by udp listener
			// (CKademliaUDPListener::processKademliaResponse)

			if ( m_type == NODE )
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Node type search result, discarding." ) ) % m_searchID );
				m_searchResults++;
				m_completedPeers ++ ;
				m_possible.clear();
				return;
			}

			if ( m_type == NODECOMPLETE )
				// Count this as a search answer, because we won't ask anything more to this peer
			{
				m_completedPeers ++;
			}

			m_responded[fromDistance] = from;

			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x added to the list of people who responded, searching %s" ) )
			                   % m_searchID % fromDest.hashCode() % m_target.toHexString() );

			// Loop through its responses

			for ( ContactList::const_iterator response = results->begin(); response != results->end();
						   ++response )
			{
				c = *response;

				distance = c.getClientID();
				distance.XOR ( m_target );

				// Ignore this contact if already know him

				if ( m_possible.count ( distance ) > 0 )
				{
					AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x already in list of possible clients searching %s" ) )
					                   % m_searchID % c.getUDPDest().hashCode() % m_target.toHexString() );
					continue;
				}

				if ( m_tried.count ( distance ) > 0 )
				{
					AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x already in list of asked    clients searching %s" ) )
					                   % m_searchID % c.getUDPDest().hashCode() % m_target.toHexString() );
					continue;
				}

				// Add to possible
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x added to the list of possible clients searching %s" ) )
				                   % m_searchID % c.getUDPDest().hashCode() % m_target.toHexString() );

				m_possible[distance] = c;

				// Try it now if in m_best list
				//if (distance < fromDistance) {

				bool top = false;

				if ( m_best.size() < ALPHA_QUERY )
				{
					AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x added to m_best list for searching %s" ) )
					                   % m_searchID % c.getUDPDest().hashCode() % m_target.toHexString() ) ;
					top = true;
					m_best[distance] = c;
				}
				else
				{
					//AddDebugLogLineM(false, logKadSearch, wxT("m_best.size() >= ALPHA_QUERY"));
					ContactMap::iterator it = m_best.end();
					--it;

					if ( distance < it->first )
					{
						// Insert c in best list
						AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %x added to m_best list searching %s, replacing %x" ) )
						                   % m_searchID % c.getUDPDest().hashCode() % m_target.toHexString() % it->second.getUDPDest().hashCode() ) ;
						m_best.erase ( it );
						m_best[distance] = c;
						top = true;
					}
					else
					{
						//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("%x not added to m_best list for search %s")) % c->getUDPDest().hashCode() % m_target.toHexString()) ;
					}
				}

				if ( top )
				{
					// Move to tried
					m_tried[distance] = c;
					m_possible.erase ( distance );

					// Send request
					sendFindPeersForTarget ( c );
				}

				//}
			}

			// We don't want anything from these people, so just increment the counter.
			if ( m_type == NODECOMPLETE )
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Search result type: Node complete" ) ) % m_searchID );
				m_searchResults++;
			}

			// We found the one that has responded, so stop looking for it in the m_tried list
			break;
		}
	}
}


// sends the effective request to peers that have responded to peer request

void CSearch::sendRequestToSelectedPeers()

{
	//wxASSERT(!m_possible.empty());
	wxASSERT ( !m_responded.empty() );

	CContact from;
	CUInt128 fromDistance;

	ContactMap::const_iterator responded;
	responded = m_responded.begin();
	fromDistance = responded->first;
	from         = responded->second;


	switch ( m_type )
	{

		case FILE:

		case KEYWORD:
		{
			if ( m_type == FILE )
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers (type: File) to %s" ) ) % m_searchID % from.getUDPDest().humanReadable() );
			}
			else
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers (type: Keyword) to %s" ) ) % m_searchID % from.getUDPDest().humanReadable() );
			}

			wxASSERT ( m_searchTerms->GetLength() > 0 );

			// The data in 'm_searchTerms' is to be sent several times, so use the don't detach flag.
			CKademlia::getUDPListener()->sendPacket ( m_searchTerms, KADEMLIA_SEARCH_REQ, from.getUDPDest() );
			m_totalRequestAnswers++;
			break;
		}

		case NOTES:
		{
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers type: Note" ) ) % m_searchID );
			CMemFile bio ( 34 );
			bio.WriteUInt128 ( m_target );
			bio.WriteUInt128 ( CKademlia::getPrefs()->getKadID() );
			CKademlia::getUDPListener()->sendPacket ( &bio, KADEMLIA_SRC_NOTES_REQ, from.getUDPDest() );
			m_totalRequestAnswers++;
			break;
		}

		case STOREFILE:
		{
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers type: StoreFile" ) ) % m_searchID );

			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: %s agreed storing me (%s) as a source for file (%s)" ) )
			                   % m_searchID
			                   % from.getUDPDest().humanReadable()
			                   % theApp.GetTcpDest().humanReadable()
			                   % m_target.toHexString()
			                   );

			if ( m_bio )
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: KadStoreFile to %s" ) ) 
						% m_searchID
						% from.getUDPDest().humanReadable() );

				CKademlia::getUDPListener()->sendPacket ( m_bio, KADEMLIA_PUBLISH_REQ, from.getUDPDest() );
				m_totalRequestAnswers++;
			}
			break;
		}

		case STOREKEYWORD:
		{
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers type: StoreKeyword" ) ) % m_searchID );

			if ( m_bio )
			{
				AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "KadStoreKeywReq to %s" ) ) % from.getUDPDest().humanReadable() );

				CKademlia::getUDPListener()->sendPacket ( m_bio, KADEMLIA_PUBLISH_REQ, from.getUDPDest() );
				m_totalRequestAnswers++;
			}


			break;
		}

		case STORENOTES:
		{
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Send sendRequestToSelectedPeers type: StoreNotes" ) ) % m_searchID );
			
			if ( m_bio )
			{
				CKademlia::getUDPListener()->sendPacket ( m_bio, KADEMLIA_PUB_NOTES_REQ, from.getUDPDest() );
				m_totalRequestAnswers++;
			}

			break;
		}

		case NODECOMPLETE:
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Search result type : NodeComplete" ) ) % m_searchID );

			break;

		case NODE:
			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Search result type : Node" ) ) % m_searchID );

			break;

		default:
			AddDebugLogLineM ( false, logKadSearch, wxString::Format ( wxT ( "#%i:Search result type: Unknown (%i)" ), m_searchID, m_type ) );

			break;
	}
}

void CSearch::processSearchResultList ( CMemFile & bio )

{
	
	// How many results.. Not supported yet..
	uint16_t count = bio.ReadUInt16();

	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: processSearchResultList (%d results)" ) )
			% m_searchID % count );
	
	while ( count > 0 )
	{
		// What is the answer
		CUInt128 answer = bio.ReadUInt128();

		// Get info about answer

		try
		{
			AddDebugLogLineM ( false, logClientKadUDP,
			                   CFormat ( wxT ( "processSearchResultList : %d response(s) remaining (rem. filesize=%s)" ) )
			                   % count %  bio.GetRemaining().ToString() );

			TagList tags ( bio );

			if ( processSearchResult ( answer, tags ) )
				m_searchResults++ ;

			count--;
		}
		catch ( ... )
		{
			AddDebugLogLineM ( false, logKadSearch, wxT ( "CSearch::processSearchResultList : error while trying to add a result" ) );
			throw;
		}
	}

	m_completedPeers ++ ;
}


bool CSearch::processSearchResult ( const CUInt128 &answer, TagList & info )
{
	bool added = false ;
	wxString type;

	switch ( m_type )
	{
		case FILE:
			type = wxT ( "File" );

			break;

		case KEYWORD:
			type = wxT ( "Keyword" );

			break;

		case NOTES:
			type = wxT ( "Notes" );

			break;

		default:
			type = wxT ( "Unknown" );
	}

	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: type=%s processSearchResult : answer = %s" ) ) % m_searchID % type  % answer.toHexString() );

	switch ( m_type )
	{
		case FILE:
			added = processResultFile ( answer, info );

			break;

		case KEYWORD:
			added = processResultKeyword ( answer, info );

			break;

		case NOTES:
			added = processResultNotes ( answer, info );

			break;

		default:
			added = false ;

			break;
	}

	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: Result %s processed" ) ) % m_searchID % type );

	return added ;
}




bool CSearch::processResultFile ( const CUInt128 &answer, TagList & info )

{
	bool added = false ;
	uint8_t       type = 0;
	CI2PAddress tcpdest;
	CI2PAddress udpdest;
	CI2PAddress serverdest;
	uint32_t      clientid = 0;

	for ( TagList::const_iterator it = info.begin(); it != info.end(); ++it )
	{
		const CTag & tag = *it;

		if ( tag . GetID() == TAG_SOURCETYPE )
		{
			type = (uint8_t) tag . GetInt().GetValue();
		}
		else if ( tag . GetID() == TAG_SOURCEDEST )
		{
			tcpdest = tag . GetAddr();
		}
		else if ( tag . GetID() == TAG_SOURCEUDEST )
		{
			udpdest = tag . GetAddr();
		}
		else if ( tag . GetID() == TAG_SERVERDEST )
		{
			serverdest = tag . GetAddr();
		}
	}

	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: source type : %d" ) ) % m_searchID % type );

	switch ( type )
	{
		case 1:
		case 3:
		{
			added = theApp.downloadqueue->KademliaSearchFile ( m_searchID, &answer, type, tcpdest, udpdest, serverdest, clientid ) ;
			break;
		}

		case 2:
		{
			//Don't use this type, some clients will process it wrong..
			break;
		}
	}

	return added ;
}

bool CSearch::processResultNotes ( const CUInt128 &answer, TagList & info )

{
	bool added = false ;
	CEntry entry;
	entry.keyID.setValue ( m_target );
	entry.sourceID.setValue ( answer );

	bool bFilterComment = false;

	TagList::const_iterator it;

	for ( it = info.begin(); it != info.end(); it++ )
	{
		const CTag & tag = *it;

		/*I2P:if (!tag . m_name.Cmp(wxT(TAG_SOURCEIP))) {
			entry.ip = tag . GetInt();
			delete tag;
		} else*/
		//if (!tag . m_name.Cmp(wxT(TAG_SOURCEPORT))) {

		if ( tag . GetID() == TAG_SOURCEDEST )
		{
			entry.tcpdest = tag . GetAddr();
		}
		else if ( tag . GetID() == TAG_FILENAME )
		{
			entry.fileName	= tag . GetStr();
		}
		else if ( tag . GetID() == TAG_DESCRIPTION )
		{
			wxString strComment ( tag . GetStr() );
			bFilterComment = thePrefs::IsMessageFiltered ( strComment );
			entry.taglist.push_front ( tag );
		}
		else if ( tag . GetID() == TAG_FILERATING )
		{
			entry.taglist.push_front ( tag );
		}
	}

	if ( bFilterComment )
	{
		return added ;
	}

	uint8_t fileid[16];

	m_target.toByteArray ( fileid );

	const CMD4Hash fileHash ( fileid );

	//Check if this hash is in our shared files..
	CKnownFile* file = theApp.sharedfiles->GetFileByID ( fileHash );

	if ( !file )
	{
		// If we didn't find anything check if it's in our download queue.
		file = ( CKnownFile* ) theApp.downloadqueue->GetFileByID ( fileHash );
	}

	if ( file )
	{
		// Ok, found the file. Add the note to it.
		added = file->AddNote ( entry ) ;
	}
	else
	{
		AddDebugLogLineM ( false, logKadSearch, wxT ( "Comment received for unknown file" ) );
	}

	return added ;
}

bool CSearch::processResultKeyword ( const CUInt128 &answer, TagList & info )

{
	bool added = false ;
	bool interested = false;
	wxString name;
	wxULongLong size = 0;
	wxString type;

	TagList taglist;

	for ( TagList::const_iterator it = info.begin(); it != info.end(); ++it )
	{
		const CTag& tag = *it;

		switch ( tag.GetID() )
		{
			case TAG_FILENAME :
				name = tag . GetStr();

				interested	= ! ( name . IsEmpty() );

				break ;

			case TAG_FILESIZE :
				size = tag . GetInt();

				break ;

			case  TAG_FILETYPE :
				type = tag . GetStr();

				break ;

			default :
				taglist. push_back ( tag );
		}
	}


	if ( interested )
	{
		added = theApp.searchlist->KademliaSearchKeyword ( m_searchID, &answer, name, size, type, taglist ) ;
	}
	else
	{
		printf ( "Not adding search results: not interested\n" );
	}

	return added ;
}

void CSearch::processPublishResult ( CMemFile & bio )

{
	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "#%i: processPublishResult" ) ) 
						% m_searchID  
			 );
	// Verify packet is expected size
	uint8_t minExpectedSize = 0 ;

	if ( m_type == STOREFILE ) minExpectedSize = 8 ;

	if ( bio.GetRemaining() < minExpectedSize )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s, at least %u expected) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % minExpectedSize % wxString::FromAscii ( __func__ ) );
	}


	if ( m_type == STOREFILE )
	{
		// update source counts in file entry
		uint32_t total = bio.ReadUInt32();
		uint32_t complete = bio.ReadUInt32();

		uint8_t fileid[16];
		m_target.toByteArray ( fileid );
		CKnownFile * file =	theApp.sharedfiles->GetFileByID ( CMD4Hash ( fileid ) ) ;

		if ( file != NULL )
		{
			file->AddTagUnique ( CTag ( TAG_COMPLETE_SOURCES, wxULongLong
			                            (std::max ( ( uint32_t ) file->GetIntTagValue ( TAG_COMPLETE_SOURCES ).GetValue(), complete )) ) );
			file->AddTagUnique ( CTag ( TAG_SOURCES, wxULongLong
			                            (std::max ( ( uint32_t ) file->GetIntTagValue ( TAG_SOURCES ).GetValue(), total )) ) );
		}
	}

	if ( bio.GetLength() > bio.GetPosition() )
	{
		uint8_t load = bio.ReadUInt8();
		updateNodeLoad ( load );
	}

	m_searchResults     ++ ;
	m_completedPeers ++ ;
}


void CSearch::addFileID ( const CUInt128& id )

{
	m_fileIDs.push_back ( id );
}


void CSearch::PreparePacketForTags ( CMemFile *bio, CKnownFile *file )

/*! Prepares a CSearch for keyword publishing.
 * Puts infos (taglist) in a packet :
 * - file name
 * - file size
 * - number of complete sources
 * - total number of sources
 * - OPT: file type
 * - OPT: file format
 * - OPT: TAG_MEDIA_ARTIST
 * - OPT: TAG_MEDIA_ALBUM
 * - OPT: TAG_MEDIA_TITLE
 * - OPT: TAG_MEDIA_LENGTH
 * - OPT: TAG_MEDIA_BITRATE
 * - OPT: TAG_MEDIA_CODEC
 */
{

	try
	{
		TagList taglist;

		if ( file && bio )
		{
			// Name, Size
			taglist.push_back ( CTag ( TAG_FILENAME, file->GetFileName() ) );
			taglist.push_back ( CTag ( TAG_FILESIZE, file->GetFileSize() ) );
			taglist.push_back ( file->GetTag ( TAG_COMPLETE_SOURCES ) );
			taglist.push_back ( file->GetTag ( TAG_SOURCES ) );

			AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "PreparePacketForTags. Filename : %s, TAG_SOURCES : %d" ) )
			                   % file->GetFileName() % file->m_nCompleteSourcesCount );

			// eD2K file type (Audio, Video, ...)
			// NOTE: Archives and CD-Images are published with file type "Pro"
			wxString strED2KFileType ( GetED2KFileTypeSearchTerm ( GetED2KFileTypeID ( file->GetFileName() ) ) );

			if ( !strED2KFileType.IsEmpty() )
			{
				taglist.push_back ( CTag ( TAG_FILETYPE, strED2KFileType ) );
			}

			// file format (filename extension)
			int iExt = file->GetFileName().Find ( wxT ( '.' ), true );

			if ( iExt != -1 )
			{
				wxString strExt ( file->GetFileName().Mid ( iExt ) );

				if ( !strExt.IsEmpty() )
				{
					strExt = strExt.Mid ( 1 );

					if ( !strExt.IsEmpty() )
					{
						taglist.push_back ( CTag ( TAG_FILEFORMAT, strExt ) );
					}
				}
			}

			// additional meta data (Artist, Album, Codec, Length, ...)
			// only send verified meta data to nodes
			if ( file->GetMetaDataVer() > 0 )
			{

				static const struct
				{
					uint8_t	nName;
					uint8_t	nType;
				}

				_aMetaTags[] =

				    {
				        { TAG_MEDIA_ARTIST,  2 },
				        { TAG_MEDIA_ALBUM,   2 },
				        { TAG_MEDIA_TITLE,   2 },
				        { TAG_MEDIA_LENGTH,  3 },
				        { TAG_MEDIA_BITRATE, 3 },
				        { TAG_MEDIA_CODEC,   2 }
				    };

				for ( int i = 0; i < ARRSIZE ( _aMetaTags ); i++ )
				{
					const ::CTag& pTag = file->GetTag ( _aMetaTags[i].nName, _aMetaTags[i].nType );

					if ( pTag.isValid() )
					{
						// skip string tags with empty string values

						if ( pTag . IsStr() && pTag . GetStr().IsEmpty() )
						{
							continue;
						}

						// skip integer tags with '0' values
						if ( pTag . IsInt() && pTag . GetInt() == 0 )
						{
							continue;
						}

						taglist.push_back ( pTag ) ;
					}
				}
			}

			taglist.WriteTo ( bio );

			taglist.deleteTags();
		}
		else if ( bio )
		{ // empty tag list
			taglist.WriteTo ( bio );
		}
		else
		{
			//If we get here.. Bad things happen.. Will fix this later if it is a real issue.
			wxASSERT ( 0 );
		}
	}
	catch ( const CSafeIOException & ioe )
	{
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "Exception in CSearch::PreparePacketForTags (IO error(%s))" ) ) % ioe.what() );
	}
}

void CSearch::PreparePublishPacket ( void )
{
	CKnownFile* file = NULL ;
	uint8_t fileid[16];
	try
	{
		switch ( m_type )
		{
			case STORENOTES :
				wxASSERT ( !m_bio );
				m_target.toByteArray ( fileid );
				file = theApp.sharedfiles->GetFileByID ( CMD4Hash ( fileid ) );

				if ( file )
				{
					m_bio = new CMemFile ( 2*1024 );
					m_bio->WriteUInt128 ( m_target );
					m_bio->WriteUInt128 ( CKademlia::getPrefs()->getKadID() );

					TagList tags ;
					tags.push_back ( CTag ( TAG_FILENAME, file->GetFileName() ) );

					if ( file->GetFileRating() != 0 )
						tags.push_back ( CTag ( TAG_FILERATING, wxULongLong(file->GetFileRating()) ) );

					if ( !file->GetFileComment().IsEmpty() )
						tags.push_back ( CTag ( TAG_DESCRIPTION, file->GetFileComment() ) );

					tags.WriteTo ( m_bio );
				}
				if ( !file ) throw CIOFailureException( _("File not found on shared when preparing STORENOTES packet!\n" ) );
				break;

			case STOREFILE :
				wxASSERT ( !m_bio );
				m_target.toByteArray ( fileid );
				file = theApp.sharedfiles->GetFileByID ( CMD4Hash ( fileid ) );

				if ( file )
				{
					m_bio = new CMemFile ( 1024 );
					m_bio->WriteUInt128 ( m_target );
					//We only use this for publishing sources now.. So we always send one here..
					m_bio->WriteUInt16 ( 1 );
					m_bio->WriteUInt128 ( CKademlia::getPrefs()->getClientHash() );
				
					// tags
					TagList taglist;
					uint32_t flags = 0;
					
					m_fileName = file->GetFileName();

					if ( !file->IsPartFile() )
						flags = flags | TAG_FLAGS_COMPLETEFILE ;

						//We can use type for different types of sources.
						//1 is reserved for highID sources..
						//2 cannot be used as older clients will not work.
						//3 Firewalled Kad Source.

					taglist.push_back ( CTag ( TAG_SOURCETYPE, wxULongLong(1) ) );
					taglist.push_back ( CTag ( TAG_SOURCEDEST, theApp.GetTcpDest() ) );
					taglist.push_back ( CTag ( TAG_FLAGS, wxULongLong(flags) ) );


					taglist.WriteTo ( m_bio );
				}
				if ( !file ) throw CIOFailureException( _("File not found on shared when preparing STOREFILE packet!\n" ) );
				break ;
				
			case STOREKEYWORD :
				
				uint16_t count = std::min ( (uint16_t) m_fileIDs.size(), ( uint16_t ) 150 );

				wxASSERT ( !m_bio );

				m_bio = new CMemFile ( 1024 * count );

				m_bio->WriteUInt128 ( m_target );

				m_bio->WriteUInt16 ( count );

				while ( count > 0 )
				{
					count--;
					m_bio->WriteUInt128 ( m_fileIDs.front() );
					m_fileIDs.front().toByteArray ( fileid );
					m_fileIDs.pop_front();
					file = theApp.sharedfiles->GetFileByID ( CMD4Hash ( fileid ) );

					if ( !file ) throw CIOFailureException( _("File not found on shared when preparing STOREKEYWORD packet!\n" ) );

					PreparePacketForTags ( m_bio, file );
				}

				break ;
		}

	}
	catch ( const CSafeIOException& ioe )
	{
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "Exception in CSearch::PreparePacket (IO error(%s))" ) ) % ioe.what() );
		throw ioe ;
	}
}

uint32_t CSearch::getNodeLoad() const
{
	if ( m_totalLoadResponses == 0 )
	{
		return 0;
	}

	return m_totalLoad / m_totalLoadResponses;
}

time_t CSearch::getSearchLifetime()

{
	switch ( getSearchTypes() )
	{
		case FILE                        : return SEARCHFILE_LIFETIME ;

		case KEYWORD              : return SEARCHKEYWORD_LIFETIME ;

		case NOTES                   : return SEARCHNOTES_LIFETIME ;

		case NODE			 : return SEARCHNODE_LIFETIME ;

		case NODECOMPLETE : return SEARCHNODECOMP_LIFETIME ;

		case STOREFILE            : return SEARCHSTOREFILE_LIFETIME ;

		case STOREKEYWORD  : return SEARCHSTOREKEYWORD_LIFETIME ;

		case STORENOTES	 : return SEARCHSTORENOTES_LIFETIME ;

		default                           : return SEARCH_LIFETIME ;
	}
}

uint32_t CSearch::getSearchMaxAnswers()
{
	switch ( getSearchTypes() )
	{
		case FILE                        : return SEARCHFILE_TOTAL ;

		case KEYWORD              : return SEARCHKEYWORD_TOTAL ;

		case NOTES                   : return SEARCHNOTES_TOTAL ;

		case NODE			 : return 1 ;

		case NODECOMPLETE : return SEARCHNODECOMP_TOTAL ;

		case STOREFILE            : return SEARCHSTOREFILE_TOTAL ;

		case STOREKEYWORD  : return SEARCHSTOREKEYWORD_TOTAL ;

		case STORENOTES	 : return SEARCHSTORENOTES_TOTAL ;

		default                           : return 1 ;
	}
}

