//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#ifndef __SEARCH_H__
#define __SEARCH_H__

#include "SearchManager.h"
#include "../routing/Maps.h"
#include "../utils/UInt128.h"
#include "../../Tag.h"

#include "i2p/CI2PAddress.h"

class CKnownFile;
class CMemFile;
class CContact;

////////////////////////////////////////
namespace Kademlia {
////////////////////////////////////////

class CSearch
{
	friend class CSearchManager;

public:
	uint32_t getSearchID() const {return m_searchID;}
	uint32_t getSearchTypes() const {return m_type;}
	void setSearchTypes( uint32_t val ) {m_type = val;}
	void setTargetID( CUInt128 val ) {m_target = val;}

	time_t getSearchLifetime() ;
	uint32_t getSearchMaxAnswers() ;
	uint32_t getSearchAnswers() const {return m_searchResults ; }
	uint32_t getCompletedPeers()       const {return m_completedPeers; }
	uint32_t getKadPacketSent() const {return m_kadPacketSent;}
	uint32_t getRequestAnswer() const {return m_totalRequestAnswers;}
	void sendRequestToSelectedPeers();

	CUInt128 m_keywordPublish; //Need to make this private...
	
	CMemFile * m_bio;
	
	const wxString& getFileName(void) const { return m_fileName; }
	void setFileName(const wxString& fileName) { m_fileName = fileName; }
	CUInt128 getTarget(void) const {return m_target;}
	void addFileID(const CUInt128& id);
	void PreparePublishPacket(void);
	void PreparePacketForTags( CMemFile* packet, CKnownFile* file );
	bool Stoping(void) const {return m_stoping;}
	uint32_t getNodeLoad() const;
	uint32_t getNodeLoadResonse() const {return m_totalLoadResponses;}
	uint32_t getNodeLoadTotal() const {return m_totalLoad;}
	void updateNodeLoad( uint8_t load ){ m_totalLoad += load; m_totalLoadResponses++; }

	enum
	{
		NODE,
		NODECOMPLETE,
		FILE,
		KEYWORD,
		NOTES,
		STOREFILE,
		STOREKEYWORD,
		STORENOTES,
	};

	CSearch();
	~CSearch();

private:
	void go(void);
	void processGotPeerListForTarget(const CI2PAddress & fromDest, ContactList *results);
	void processSearchResultList ( CMemFile & bio ) ;
	void processPublishResult ( CMemFile & bio ) ;
	bool processSearchResult( const CUInt128 &answer, TagList & info ) ;
	bool processResultFile( const CUInt128 &answer, TagList &info );
	bool processResultKeyword( const CUInt128 &answer, TagList &info);
	bool processResultNotes( const CUInt128 &answer, TagList &info);
	bool jumpStart(void);
	void sendFindPeersForTarget(CContact contact);
	void stopLookingForPeers(void);

	bool		m_stoping;
	time_t		m_created;
	uint32_t		m_type;
	uint32_t		m_completedPeers;         // counts the peers from which we don't wait for anything more
	uint32_t		m_searchResults ;            // counts the number of results received
	uint32_t		m_totalRequestAnswers;
	uint32_t		m_kadPacketSent; //Used for gui reasons.. May not be needed later..
	uint32_t		m_totalLoad;
	uint32_t		m_totalLoadResponses;
	uint32_t		m_lastResponse;

	uint32_t		m_searchID;
	CUInt128	m_target;
	CMemFile*	m_searchTerms;
	WordList	m_words;
	wxString	m_fileName;
	UIntList	m_fileIDs;

	ContactMap	m_possible;                // fallback virtual contacts
	ContactMap	m_tried;                   // virtual contacts asked for closer virtual contacts
	ContactMap	m_best;                    // list of ALPHA_QUERY closest virtual contacts
	ContactMap	m_responded;               // contacts that replied with their list of closer virtual contacts
	ContactList	m_delete;                  // contacts to be deleted on destruction if not used elsewhere
};

} // End namespace

#endif //__SEARCH_H__
