//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

//#include "stdafx.h"
//#include "resource.h"
#include "SearchManager.h"
#include "Search.h"
#include "Kademlia.h"
#include "Indexed.h"
#include "../../OPCodes.h"
#include "Defines.h"
#include "../routing/Contact.h"
#include "../utils/UInt128.h"
#include "../kademlia/Prefs.h"
#include "MemFile.h"
#include "OtherFunctions.h"
#include "Logger.h"
#include "common/Format.h"

#include <wx/tokenzr.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif


wxString InvKadKeywordChars = wxT ( " ()[]{}<>,._-!?:;\\/" );

////////////////////////////////////////

using namespace Kademlia;

////////////////////////////////////////

uint32_t CSearchManager::m_nextID = 0;

SearchMap CSearchManager::m_searches;

void CSearchManager::stopSearch ( uint32_t searchID, bool delayDelete )
{
	if ( searchID == ( uint32_t ) - 1 )
	{
		return;
	}

	SearchMap::iterator it = m_searches.begin();

	while ( it != m_searches.end() )
	{
		if ( it->second->m_searchID == searchID )
		{
			if ( delayDelete )
			{
				it->second->stopLookingForPeers();
				++it;
			}
			else
			{
				delete it->second;
				m_searches.erase ( it++ );
			}
		}
		else
		{
			++it;
		}
	}
}

bool CSearchManager::isNodeSearch ( const CUInt128 &target )

{
	SearchMap::iterator it = m_searches.begin();

	while ( it != m_searches.end() )
	{
		if ( it->second->m_target == target )
		{
			return 	it->second->m_type == CSearch::NODE ||
			        it->second->m_type == CSearch::NODECOMPLETE;
		}
		else
		{
			++it;
		}
	}

	return false;
}

void CSearchManager::stopAllSearches ( void )

{
	SearchMap::iterator it;

	for ( it = m_searches.begin(); it != m_searches.end(); ++it )
	{
		delete it->second;
	}

	m_searches.clear();
}

bool CSearchManager::startSearch ( CSearch* pSearch )
{
	if ( alreadySearchingFor ( pSearch->m_target ) )
	{
		delete pSearch;
		return false;
	}

	m_searches[pSearch->m_target] = pSearch;

	pSearch->go();
	return true;
}

void CSearchManager::deleteSearch ( CSearch* pSearch )
{
	delete pSearch;
}

CSearch* CSearchManager::prepareFindKeywords ( const wxString& keyword, CMemFile* ed2k_packet, uint32_t searchid )
{
	CSearch *s = new CSearch;
#ifdef __DEBUG__
	wxString debugStr;
	int line = 0;
#define DEBSTR(X) debugStr << (X) << wxT(" ")
#else
#define DEBSTR(X)
#endif

	try
	{
		DEBSTR(0);
		s->m_type = CSearch::KEYWORD;

		// This will actually get the first word.
		getWords ( keyword, &s->m_words );

		if ( s->m_words.size() == 0 )
		{
			throw wxString ( _ ( "Kademlia: search keyword too short" ) );
		}

		wxString wstrKeyword = s->m_words.front();

		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "Keyword for search: %s" ) ) % wstrKeyword );
		DEBSTR(1);
		
		KadGetKeywordHash ( wstrKeyword, &s->m_target );
		DEBSTR(2);

		if ( alreadySearchingFor ( s->m_target ) )
		{
			DEBSTR(3);
			throw wxT ( "Kademlia: Search keyword is already on search list: " ) + wstrKeyword;
		}

		// Write complete packet
		s->m_searchTerms = ed2k_packet;

		DEBSTR(4);
		s->m_searchTerms->Seek ( 0, wxFromStart );
		DEBSTR(5);
		s->m_searchTerms->WriteUInt128 ( s->m_target );
		DEBSTR(6);
		if ( s->m_searchTerms->GetLength() > ( 16 /*uint128*/ + 1 /*uint8_t*/ ) )
		{
		    DEBSTR(7);
			// There is actually ed2k search data
			s->m_searchTerms->WriteUInt8 ( 1 );
			DEBSTR(8);
		} // 0 is default, no need for else branch
		DEBSTR(9);
		// if called from external client - use predefined search id
		s->m_searchID = ( ( searchid & 0xffffff00 ) == 0xffffff00 ) ? searchid : ++m_nextID;
		DEBSTR(10);
		m_searches[s->m_target] = s;
		DEBSTR(11);
		s->go();
		DEBSTR(12);
	}
	catch ( const CSafeIOException & ioe )
	{
		wxString strError = wxString ( wxT ( "IO-Exception in " ) ) + wxString::FromAscii ( __FUNCTION__ ) + ( CFormat ( wxT ( ": Error %s" ) ) % ioe.what() );
		delete s;
		throw strError;
	}
	catch ( ... )
	{
		delete s;
#ifdef __DEBUG__
		AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "Error trace: %s" ) ) % debugStr );
#endif
		throw;
	}

#ifdef __DEBUG__
	AddDebugLogLineM ( false, logKadSearch, CFormat ( wxT ( "No Error trace: %s" ) ) % debugStr );
#endif
	return s;
}

CSearch* CSearchManager::prepareLookup ( uint32_t type, bool start, const CUInt128 &id )

{

	if ( alreadySearchingFor ( id ) )
	{
		return NULL;
	}

	CSearch *s = new CSearch;

	try
	{
		// do not lookup keyword if it is in m_Load of getIndexed()
		if ( type == CSearch::STOREKEYWORD )
			if ( !Kademlia::CKademlia::getIndexed()->checkNotInLoadStatus ( id ) )
			{
				delete s;
				return NULL;
			}

		s->m_type = type;

		s->m_target = id;

		// Write complete packet
		s->m_searchTerms = new CMemFile();
		s->m_searchTerms->WriteUInt128 ( s->m_target );
		s->m_searchTerms->WriteUInt8 ( 1 );

		s->m_searchID = ++m_nextID;

		if ( start )
		{
			m_searches[s->m_target] = s;
			s->go();
		}
	}
	catch ( const CSafeIOException & ioe )
	{
		AddDebugLogLineM ( false, logKadSearch,
		                   CFormat ( wxT ( "Exception in CSearchManager::prepareLookup (IO error(%s))" ) ) % ioe.what() );
		delete s;
		return NULL;
	}
	catch ( ... )
	{
		AddDebugLogLineM ( false, logKadSearch,
		                   wxT ( "Exception in CSearchManager::prepareLookup" ) );
		delete s;
		throw;
	}

	return s;
}

void CSearchManager::findNode ( const CUInt128 &id )

{
	if ( alreadySearchingFor ( id ) )
	{
		return;
	}

	CSearch *s = new CSearch;

	s->m_type = CSearch::NODE;
	s->m_target = id;
	s->m_searchTerms = NULL;
	s->m_searchID = ++m_nextID + 100000;
	m_searches[s->m_target] = s;
	s->go();
}

void CSearchManager::findNodeComplete ( const CUInt128 &id )

{
	if ( alreadySearchingFor ( id ) )
	{
		return;
	}

	CSearch *s = new CSearch;

	s->m_type = CSearch::NODECOMPLETE;
	s->m_target = id;
	s->m_searchTerms = NULL;
	s->m_searchID = ++m_nextID + 200000;
	m_searches[s->m_target] = s;
	s->go();
}

bool CSearchManager::alreadySearchingFor ( const CUInt128 &target )

{
	return m_searches.count ( target );
}

void CSearchManager::getWords ( const wxString& str, WordList *words )

{
	int len = 0;
	wxString current_word;
	wxStringTokenizer tkz ( str, InvKadKeywordChars );

	while ( tkz.HasMoreTokens() )
	{
		current_word = tkz.GetNextToken();

		if ( ( len = current_word.Length() ) > 2 )
		{
			current_word.LowerCase();

			words->remove ( current_word );

			words->push_back ( current_word );
		}
	}

	// If the last word is 3 bytes long, chances are it's a file extension.
#warning File extensions have to be removed from keywords when the network is big
	/*if ( words->size() > 1 && len == 3 )
	{
		words->pop_back();
	}*/
}

void CSearchManager::jumpStart ( void )

{
	time_t now = time ( NULL );
	//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("jumpstart. %d searches.")) % m_searches.size()) ;

	std::set<CUInt128> toDelete ;

	for ( SearchMap::iterator current_it = m_searches.begin(); current_it != m_searches.end(); current_it++ )
	{

		//AddDebugLogLineM(false, logKadSearch, CFormat(wxT("#%i: CSearchManager::jumpstart.")) % current_it->second->getSearchID() ) ;

		if ( ! current_it->second->jumpStart() )
			toDelete.insert ( current_it->first );
	}

	for ( std::set<CUInt128>::iterator it = toDelete.begin(); it != toDelete.end(); it++ ) {
			delete m_searches[*it] ;
			m_searches.erase ( *it );
	}
}

void CSearchManager::updateStats ( void )
{
	uint8_t m_totalFile = 0;
	uint8_t m_totalStoreSrc = 0;
	uint8_t m_totalStoreKey = 0;
	uint8_t m_totalSource = 0;
	uint8_t m_totalNotes = 0;
	uint8_t m_totalStoreNotes = 0;

	for ( SearchMap::iterator it = m_searches.begin(); it != m_searches.end(); ++it )
	{
		switch ( it->second->getSearchTypes() )
		{
			case CSearch::FILE:
				{
					m_totalFile++;
					break;
				}

			case CSearch::STOREFILE:
				{
					m_totalStoreSrc++;
					break;
				}

			case CSearch::STOREKEYWORD:
				{
					m_totalStoreKey++;
					break;
				}

				/* NOTE: not for I2P ?? case CSearch::FINDSOURCE: {
					m_totalSource++;
					break;
				}*/
			case CSearch::STORENOTES:
				{
					m_totalStoreNotes++;
					break;
				}

			case CSearch::NOTES:
				{
					m_totalNotes++;
					break;
				}

			default:
				break;
		}
	}

	CPrefs *prefs = CKademlia::getPrefs();

	prefs->setTotalFile ( m_totalFile );
	prefs->setTotalStoreSrc ( m_totalStoreSrc );
	prefs->setTotalStoreKey ( m_totalStoreKey );
	prefs->setTotalSource ( m_totalSource );
	prefs->setTotalNotes ( m_totalNotes );
	prefs->setTotalStoreNotes ( m_totalStoreNotes );
}

void CSearchManager::processPublishResult ( CMemFile & bio )
{

	CUInt128 key = bio.ReadUInt128();

	CSearch *s = NULL;
	SearchMap::const_iterator it = m_searches.find ( key );

	if ( it != m_searches.end() )
	{
		s = it->second;
	}

	if ( s == NULL )
	{
		AddDebugLogLineM(false, logKadSearch,
			wxT("Search either never existed or receiving late results (CSearchManager::processPublishResults)"));
		return;
	}

	s->processPublishResult( bio ) ;
}


void CSearchManager::processGotPeerListForTarget ( const CUInt128 &target, const CI2PAddress & fromDest, ContactList *results )

{
	CSearch *s = NULL;
	SearchMap::const_iterator it = m_searches.find ( target );

	if ( it != m_searches.end() )
	{
		s = it->second;
	}

	if ( s == NULL )
	{
		AddDebugLogLineM ( false, logKadSearch,
				   wxString(wxT ( "CSearchManager::processGotPeerListForTarget : Search either "))
				   + wxT("never existed or receiving late results" ) );
	}
	else
	{
		s->processGotPeerListForTarget ( fromDest, results );
	}
}

void CSearchManager::processSearchResultList(  CMemFile & bio )
{
	// What search does this relate to
	CUInt128 target = bio.ReadUInt128();

	CSearch * s = NULL;

	AddDebugLogLineM ( false, logKadSearch,
	                   CFormat ( wxT ( "CSearchManager::processSearchResultList(search = %s)" ) )
	                   % target.toHexString() );

	SearchMap::const_iterator it = m_searches.find ( target );

	if ( it != m_searches.end() )
	{
		s = it->second;
	}

	if ( s == NULL )
	{
		AddDebugLogLineM ( false, logKadSearch,
		                   wxT ( "Search either never existed or receiving late results (CSearchManager::processResult)" ) );
	}
	else
	{
		s->processSearchResultList ( bio );
	}

}
