//
// This file is part of imule Project
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Project ( http://www.imule-project.net )
// Copyright (C)2003 Barry Dunne (http://www.emule-project.net)

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#include "KademliaUDPListener.h"
#include "../kademlia/Prefs.h"
#include "../kademlia/Kademlia.h"
#include "../kademlia/SearchManager.h"
#include "../routing/Contact.h"
#include "../routing/RoutingZone.h"
#include "../../NetworkFunctions.h"
#include "../../KnownFile.h"
#include "../../KnownFileList.h"
#include "../../OtherFunctions.h"
#include "../kademlia/Indexed.h"
#include "../../OPCodes.h"
#include "../kademlia/Defines.h"
#include "../../imule.h"
#include "../../ClientUDPSocket.h"
#include "../../Packet.h"
#include "../../ClientList.h"
#include "../../Statistics.h"
#include "../../MemFile.h"
#include "../../updownclient.h"
#include "../../ClientTCPSocket.h"
#include "../../Logger.h"
#include <common/Format.h>
#include "../../Preferences.h"

#include <wx/tokenzr.h>
#include <wx/arrstr.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif

#ifndef IMULE_DAEMON
#include "imuleDlg.h"
#include "KadDlg.h"
#endif

#include "OtherStructs.h"

//#define __PACKET_SEND_DUMP__   see also in ClientUDPSocket.cpp
//#define __PACKET_RECV_DUMP__


extern wxChar* InvKadKeywordChars;


const unsigned int & detailslen()
{
	static const unsigned int l = 16/*ID*/ + 2 * CI2PAddress::validLength() /*TCP+UDP*/ + 1/*0*/;

	return l;
}

const uint32_t & maxMessageSize()
{
	static const uint32_t l = CPacket::UDPPacketMaxDataSize() ;
	return l;
}

////////////////////////////////////////

using namespace Kademlia;

////////////////////////////////////////

void CKademliaUDPListener::bootstrap ( const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, wxT ( "bootstrap" ) );
	wxASSERT ( dest.isValid() );
	sendMyDetails ( KADEMLIA_BOOTSTRAP_REQ, dest );
}

void CKademliaUDPListener::sendMyDetails ( uint8_t opcode, const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "CKademliaUDPListener::sendMyDetails(%d, %s)" ) )
			% opcode % dest.humanReadable() );
	
	CMemFile bio ( detailslen() );
	CContact::Self().writeToFile(bio);


	sendPacket ( &bio, opcode, dest );
}

void CKademliaUDPListener::sendNullPacket ( uint8_t opcode, const CI2PAddress & dest )

{
	CMemFile bio ( 0 );
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadNullPacket sent to %x" ) ) % dest.hashCode() );
	sendPacket ( &bio, opcode, dest );
}

void CKademliaUDPListener::processPacket ( const uint8_t* data, uint32_t lenData, const CI2PAddress & dest )

{
#ifdef __PACKET_RECV_DUMP__
//	printf("processPacket, taille %d :\n", lenPacket);
//	DumpMem(packetData, 100);
#endif
	//Update connection state only when it changes.
	bool curCon = CKademlia::getPrefs()->hasHadContact();
	CKademlia::getPrefs()->setLastContact();

	if ( curCon != CKademlia::getPrefs()->hasHadContact() )
	{
		theApp.ShowConnectionState();
	}

	CMemFile bio ( data, lenData );

	bio.ReadByte();
	uint8_t opcode = bio.ReadByte();//data[1];

	switch ( opcode )
	{
		case KADEMLIA_BOOTSTRAP_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadBootstrapReq from %x" ) ) % dest.hashCode() );

			processBootstrapRequest ( bio, dest );

			break;

		case KADEMLIA_BOOTSTRAP_RES:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadBootstrapRes from %x" ) ) % dest.hashCode() );

			processBootstrapResponse ( bio, dest );

			break;

		case KADEMLIA_HELLO_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadHelloReq from %x" ) ) % dest.hashCode() );

			processHelloRequest ( bio, dest );

			break;

		case KADEMLIA_HELLO_RES:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadHelloRes from %x" ) ) % dest.hashCode() );

			processHelloResponse ( bio, dest );

			break;

		case KADEMLIA_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadReq from %x" ) ) % dest.hashCode() );

			processKademliaRequest ( bio, dest );

			break;

		case KADEMLIA_RES:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadRes from %x" ) ) % dest.hashCode() );

			processKademliaResponse ( bio, dest );

			break;

		case KADEMLIA_SEARCH_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadSearchReq from %x" ) ) % dest.hashCode() );

			processSearchRequest ( bio, dest );

			break;

		case KADEMLIA_SEARCH_RES:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadSearchRes from %x" ) ) % dest.hashCode() );

			processSearchResponse ( bio, dest );

			break;

		case KADEMLIA_PUBLISH_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadPublishReq from %x" ) ) % dest.hashCode() );

			processPublishRequest ( bio, dest );

			break;

		case KADEMLIA_PUBLISH_RES:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadPublishRes from %x" ) ) % dest.hashCode() );

			processPublishResponse ( bio, dest );

			break;

		case KADEMLIA_SRC_NOTES_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadSearchNotesReq from %x" ) ) % dest.hashCode() );

			processSearchNotesRequest ( bio, dest );

			break;

		case KADEMLIA_PUB_NOTES_REQ:
			AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadPublishNotesReq from %x" ) ) % dest.hashCode() );

			processPublishNotesRequest ( bio, dest );

			break;


		default:
		{
			throw CFormat ( wxT ( "Unknown opcode %02x on CKademliaUDPListener::processPacket" ) )
			% opcode ;
		}
	}

	// signals the routing zone that the contact is alive
	CKademlia::getRoutingZone()->setAlive( dest );
	
	// bootstrap if necessary
	if ( CKademlia::getRoutingZone()->getNumContacts() == 0 )
		bootstrap ( dest );
}



void CKademliaUDPListener::addContact ( CMemFile & bio, const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "addContact : called with (udp=%s)" ) )
			% dest.humanReadable() );
	
	// read the contact from bio
	CContact new_contact(bio);
	// adjust its destinations if it is lying (?)
	if ( dest.isValid() ) 
	{
		new_contact.setUDPDest( dest );
		CKademlia::getRoutingZone()->add ( new_contact, true );
	}
}



/**
 * <b>Add  a list of contacts</b> to Kademlia RoutingZone
 * @param bio 
 * @param numContacts 
 */
void CKademliaUDPListener::addContacts ( CMemFile & bio, uint16_t numContacts )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "addContacts : called with (numContacts=%d)" ) )
			% numContacts );
	
	CRoutingZone *routingZone = CKademlia::getRoutingZone();

	for ( uint16_t i = 0; i < numContacts; i++ )
	{
		CContact new_contact( bio );

		if ( new_contact.getUDPDest().isValid() )
		{
			routingZone->add ( new_contact );
		}
	}
}

//KADEMLIA_BOOTSTRAP_REQ

void CKademliaUDPListener::processBootstrapRequest ( CMemFile & bio, const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "processBootstrapRequest(bio, %s)" ) ) % dest.humanReadable() );

	// Verify packet is expected size

	if ( bio.GetRemaining() != detailslen() )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s, expected %u) packet in %s" ) )
		                     % bio.GetRemaining().ToString() % detailslen() % wxString::FromAscii ( __func__ ) );
	}


	// Add the sender to the list of contacts
	addContact ( bio, dest );

	// Get some contacts to return
	ContactList contacts;
    uint32_t max_contacts = (CPacket::UDPPacketMaxDataSize()-2/*number of contacts*/)
                            / detailslen();

	uint16_t numContacts = (uint16_t) (1 + CKademlia::getRoutingZone()->getBootstrapContacts ( contacts, max_contacts - 1 /* -1 : my own node will be added at the end of the list*/ ));

	// Create response packet
	//We only collect a max of numContacts contacts here.. Max size is 527.
	
    CMemFile res ( 2 + detailslen() * (numContacts+1) );

	// Write packet info
	res.WriteUInt16 ( numContacts ); // size : 2

	CContact contact;

	ContactList::const_iterator it;

	// write contacts taken from our routing zone
	for ( it = contacts.begin(); it != contacts.end(); it++ )
	{
		contact = *it;
		contact.writeToFile(res); // size : detailslen() ( * numContacts)
	}

	// write ourself
	CContact::Self().writeToFile(res); // size : detailslen()

	// Send response
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadBootstrapRes %x" ) ) % dest.hashCode() );

	sendPacket ( &res, KADEMLIA_BOOTSTRAP_RES, dest );
}

//KADEMLIA_BOOTSTRAP_RES
void CKademliaUDPListener::processBootstrapResponse ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() < 2 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	// How many contacts were given

	uint16_t numContacts = bio.ReadUInt16();

	// Verify packet is expected size
	if ( bio.GetRemaining() != ( uint32_t ) ( detailslen() *numContacts ) )
	{
		return;
	}

	// Add these contacts to the list.
	addContacts ( bio, numContacts );
}

//KADEMLIA_HELLO_REQ
void CKademliaUDPListener::processHelloRequest ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() != detailslen() )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	// Add the sender to the list of contacts
	addContact ( bio, dest );

	// Send response
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadHelloRes to   %x" ) ) % dest.hashCode() );

	sendMyDetails ( KADEMLIA_HELLO_RES, dest );

}

//KADEMLIA_HELLO_RES
void CKademliaUDPListener::processHelloResponse ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() != detailslen() )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	// Add or Update contact.
	addContact ( bio, dest );
}

//KADEMLIA_REQ
void CKademliaUDPListener::processKademliaRequest ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() != 1/*m_type*/ + 16 /*m_target*/ + 16 /*my ID*/ + 16 /*peer ID*/ )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	// Get type
	uint8_t type = bio.ReadUInt8();

	// This is the target node trying to be found.
	CUInt128 target = bio.ReadUInt128();

	// This should be my ID
	CUInt128 check  = bio.ReadUInt128();

	// This is the ID of the requesting peer
	CUInt128 destId = bio.ReadUInt128();

	type = (uint8_t) (type & (uint8_t)0x1F);

	if ( type == 0 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong type (0x%02x) in %s" ) )
		                     % type % wxString::FromAscii ( __func__ ) );
	}


	CUInt128 distance ( CKademlia::getPrefs()->getKadID() );

	distance.XOR ( target );

	//This makes sure we are not mistaken identify. Some client may have fresh installed and have a new hash.

	if ( CKademlia::getPrefs()->getKadID().compareTo ( check ) )
	{
		return;
	}

	// Get required number closest to target
	ContactMap results;

	uint32_t max_results = std::min<uint32_t>( type, (maxMessageSize() - 16 - 1) / detailslen() ) ;
	
	CKademlia::getRoutingZone()->getClosestTo ( 2, target, distance, ( int ) max_results, results );

	uint16_t count = ( uint16_t ) results.size();

	// Write response
	// Max count is 32. size 817..
	// 16 + 1 + 817(32)
	 
	CMemFile res ( 16 + 1 + detailslen() * count );

	res.WriteUInt128 ( target );

	res.WriteUInt8 ( ( uint8_t ) count );

	CContact c;

	ContactMap::const_iterator it;

	for ( it = results.begin(); it != results.end(); it++ )
	{
		c = it->second;
		c.writeToFile( res );
	}

	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadRes %x Count=%u Req=%u" ) ) % dest.hashCode() % count % type );

	sendPacket ( &res, KADEMLIA_RES, dest );


	// Try to add the requesting peer to our contact nodes

	if ( CKademlia::getRoutingZone()->wouldLikeToAdd ( destId ) )
		sendMyDetails ( KADEMLIA_HELLO_REQ, dest );
}

//KADEMLIA_RES
void CKademliaUDPListener::processKademliaResponse ( CMemFile & bio, const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "processKademliaResponse(bio, %s)" ) ) % dest.humanReadable() );
	
	// Verify packet is expected size

	if ( bio.GetRemaining() < 17 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) ) ;
	}

	//Used Pointers
	CRoutingZone *routingZone = CKademlia::getRoutingZone();

	// What search does this relate to
	CUInt128 target = bio.ReadUInt128();

	uint16_t numContacts = bio.ReadUInt8();

	// Verify packet is expected size
	if ( bio.GetRemaining() != ( uint32_t ) ( detailslen() * numContacts ) )
	{
		//if (lenPacket != (uint32_t)(16+1 + (16+4+2+2+1)*numContacts)) {
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) )
		                     % bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	ContactList results;

	try
	{
		CContact myself = CContact::Self() ;
		for ( uint16_t i = 0; i < numContacts; i++ )
		{
			CContact new_contact( bio );

			if ( ::IsGoodDest ( new_contact.getUDPDest() ) )
			{
				CContact kad_contact = routingZone->add ( new_contact ) ;
				
				// if its ID is mine
				// Let's check if the client has registered my current UDP dest
				if (  new_contact.getClientID() == myself.getClientID()
					   && new_contact.getUDPDest()  != myself.getUDPDest() )
					{ // Let's recall him who I am !
						sendMyDetails(KADEMLIA_HELLO_RES, dest) ;
					}
				else if ( kad_contact.isInvalid())
				{
					results.push_back( new_contact );
				}
				else
				{
					results.push_back ( kad_contact );
				}
			}
		}
	}
	catch ( ... )
	{
		throw;
	}

	CSearchManager::processGotPeerListForTarget ( target, dest, &results );
}


void CKademliaUDPListener::Free ( SSearchTerm* pSearchTerms )

{
	if ( !pSearchTerms )
	{
		return;
	}

	Free ( pSearchTerms->left );

	Free ( pSearchTerms->right );
	delete pSearchTerms;
}



SSearchTerm* CKademliaUDPListener::CreateSearchExpressionTree ( CMemFile& bio, int iLevel )
{
	// the max. depth has to match our own limit for creating the search expression
	// (see also 'ParsedSearchExpression' and 'GetSearchPacket')

	if ( iLevel >= 24 )
	{
		AddDebugLogLineM ( false, logClientKadUDP, wxT ( "***NOTE: Search expression tree exceeds depth limit!" ) );
		return NULL;
	}

	iLevel++;

	uint8_t op = bio.ReadUInt8();

	if ( op == 0x00 )
	{
		uint8_t boolop = bio.ReadUInt8();

		if ( boolop == 0x00 )
		{ // AND
			SSearchTerm* pSearchTerm = new SSearchTerm;
			pSearchTerm->type = SSearchTerm::AND;
			//TRACE(" AND");

			if ( ( pSearchTerm->left = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				delete pSearchTerm;
				return NULL;
			}

			if ( ( pSearchTerm->right = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				Free ( pSearchTerm->left );
				delete pSearchTerm;
				return NULL;
			}

			return pSearchTerm;
		}
		else if ( boolop == 0x01 )
		{ // OR
			SSearchTerm* pSearchTerm = new SSearchTerm;
			pSearchTerm->type = SSearchTerm::OR;
			//TRACE(" OR");

			if ( ( pSearchTerm->left = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				delete pSearchTerm;
				return NULL;
			}

			if ( ( pSearchTerm->right = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				Free ( pSearchTerm->left );
				delete pSearchTerm;
				return NULL;
			}

			return pSearchTerm;
		}
		else if ( boolop == 0x02 )
		{ // NAND
			SSearchTerm* pSearchTerm = new SSearchTerm;
			pSearchTerm->type = SSearchTerm::NAND;
			//TRACE(" NAND");

			if ( ( pSearchTerm->left = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				delete pSearchTerm;
				return NULL;
			}

			if ( ( pSearchTerm->right = CreateSearchExpressionTree ( bio, iLevel ) ) == NULL )
			{
				wxASSERT ( 0 );
				Free ( pSearchTerm->left );
				delete pSearchTerm;
				return NULL;
			}

			return pSearchTerm;
		}
		else
		{
			AddDebugLogLineM ( false, logKadSearch, ( wxString ) ( CFormat ( wxT ( "*** Unknown boolean search operator 0x%02x (CreateSearchExpressionTree)" ) ) % boolop ) );
			return NULL;
		}
	}
	else if ( op == 0x01 )
	{ // String
		wxString str ( bio.ReadString ( true ) );

		str.MakeLower(); // make lowercase, the search code expects lower case strings!
		//TRACE(" \"%ls\"", str);

		SSearchTerm* pSearchTerm = new SSearchTerm;
		pSearchTerm->type = SSearchTerm::String;
		pSearchTerm->astr = new wxArrayString;

		// pre-tokenize the string term
		wxStringTokenizer token ( str, InvKadKeywordChars, wxTOKEN_DEFAULT );

		while ( token.HasMoreTokens() )
		{
			wxString strTok ( token.GetNextToken() );

			if ( !strTok.IsEmpty() )
			{
				pSearchTerm->astr->Add ( strTok );
			}
		}

		return pSearchTerm;
	}
	else if ( op == 0x02 )
	{ // Meta tag
		// read tag value
		wxString strValue(bio.ReadString(true));

		strValue.LowerCase(); // make lowercase, the search code expects lower case strings!

		// read tag ID
		//wxString strTagName =  bio.ReadString(true);
		bio.ReadUInt16() ;
		uint8_t tagid = bio.ReadUInt8() ;
		//wxString strValue = metaTag.GetStr();
		//strValue.LowerCase() ; // make lowercase, the search code expects lower case strings!

		SSearchTerm* pSearchTerm = new SSearchTerm;
		pSearchTerm->type = SSearchTerm::MetaTag;
		//pSearchTerm->tag = CTag ( metaTag.GetID(), strValue );
		pSearchTerm->tag = CTag ( tagid, strValue );
		return pSearchTerm;
	}
	else if ( op == 0x03 )
	{ // Min/Max

		static const struct
		{
			SSearchTerm::ESearchTermType eSearchTermOp;
			wxString pszOp;
		}

		_aOps[] =

		    {
		        { SSearchTerm::OpEqual,			wxT ( "=" )		}, // mmop=0x00
		        { SSearchTerm::OpGreaterEqual,	wxT ( ">=" )	}, // mmop=0x01
		        { SSearchTerm::OpLessEqual,		wxT ( "<=" )	}, // mmop=0x02
		        { SSearchTerm::OpGreater,		wxT ( ">" )		}, // mmop=0x03
		        { SSearchTerm::OpLess,			wxT ( "<" )		}, // mmop=0x04
		        { SSearchTerm::OpNotEqual,		wxT ( "!=" )	}  // mmop=0x05
		    };

		// read integer operator
		uint8_t mmop = bio.ReadUInt8();

		if ( mmop >= ARRSIZE ( _aOps ) )
		{
			AddDebugLogLineM ( false, logClientKadUDP, ( wxString ) ( CFormat ( wxT ( "*** Unknown integer search op=0x%02x (CreateSearchExpressionTree)" ) ) % mmop ) );
			return NULL;
		}

		// read tag
		SSearchTerm* pSearchTerm = new SSearchTerm;

		pSearchTerm->type = _aOps[mmop].eSearchTermOp;

		pSearchTerm->tag = CTag ( bio );

		return pSearchTerm;
	}
	else
	{
		AddDebugLogLineM ( false, logClientKadUDP,
		                   ( wxString ) ( CFormat ( wxT ( "*** Unknown search op=0x%02x (CreateSearchExpressionTree)" ) )
		                                  % op ) );
		return NULL;
	}
}

//KADEMLIA_SEARCH_REQ

void CKademliaUDPListener::processSearchRequest ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() < 17 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
		                     bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	CUInt128 target = bio.ReadUInt128();

	uint8_t restrictive = bio.ReadUInt8();

#ifdef _DEBUG
	//DWORD dwNow = GetTickCount();
#endif

	if ( bio.GetRemaining() == 0 )
	{
		if ( restrictive )
		{
			//Source request
			CKademlia::getIndexed()->SendValidSourceResult ( target, dest );
			//DEBUG_ONLY( Debug("SendValidSourceResult: Time=%.2f sec\n", (GetTickCount() - dwNow) / 1000.0) );
		}
		else
		{
			//Single keyword request
			CKademlia::getIndexed()->SendValidKeywordResult ( target, NULL, dest );
			//DEBUG_ONLY( Debug("SendValidKeywordResult (Single): Time=%.2f sec\n", (GetTickCount() - dwNow) / 1000.0) );
		}
	}
	else
	{
		SSearchTerm* pSearchTerms = NULL;

		if ( restrictive )
		{
			try
			{
				pSearchTerms = CreateSearchExpressionTree ( bio, 0 );
				//TRACE("\n");
			}
			catch ( ... )
			{
				Free ( pSearchTerms );
				throw;
			}
		}

		//Keyword request with added options.
		CKademlia::getIndexed()->SendValidKeywordResult ( target, pSearchTerms, dest );

		Free ( pSearchTerms );

		//DEBUG_ONLY( Debug("SendValidKeywordResult: Time=%.2f sec\n", (GetTickCount() - dwNow) / 1000.0) );
	}
}

//KADEMLIA_SEARCH_RES
void CKademliaUDPListener::processSearchResponse ( CMemFile & bio, const CI2PAddress & dest )
{
	AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "processSearchResponse(%s, %s)" ) )
						% bio.GetRemaining().ToString() % dest.humanReadable() );
	
	// Verify packet is expected size

	if ( bio.GetRemaining() < 18 ) // at least the key (16) and the number of results (2)
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) )
		                     % bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}


	CSearchManager::processSearchResultList ( bio );

}

//KADEMLIA_PUBLISH_REQ
void CKademliaUDPListener::processPublishRequest ( CMemFile & bio, const CI2PAddress & dest, bool self )
{
	//There are different types of publishing..
	//Keyword and File are Stored..
	
	// only send the response once
	bool responseSent = false ;
	
	// Verify packet is expected size

	if ( bio.GetRemaining() < 37 ) // at least the key (16), the number of sources (2), one source (16), the number of tags (2), one tag (1)
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
				bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	//Used Pointers
	CIndexed *indexed = CKademlia::getIndexed();

	/*
	if( CKademlia::getPrefs()->getFirewalled() ) {
		//We are firewalled. We should not index this entry and give publisher a false report.
		return;
	}
	*/

	CUInt128 key = bio.ReadUInt128();

	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : dest=%s, key=%s" ) ) % dest.humanReadable() % key.toHexString() );

	CUInt128 distance ( CKademlia::getPrefs()->getKadID() );

	distance.XOR ( key );

	// TODO: check if our node is a good place for publishing the key (do not trust the requesting node)

	uint16_t count = bio.ReadUInt16();

	AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : count=%d" ) ) % count );

	uint8_t load = 0;

	while ( count > 0 )
	{
		CUInt128 target = bio.ReadUInt128();
		AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : target=%s" ) ) % target.toHexString() );

		Kademlia::CEntry entry;

		try
		{
			entry.udpdest = dest;
			entry.keyID.setValue ( key );
			entry.sourceID.setValue ( target );

			TagList taglist ( bio ) ;

			AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : #tags=%d" ) ) % taglist.size() );

			while ( !taglist.empty() )
			{
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : #tags remaining=%d" ) ) % taglist.size() );
				CTag tag = taglist.back() ;
				taglist.pop_back();

				if ( tag.isValid() )
				{
					if ( tag . GetID() == TAG_SOURCETYPE )
					{
						if ( entry.source == false )
						{
							entry.taglist.push_back ( tag ); /* record the type, which is checked in CSearch::processResultFile
																									    even if it is of no use (no firewalled source with i2p) */
							entry.taglist.push_back ( CTag ( TAG_SOURCEUDEST, entry.udpdest ) );
						}

						entry.source = true;

						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : TAG_SOURCETYPE" ) ) );
					}

					else if ( tag . GetID() == TAG_FILENAME )
					{
						if ( entry.fileName.IsEmpty() )
						{
							entry.fileName = tag . GetStr();
							AddDebugLogLineM ( false, logKadIndex,
							                   CFormat ( wxT ( "processPublishRequest : TAG_FILENAME, filename : %s" ) ) % entry.fileName );

							entry.fileName.LowerCase(); // make lowercase, the search code expects lower case strings!
							// NOTE: always add the 'name' tag, even if it's stored separately in 'fileName'. the tag is still needed for answering search request
							entry.taglist.push_back ( tag );
						}
					}
					else if ( tag . GetID() == TAG_FILESIZE )
					{
						if ( entry.size == 0 )
						{
							entry.size = tag . GetInt();
							// NOTE: always add the 'size' tag, even if it's stored separately in 'size'. the tag is still needed for answering search request
							entry.taglist.push_back ( tag );
						}

						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : TAG_FILESIZE : %llu" ) )

						                   % entry.size.GetValue() );

					}
					else if ( tag . GetID() == TAG_SOURCEDEST )
					{
						if ( !entry.tcpdest )
						{
							entry.tcpdest = tag . GetAddr();
							entry.taglist.push_back ( tag );
						}

						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : TAG_SOURCEDEST %s" ) )

						                   % entry.tcpdest.humanReadable() );

					}
					else
					{
						//TODO: Filter tags
						AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : UNKNOWN TAG, id: %X, type:%d" ) )
						                   % ( int ) tag . GetID() % tag . GetType() );

						entry.taglist.push_back ( tag );
					}
				}
			}
		}
		catch ( ... )
		{
			DebugClientOutput ( wxT ( "CKademliaUDPListener::processPublishRequest - ERROR" ), dest );
			throw;
		}

		if ( entry.source == true )
		{
			AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : source indexing request" ) ) );
			entry.lifetime = ( uint32_t ) time ( NULL ) + KADEMLIAREPUBLISHTIMES;

			if ( indexed->AddSources ( key, target, entry, load ) )
			{
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : source indexed : %s" ) ) %  key.toHexString() );

				if ( !self  && !responseSent )
				{
					CMemFile res ( 25 );
					res.WriteUInt128 ( key );
					res.WriteUInt32 ( indexed->GetSourceCount ( key ) );
					res.WriteUInt32 ( indexed->GetCompleteSourceCount ( key ) );
					res.WriteUInt8 ( load );
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "KadPublishRes (source) to %s" ) ) % dest.humanReadable() );
					AddDebugLogLineM ( false, logKadSourceCount,
					                   CFormat ( wxT ( "KadPublishRes (source) to %s : %s (complete=%d, total=%d)" ) )
					                   % dest.humanReadable() % key.toHexString()
					                   % indexed->GetCompleteSourceCount ( key )
					                   % indexed->GetSourceCount ( key ) );

					sendPacket ( &res, KADEMLIA_PUBLISH_RES, dest );
					responseSent = true ;
				}
				else if ( self )
				{
					AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : source indexed from myself" ) ) );
				}
			}
			else
			{
				AddDebugLogLineM ( true, logKadIndex, CFormat ( wxT ( "processPublishRequest : source NOT indexed : %s" ) ) %  key.toHexString() );
			}
		}
		else
		{
			entry.lifetime = ( uint32_t ) time ( NULL ) + KADEMLIAREPUBLISHTIMEK;

			if ( ! indexed->AddKeyword ( key, target, entry, load ) )
			{
				//We already indexed the maximum number of keywords.
				//We do not index anymore but we still send a success..
				//Reason: Because if a VERY busy node tells the publisher it failed,
				//this busy node will spread to all the surrounding nodes causing popular
				//keywords to be stored on MANY nodes..
				//So, once we are full, we will periodically clean our list until we can
				//begin storing again..
			}

			if ( !self && !responseSent )
			{
				CMemFile res ( 17 );
				res.WriteUInt128 ( key );
				res.WriteUInt8 ( load );

				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "KadPublishRes (keyword) to %x" ) ) % dest.hashCode() );

				sendPacket ( &res, KADEMLIA_PUBLISH_RES, dest );
				responseSent = true ;
			}
			else if ( self )
			{
				AddDebugLogLineM ( false, logKadIndex, CFormat ( wxT ( "processPublishRequest : keyword indexed from myself" ) ) );
			}
		}

		count--;
	}

}

// KADEMLIA_PUBLISH_RES
void CKademliaUDPListener::processPublishResponse ( CMemFile & bio, const CI2PAddress & dest )
{
	CSearchManager::processPublishResult ( bio );
}

//KADEMLIA_SRC_NOTES_REQ
void CKademliaUDPListener::processSearchNotesRequest ( CMemFile & bio, const CI2PAddress & dest )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() < 32 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
				bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	CUInt128 target = bio.ReadUInt128();

	CUInt128 source = bio.ReadUInt128();

	CKademlia::getIndexed()->SendValidNoteResult ( target, source, dest );
}


//KADEMLIA_PUB_NOTES_REQ
void CKademliaUDPListener::processPublishNotesRequest ( CMemFile & bio, const CI2PAddress & dest, bool self )
{
	// Verify packet is expected size

	if ( bio.GetRemaining() < 37 )
	{
		throw ( wxString ) ( CFormat ( wxT ( "***NOTE: Received wrong size (%s) packet in %s" ) ) %
				bio.GetRemaining().ToString() % wxString::FromAscii ( __func__ ) );
	}

	/*
	if( CKademlia::getPrefs()->getFirewalled() ) {
		//We are firewalled. We should not index this entry and give publisher a false report.
		return;
	}
	*/

	CUInt128 target = bio.ReadUInt128();

	CUInt128 distance ( CKademlia::getPrefs()->getKadID() );

	distance.XOR ( target );

	CUInt128 source = bio.ReadUInt128();

	Kademlia::CEntry entry;

	try
	{
		//entry.ip = ip;
		entry.udpdest = dest;
		entry.keyID.setValue ( target );
		entry.sourceID.setValue ( source );
		entry.taglist = TagList ( bio ); // true added : for UTF8 strings
		entry.source = false;
	}
	catch ( ... )
	{
		DebugClientOutput ( wxT ( "CKademliaUDPListener::processPublishNotesRequest" ), dest );
		throw;
	}

	if ( entry.taglist.size() == 0 || entry.taglist.size() > 5 )
	{
		throw wxString ( wxT ( "CKademliaUDPListener::processPublishNotesRequest: entry.taglist.size() == 0 || entry.taglist.size() > 5" ) );
	}

	if ( !self )
	{
		uint8_t load = 0;
		CMemFile res ( 17 );
		res.WriteUInt128 ( target );
		res.WriteUInt8 ( load );
		AddDebugLogLineM ( false, logClientKadUDP, CFormat ( wxT ( "KadPublishNotesRes %x" ) ) % dest.hashCode() );
		sendPacket ( &res, KADEMLIA_PUBLISH_RES, dest );
	}
}




void CKademliaUDPListener::sendPacket ( CMemFile *data, uint8_t opcode, const CI2PAddress & dest )

{
	//AddDebugLogLineM(false, logClientKadUDP, CFormat(wxT("sendPacket")));
	CPacket* packet = new CPacket ( data, OP_KADEMLIAHEADER, opcode );
#ifdef __PACKET_SEND_DUMP__
	//bool deb = (opcode == KADEMLIA_PUBLISH_REQ) ;
	bool deb = packet->GetPacketSize() > 200 ;
	uint32_t siz = packet->GetPacketSize();
#endif

	if ( packet->GetPacketSize() > 200 )
	{
		//wxASSERT( packet->GetPacketSize() < 50000 );
		packet->PackPacket();
#ifdef __PACKET_SEND_DUMP__

		if ( siz > packet->GetPacketSize() )
		{
			printf ( "Paquet COMPRESSE envoyé : opcode = %x taille = %d\n", opcode, packet->GetPacketSize() );
			DumpMem ( packet->GetDataBuffer(), packet->GetPacketSize() );
		}

#endif
	}

#ifdef __PACKET_SEND_DUMP__
	if ( false && siz <= packet->GetPacketSize() )
	{
		printf ( "Paquet non compressé envoyé :\n" );
		DumpMem ( packet->GetPacket(), 100 );
	}

#endif

	theStats::AddUpOverheadKad ( packet->GetPacketSize() );

	theApp.clientudp->SendPacket ( packet, dest );
}

void CKademliaUDPListener::DebugClientOutput ( const wxString& place, const CI2PAddress & dest )

{

	printf ( "Error on %s received from: %x\n", ( const char* ) unicode2char ( place ),/*(const char*)unicode2char(*/dest.hashCode() /*)*/ );

	CClientList::SourceList clientslist = theApp.clientlist->GetClientsByDest ( dest );

	if ( !clientslist.empty() )
	{
		for ( CClientList::SourceList::iterator it = clientslist.begin(); it != clientslist.end(); ++it )
		{

			printf ( "Ip Matches: %s\n", ( const char* ) unicode2char ( ( *it )->GetClientFullInfo() ) );
		}
	}
	else
	{
		printf ( "No ip match\n" );
	}
}
