//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#ifndef __KAD_UDP_LISTENER_H__
#define __KAD_UDP_LISTENER_H__

#include "../utils/UInt128.h"
//#include "../io/ByteIO.h"
#include "../../Types.h"
#include "../../Tag.h"
#include "i2p/CI2PAddress.h"

class CKnownFile;
class CMemFile;
struct SSearchTerm;


////////////////////////////////////////
namespace Kademlia {
////////////////////////////////////////

class CSearch;

class CKademliaUDPListener
{
	friend class CSearch;

public:
	virtual ~CKademliaUDPListener() {};
	void bootstrap(const CI2PAddress & dest);
	void sendMyDetails(uint8_t opcode, const CI2PAddress & dest);
	void sendNullPacket(uint8_t opcode, const CI2PAddress & dest);
	virtual void processPacket(const uint8_t* data, uint32_t lenData, const CI2PAddress & dest);
	void sendPacket(CMemFile* data, uint8_t opcode, const CI2PAddress & dest);

	void processPublishRequest		(CMemFile & packetData, const CI2PAddress & dest, bool self=false);
	void processPublishNotesRequest		(CMemFile & packetData, const CI2PAddress & dest, bool self=false);
	private:
	void addContact (CMemFile & data, const CI2PAddress & dest);
	void addContacts(CMemFile & data, uint16_t numContacts);
	static SSearchTerm* CreateSearchExpressionTree(CMemFile& bio, int iLevel);
	static void Free(SSearchTerm* pSearchTerms);

	void processBootstrapRequest		(CMemFile & packetData, const CI2PAddress & dest);
	void processBootstrapResponse		(CMemFile & packetData, const CI2PAddress & dest);
	void processHelloRequest			(CMemFile & packetData, const CI2PAddress & dest);
	void processHelloResponse			(CMemFile & packetData, const CI2PAddress & dest);
	void processKademliaRequest			(CMemFile & packetData, const CI2PAddress & dest);
	void processKademliaResponse		(CMemFile & packetData, const CI2PAddress & dest);
	void processSearchRequest			(CMemFile & packetData, const CI2PAddress & dest);
	void processSearchResponse			(CMemFile & packetData, const CI2PAddress & dest);
	void processPublishResponse			(CMemFile & packetData, const CI2PAddress & dest);
	void processSearchNotesRequest		(CMemFile & packetData, const CI2PAddress & dest);
	void processPublishNotesResponse	(CMemFile & packetData, const CI2PAddress & dest);
	
	void DebugClientOutput(const wxString& place, const CI2PAddress & dest);
};

} // End namespace

#endif //__KAD_UDP_LISTENER_H__
