//
// This file is part of imule Project
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Project ( http://www.imule-project.net )
// Copyright (C)2003 Barry Dunne (http://www.emule-project.net)

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

// This work is based on the java implementation of the Kademlia protocol.
// Kademlia: Peer-to-peer routing based on the XOR metric
// Copyright (C) 2002  Petar Maymounkov [petar@post.harvard.edu]
// http://kademlia.scs.cs.nyu.edu

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#include "Contact.h"
#include "RoutingZone.h"
#include "../../imule.h"
#include "../../OPCodes.h" // Neededf for MIN2MS and such stuff.
#include "../../Statistics.h"
#include <Logger.h>
#include <common/Format.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

////////////////////////////////////////
using namespace Kademlia;
////////////////////////////////////////

std::set<CContact> CContact::s_kadContacts ;

CContact::~CContact()
{
/*#ifdef __WXDEBUG__
	if ( ! isInvalid() )
	{
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "deleting ref. n° %d of contact %s") )
				% GetData() -> GetRefCount() % getInfoString() );
	}
#endif
*/
}

void CContact::initContact() 
{
	GetData()->m_type = 3;
	GetData()->m_expires = 0;
	GetData()->m_lastTypeSet = time(NULL);
	GetData()->m_lastTypeUpdate = 0;
	GetData()->m_created = time(NULL);
	GetData()->m_inUse = 0;
	GetData()->m_inKadNodes = false;
}

CContact::CContact() : wxObject()
{
	SetRefData( NULL );
}

wxMutex & CContact::GetDataMutex()
{
	wxASSERT( GetData() );
	return GetData()->m_mutex ;
}

CContact CContact::Self()
{
	static CContact me;
	if ( me.isInvalid() ) me.SetRefData( new Data() );
	//wxMutexLocker lock( me.GetDataMutex() );
	me.initContact();
	me.GetData()->m_clientID = CKademlia::getPrefs()->getKadID() ;
	me.GetData()->m_udpDest  = theApp.GetUdpDest();
	me.GetData()->m_tcpDest  = theApp.GetTcpDest();
	me.GetData()->m_distance = 0 ;
	me.GetData()->m_type     = 0 ;
	return me ;
}

bool CContact::isInvalid() const
{
	return GetData()==NULL ;
}


bool CContact::operator< (const CContact & other) const
{
	return GetData()->m_clientID < other.GetData()->m_clientID ;
}

bool CContact::operator==(const CContact & other) const
{
	return GetData()->m_clientID == other.GetData()->m_clientID ;
}


CContact::CContact ( CFileDataIO & file )
{
	SetRefData( new Data() );
	//wxMutexLocker lock( GetDataMutex() );
	initContact();
	
	GetData()->m_clientID = file.ReadUInt128();
	GetData()->m_udpDest  = file.ReadAddress();
	GetData()->m_tcpDest  = file.ReadAddress();
	GetData()->m_type     = file.ReadUInt8()  ;
	
	GetData()->m_distance.setValue( CKademlia::getPrefs()->getKadID() );
	GetData()->m_distance.XOR( GetData()->m_clientID );
	
	if ( GetData()->m_type > 4 ) 
	{
		AddDebugLogLineM ( true, logKadRouting, CFormat ( wxT ( "CContact::CContact(stream) -> %s") )
			% getInfoString() );
		throw BadTypeException(GetData()->m_type);
	}
	else
	{
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CContact::CContact(stream) -> %s") )
			% getInfoString() );
	}
}

void  CContact::writeToFile( CFileDataIO & file )
{
	//wxMutexLocker lock( GetDataMutex() );
	file.WriteUInt128 ( GetData()->m_clientID );
	file.WriteAddress ( GetData()->m_udpDest );
	file.WriteAddress ( GetData()->m_tcpDest );
	file.WriteUInt8   ( GetData()->m_type );
}

void CContact::addedToKadNodes()
{
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "(1) Adding contact %s, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );

	wxASSERT( !GetData()->m_inKadNodes );
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "(2) Adding contact %s, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );
	{
		//wxMutexLocker lock( GetDataMutex() );
		GetData()->m_inKadNodes = true ;
	}
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "(3) Adding contact %s, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );
	CRoutingZone::s_ContactsSet . insert( *this ) ;

	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "(4) Adding contact %s, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );
	
	theStats::AddKadNode();
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "(5) Adding contact %s, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );
	
	Notify_KadNode_Added( new CContact(*this) );
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Contact %s added, with type %d, %d remaining" ) )
			% getInfoString() % getType() % CRoutingZone::s_ContactsSet.size() );
}

void CContact::removedFromKadNodes()
{
	wxASSERT( GetData()->m_inKadNodes );
	{
		//wxMutexLocker lock( GetDataMutex() );
		GetData()->m_inKadNodes = false ;
	}
	theStats::RemoveKadNode();
	Notify_KadNode_Deleted( new CContact(*this) );

	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Contact %s removed, %d remaining" ) )
			% getInfoString() % CRoutingZone::s_ContactsSet.size() );

	wxASSERT( CRoutingZone::s_ContactsSet . count( *this ) );
	CRoutingZone::s_ContactsSet . erase( *this ) ;
}

const wxString CContact::getInfoString( void ) const
{
	return CFormat( wxT("(kad: %s) (dist: %s) (udp: %s) (tcp: %s)") )
			% GetData()->m_clientID.toBinaryString().Mid(0,12)
			% GetData()->m_distance.toBinaryString().Mid(0,12)
			% getUDPDest().humanReadable()
			% getTCPDest().humanReadable() ;
}

const wxString CContact::getClientIDString(void) const
{
	return GetData()->m_clientID.toHexString();
}

#ifndef CLIENT_GUI
void CContact::setClientID(const CUInt128 &clientID)
{
	//wxMutexLocker lock( GetDataMutex() );
	GetData()->m_clientID = clientID;
	GetData()->m_distance = CKademlia::getPrefs()->getKadID();
	GetData()->m_distance.XOR(clientID);
}
#endif

void CContact::setDistanceFrom( const CUInt128 & other )
{
	//wxMutexLocker lock( GetDataMutex() );
	GetData()->m_distance = other ;
	GetData()->m_distance.XOR ( GetData()->m_clientID );
}


const wxString CContact::getDistanceString(void) const
{
	return GetData()->m_distance.toBinaryString();
}

CI2PAddress CContact::getTCPDest(void) const
{
	return GetData()->m_tcpDest;
}

void CContact::setTCPDest(const CI2PAddress & dest)
{
	//wxMutexLocker lock( GetDataMutex() );
	GetData()->m_tcpDest = dest;
}

CI2PAddress CContact::getUDPDest(void) const
{
	return GetData()->m_udpDest;
}


void CContact::setUDPDest(const CI2PAddress & dest)
{
	//wxMutexLocker lock( GetDataMutex() );
	GetData()->m_udpDest = dest;
}
//

void CContact::setExpireTime ( time_t value )
{
	//wxMutexLocker lock( GetDataMutex() );
	GetData()->m_expires = value;
}

uint8_t CContact::getType(void) const
{
	return GetData()->m_type;
}

//! Increments m_type up to 4 (dead contact) (processed max every 10 seconds)
//! and gives at least a 2mn delay to the contact to answer.

void CContact::checkingType()
{
	if (!GetData()->m_inKadNodes)
		return ;
	
	if(time(NULL) - GetData()->m_lastTypeSet < 10 || GetData()->m_type == 4) {
		return;
	}
	{
		//wxMutexLocker lock( GetDataMutex() );
		GetData()->m_lastTypeSet = time(NULL);
	}
	if (GetData()->m_expires < GetData()->m_lastTypeSet) GetData()->m_expires = GetData()->m_lastTypeSet + MIN2S(2);
	GetData()->m_type++;
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "checking type of contact %s : type set to %d" ) )
			% GetData()->m_udpDest.humanReadable() % GetData()->m_type );
	
	Notify_KadNode_Added( new CContact(*this));
}


void CContact::updateType()
{
	{
		//wxMutexLocker lock( GetDataMutex() );
		GetData()->m_lastTypeUpdate = time(NULL);
	
		uint32_t hours = (GetData()->m_lastTypeUpdate-GetData()->m_created)/HR2S(1);
		switch(hours) {
			case 0:
				GetData()->m_type = 2;
				GetData()->m_expires = GetData()->m_lastTypeUpdate + HR2S(1);
				break;
			case 1:
				GetData()->m_type = 1;
				GetData()->m_expires = GetData()->m_lastTypeUpdate + (int)HR2S(1.5);
				break;
			default:
				GetData()->m_type = 0;
				GetData()->m_expires = GetData()->m_lastTypeUpdate + HR2S(2);
		}
	}
	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "updating type of contact %s : type set to %d" ) )
			% GetData()->m_udpDest.humanReadable() % GetData()->m_type );
	
	Notify_KadNode_Added(new CContact(*this));
}

