//
// This file is part of the imule Project.
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2003 Barry Dunne (http://www.emule-project.net)
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

#ifndef __CONTACT_H__
#define __CONTACT_H__

#include <set>

#include "../utils/UInt128.h"
#include "../kademlia/Prefs.h"
#include "../kademlia/Kademlia.h"
#include "../../SafeFile.h"
#include "i2p/CI2PAddress.h"
#include "wx/object.h"
#include "wx/thread.h"

////////////////////////////////////////

namespace Kademlia
{
////////////////////////////////////////

	class CContact : public wxObject
	{
		public:
			CContact();

			static CContact Self();
			
			~CContact();

			CContact ( CFileDataIO & file );

			class BadTypeException{ public: BadTypeException(uint8_t i) {type=i;} uint32_t type;};
			
			wxMutex &       GetDataMutex();
			
			bool            isInvalid() const ;
			
			void            writeToFile( CFileDataIO & file );
			
			void            addedToKadNodes();
			
			void            removedFromKadNodes();
			
			bool            inKadNodes() const { return GetData()->m_inKadNodes; }
			
			const wxString  getInfoString( void ) const;

			const CUInt128& getClientID ( void ) const { return GetData()->m_clientID; };

			const wxString  getClientIDString ( void ) const;

			void            setClientID ( const CUInt128 &clientID );

			const CUInt128& getDistance ( void ) const { return GetData()->m_distance; }

			const wxString  getDistanceString ( void ) const;

			void            setDistanceFrom( const CUInt128 & other ) ;

			CI2PAddress     getTCPDest ( void ) const;

			void            setTCPDest ( const CI2PAddress & dest );

			CI2PAddress     getUDPDest ( void ) const;

			void            setUDPDest ( const CI2PAddress & dest );
			
			uint8_t            getType ( void ) const;

			void            updateType();

			void            checkingType();

			time_t    	getCreatedTime() const         { return GetData()->m_created;     }

			time_t    	getExpireTime() const          { return GetData()->m_expires;     }

			void            setExpireTime ( time_t value ) ;

			time_t    	getLastTypeSet() const         { return GetData()->m_lastTypeSet; }
			
			time_t    	getLastTypeUpdate() const         { return GetData()->m_lastTypeUpdate; }
			
			static const std::set<CContact> getKadContacts()
			{
				return s_kadContacts ;
			}

			bool operator< (const CContact& ) const;

			bool operator==(const CContact& ) const;


		private:

			void            initContact(); // Common initialization goes here

			class Data : public wxObjectRefData
			{
				public:
				bool            m_inKadNodes;               // flag indicating if contact is to be counted in statistics
				CUInt128	m_clientID;
				CUInt128	m_distance;
				CI2PAddress	m_tcpDest;
				CI2PAddress	m_udpDest;

				uint8_t		m_type; /* set to 3 in initContact,
						   set to 2, 1 or 0 in updateType according to time since contact creation
						   (called by SetAlive, when receiving a response from contact)
						   incremented in checkingType */
				time_t		m_lastTypeSet;
				time_t		m_lastTypeUpdate;
				time_t		m_expires;
				time_t		m_created;
				uint32_t		m_inUse;

				wxMutex         m_mutex;
			};			
			static std::set<CContact>    s_kadContacts ;
			Data * GetData() const   { return wx_static_cast(Data *, m_refData); }
	};


} // End namespace

#endif // __CONTACT_H__
