//
// This file is part of imule Project
//
// Copyright (c) 2004-2006 Angel Vidal (Kry) ( kry@amule.org )
// Copyright (c) 2004-2006 imule Project ( http://www.imule-project.net )
// Copyright (C)2003 Barry Dunne (http://www.emule-project.net)

// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either
// version 2 of the License, or (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA

// This work is based on the java implementation of the Kademlia protocol.
// Kademlia: Peer-to-peer routing based on the XOR metric
// Copyright (C) 2002  Petar Maymounkov [petar@post.harvard.edu]
// http://kademlia.scs.cs.nyu.edu

// Note To Mods //
/*
Please do not change anything here and release it..
There is going to be a new forum created just for the Kademlia side of the client..
If you feel there is an error or a way to improve something, please
post it in the forum first and let us look at it.. If it is a real improvement,
it will be added to the offical client.. Changing something without knowing
what all it does can cause great harm to the network if released in mass form..
Any mod that changes anything within the Kademlia side will not be allowed to advertise
there client on the eMule forum..
*/

/**
 * The *Zone* is just a node in a binary tree of *Zone*s.
 * Each zone is either an internal node or a leaf node.
 * Internal nodes have "bin == null" and "subZones[i] != null",
 * leaf nodes have "subZones[i] == null" and "bin != null".
 *
 * All key unique id's are relative to the center (self), which
 * is considered to be 000..000
 */
//#include "stdafx.h"
#include "RoutingZone.h"
#include "Contact.h"
#include "RoutingBin.h"
#include "../utils/UInt128.h"
//#include "../utils/MiscUtils.h"
#include "../kademlia/Kademlia.h"
#include "../kademlia/Prefs.h"
#include "../kademlia/SearchManager.h"
#include "../kademlia/Defines.h"
#include "../kademlia/Error.h"
#include "../net/KademliaUDPListener.h"
#include "../../OtherFunctions.h"
#include "../../OPCodes.h"
#include "../../imule.h"
#include "../../CFile.h"
#include "../../Logger.h"
#include "../../NetworkFunctions.h"
#include "../../ArchSpecific.h"
#include "../../Preferences.h"
#include "../../Statistics.h"

#include <common/Format.h>

#include <wx/file.h>

//#warning EC
#ifndef IMULE_DAEMON
#include "../../imuleDlg.h"
#include "../../KadDlg.h"
#endif

#include <cmath>
#include <algorithm>		// Needed for std::min

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE

static char THIS_FILE[] = __FILE__;

#endif


////////////////////////////////////////

using namespace Kademlia;

////////////////////////////////////////

// This is just a safety precaution
#define CONTACT_FILE_LIMIT 5000

wxString	CRoutingZone::m_filename;

CUInt128	CRoutingZone::me ( ( uint32_t ) 0 );

ContactSet      CRoutingZone::s_ContactsSet ;


CRoutingZone::CRoutingZone()
{
	// Can only create routing zone after prefs
	me = CKademlia::getPrefs()->getKadID();
	m_filename = theApp.ConfigDir + wxT ( "nodes.dat" );
	CUInt128 zero ( ( uint32_t ) 0 );
	init ( NULL, 0, zero );
}

CRoutingZone::CRoutingZone ( CRoutingZone *super_zone, int level, const CUInt128 &zone_index )

{
	init ( super_zone, level, zone_index );
}

void CRoutingZone::init ( CRoutingZone *super_zone, int level, const CUInt128 &zone_index )

{
	m_superZone = super_zone;
	m_level = level;
	m_zoneIndex = zone_index;
	m_subZones[0] = NULL;
	m_subZones[1] = NULL;
	m_bin = new CRoutingBin();

	m_nextSmallTimer = time ( NULL ) + m_zoneIndex.get32BitChunk ( 3 );

	if ( ( m_superZone == NULL ) && ( m_filename.Length() > 0 ) )
	{
		readFile();
	}

	dirty = false;

	startTimer();
}

CRoutingZone::~CRoutingZone()

{
	if ( ( m_superZone == NULL ) && ( m_filename.Length() > 0 ) )
	{
		writeFile();
		ContactSet toRemove = s_ContactsSet ;
		for ( ContactSet::iterator it=toRemove.begin(); it!=toRemove.end(); it++)
		{
			CContact c(*it);
			wxASSERT( c.inKadNodes() );
			c.removedFromKadNodes();
		}
		wxASSERT( s_ContactsSet.size()==0 );
		//s_ContactsSet.clear();
	}

	if ( isLeaf() )
	{
		delete m_bin;
	}
	else
	{
		delete m_subZones[0];
		delete m_subZones[1];
	}
}

void CRoutingZone::readFile ( void )
{
	AddDebugLogLineM ( false, logKadRouting, _ ( "CRoutingZone::readFile called" ) );
	try
	{
		uint32_t numContacts = 0;
		CFile file;

		if ( file.Open ( m_filename, CFile::read ) )
		{

			numContacts = file.ReadUInt32();

			for ( uint32_t i = 0; i < numContacts; i++ )
			{
				AddDebugLogLineM ( false, logKadRouting, wxString::Format ( _ ( "Reading Kad contact %u" ), i+1 ) );
				try {
					CContact contact( file );
					AddDebugLogLineM ( false, logKadRouting, wxString::Format ( _ ( "Kad contact %u read" ), i+1 ) );

					if ( contact . getUDPDest().isValid() )
					{
						AddDebugLogLineM ( false, logKadRouting, wxString::Format ( _ ( "Adding Kad contact %u" ), i+1 ) );
						add ( contact );
					}
				} catch ( CContact::BadTypeException e ) {
					AddDebugLogLineM ( true, logKadRouting, wxString::Format ( _ ( "Kad contact %u has a bad type (%u). Discarded." ), i+1, e.type ) );
				}
			}

			AddLogLineM ( false, wxString::Format ( _ ( "Read %u Kad contacts" ), numContacts ) );
		}

		if ( numContacts == 0 )
		{
			AddDebugLogLineM ( false, logKadRouting, _ ( "Error while reading Kad contacts - 0 entries" ) );
		}
	}
	catch ( const CSafeIOException& e )
	{
		AddLogLineM ( false, _( "IO error in CRoutingZone::readFile: " ) + e.what() );
	}
	catch ( std::exception& e)
	{
		AddLogLineM ( false, wxString(_( "Error in CRoutingZone::readFile: " )) << wxString::FromAscii(e.what()) );
	}
}

void CRoutingZone::exportFile ( wxString filename )
{
	try
	{
		unsigned int count = 0;
		CContact c;
		wxFile file ;

		if ( file.Open ( filename, wxFile::write ) )
		{
			ContactList contacts;
			getBootstrapContacts ( contacts, 200 );
			uint32_t contacts_size = contacts.size() ;
			AddDebugLogLineM ( false, logKadRouting, CFormat ( _ ( "Export file : %d kad contacts, %d bootstrap contacts." ) )
			                   % CContact::getKadContacts().size()
			                   % contacts_size );

			if ( thePrefs::GetWriteMyselfToNodesDat() ) contacts_size++ ;

			ContactList::const_iterator it;

			if ( thePrefs::GetWriteMyselfToNodesDat() )
			{
				CContact myself = CContact::Self() ;
				file.Write(myself.getUDPDest().toString()+wxT("\n")) ;
				AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Writing self kad info : %s" ) ) % myself.getInfoString() );
			}

			for ( it = contacts.begin(); it != contacts.end() && count < CONTACT_FILE_LIMIT; ++it )
			{
				count++;
				c = *it;
				file.Write( c.getUDPDest().toString() + wxT("\n") ) ;
				AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Writing node kad info : %s" ) ) % c . getInfoString() );
			}
		}

		AddDebugLogLineM ( false, logKadRouting, wxString::Format ( wxT ( "Wrote %d contacts to file." ), count ) );
	}
	catch ( const CIOFailureException& e )
	{
		AddDebugLogLineM ( false, logKadRouting, wxT ( "IO failure in CRoutingZone::writeFile: " ) + e.what() );
	}
}

void CRoutingZone::writeFile ( void )
{
	try
	{
		unsigned int count = 0;
		CContact c;
		CFile file;

		if ( file.Open ( m_filename, CFile::write ) )
		{
			ContactList contacts;
			getBootstrapContacts ( contacts, 200 );
			uint32_t contacts_size = contacts.size() ;
			AddDebugLogLineM ( false, logKadRouting, CFormat ( _ ( "Write file : %d kad contacts, %d bootstrap contacts." ) )
			                   % CContact::getKadContacts().size()
			                   % contacts_size );

			if ( thePrefs::GetWriteMyselfToNodesDat() ) contacts_size++ ;

			file.WriteUInt32 ( std::min ( contacts_size, ( uint32_t ) CONTACT_FILE_LIMIT ) );

			ContactList::const_iterator it;

			if ( thePrefs::GetWriteMyselfToNodesDat() )
			{
				count++ ;
				CContact myself = CContact::Self() ;
				myself.writeToFile( file ) ;
				AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Writing self kad info : %s" ) ) % myself.getInfoString() );
			}

			for ( it = contacts.begin(); it != contacts.end() && count < CONTACT_FILE_LIMIT; ++it )
			{
				count++;
				c = *it;
				c . writeToFile( file ) ;
				AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "Writing node kad info : %s" ) ) % c . getInfoString() );
			}
		}

		AddDebugLogLineM ( false, logKadRouting, wxString::Format ( wxT ( "Wrote %d contacts to file." ), count ) );
	}
	catch ( const CIOFailureException& e )
	{
		AddDebugLogLineM ( false, logKadRouting, wxT ( "IO failure in CRoutingZone::writeFile: " ) + e.what() );
	}
}

bool CRoutingZone::canSplit ( void ) const
{
	if ( m_level >= 127 )
	{
		return false;
	}

	/* Check if we are close to the center */
	if ( ( m_zoneIndex < KK || m_level < KBASE ) && m_bin->getSize() == K )
	{
		return true;
	}

	return false;
}

CContact CRoutingZone::add ( CContact contact, bool trusted )
{

	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::add(%s)" ) ) % contact . getInfoString() );
	
	CContact contactInZone = getContact( contact . getClientID() );

	if ( contact . getClientID()==me)
	{
		if ( contactInZone.isInvalid() )
		{
			remove(contact . getClientID());
		}
	}
	
	else if ( ! contactInZone.isInvalid() ) 
	{
		if (contactInZone . getUDPDest()!=contact . getUDPDest() || contactInZone . getTCPDest()!=contact . getTCPDest())
			if ( trusted )
			{
				contactInZone . setUDPDest(contact . getUDPDest());
				contactInZone . setTCPDest(contact . getTCPDest());
			}
			else
			{
				CKademlia::getUDPListener()->sendMyDetails(KADEMLIA_HELLO_REQ, contact . getUDPDest());
				contactInZone = CContact() ;
			}
	}
	else
	{
		contact . setDistanceFrom( me );
	
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "--------------1->(%s)" ) ) % contact . getInfoString() );

		if ( addByDistance ( contact ) )
		{
			contactInZone = contact ;
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "--------------2->(%s)" ) ) % contact . getInfoString() );
			contact . addedToKadNodes();
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "--------------3->(%s)" ) ) % contact . getInfoString() );
		}
	}
	return contactInZone ;
}


bool CRoutingZone::wouldLikeToAdd ( const CUInt128 &id ) const

{
	if ( id == me ) return false;

	CUInt128 distance ( me );

	distance.XOR ( id );

	return wouldLikeToAddByDistance ( distance, id );
}



/**
 * Adds a contact to the routing zone, according to its distance from self Kad node
 * @param CContact *  : the contact to add
 * @return bool : true iff the contact has been added
 */

bool CRoutingZone::addByDistance ( CContact contact )
{

	bool retVal = false;

	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 1 (%s)" ) ) % contact . getInfoString() );

	if ( !isLeaf() )
	{
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 2 (%s)" ) ) % contact . getInfoString() );
		retVal = m_subZones[contact . getDistance().getBitNumber ( m_level ) ]->addByDistance ( contact );
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 3 (%s)" ) ) % contact . getInfoString() );
	}
	else
	{
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 4 (%s)" ) ) % contact . getInfoString() );
		CContact c = m_bin->getContact ( contact . getClientID() );
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 5 (%s)" ) ) % contact . getInfoString() );
		
		// if id already stored, update its destinations
		if ( ! c . isInvalid() )
		{
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 6 (%s)" ) ) % contact . getInfoString() );
			c . setUDPDest ( contact . getUDPDest() );
			c . setTCPDest ( contact . getTCPDest() );
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 7 (%s)" ) ) % contact . getInfoString() );
		}
		// else try to add the contact (after splitting or merging if necessary)
		else if ( m_bin->getRemaining() > 0 )
		{
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 8 (%s)" ) ) % contact . getInfoString() );
			retVal = m_bin->add ( contact, false );
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 9 (%s)" ) ) % contact . getInfoString() );
		}
		else if ( canSplit() )
		{
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 10 (%s)" ) ) % contact . getInfoString() );
			split();
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 11 (%s)" ) ) % contact . getInfoString() );
			retVal = m_subZones[contact . getDistance().getBitNumber ( m_level ) ]->addByDistance ( contact );
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 12 (%s)" ) ) % contact . getInfoString() );
		}
		else
		{
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 13 (%s)" ) ) % contact . getInfoString() );
			retVal = m_bin->add ( contact, false );
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 14 (%s)" ) ) % contact . getInfoString() );
			merge();
			AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 15 (%s)" ) ) % contact . getInfoString() );
		}
	}

	AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "CRoutingZone::addByDistance 16 (%s)" ) ) % contact . getInfoString() );
	return retVal;
}


bool CRoutingZone::wouldLikeToAddByDistance ( const CUInt128 &distance, const CUInt128 &id ) const

{
	if ( !isLeaf() )
	{
		return m_subZones[distance.getBitNumber ( m_level ) ]->wouldLikeToAddByDistance ( distance, id );
	}
	else
	{
		if ( ! m_bin->getContact ( id ) . isInvalid() )
		{
			return false;
		}
		else if ( m_bin->getRemaining() > 0 )
		{
			return true;
		}
		else if ( canSplit() )
		{
			return true;
		}
		else if ( m_superZone != NULL )
		{
			if ( m_superZone->m_subZones[0]->isLeaf()
			        && m_superZone->m_subZones[1]->isLeaf()
			        && ( m_superZone->getNumContacts() < K / 2 ) )
				// merge would be possible
				return true;
		}
	}

	return false;
}


void CRoutingZone::remove ( const CUInt128 &id )
{
	CUInt128 distance ( me );
	distance.XOR ( id );

	if ( !isLeaf() )
	{

		m_subZones[distance.getBitNumber ( m_level ) ]->remove ( id );
	}
	else
	{
		CContact c = m_bin->getContact ( id );

		if ( ! c . isInvalid() )
		{

			m_bin->remove ( c );
		}
	}
}

bool CRoutingZone::setAlive ( const CI2PAddress & dest )

{
	if ( isLeaf() )
	{
		return m_bin->setAlive ( dest );
	}
	else
	{
		if ( m_subZones[0]->setAlive ( dest ) )
			return true;

		return m_subZones[1]->setAlive ( dest );
	}
}

CContact CRoutingZone::getContact ( const CUInt128 &id ) const

{
	if ( isLeaf() )
	{
		return m_bin->getContact ( id );
	}
	else
	{
		return m_subZones[id.getBitNumber ( m_level ) ]->getContact ( id );
	}
}

void CRoutingZone::getClosestTo ( uint32_t maxType, const CUInt128 &target, const CUInt128 &distance, uint32_t maxRequired, ContactMap & result, bool emptyFirst ) const

{
	// If leaf zone, do it here

	if ( isLeaf() )
	{
		m_bin->getClosestTo ( maxType, target, maxRequired, result, emptyFirst );
		return;
	}

	// otherwise, recurse in the closer-to-the-target subzone first
	int closer = distance.getBitNumber ( m_level );

	m_subZones[closer]->getClosestTo ( maxType, target, distance, maxRequired, result, emptyFirst );

	// if still not enough tokens found, recurse in the other subzone too
	if ( result . size() < maxRequired )
	{
		m_subZones[1-closer]->getClosestTo ( maxType, target, distance, maxRequired, result, false );
	}
}

void CRoutingZone::getAllEntries ( ContactList & result, bool emptyFirst )

{
	if ( isLeaf() )
	{
		m_bin->getEntries ( result, emptyFirst );
	}
	else
	{
		m_subZones[0]->getAllEntries ( result, emptyFirst );
		m_subZones[1]->getAllEntries ( result, false );
	}
}

void CRoutingZone::topDepth ( int depth, ContactList & result, bool emptyFirst )
{
	if ( isLeaf() )
	{
		m_bin->getEntries ( result, emptyFirst );
	}
	else if ( depth <= 0 )
	{
		randomBin ( result, emptyFirst );
	}
	else
	{
		m_subZones[0]->topDepth ( depth - 1, result, emptyFirst );
		m_subZones[1]->topDepth ( depth - 1, result, false );
	}
}

void CRoutingZone::randomBin ( ContactList & result, bool emptyFirst )
{
	if ( isLeaf() )
	{
		m_bin->getEntries ( result, emptyFirst );
	}
	else
	{
		m_subZones[rand() &1]->randomBin ( result, emptyFirst );
	}
}

uint32_t CRoutingZone::getMaxDepth ( void ) const
{
	if ( isLeaf() )
	{
		return 0;
	}

	return 1 + std::max ( m_subZones[0]->getMaxDepth(), m_subZones[1]->getMaxDepth() );
}

void CRoutingZone::split ( void )

{
	stopTimer();

	m_subZones[0] = genSubZone ( 0 );
	m_subZones[1] = genSubZone ( 1 );

	ContactList entries;
	m_bin->getEntries ( entries );
	ContactList::const_iterator it;

	for ( it = entries.begin(); it != entries.end(); ++it )
	{
		int sz = it->getDistance().getBitNumber ( m_level );

		m_subZones[sz]->m_bin->add ( *it );
	}

	delete m_bin;
	m_bin = NULL;
}

void CRoutingZone::merge ( void )
{
	dirty = false; /* This subzone/superzone won't be re-checked */

	AddDebugLogLineM ( false, logKadRouting, _ ( "Merge attempt" ) );

	if ( isLeaf() && m_superZone != NULL )
	{
		AddDebugLogLineM ( false, logKadRouting, _ ( "Recursive merge" ) );
		m_superZone->merge();
	}
	else if ( ( !isLeaf() )
	          && ( m_subZones[0]->isLeaf() && m_subZones[1]->isLeaf() )
	          &&	( getNumContacts() ) < ( K / 2 ) )
	{
		m_subZones[0]->stopTimer();
		m_subZones[1]->stopTimer();
		
		ContactList list0;
		ContactList list1;
		m_subZones[0]->m_bin->getEntries ( list0 );
		m_subZones[1]->m_bin->getEntries ( list1 );
		ContactList::const_iterator it;

		m_bin = new CRoutingBin();

		for ( it = list0.begin(); it != list0.end(); ++it )
		{

			m_bin->add ( *it );
		}

		for ( it = list1.begin(); it != list1.end(); ++it )
		{

			m_bin->add ( *it );
		}


		delete m_subZones[0];

		delete m_subZones[1];

		m_subZones[0] = NULL;

		m_subZones[1] = NULL;

		startTimer();

		AddDebugLogLineM ( false, logKadRouting, _ ( "Sucessful merge!" ) );

		if ( m_superZone != NULL )
		{
			m_superZone->merge();
		}
	}
	else
	{
		AddDebugLogLineM ( false, logKadRouting, _ ( "No merge possible" ) );
	}
}

bool CRoutingZone::isLeaf ( void ) const
{
	return ( m_bin != NULL );
}

CRoutingZone *CRoutingZone::genSubZone ( int side )

{
	CUInt128 newIndex ( m_zoneIndex );
	newIndex.shiftLeft ( 1 );

	if ( side != 0 )
	{
		newIndex.add ( 1 );
	}

	CRoutingZone *retVal = new CRoutingZone ( this, m_level + 1, newIndex );

	return retVal;
}

void CRoutingZone::startTimer ( void )
{
	time_t now = time ( NULL );
	// Start filling the tree, closest bins first.
	m_nextBigTimer = now + ( MIN2S ( 1 ) * m_zoneIndex.get32BitChunk ( 3 ) ) + SEC ( 10 );
	CKademlia::addEvent ( this );
}

void CRoutingZone::stopTimer ( void )
{
	CKademlia::removeEvent ( this );
}

bool CRoutingZone::onBigTimer ( void )
{
	if ( !isLeaf() )
	{
		return false;
	}

	if ( ( m_zoneIndex < KK || m_level < KBASE || m_bin->getRemaining() >= ( K*.4 ) ) )
	{
		randomLookup();
		return true;
	}

	return false;
}

//This is used when we find a leaf and want to know what this sample looks like.
//We fall back two levels and take a sample to try to minimize any areas of the
//tree that will give very bad results.
uint32_t CRoutingZone::estimateCount()
{
	if ( !isLeaf() )
	{
		return 0;
	}

	if ( m_level < KBASE )
	{
		return ( uint32_t ) ( pow ( 2, m_level ) *10 );
	}

	CRoutingZone* curZone = m_superZone->m_superZone->m_superZone;

	float modify = ( ( float ) curZone->getNumContacts() ) / 20.0F;
	return ( uint32_t ) ( pow ( 2, m_level - 2 ) *10* ( modify ) );
}

void CRoutingZone::onSmallTimer ( void )
{
	if ( !isLeaf() )
	{
		return;
	}

	dirty = false;

	wxString test ( m_zoneIndex.toBinaryString() );

	CContact c ;
	time_t now = time ( NULL );
	ContactList entries;
	ContactList::iterator it;

	// Remove dead entries
	m_bin->getEntries ( entries );

	for ( it = entries.begin(); it != entries.end(); ++it )
	{
		c = *it;

		if ( c . getType() == 4
		     && CRoutingZone::s_ContactsSet.size() > thePrefs::GetKeepKadContacts() )
		{
			if ( ( ( c . getExpireTime() > 0 ) && ( c . getExpireTime() <= now ) ) )
			{
				m_bin->remove ( c );
				dirty = true;
				continue;
			}
		}

		if ( c . getExpireTime() == 0 )
		{
			c . setExpireTime ( now );
		}
	}

	//Ping only contacts that are in the branches that meet the set level and are not close to our ID.
	//The other contacts are checked with the big timer.
	c = CContact();

	// Ping a contact only if this bin is more than half full
	if ( m_bin->getRemaining() < ( K* ( .4 ) ) )
	{
		c = m_bin->getOldest();
	}

	if ( ! c . isInvalid() )
	{
		//if ( c->getExpireTime() >= now || c->getType() == 4) { // NOTE: mkvore : I don't understand this logic
		// Do not ping the oldest contact if it is not close to be deleted

		if ( c . getExpireTime() >= now || c . getType() != 4 )
		{
			//dirty = true;

			m_bin->m_entries.remove ( c );
			m_bin->m_entries.push_back ( c );

			c = CContact();
		}
	}

	if ( ! c . isInvalid() )
	{
		AddDebugLogLineM ( false, logKadRouting, CFormat ( wxT ( "ping %s with KadHelloReq to keep it alive (onSmallTimer of routing zone %s)" ) )
		                   % c . getUDPDest().humanReadable() % test );
		c . checkingType();
		CKademlia::getUDPListener()->sendMyDetails ( KADEMLIA_HELLO_REQ, c . getUDPDest() );
	}
}

void CRoutingZone::randomLookup ( void )
{
	// Look-up a random client in this zone
	CUInt128 prefix ( m_zoneIndex );
	prefix.shiftLeft ( 128 - m_level );
	CUInt128 random ( prefix, m_level );
	random.XOR ( me );
	CSearchManager::findNode ( random );
}

uint32_t CRoutingZone::getNumContacts ( void ) const
{
	if ( isLeaf() )
	{
		return m_bin->getSize();
	}
	else
	{
		return m_subZones[0]->getNumContacts() + m_subZones[1]->getNumContacts();
	}
}

uint32_t CRoutingZone::getBootstrapContacts ( ContactList & results, uint32_t maxRequired )

{
	wxASSERT ( m_superZone == NULL );

	results . clear();

	uint32_t retVal = 0;
	ContactList top;
	topDepth ( LOG_BASE_EXPONENT, top );

	if ( top.size() > 0 )
	{
		ContactList::const_iterator it;

		for ( it = top.begin(); it != top.end(); ++it )
		{
			results . push_back ( *it );
			retVal++;

			if ( retVal == maxRequired )
			{
				break;
			}
		}
	}

	return retVal;
}
