//
// C++ Interface: CI2PAddress
//
// Description: 
//
//
// Author: MKVore <mkvore@mail.i2p>, (C) 2007
//
// Copyright: See COPYING file that comes with this distribution
//
//


#ifndef __address_h__
#define __address_h__

#include <wx/wx.h>
#include <stdint.h>

#include <inttypes.h>

#define PUBKEY_LEN 387
/* The length of a base 64 public key - it's actually 516, but +1 for '\0' */
#define BASE64_PUBKEY_LEN 516
typedef char sam_pubkey_t[BASE64_PUBKEY_LEN+1];  /* base 64 public key, followed by '\0' */

class CI2PAddress
{
	public:
		static uint16_t validLength();
		
		void init();
		CI2PAddress(const char & type = 0);
		CI2PAddress(const sam_pubkey_t);
		CI2PAddress(const CI2PAddress & src);		
		CI2PAddress& operator=( const CI2PAddress & src );
		
		bool isValid() const;
		bool operator!() const;
		
		const char * toBase64CharArray() const;
		const wxString toString() const {return wxString::FromAscii(m_b64);}
		static CI2PAddress fromString(wxString);
		
		wxString humanReadable() const;
		const wxString& getAlias() const {return m_alias;}
		void setAlias(wxString);
		
		bool operator== (const CI2PAddress &) const;
		bool operator!= (const CI2PAddress & other) const {return !(*this==other);}
		bool operator<  (const CI2PAddress &) const;
		
		wxUint32 size() const;
		wxUint32 writeBytes(char * buffer) const;
		wxUint32 readBytes(char * buffer);
		wxUint32 hashCode() const {return m_hashCode;}
		
	protected:
		void setHashCode();
		
		sam_pubkey_t m_b64;
		char         m_bytes[PUBKEY_LEN];
		wxString     m_alias;
		wxUint32     m_hashCode;
		char         m_type;  // special uses in imule
		
	public:
		static const CI2PAddress null;
		static const CI2PAddress complete_file;
		static const CI2PAddress incomplete_file;

	private:
		static uint16_t s_validLength;
};

#endif
