//
// This file is part of the imule Project.
//
// Copyright (c) 2003-2006 imule Team ( mkvore@mail.i2p / http://www.imule.i2p )
// Copyright (c) 2002 Merkur ( devs@emule-project.net / http://www.emule-project.net )
//
// Any parts of this program derived from the xMule, lMule or eMule project,
// or contributed by third-party developers are copyrighted by their
// respective authors.
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA
//

#ifndef UPDOWNCLIENT_H
#define UPDOWNCLIENT_H

#include <map>
#include <vector>

#include <wx/defs.h>		// Needed before any other wx/*.h
#include <wx/string.h>		// Needed for wxString
#include <wx/intl.h>

#include "Types.h"		// Needed for uint8_t, uint16_t, uint8_t, uint16_t, uint32_t and wxULongLong
#include "CTypedPtrList.h"	// Needed for CTypedPtrList
#include "GetTickCount.h"	// Needed for GetTickCount
#include "MD4Hash.h"
#include "common/StringFunctions.h"
#include "NetworkFunctions.h"


#include "i2p/CI2PAddress.h"

typedef std::vector<bool> BitVector;


class CPartFile;
class CClientTCPSocket;
class CClientCredits;
class CPacket;
class CFriend;
class Requested_Block_Struct;
class CKnownFile;
class Pending_Block_Struct;
class CMemFile;
class CMemFile;
class Requested_File_Struct;
class CAICHHash;


// uploadstate
enum EUploadState {
    US_UPLOADING = 0,
    US_ONUPLOADQUEUE,
//#define   US_WAITCALLBACK     2
    US_CONNECTING,
    US_PENDING,
//#define   US_LOWTOLOWIP       5
    US_BANNED,
    US_ERROR,
    US_NONE
};

// downloadstate
enum EDownloadState {
	DS_DOWNLOADING = 0,
	DS_ONQUEUE,
	DS_CONNECTED,
	DS_CONNECTING,
	//DS_WAITCALLBACK,
	//DS_WAITCALLBACKKAD,
	DS_REQHASHSET,
	DS_NONEEDEDPARTS,
	DS_TOOMANYCONNS,
	DS_TOOMANYCONNSKAD,
	//DS_LOWTOLOWIP,
	DS_BANNED,
	DS_ERROR,
	DS_NONE,
	DS_REMOTEQUEUEFULL  // not used yet, except in statistics
};

// m_byChatstate
enum EChatState {
	MS_NONE = 0,
	MS_CHATTING,
	MS_CONNECTING,
	MS_UNABLETOCONNECT
};

enum ESourceFrom {
	SF_NONE,
	SF_LOCAL_SERVER,
	SF_REMOTE_SERVER,
	SF_KADEMLIA,
	SF_SOURCE_EXCHANGE,
	SF_PASSIVE,
	SF_LINK,
	SF_SOURCE_SEEDS
};

enum ESecureIdentState{
	IS_UNAVAILABLE		= 0,
	IS_ALLREQUESTSSEND	= 0,
	IS_SIGNATURENEEDED	= 1,
	IS_KEYANDSIGNEEDED	= 2
};

enum EInfoPacketState{
	IP_NONE			= 0,
	IP_EDONKEYPROTPACK	= 1,
	IP_EMULEPROTPACK	= 2,
	IP_BOTH			= 3
};

// NOTE : not for I2P
//enum EKadState{
	//KS_NONE,
	//KS_QUEUED_FWCHECK,
	//KS_CONNECTING_FWCHECK,
	//KS_CONNECTED_FWCHECK,
	//KS_QUEUED_BUDDY,
	//KS_INCOMING_BUDDY,
	//KS_CONNECTING_BUDDY,
	//KS_CONNECTED_BUDDY,
	//KS_NONE_LOWID,
	//KS_WAITCALLBACK_LOWID,
	//KS_QUEUE_LOWID
//};

//! Used to keep track of the state of the client
enum ClientState
{
	//! New is for clients that have just been created.
	CS_NEW = 0,
	//! Listed is for clients that are on the clientlist
	CS_LISTED,
	//! Dying signifies clients that have been queued for deletion
	CS_DYING
};


class CUpDownClient
{
	friend class CClientList;
	friend class CUpDownClientListRem;
private:
	/**
	 * Please note that only the ClientList is allowed to delete the clients.
	 * To schedule a client for deletion, call the CClientList::AddToDeleteQueue
	 * funtion, which will safely remove dead clients once every second.
	 */
	~CUpDownClient();
	
public:
#ifdef CLIENT_GUI
	CUpDownClient(class CEC_UpDownClient_Tag *);
#else
	//base
	CUpDownClient(CClientTCPSocket* sender = 0);
	
	// I2P
	// CUpDownClient(uint16_t in_port, uint32_t in_userid, uint32_t in_serverup, uint16_t in_serverport,CPartFile* in_reqfile, bool ed2kID, bool checkfriend);
	CUpDownClient(const CI2PAddress & in_dest, const CI2PAddress & in_serverdest, 
		      CPartFile* in_reqfile, bool ed2kID, bool checkfriend);
			
#endif
	/**
	 * This function should be called when the client object is to be deleted.
	 * It'll close the socket of the client and add it to the deletion queue
	 * owned by the CClientList class. However, if the CUpDownClient isn't on 
	 * the normal clientlist, it will be deleted immediatly.
	 * 
	 * The purpose of this is to avoid clients suddenly being removed due to 
	 * asyncronous events, such as socket errors, which can result in the 
	 * problems, as each CUpDownClient object is often kept in multiple lists,
	 * and instantly removing the client poses the risk of invalidating 
	 * currently used iterators and/or creating dangling pointers.
	 * 
	 * @see CClientList::AddToDeleteQueue
	 * @see CClientList::Process
	 */	
	void		Safe_Delete();

	/**
	 * Specifies if the client has been queued for deletion.
	 *
	 * @return True if Safe_Delete has been called, false otherwise.
	 */
	bool		HasBeenDeleted()		{ return m_clientState == CS_DYING; }

	ClientState	GetClientState()		{ return m_clientState; }

	bool		Disconnected(const wxString& strReason, bool bFromSocket = false);
	bool		TryToConnect(bool bIgnoreMaxCon = false);
	void		ConnectionEstablished();
	const wxString&	GetUserName() const		{ return m_Username; }
	
	void		SetTCPDest        ( const CI2PAddress & val );
	void		SetUDPDest        ( const CI2PAddress & val ) { m_nUDPDest    = val; }
	void		SetServerDest     ( const CI2PAddress & val ) { m_nServerDest = val; }

	const CI2PAddress & GetTCPDest       () const	{ return m_nTCPDest;     }
	const CI2PAddress & GetUDPDest       () const	{ return m_nUDPDest; }
	const CI2PAddress & GetServerDest    () const	{ return m_nServerDest; }
	
	
	wxULongLong	GetTransferedDown() const	{ return m_nTransferedDown; }
	
	
	const CMD4Hash&	GetUserHash() const		{ return m_UserHash; }
	void		SetUserHash(const CMD4Hash& userhash);
	void		ValidateHash()			{ m_HasValidHash = !m_UserHash.IsEmpty(); }
	bool		HasValidHash() const		{ return m_HasValidHash; }
	uint32_t		GetVersion() const		{ return m_nClientVersion;}
	uint8_t		GetMuleVersion() const		{ return m_byEmuleVersion;}
	bool		ExtProtocolAvailable() const	{ return m_bEmuleProtocol;}
	bool		IsEmuleClient()	const		{ return m_byEmuleVersion;}
	bool		IsBanned() const;
	const wxString&	GetClientFilename() const	{ return m_clientFilename; }
	
	
	
	uint8_t		GetUDPVersion() const		{ return m_byUDPVer; }
	uint8_t		GetExtendedRequestsVersion() const { return m_byExtendedRequestsVer; }
	bool		IsFriend() const 		{ return m_Friend != NULL; }
	bool		IsML() const			{ return m_bIsML; }
	bool		IsHybrid() const		{ return m_bIsHybrid; }
	uint32_t		GetCompatibleClient() const	{ return m_byCompatibleClient; }

	void		ClearDownloadBlockRequests();
	void		RequestSharedFileList();
	void		ProcessSharedFileList(const char* pachPacket, uint32_t nSize, wxString& pszDirectory);

	wxString	GetUploadFileInfo();

	void		SetUserName(const wxString& NewName) { m_Username = NewName; }

	uint8_t		GetClientSoft() const		{ return m_clientSoft; }
	void		ReGetClientSoft();
	bool		ProcessHelloAnswer(const char *pachPacket, uint32_t nSize);
	bool		ProcessHelloPacket(const char *pachPacket, uint32_t nSize);
	void		SendHelloAnswer();
	bool		SendHelloPacket();
	void		SendMuleInfoPacket(bool bAnswer, bool OSInfo = false);
	bool		ProcessMuleInfoPacket(const char* pachPacket, uint32_t nSize);
	void		ProcessMuleCommentPacket(const char *pachPacket, uint32_t nSize);
	bool		Compare(const CUpDownClient* tocomp, bool bIgnoreUserhash = false) const;
	void		SetLastSrcReqTime()		{ m_dwLastSourceRequest = ::GetTickCount(); }
	void		SetLastSrcAnswerTime()		{ m_dwLastSourceAnswer = ::GetTickCount(); }
	void		SetLastAskedForSources()	{ m_dwLastAskedForSources = ::GetTickCount(); }
	uint32_t		GetLastSrcReqTime() const 	{ return m_dwLastSourceRequest; }
	uint32_t		GetLastSrcAnswerTime() const	{ return m_dwLastSourceAnswer; }
	uint32_t		GetLastAskedForSources() const	{ return m_dwLastAskedForSources; }
	bool		GetFriendSlot() const 		{ return m_bFriendSlot; }
	void		SetFriendSlot(bool bNV)		{ m_bFriendSlot = bNV; }
	void		SetCommentDirty(bool bDirty = true)	{ m_bCommentDirty = bDirty; }
	uint8_t		GetSourceExchangeVersion() const	{ return m_bySourceExchangeVer; }
	bool		SafeSendPacket(CPacket* packet);

	void		SendPublicKeyPacket();
	void		SendSignaturePacket();
	void		ProcessPublicKeyPacket(const uint8_t* pachPacket, uint32_t nSize);
	void		ProcessSignaturePacket(const uint8_t* pachPacket, uint32_t nSize);
	ESecureIdentState   GetSecureIdentState();

	void		SendSecIdentStatePacket();
	void		ProcessSecIdentStatePacket(const uint8_t* pachPacket, uint32_t nSize);

	uint8_t		GetInfoPacketsReceived() const	{ return m_byInfopacketsReceived; }
	void		InfoPacketsReceived();

	//upload
	EUploadState    GetUploadState() const		{ return m_nUploadState; }
	void            SetUploadState(EUploadState news);
	wxULongLong	    GetTransferredUp() const	{ return m_nTransferredUp; }
	wxULongLong	    GetSessionUp() const		{ return m_nTransferredUp - m_nCurSessionUp; }
	void		    ResetSessionUp() {
						m_nCurSessionUp = m_nTransferredUp;
						m_addedPayloadQueueSession = 0;
						m_nCurQueueSessionPayloadUp = 0;
					}
	uint32_t		GetUploadDatarate() const	{ return m_nUpDatarate; }

#ifndef CLIENT_GUI
	uint32_t		GetWaitTime() const 		{ return m_dwUploadTime - GetWaitStartTime(); }
	uint32_t		GetUpStartTimeDelay() const	{ return ::GetTickCount() - m_dwUploadTime; }
	uint32_t		GetWaitStartTime() const;
#else
	uint32_t		m_WaitTime, m_UpStartTimeDelay, m_WaitStartTime;
	uint32_t		GetWaitTime() const		{ return m_WaitTime; }
	uint32_t		GetUpStartTimeDelay() const	{ return m_UpStartTimeDelay; }
	uint32_t		GetWaitStartTime() const	{ return m_WaitStartTime; }
#endif

	bool		IsDownloading()	const 		{ return (m_nUploadState == US_UPLOADING); }

#ifdef CLIENT_GUI
	uint32_t m_base_score, m_score;
	uint32_t		GetScore(
				bool WXUNUSED(sysvalue),
				bool WXUNUSED(isdownloading) = false,
				bool WXUNUSED(onlybasevalue) = false) const
	{
		// lfroen: it's calculated
		return 0;
	}
#else
	uint32_t		GetScore(
				bool sysvalue,
				bool isdownloading = false,
				bool onlybasevalue = false) const;
#endif

	void		AddReqBlock(Requested_Block_Struct* reqblock);
	void		CreateNextBlockPackage();
	void		SetUpStartTime() 		{ m_dwUploadTime = ::GetTickCount(); }
	void		SetWaitStartTime();
	void		ClearWaitStartTime();
	void		SendHashsetPacket(const CMD4Hash& forfileid);
	bool		SupportMultiPacket() const	{ return m_bMultiPacket; }

	void		SetUploadFileID(CKnownFile *newreqfile);

	/**
	 *Gets the file actually on upload
	 *
	 */
	const CKnownFile* GetUploadFile() const		{ return m_uploadingfile; }

	void		SendOutOfPartReqsAndAddToWaitingQueue();
	void		ProcessExtendedInfo(CMemFile *data, CKnownFile *tempreqfile);
	void		ProcessFileInfo(CMemFile* data, const CPartFile* file);
	void		ProcessFileStatus(bool bUdpPacket, CMemFile* data, const CPartFile* file);

	const CMD4Hash&	GetUploadFileID() const		{ return m_requpfileid; }
	void		SetUploadFileID(const CMD4Hash& new_id);
	void		ClearUploadFileID()		{ m_requpfileid.Clear(); m_uploadingfile = NULL;};
	uint32_t		SendBlockData();
	void		ClearUploadBlockRequests();
	void		SendRankingInfo();
	void		SendCommentInfo(CKnownFile *file);
	bool 		IsDifferentPartBlock() const;
	void		UnBan();
	void		Ban();
	bool		m_bAddNextConnect;      // VQB Fix for LowID slots only on connection
	uint32_t		GetAskedCount() const 		{ return m_cAsked; }
	void		AddAskedCount()			{ m_cAsked++; }
	void		ClearAskedCount()		{ m_cAsked = 1; }	// 1, because it's cleared *after* the first request...
	void		FlushSendBlocks();	// call this when you stop upload, 
						// or the socket might be not able to send
	void		SetLastUpRequest()		{ m_dwLastUpRequest = ::GetTickCount(); }
	uint32_t		GetLastUpRequest() const 	{ return m_dwLastUpRequest; }
	wxULongLong	GetUpPartCount() const 		{ return m_upPartStatus.size(); }


	//download
	void 		SetRequestFile(CPartFile* reqfile); 
	CPartFile*	GetRequestFile() const		{ return m_reqfile; }

	EDownloadState  GetDownloadState() const	{ return m_nDownloadState; }
	void            SetDownloadState(EDownloadState byNewState);
	uint32_t		GetLastAskedTime() const	{ return m_dwLastAskedTime; }

	bool		IsPartAvailable(size_t iPart) const
					{ return ( iPart < m_downPartStatus.size() ) ? m_downPartStatus[iPart] : false; }
	bool		IsUpPartAvailable(size_t iPart) const
					{ return ( iPart < m_upPartStatus.size() ) ? m_upPartStatus[iPart] : false;}

	const BitVector& GetPartStatus() const		{ return m_downPartStatus; }
	const BitVector& GetUpPartStatus() const	{ return m_upPartStatus; }
	float		GetKBpsDown() const		{ return kBpsDown; }
	float		CalculateKBpsDown();
	uint16_t		GetRemoteQueueRank() const	{ return m_nRemoteQueueRank; }
	uint16_t		GetOldRemoteQueueRank() const	{ return m_nOldRemoteQueueRank; }
	void		SetRemoteQueueFull(bool flag)	{ m_bRemoteQueueFull = flag; }
	bool		IsRemoteQueueFull() const 	{ return m_bRemoteQueueFull; }
	void		SetRemoteQueueRank(uint16_t nr);
	bool		AskForDownload();
	void		SendStartUploadReq();
	void		SendFileRequest();
	void		ProcessHashSet(const char *packet, uint32_t size);
	bool		AddRequestForAnotherFile(CPartFile* file);
	bool		DeleteFileRequest(CPartFile* file);
	void		DeleteAllFileRequests();
	void		SendBlockRequests();
	void		ProcessBlockPacket(const char* packet, uint32_t size, bool packed = false);

#ifndef CLIENT_GUI
	wxULongLong	GetAvailablePartCount() const;
#else
	wxULongLong	m_AvailPartCount;
	wxULongLong	GetAvailablePartCount() const	{ return m_AvailPartCount; }
#endif

	bool		SwapToAnotherFile(bool bIgnoreNoNeeded, bool ignoreSuspensions, bool bRemoveCompletely, CPartFile* toFile = NULL);
	void		UDPReaskACK(uint16_t nNewQR);
	void		UDPReaskFNF();
	void		UDPReaskForDownload();
	bool		IsSourceRequestAllowed();
	uint16_t		GetUpCompleteSourcesCount() const	{ return m_nUpCompleteSourcesCount; }
	void		SetUpCompleteSourcesCount(uint16_t n)	{ m_nUpCompleteSourcesCount = n; }

	//chat
	EChatState  GetChatState()			{ return m_byChatstate; }
    void		SetChatState(EChatState nNewS)	{ m_byChatstate = nNewS; }

	//File Comment
	const wxString&	GetFileComment() const 		{ return m_strComment; }
	uint8_t		GetFileRating() const		{ return m_iRating; }

	const wxString&	GetSoftStr() const 		{ return m_clientSoftString; }
	const wxString&	GetSoftVerStr() const		{ return m_clientVerString; }
	
	/* I2P
	uint16_t		GetKadPort() const		{ return m_nKadPort; }
	void		SetKadPort(uint16_t nPort)	{ m_nKadPort = nPort; }
	*/
	//
	
	// Kry - AICH import
	void		SetReqFileAICHHash(CAICHHash* val);
	CAICHHash*	GetReqFileAICHHash() const	{return m_pReqFileAICHHash;}
	bool		IsSupportingAICH() const	{return m_fSupportsAICH & 0x01;}
	void		SendAICHRequest(CPartFile* pForFile, wxULongLong nPart);
	bool		IsAICHReqPending() const	{return m_fAICHRequested; }
	void		ProcessAICHAnswer(const char* packet, uint32_t size);
	void		ProcessAICHRequest(const char* packet, uint32_t size);
	void		ProcessAICHFileHash(CMemFile* data, const CPartFile* file);	

	EUtf8Str	GetUnicodeSupport() const;

	// Barry - Process zip file as it arrives, don't need to wait until end of block
	int 		unzip(Pending_Block_Struct *block, uint8_t *zipped, uint32_t lenZipped, uint8_t **unzipped, uint32_t *lenUnzipped, int iRecursion = 0);
	// Barry - Sets string to show parts downloading, eg NNNYNNNNYYNYN
	wxString	ShowDownloadingParts() const;
	void 		UpdateDisplayedInfo(bool force = false);
	int 		GetFileListRequested() const 	{ return m_iFileListRequested; }
	void 		SetFileListRequested(int iFileListRequested) { m_iFileListRequested = iFileListRequested; }

	void		ResetFileStatusInfo();

	bool		CheckHandshakeFinished(uint32_t protocol, uint32_t opcode) const;

	bool		GetSentCancelTransfer() const	{ return m_fSentCancelTransfer; }
	void		SetSentCancelTransfer(bool bVal)	{ m_fSentCancelTransfer = bVal; }

	wxString	GetClientFullInfo();

	const wxString& GetClientOSInfo() const		{ return m_sClientOSInfo; }

	void		ProcessPublicIPAnswer(const uint8_t* pbyData, uint32_t uSize);
	void		SendPublicIPRequest();

	/**
	 * Sets the current socket of the client.
	 * 
	 * @param socket The pointer to the new socket, can be NULL.
	 *
	 * Please note that this function DOES NOT delete the old socket.
	 */
	void 		SetSocket(CClientTCPSocket* socket);

	/**
	 * Function for accessing the socket owned by a client.
	 * 
	 * @return The pointer (can be NULL) to the socket used by this client.
	 *
	 * Please note that the socket object is quite volatile and can be removed
	 * from one function call to the next, therefore, you should normally use 
	 * the safer functions below, which all check if the socket is valid before
	 * deferring it.
	 */
	CClientTCPSocket* GetSocket() const		{ return m_socket; }

	/**
	 * Safe function for checking if the socket is connected.
	 *
	 * @return True if the socket exists and is connected, false otherwise.
	 */
#ifndef CLIENT_GUI
	bool		IsConnected() const;
#else
	bool		m_IsConnected;
	bool		IsConnected() const		{ return m_IsConnected; }
#endif

	/**
	 * Safe function for sending packets.
	 *
	 * @return True if the socket exists and the packet was sent, false otherwise.
	 */
	bool		SendPacket(CPacket* packet, bool delpacket = true, bool controlpacket = true);

	/**
	 * Safe function for setting the download limit of the socket.
	 *
	 * @return Current download speed of the client.
	 */
	float		SetDownloadLimit(uint32_t reducedownload);

	/**
	 * Sends a message to a client
	 *
	 * @return True if sent, false if connecting
	 */
	bool	SendMessage(const wxString& message);

	wxULongLong	GetPayloadInBuffer() const	{ return m_addedPayloadQueueSession - GetQueueSessionPayloadUp(); }
	wxULongLong	GetQueueSessionPayloadUp() const	{ return m_nCurQueueSessionPayloadUp; }
	void		SendCancelTransfer(CPacket* packet = NULL);
	bool		HasBlocks() const		{ return !m_BlockRequests_queue.IsEmpty(); }

	/* Source comes from? */
	ESourceFrom		GetSourceFrom() const	{ return (ESourceFrom)m_nSourceFrom; }
	void			SetSourceFrom(ESourceFrom val)	{ m_nSourceFrom = val; }

	uint8_t		GetKadVersion()			{ return m_byKadVersion; }
	
	/* Returns the client hash type (SO_EMULE, mldonkey, etc) */
	int		GetHashType() const;

	/**
	 * Checks that a client isn't aggressively re-asking for files.
	 * 
	 * Call this when a file is requested. If the time since the last request is
	 * less than MIN_REQUESTTIME, 3 is added to the m_Aggressiveness variable.
	 * If the time since the last request is >= MIN_REQUESTTIME, the variable is
	 * decremented by 1. The client is banned if the variable reaches 10 or above.
	 *
	 * To check if a client is aggressive use the IsClientAggressive() function.
	 * 
	 * Currently this function is called when the following packets are recieved:
	 *  - OP_STARTUPLOADREQ
	 *  - OP_REASKFILEPING
	 */
	void		CheckForAggressive();

	const wxString&	GetClientModString() const	{ return m_strModVersion; }

	const wxString&	GetClientVerString() const	{ return m_fullClientVerString; }

	const wxString&	GetVersionString() const	{ return m_clientVersionString; }

	void		UpdateStats();

	/* Returns a pointer to the credits, only for hash purposes */
	void*		GetCreditsHash() const { return (void*)credits; }
	
	wxULongLong	GetLastBlockOffset() const { return m_nLastBlockOffset; }
	
	bool		GetOSInfoSupport() const { return m_fOsInfoSupport; }
	
	wxULongLong	GetLastPartAsked() const { return m_lastPartAsked; }
	
	void		SetLastPartAsked(wxULongLong nPart) { m_lastPartAsked = nPart; }
	
	CFriend*	GetFriend() const { return m_Friend; }
	
	void		SetFriend(CFriend* newfriend) { m_Friend = newfriend; }
	
	bool		IsIdentified() const;
	
	bool		IsBadGuy() const;
	
	bool		SUIFailed() const;
	
	bool		SUINeeded() const;
	
	bool		SUINotSupported() const;

	wxULongLong		GetDownloadedTotal() const;
	
	wxULongLong		GetUploadedTotal() const;
	
	float		GetScoreRatio() const;
	
	uint32_t		GetCreationTime() const { return m_nCreationTime; }
	
private:
	
	CClientCredits	*credits;
	CFriend 	*m_Friend;

	wxULongLong	m_nTransferredUp;
	wxULongLong	m_nCurQueueSessionPayloadUp;
	wxULongLong	m_addedPayloadQueueSession;

	struct TransferredData {
		uint32_t	datalen;
		uint32_t	timestamp;
	};

	//////////////////////////////////////////////////////////
	// Upload data rate computation
	//
	uint32_t		m_nUpDatarate;
	wxULongLong	m_nSumForAvgUpDataRate;
	CList<TransferredData> m_AvarageUDR_list;


	/**
	 * This struct is used to keep track of CPartFiles which this source shares.
	 */
	struct A4AFStamp {
		//! Signifies if this sources has needed parts for this file. 
		bool NeededParts;
		//! This is set when we wish to avoid swapping to this file for a while.
		uint32_t timestamp;
	};

	//! I typedef in the name of readability!
	typedef std::map<CPartFile*, A4AFStamp> A4AFList;
	//! This list contains all PartFiles which this client can be used as a source for.
	A4AFList m_A4AF_list;

	/**
	 * Helper function used by SwapToAnotherFile().
	 *
	 * @param it The iterator of the PartFile to be examined.
	 * @param ignorenoneeded Do not check for the status NoNeededParts when checking the file.
	 * @param ignoresuspended Do not check the timestamp when checking the file.
	 * @return True if the file is a viable target, false otherwise.
	 *
	 * This function is used to perform checks to see if we should consider 
	 * this file a viable target for A4AF swapping. Unless ignoresuspended is
	 * true, it will examine the timestamp of the file and reset it if needed.
	 */
	bool		IsValidSwapTarget( A4AFList::iterator it, bool ignorenoneeded = false, bool ignoresuspended = false );

	CPartFile*	m_reqfile;

	// base
	void		Init();
	bool		ProcessHelloTypePacket(CMemFile& data);
	void		SendHelloTypePacket(CMemFile* data);
	void		ClearHelloProperties(); // eMule 0.42
	
	/*I2P	
	uint32_t		m_dwUserIP;
	uint32_t		m_nConnectIP;		// holds the supposed IP or (after we had a connection) the real IP
	uint32_t		m_dwServerIP;
	uint32_t		m_nUserIDHybrid;
	uint16_t		m_nUserPort;
	uint16_t		m_nServerPort;
	*/
	
	CI2PAddress	m_nTCPDest;
	CI2PAddress	m_nUDPDest;
	CI2PAddress	m_nServerDest;

	
	uint32_t		m_nClientVersion;
	uint32_t		m_cSendblock;
	uint8_t		m_byEmuleVersion;
	uint8_t		m_byDataCompVer;
	bool		m_bEmuleProtocol;
	wxString	m_Username;
	CMD4Hash	m_UserHash;
	bool		m_HasValidHash;
	
	uint8_t		m_byUDPVer;
	uint8_t		m_bySourceExchangeVer;
	uint8_t		m_byAcceptCommentVer;
	uint8_t		m_byExtendedRequestsVer;
	uint8_t		m_clientSoft;
	uint32_t		m_dwLastSourceRequest;
	uint32_t		m_dwLastSourceAnswer;
	uint32_t		m_dwLastAskedForSources;
	int		m_iFileListRequested;
	bool		m_bFriendSlot;
	bool		m_bCommentDirty;
	bool		m_bIsHybrid;
	bool		m_bIsML;
 	bool		m_bSupportsPreview;
	bool		m_bUnicodeSupport;	
	bool		m_bMultiPacket;
	ClientState	m_clientState;
	CClientTCPSocket*m_socket;		
	bool		m_fNeedOurPublicIP; // we requested our IP from this client

	// Kry - Secure User Ident import
	ESecureIdentState	m_SecureIdentState; 
	uint8_t		m_byInfopacketsReceived;		// have we received the edonkeyprot and emuleprot packet already (see InfoPacketsReceived() )
	//I2P:uint32_t		m_dwLastSignatureIP;
	CI2PAddress	m_dwLastSignatureDest;
	uint8_t		m_bySupportSecIdent;

	uint32_t		m_byCompatibleClient;
	CList<CPacket*>	m_WaitingPackets_list;
	uint32_t		m_lastRefreshedDLDisplay;

	//upload
	void CreateStandartPackets(const unsigned char* data,uint32_t togo, Requested_Block_Struct* currentblock);
	void CreatePackedPackets(const unsigned char* data,uint32_t togo, Requested_Block_Struct* currentblock);

	EUploadState    m_nUploadState;
	uint32_t		m_dwUploadTime;
	uint32_t		m_cAsked;
	uint32_t		m_dwLastUpRequest;
	wxULongLong	m_nCurSessionUp;
	wxULongLong	m_nUpPartCount;
	CMD4Hash	m_requpfileid;
	uint16_t		m_nUpCompleteSourcesCount;

	//! This vector contains the avilability of parts for the file that the user
	//! is requesting. When changing it, be sure to call CKnownFile::UpdatePartsFrequency
	//! so that the files know the actual availability of parts.
	BitVector	m_upPartStatus;
	wxULongLong	m_lastPartAsked;
	wxString	m_strModVersion;

	CList<Requested_Block_Struct*>	m_BlockRequests_queue;
	CList<Requested_Block_Struct*>	m_DoneBlocks_list;

	//download
    EDownloadState  m_nDownloadState;
    bool		    m_bRemoteQueueFull;
	size_t		    m_nPartCount;
	uint32_t	    m_dwLastAskedTime;
	wxString	    m_clientFilename;
	wxULongLong	    m_nTransferedDown;
	wxULongLong	    m_nLastBlockOffset;   // Patch for show parts that you download [Cax2]
	uint16_t		m_cShowDR;
	uint32_t		m_dwLastBlockReceived;
	uint16_t		m_nRemoteQueueRank;
	uint16_t		m_nOldRemoteQueueRank;
	bool		m_bCompleteSource;
	bool		m_bReaskPending;
	bool		m_bUDPPending;
	bool		m_bHashsetRequested;

	CList<Pending_Block_Struct*>	m_PendingBlocks_list;
	CList<Requested_Block_Struct*>	m_DownloadBlocks_list;

	float		kBpsDown;
	float		fDownAvgFilter;
	uint32_t		msReceivedPrev;
	uint32_t		bytesReceivedCycle;
	// chat
    EChatState  m_byChatstate;
	wxString	m_strComment;
	uint8_t		m_iRating;
	bool		m_fHashsetRequesting; // we have sent a hashset request to this client
	bool		m_fNoViewSharedFiles; // client has disabled the 'View Shared Files' feature,
						// if this flag is not set, we just know that we don't know
						// for sure if it is enabled
	bool		m_fSupportsPreview   ;
	bool		m_fSentCancelTransfer; // we have sent an OP_CANCELTRANSFER in the current connection
	bool		m_fSharedDirectories ; // client supports OP_ASKSHAREDIRS opcodes
	uint8_t		m_fSupportsAICH      ; // 3 bits used
	bool		m_fAICHRequested     ;
	bool		m_fSentOutOfPartReqs ;

	bool		m_fOsInfoSupport     ;

	/* Razor 1a - Modif by MikaelB */

	bool		m_bHelloAnswerPending;

	//! This vector contains the availability of parts for the file we requested 
	//! from this user. When changing it, be sure to call CPartFile::UpdatePartsFrequency
	//! so that the files know the actual availability of parts.
	BitVector	m_downPartStatus;

	CAICHHash* 	m_pReqFileAICHHash; 

	ESourceFrom	m_nSourceFrom;

	/* Kad Stuff */
	uint8_t		m_byKadVersion;

	//! This keeps track of aggressive requests for files. 
	uint16_t		m_Aggressiveness;
	
	//! This tracks the time of the last time since a file was requested
	uint32_t		m_LastFileRequest;

	bool 		m_OSInfo_sent;
	
	wxString	m_clientSoftString;	/* software name */
	wxString	m_clientVerString;	/* version + optional mod name */
	wxString	m_clientVersionString;	/* version string */
	wxString	m_fullClientVerString;	/* full info string */
	wxString	m_sClientOSInfo;
	wxString	m_pendingMessage;

	int		SecIdentSupRec;

	CKnownFile*	m_uploadingfile;

	// needed for stats
	uint32_t		m_lastClientSoft;
	uint32_t		m_lastClientVersion;
	wxString	m_lastOSInfo;
	
	/* For buddies timeout */
	uint32_t m_nCreationTime;
};


#define	MAKE_CLIENT_VERSION(mjr, min, upd) \
	((uint32_t)(mjr)*100U*10U*100U + (uint32_t)(min)*100U*10U + (uint32_t)(upd)*100U)


#endif // UPDOWNCLIENT_H
